Require Import digit_big_num. 


Module Type BASE.
  Parameter b : Z.
  Axiom base_positive : (4 < b)%Z.
End BASE.



(*Exemple of implementation using Coq standard integers for both types*)


Module ZZNUMS (Ba: BASE) <: NUMS.

Definition b := Ba.b.
Theorem base_positive : (4 < b)%Z.
  exact Ba.base_positive.
Qed.

Definition digit := Z.
Definition D0 := 0%Z.   
Definition D1 := 1%Z.
Definition Dbase := b.
Definition Dmax := (b-1)%Z.
Definition Dplus := Zplus.
Definition Dminus := Zminus.
Definition Dmult := Zmult.
Definition Ddiv_eucl := Zdiv_eucl.
Definition Dopp := Zopp.
Definition Dlog2 := N_digits.
Definition Dcompare := Zcompare.
Definition IDN := Zabs_nat.

(* Definition t_ := (Z * Z)%type. *)

(* Definition Dlow := (@fst Z Z). *)
(* Definition Dhigh := (@snd Z Z). *)

(* Definition Dplus_ (m n : Z) := *)
(*   (let s := m + n in  *)
(*     if Zle_bool b s *)
(*       then (s-b, 1) *)
(*       else if Zle_bool s (-b)  *)
(* 	then (s+b, -1) *)
(* 	else (s, 0))%Z. *)

Delimit Scope digit_scope with D.

Infix "+" := Dplus : digit_scope.
Infix "-" := Dminus : digit_scope.
Infix "*" := Dmult : digit_scope.
Infix "?=" := Dcompare : digit_scope.
Notation "- x" := (Dopp x) : digit_scope.
Notation "0" := D0 : digit_scope.
Notation "1" := D1 : digit_scope.
Notation "4" := (1+1+1+1)%D : digit_scope.

Definition IDZ : Z -> Z. intros z; exact z. Defined.
Theorem D0_IDZ : IDZ 0 = 0%Z. auto. Qed.
Theorem D1_IDZ : IDZ 1 = 1%Z. auto. Qed.
Theorem Dmax_IDZ : IDZ Dmax = (b -1)%Z. auto. Qed.
Theorem Dbase_IDZ : IDZ Dbase = b. auto. Qed.

(*
Theorem plus_IDZ_0 : 
    forall m n, 
      (-b < IDZ m + IDZ n < b ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n /\ 
	Dhigh (Dplus_ m n) = D0)%Z.
Proof.
unfold IDZ, Dlow, Dplus_, Dhigh.
intros m n Hsum.
Zbool G1.
absurd (b <= m + n)%Z; omega.
Zbool G2.
absurd (m + n <= - b)%Z; omega.
simpl; auto.
Qed.

Theorem plus_IDZ_1 : 
  forall m n, 
    (b <= IDZ m + IDZ n <= 2*b-2 ->
      IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n -b /\ 
      Dhigh (Dplus_ m n) = D1)%Z.
Proof.
unfold IDZ, Dlow, Dplus_, Dhigh.
intros m n Hsum.
Zbool G1.
simpl; auto.
Zbool G2.
absurd (m + n <= - b)%Z; omega.
absurd (m + n > - b)%Z; omega.
Qed.

Theorem plus_IDZ_m1 : 
  forall m n, 
    (-2*b+2 <= IDZ m + IDZ n <= - b ->
      IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n + b/\ 
      Dhigh (Dplus_ m n) = (-D1)%D)%Z.
Proof.
unfold IDZ, Dlow, Dplus_, Dhigh.
intros m n Hsum.
Zbool G1.
absurd (b <= m + n)%Z; omega.
Zbool G2.
simpl; auto.
absurd (m + n > - b)%Z; omega.
Qed.
*)


Theorem plus_IDZ: forall m n : digit, IDZ (m + n)%D = (IDZ m + IDZ n)%Z.
     auto. Qed.
Theorem minus_IDZ: forall m n : digit, IDZ (m - n)%D = (IDZ m - IDZ n)%Z.
  auto. Qed.
Theorem mult_IDZ: forall m n : digit, IDZ (m * n)%D = (IDZ m * IDZ n)%Z.
  auto. Qed.
(* Theorem div_IDZ : forall m n, let (k,r) := Ddiv_eucl m n in  *)
(*   IDZ k = (IDZ m/IDZ n)%Z. *)
(* Proof. *)
(*   unfold Ddiv_eucl, Zdiv, IDZ. *)
(*   intros m n; caseEq (Zdiv_eucl m n); tauto. Qed. *)
(* Theorem mod_IDZ : forall m n, let (k,r) := Ddiv_eucl m n in  *)
(*   IDZ r = (IDZ m mod IDZ n)%Z.   *)
(* Proof. *)
(*   unfold Ddiv_eucl, Zmod, IDZ. *)
(*   intros m n; caseEq (Zdiv_eucl m n); tauto. Qed. *)
Theorem div_eucl_IDZ : forall m n, 
  let (q, r) := Ddiv_eucl m n in 
    (IDZ q, IDZ r) = Zdiv_eucl (IDZ m) (IDZ n).
Proof.
intros m n.
unfold Ddiv_eucl, IDZ.
caseEq  (Zdiv_eucl m n).
tauto.
Qed.
 Theorem opp_IDZ: forall n, IDZ (- n) = (- IDZ n)%Z.
  auto. Qed.
Theorem log_IDZ : forall n : digit, IDZ (Dlog2 n) = N_digits (IDZ n). 
  auto. Qed.
Theorem IDZ_IDN : forall n, IDN n = Zabs_nat (IDZ n). 
  auto. Qed.
Theorem compare_IDZ:
  forall x y : digit,
    match (x ?= y)%D with
      | Eq => IDZ x = IDZ y
      | Lt => (IDZ x < IDZ y)%Z
      | Gt => (IDZ x > IDZ y)%Z
    end.
  intros x y.
unfold Dcompare, IDZ.
elim (Ztrichotomy x y); intros G1.
replace (x ?= y)%Z with Lt; assumption.
elim G1; clear G1; intros G1.
subst; rewrite Zcompare_refl; auto.
replace (x ?= y)%Z with Gt; assumption.
Qed.


Definition big := Z.
Definition B0 := 0%Z.  
Definition B1 := 1%Z.
Definition Bbase := b.
Definition Bplus := Zplus.
Definition Bmult := Zmult.
Definition Bpower := Zpower.
Definition Bminus := Zminus.
Definition Bdiv_eucl := Zdiv_eucl.
Definition Bopp := Zopp.
Definition Ble_bool := Zle_bool.
Infix "+" := Bplus : big_scope.
Infix "-" := Bminus : big_scope.
Infix "*" := Bmult : big_scope.
Infix "^" := Bpower : big_scope.
Infix "<=" := Ble_bool : big_scope.
Notation "- x" := (Bopp x) : big_scope.
Notation "0" := B0 : big_scope.
Notation "1" := B1 : big_scope.

Delimit Scope big_scope with B.  
Definition IBZ : big -> Z. intros z; exact z. Defined.
Definition IZB : big -> Z. intros z; exact z. Defined.

Theorem B0_IBZ : IBZ 0%B = 0%Z. auto. Qed.
Theorem B1_IBZ : IBZ 1%B =1%Z. auto. Qed.
Theorem Bbase_IBZ : IBZ Bbase = b. auto. Qed.

Theorem IZB_IBZ_eq : forall b c, IZB b = c <-> b = IBZ c. split;  auto. Qed.
Theorem plus_IBZ: forall m n : big, IBZ (m + n)%B = (IBZ m + IBZ n)%Z.
  auto. Qed.
Theorem minus_IBZ: forall m n : big, IBZ (m - n)%B = (IBZ m - IBZ n)%Z.
  auto. Qed.
Theorem mult_IBZ: forall m n : big, IBZ (m * n)%B = (IBZ m * IBZ n)%Z.
  auto. Qed.
Theorem power_IBZ: forall m n : big, IBZ (m ^ n)%B = (IBZ m ^ IBZ n)%Z.
  auto. Qed.
Theorem Ropp_Ropp_IBZ: forall n: big, IBZ (- n)%B = (- IBZ n)%Z.
  auto. Qed.
Theorem div_IBZ : forall m n, let (k,r) := Bdiv_eucl m n in 
  IBZ k = (IBZ m/IBZ n)%Z.
Proof.
unfold Bdiv_eucl, Zdiv, IBZ.
intros m n; caseEq (Zdiv_eucl m n); tauto. Qed.
Theorem mod_IBZ : forall m n, let (k,r) := Bdiv_eucl m n in 
  IBZ r = (IBZ m mod IBZ n)%Z.  
Proof.
unfold Bdiv_eucl, Zmod, IBZ.
intros m n; caseEq (Zdiv_eucl m n); tauto. Qed.
Theorem opp_IBZ: forall n, IBZ (- n)%B = (- IBZ n)%Z.
  auto. Qed.
Theorem le_IBZ: forall m n : big, (m <= n)%B = (Zle_bool (IBZ m) (IBZ n)).
  auto. Qed.

Definition digit_of_big (x: big) : digit := x.
Definition big_of_digit (x: digit) : big := x.

Theorem digit_of_big_of_digit_inv : forall d, digit_of_big (big_of_digit d) = d.
unfold digit_of_big, big_of_digit; auto. Qed.
Theorem big_of_digit_of_big_inv : forall x, big_of_digit (digit_of_big x) = x.
unfold digit_of_big, big_of_digit; auto. Qed.
Theorem IBZ_IDZ : forall d, IBZ (big_of_digit d) = IDZ d.
unfold IBZ, big_of_digit; auto. Qed.
End ZZNUMS.


(*Exemple of implementation using BigZ for both types*)

Require Export BigZ.

Module BIGZNUMS (Ba : BASE)<: NUMS.
Open Scope bigZ_scope.

Definition b : Z := Ba.b.
Theorem base_positive : (4 < b)%Z.
  exact Ba.base_positive.
Qed.

Definition digit : Type := bigZ.
Definition D0    : digit := 0.
Definition D1    : digit := 1.
Definition Dbase : digit := BigZ.of_Z b.
Definition Dmax  : digit := Dbase-1.
Definition Dplus : digit -> digit -> digit := BigZ.add.
Definition Dminus: digit -> digit -> digit := BigZ.sub.
Definition Dmult : digit -> digit -> digit := BigZ.mul.
Definition Ddiv_eucl : digit -> digit -> digit*digit := BigZ.div_eucl.
Definition Dopp  : digit -> digit := BigZ.opp.

Definition Nlog2 x := 
  match x with
  | BigN.N0 x => 30 - BigZ.Pos (BigN.N0 (head031 x))
  | BigN.N1 x => 61 - BigZ.Pos (BigN.N1 (BigN.w1_op.(CyclicAxioms.znz_head0) x))
  | BigN.N2 x => 123 - BigZ.Pos (BigN.N2 (BigN.w2_op.(CyclicAxioms.znz_head0) x))
  | BigN.N3 x => 247 - BigZ.Pos (BigN.N3 (BigN.w3_op.(CyclicAxioms.znz_head0) x))
  | BigN.N4 x => 495 - BigZ.Pos (BigN.N4 (BigN.w4_op.(CyclicAxioms.znz_head0) x)) 
  | BigN.N5 x => 991 - BigZ.Pos (BigN.N5 (BigN.w5_op.(CyclicAxioms.znz_head0) x))
  | BigN.N6 x => 1983 - BigZ.Pos (BigN.N6 (BigN.w6_op.(CyclicAxioms.znz_head0) x))
  | _ => 0
  end.

Definition Dlog2 (x : digit) : digit :=
 match x with
 | BigZ.Pos x | BigZ.Neg x => Nlog2 x
 end.


Definition Dcompare : digit -> digit -> comparison := BigZ.compare.
Definition IDN (x :digit) : nat := Zabs_nat (BigZ.to_Z x).
 

(*
Definition le31 x y :=
 match compare31 x y with
 | Gt => false
 | _ => true
 end.
*)
(* Definition split (f:BigN.t -> BigZ.t) n := *)
(*  match n with *)
(*  | BigN.N0 _ | BigN.N1 _ => (f 0%bigN, f n) *)
(*  | BigN.N2 W0 => (f 0%bigN, f 0%bigN) *)
(*  | BigN.N2 (WW h l) => (f (BigN.N1 h), f (BigN.N1 l)) *)
(*  | _ =>  (f 0%bigN, f 0%bigN) *)
(*  end. *)
(*
Definition Dmultb (b x y:digit) :=
  match x * y with
  | BigZ.Pos n => split BigZ.Pos n
  | BigZ.Neg n => split BigZ.Neg n
  end. 
*)


Definition IDZ : digit -> Z := BigZ.to_Z.

Delimit Scope digit_scope with D.

Infix "+" := Dplus : digit_scope.
Infix "-" := Dminus : digit_scope.
Infix "*" := Dmult : digit_scope.
Notation "- x" := (Dopp x) : digit_scope.
Notation "0" := D0 : digit_scope.
Notation "1" := D1 : digit_scope.
Notation "4" := (1+1+1+1)%D : digit_scope.

Theorem D0_IDZ : IDZ D0 = 0%Z. 
reflexivity.
Qed.

Theorem D1_IDZ : IDZ D1 = 1%Z. 
reflexivity.
Qed.

Theorem Dbase_IDZ : IDZ Dbase = b.
unfold Dmax, IDZ, Dbase.
apply BigZ.spec_of_Z.
Qed.

Theorem Dmax_IDZ : IDZ Dmax = (b -1)%Z.
unfold Dmax, Dbase, IDZ, D1.
rewrite BigZ.spec_sub.
rewrite BigZ.spec_of_Z.
fold IDZ.
fold D1.
rewrite D1_IDZ.
reflexivity.
Qed.

Theorem plus_IDZ : forall m n : digit, IDZ (m + n)%D = (IDZ m + IDZ n)%Z.
  exact BigZ.spec_add.
Qed.

Theorem minus_IDZ: forall m n : digit, IDZ (m - n)%D = (IDZ m - IDZ n)%Z.
exact BigZ.spec_sub.
Qed.

Theorem mult_IDZ: forall m n : digit, IDZ (m * n)%D = (IDZ m * IDZ n)%Z.
exact BigZ.spec_mul.
Qed.

Theorem div_eucl_IDZ : forall m n, 
  let (q, r) := Ddiv_eucl m n in 
    (IDZ q, IDZ r) = Zdiv_eucl (IDZ m) (IDZ n).
Proof.
intros m n.
unfold Ddiv_eucl, IDZ.
Admitted.

Theorem opp_IDZ: forall n, IDZ (- n)%D = (- IDZ n)%Z.
exact BigZ.spec_opp.
Qed.

Parameter log_IDZ : forall n : digit, IDZ (Dlog2 n) = N_digits (IDZ n). 

Theorem compare_IDZ:
  forall x y : digit,
    match (x ?= y)%D with
      | Eq => IDZ x = IDZ y
      | Lt => (IDZ x < IDZ y)%Z
      | Gt => (IDZ x > IDZ y)%Z
    end.
exact  BigZ.spec_compare.
Qed.

Theorem IDZ_IDN : forall n, IDN n = Zabs_nat (IDZ n). 
unfold IDN, IDZ; intros n; reflexivity.
Qed.

Definition big : Type := bigZ.
Definition B0  : big  := 0%bigZ.  
Definition B1  : big  := 1%bigZ.
Definition Bbase : big := BigZ.of_Z b.
Definition Bplus : big -> big -> big := BigZ.add.
Definition Bmult : big -> big -> big := BigZ.mul.
Definition Bminus : big -> big -> big := BigZ.sub.
Definition Bpower (x y : big) : big := 
  match BigZ.to_Z y with
  | Zpos p => BigZ.power_pos x p
  | Z0 => B1
  | Zneg _ => B0
  end.
Definition Bdiv_eucl : big -> big -> big*big := BigZ.div_eucl.
Definition Bopp : big -> big := BigZ.opp.
Definition Ble_bool (x y : big) : bool := 
  match BigZ.compare x y with
  | Gt => false
  | _ => true
  end.

Infix "+" := Bplus : big_scope.
Infix "-" := Bminus : big_scope.
Infix "*" := Bmult : big_scope.
Infix "^" := Bpower : big_scope.
Infix "<=" := Ble_bool : big_scope.
Notation "- x" := (Bopp x) : big_scope.
Notation "0" := B0 : big_scope.
Notation "1" := B1 : big_scope.

Delimit Scope big_scope with B.  

Definition IBZ : big -> Z := BigZ.to_Z.
Definition IZB : Z -> big := BigZ.of_Z.

Theorem IZB_IBZ_eq : forall m n, IZB m = n <-> m = IBZ n. 
unfold IBZ, IZB.
intros m n.
split; intros G1; subst.
rewrite BigZ.spec_of_Z; auto.
Admitted.

Theorem Bbase_IBZ : IBZ Bbase = b.
unfold IBZ, Bbase; rewrite BigZ.spec_of_Z; auto.
Qed.

Theorem B0_IBZ : IBZ 0%B = 0%Z. 
reflexivity.
Qed.

Theorem B1_IBZ : IBZ 1%B =1%Z. 
reflexivity.
Qed.

Theorem plus_IBZ: forall m n : big, IBZ (m + n)%B = (IBZ m + IBZ n)%Z.
unfold IBZ; exact BigZ.spec_add.
Qed.

Theorem mult_IBZ: forall m n : big, IBZ (m * n)%B = (IBZ m * IBZ n)%Z.
exact BigZ.spec_mul.
Qed.

Theorem minus_IBZ: forall m n : big, IBZ (m - n)%B = (IBZ m - IBZ n)%Z.
exact BigZ.spec_sub.
Qed.

Theorem power_IBZ: forall m n : big, IBZ (m ^ n)%B = (IBZ m ^ IBZ n)%Z.
intros m n.
unfold IBZ, Bpower, B1.
case (BigZ.to_Z n).
simpl Zpower; auto.
intros p.
apply BigZ.spec_power_pos.
intros; simpl Zpower; auto.
Qed.

Theorem div_IBZ : forall m n, let (k,r) := Bdiv_eucl m n in 
  IBZ k = (IBZ m/IBZ n)%Z.
Proof.
intros m n.
unfold Bdiv_eucl, IBZ.
generalize (BigZ.spec_div_eucl m n).
caseEq  (BigZ.div_eucl m n).
intros q r.
Admitted.

Parameter mod_IBZ : forall m n, let (k,r) := Bdiv_eucl m n in 
  IBZ r = (IBZ m mod IBZ n)%Z.  

Theorem opp_IBZ: forall n, IBZ (- n)%B = (- IBZ n)%Z.
exact BigZ.spec_opp.
Qed.

Theorem le_IBZ: forall m n : big, (m <= n)%B = (Zle_bool (IBZ m) (IBZ n)).
intros m n.
unfold Ble_bool, Zle_bool, IBZ.
Admitted.


(*
Definition digit_of_Z (z:Z) :=
 match z with
 | Zpos _ =>phi_inv z
 | Z0 => 0%int31
 | Zneg p =>  Dopp (phi_inv (Zpos p))
 end.
*)

Definition digit_of_big (x: big) : digit := x.
Definition big_of_digit (d:digit) : big := d.

Theorem digit_of_big_of_digit_inv : forall d, digit_of_big (big_of_digit d) = d.
intros d; reflexivity.
Qed.

Theorem big_of_digit_of_big_inv : forall b, big_of_digit (digit_of_big b) = b.
intros d; reflexivity.
Qed.

Theorem IBZ_IDZ : forall d, IBZ (big_of_digit d) = IDZ d.
intros d; reflexivity.
Qed.

(*

  Axiom plus_IDZ_0 : 
    forall m n, 
      (-b < IDZ m + IDZ n < b ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n /\ 
	Dhigh (Dplus_ m n) = D0)%Z.

  Axiom plus_IDZ_1 : 
    forall m n, 
      (b <= IDZ m + IDZ n <= 2*b-2 ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n -b /\ 
	Dhigh (Dplus_ m n) = D1)%Z.

  Axiom plus_IDZ_m1 : 
    forall m n, 
      (-2*b+2 <= IDZ m + IDZ n <= - b ->
	IDZ (Dlow (Dplus_ m n)) = IDZ m + IDZ n + b/\ 
	Dhigh (Dplus_ m n) = (-D1)%D)%Z.
*)

End BIGZNUMS.






