Require Export inverse_correct. 
Require Import kanto_exists_uniq kanto_speed.
Require Import Streams.

Module NEWTON(Nums : NUMS).

Module Inv_correct := INV_CORRECT(Nums).
Export Inv_correct.

Fixpoint EXn g x0 (n : nat)  {struct n} : Stream digit :=
match n with 
| 0 => x0
|S n => let xn := (EXn g x0 n) in  Ssub xn (g xn)
end.

Lemma disc_b2:
forall r b0, -1 <= r <= 1 ->
0<= b0 ->
included (c_disc r (2*b0)) (c_I (-1) 1) -> b0 <= 1/2.
Proof.
intros r b0 Hr Hb0 H1.
assert (Hd: c_disc r (2 * b0) (r + 2 * b0)).
unfold c_disc; rewrite Rabs_left1;
ring_simplify;  fourier.
assert (H2:= H1 (r  + 2*b0) Hd).
assert (Hdm: c_disc r (2 * b0) (r - 2 * b0)).
unfold c_disc; rewrite Rabs_right;
ring_simplify;  fourier.
assert (H3:= H1 (r  - 2*b0) Hdm).
unfold c_I in *; dec_tac.
apply Rmult_le_reg_l with 2; try fourier.
Qed.


Theorem EXn_correct : 
forall g EX0 f f' X0 a0 b0 c n
  (A0_b0_c: 2*a0*b0*c <=1 )
  (pr: forall t:R, c_I (-1) 1 t -> derivable_pt f t)
  (Hder_f: forall (t:R)(H:c_I (-1) 1 t), derive_pt f t (pr t H)=f' t) 
  (Hcont_f':forall y, (c_I (-1) 1 y) -> continuity_pt f' y) 
  (Hlim_f':forall y1 y2:R, c_I (-1) 1 y1 -> c_I (-1) 1 y2 -> 
    Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2))) 
  (Hincl: -1 < X0 < 1) 
  (Hdif_f':f' X0 <> 0 ) 
  (Habs_a: Rabs (/(f' X0)) <= a0) 
  (Habs_b:Rabs (f X0 / (f' X0)) <= b0 ) 
  ( Hb_eps: included (c_disc X0 (2*b0)) (c_I (-1) 1)),
  Srepresents EX0 X0 ->
  (forall x vx, Srepresents x vx -> 
    (c_disc X0 (2*b0) vx) -> f' vx <>0 -> -1 <= f vx / f' vx <= 1 ->
    Srepresents (g x) (f vx/ f' vx)) -> 
  Srepresents (EXn g EX0 n) (Xn X0 f f' n).
Proof.
intros g EX0 f f' X0 a0 b0 c n  A0_b0_c  pr 
  Hder_f Hcont_f'   Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_ep
Hrepxo Hrepgf ; induction n. 
simpl; tauto.

simpl.
apply Ssub_correct.
apply IHn; assumption.
apply Hrepgf.
apply IHn; assumption.
apply (xn_in_disc a0 b0 c X0 (-1) 1 f f' A0_b0_c pr
Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_ep  n).
destruct (lemma_princ a0 b0 c X0 (-1) 1 f f' A0_b0_c pr
Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_ep n); 
intuition.
rewrite <-Rabs_le_le_eq.

apply Rle_trans with (Bn a0 b0 c n).
destruct (lemma_princ a0 b0 c X0 (-1) 1 f f' A0_b0_c pr
Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_ep  n)
; intuition.
apply Rle_trans with b0.
pattern b0 at 2; replace b0 with (Bn a0 b0 c 0);[|simpl; trivial].
case (O_or_S n); intro H; elim H;  [intros m Hm;
apply Bn_monot3  with X0 (-1) 1 f f'; 
  try assumption; omega| right; trivial].
apply Rle_trans with (1/2); try fourier.
apply disc_b2 with X0.
elim Hincl; intros; split; fourier.
apply H_b0 with X0 f f'; assumption.
assumption.
replace (Xn X0 f f' n - f (Xn X0 f f' n) / f' (Xn X0 f f' n)) 
with (Xn X0 f f' (S n));[| simpl; trivial].
assert (-1 <  Xn X0 f f' (S n) < 1); [|intuition].
fold (o_I (-1) 1 (Xn X0 f f' (S n))).
apply (in_open_xn a0 b0 c X0 (-1) 1 f f' A0_b0_c pr
Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_ep (S n))
; intuition.
Qed.

Theorem kantoro_digit_approx : 
  forall f f' X0 A B a0 b0 c n xs0
    (A0_b0_c: 2*a0*b0*c <=1 )
    (pr: forall t:R, c_I A B t -> derivable_pt f t)
    (Hder_f: forall (t:R)(H:c_I A B t), derive_pt f t (pr t H)=f' t) 
    (Hcont_f':forall y, (c_I A B y) -> continuity_pt f' y) 
    (Hlim_f':forall y1 y2:R, c_I A B y1 -> c_I A B y2 -> 
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2))) 
    (Hincl: A < X0 < B) 
    (Hdif_f':f' X0 <> 0 ) 
    (Habs_a: Rabs (/(f' X0)) <= a0) 
    (Habs_b:Rabs (f X0 / (f' X0)) <= b0 ) 
    ( Hb_eps: included (c_disc X0 (2*b0)) (c_I A B)),
    (Z2R b - 2) * Rpower 2 (INR n-1) >= b0 * (2* Z2R b * Z2R b) ->     
    Un_cv (Xn X0 f f') xs0 ->
    
    Rabs (xs0 -  (Xn X0 f f' n)) * (2* Z2R b * Z2R b)
    <=(Z2R b -2).
Proof.
assert (Hbpos:= base_positive).
intros  f f' X0 A B a0 b0 c n xs0 A0_b0_c  pr 
  Hder_f Hcont_f'   Hlim_f' Hincl Hdif_f' Habs_a Habs_b 
    Hb_eps Hspeed Hcv1.
apply Rdiv_le_l; [ZR0compat|].
apply Rle_trans with 
((/ (Rpower 2 (INR n - 1))) * ((mu a0 b0 c 0%nat) ^ (Npow 2 n - 1)) * b0).

assert (R1 : xs a0 b0 c X0 A B f f' A0_b0_c  pr 
      Hder_f Hcont_f' Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps = xs0).
apply kantoro_unic2.
rewrite rewrite_c_disc.
  apply lim_in_interv with X0 A B (Xn X0 f f'); try assumption.
intros; rewrite <- rewrite_c_disc; apply xn_in_disc with a0  c A B pr; 
  try assumption.
elim (conv_root a0 b0 c X0 A B f f' A0_b0_c  pr 
      Hder_f Hcont_f' Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps xs0 Hcv1);
 intros ; assumption.

rewrite <- R1.
apply conv_speed.
apply Rle_trans with ( / Rpower 2 (INR n - 1) * b0).
rewrite Rmult_assoc.
apply Rmult_le_compat_l.
left.
apply Rinv_0_lt_compat.
apply Rpow_pos; fourier.
pattern b0 at 3; replace b0 with (1*b0);[|ring].

apply Rmult_le_compat_r.
apply (H_b0 _ _ _ _ Habs_b) .
assert (0 <=  mu a0 b0 c 0 ^ (Npow 2 n - 1) <= 1); [|intuition].
apply pow_le1; unfold mu; simpl; split; 
[apply (greater0 a0 b0 c X0 A B f f'); assumption| trivial].

replace (/ Rpower 2 (INR n - 1) * b0) with 
  (b0 / Rpower 2 (INR n - 1)) by (unfold Rdiv; ring).
apply Rdiv_mult_compat.
ZR0compat.
apply Rpow_pos; fourier.
fourier.
Qed.

Lemma Hbgt0: 0 < Z2R b.
Proof. assert (Hbp:=base_positive); ZR0compat. Qed.

Definition nb:nat:= proj1_sig (archim (Z2R b) Hbgt0).

CoFixpoint exact_newton (g : Stream digit -> Stream digit) EX0 n :=
  match (make_digit (EXn g EX0 n)) with
    | x1::x' => x1::exact_newton (fun x => multb (g (x1::x))) x' n
  end.

Lemma sig_proj : forall (A: Type)(P:A->Type) 
  (y: {x:A & P x}) (x:A), x = projT1 y -> P x.
intros A P [y p] x q.
rewrite q.
trivial.
Defined.

Theorem kanto_correct : 
  forall g EX0 f f' X0 a0 b0 c n xs0
    (pr : forall t:R, c_I (-1) 1 t -> derivable_pt f t), 
    (forall (t:R)(H: c_I (-1) 1 t), derive_pt f t (pr t H)=f' t) ->
    (forall y, (c_I (-1) 1 y) -> continuity_pt f' y) -> 
    (forall y1 y2:R, c_I (-1) 1 y1 -> c_I (-1) 1 y2 -> 
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2))) -> 
    2*a0*b0*c <=1 ->
    -1 < X0 < 1 -> 
    f' X0 <> 0 -> 
    Rabs (/(f' X0)) <= a0 -> 
    Rabs (f X0 / (f' X0)) <= b0 ->
    included (c_disc X0 (2*b0)) (c_I (-1) 1) ->
    (Z2R b - 2) * Rpower 2 (INR n-1) >= b0 * (2* Z2R b * Z2R b) ->  
    (Z2R  b)  <=    2^n ->
    (forall x vx, Srepresents x vx ->
      (c_disc X0 (2*b0) vx) -> f' vx <>0 -> -1 <= f vx / f' vx <= 1 ->
      Srepresents (g x) (f vx/ f' vx)) -> 
    Srepresents EX0 X0 ->
    Un_cv (Xn X0 f f') xs0 ->
    (*c_disc X0 (2*b0) xs0 -> 
    f xs0 = 0 ->*)
    Srepresents (exact_newton g EX0 n) xs0.
Proof.
assert (Hbpos:= base_positive).
cofix.
intros g EX0 f f' X0 a0 b0 c n xs0 pr Hder_f Hcont_f' Hlim_f' 
   A0_b0_c Hincl Hdif_f' Habs_a Habs_b Hb_eps 
  Hspeed Hnb  Hgeq HrepX0 Hcv1.
str_unfold (exact_newton g EX0 n).
simpl.
fold_decomp.
assert (HrepEXn : Srepresents (EXn g EX0 n) (Xn X0 f f' n)).
   clear kanto_correct.
    abstract (apply (EXn_correct g EX0 f f' X0 a0 b0 c n  A0_b0_c  pr 
    Hder_f Hcont_f'   Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps)
  ; try assumption).

assert (Hxs0 : xs0 = (Xn X0 f f' n + (xs0 - (Xn X0 f f' n)))) by 
  (clear kanto_correct; abstract ring).

rewrite Hxs0.

assert (Hmd_eq := make_digit_eq (EXn g EX0 n) (Xn X0 f f' n) HrepEXn). 
inversion Hmd_eq.

assert (Hmd_add : -1 <= xs0 * Z2R b - IDR k <= 1).
clear kanto_correct.
abstract (rewrite Hxs0;
  apply (make_digit_add (EXn g EX0 n) k s); try assumption; [
    apply sym_eq; assumption |
    rewrite <- Hxs0;
    apply Hb_eps;
rewrite rewrite_c_disc;
  apply lim_in_interv with X0 (-1) 1 (Xn X0 f f'); try assumption;
intros; rewrite <- rewrite_c_disc; apply xn_in_disc with a0  c (-1) 1 pr; 
  try assumption|
    rewrite <-Rabs_le_le_eq;
    rewrite Rabs_mult;
    rewrite (Rabs_right (2 * Z2R b * Z2R b)); [| ZR0compat];
    apply (kantoro_digit_approx  f f' X0 (-1) 1 a0 b0 c n xs0 A0_b0_c  pr 
  Hder_f Hcont_f'   Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps Hspeed Hcv1) ]).

replace ((IDR k + r) / Z2R b + (xs0 - (IDR k + r) / Z2R b)) with 
  ((IDR k + (xs0*Z2R b-IDR k)) / Z2R b) by ( clear kanto_correct; 
  abstract field_Z).
constructor; try assumption.

assert (A0_b0_c2 : 2 *  (Z2R b * An a0 n) * ( Z2R b * Bn a0 b0 c n)  *
     (c/Z2R b^2) <= 1).
  clear kanto_correct.
  abstract ( 
    replace ( 2 * (Z2R b * An a0 n) * (Z2R b * Bn a0 b0 c n) * (c / Z2R b ^ 2)) 
     with ( 2  * An a0 n   * Bn a0 b0 c n * c ) by field_Z;
      apply An_bn_c1 with X0 (-1) 1 f f'; trivial).

assert (HinI: forall t, c_I (-1) 1 t -> c_I (-1) 1 ((IDR k + t) */ Z2R b)).
  clear kanto_correct.
  abstract (intros t Ht; unfold c_I in *;
  fold (Rdiv (IDR k +t) (Z2R b));
  apply Rle_div; [ZR0compat|];
  assert (- Z2R b +1  <= IDR k <= Z2R b -1) by (split; IDR_to_Z);
  dec_tac; split; fourier).

assert (pr2:  forall t : R, c_I (-1) 1 t -> derivable_pt 
  (fun x => f ((IDR k + x)/Z2R b)) t).
  clear kanto_correct.
  abstract (intros t HAB;  reg; apply pr;  
  apply HinI; assumption).
    
assert  (Hder_f2 : forall (t : R) (H : c_I (-1) 1 t),  
  derive_pt  (fun x => f ((IDR k + x)/Z2R b)) t (pr2 t H) 
  =  (fun x => 1/Z2R b*f'((IDR k + x)/Z2R b)) t).
  clear kanto_correct.
  abstract (intros t Hc; reg; [
    rewrite Rmult_comm; unfold Rdiv; rewrite Rmult_1_l;
      apply Rmult_eq_compat_l;
	replace ( derive_pt f ((IDR k + t) * / Z2R b) H4) with 
          ( derive_pt f ((IDR k + t) * / Z2R b)
	(pr ((IDR k + t) */ Z2R b) (HinI t Hc))); [|apply pr_nu];
	apply Hder_f 
    | apply pr; apply HinI; try assumption]).

assert (Hcont_f'2 : forall y : R, c_I (-1) 1 y -> 
  continuity_pt (fun x => 1/Z2R b*f'((IDR k + x)/Z2R b))  y).
  clear kanto_correct.
  abstract (intros t Ht; reg; apply Hcont_f'; apply HinI; assumption).

assert  (Hlim_f'2 : forall y1 y2 : R,
        c_I (-1) 1 y1 ->
        c_I (-1) 1 y2 ->
     Rabs ( 1/Z2R b*f'((IDR k + y1)/Z2R b) -  1/Z2R b*
    f'((IDR k + y2)/Z2R b)) <= c / Z2R b^2 * Rabs (y1 - y2)).
  clear kanto_correct.
abstract(  intros y1 y2 Hcy1 Hcy2;
 (rewrite <-Rmult_minus_distr_l; rewrite Rabs_mult; rewrite Rabs_right;
   [|ZR0compat];
  apply Rle_trans with ((1/Z2R b)*(c* Rabs ((IDR k + y1) / Z2R b - (IDR k + y2) / Z2R b)));
    [apply Rmult_le_compat_l;
   [ZR0compat| apply Hlim_f'; unfold Rdiv; apply HinI; assumption]|
  replace ((IDR k + y1) / Z2R b - (IDR k + y2) / Z2R b) with 
    ((y1 - y2)/Z2R b) by field_IDR;
  rewrite Rabs_div; [|R_to_Z];
  rewrite (Rabs_right (Z2R b)); [|ZR0compat];
  right; field_IDR])).

assert (Hrin: -1 < r < 1).
  clear kanto_correct;
  abstract(generalize (make_digit_bounds (EXn g EX0 n) (Xn X0 f f' n) r HrepEXn);
  caseEq (make_digit (EXn g EX0 n));
  rewrite <- H;
  intros d s0 Heq5 U1;
  injection Heq5; intros R1 R2;
  apply U1;[
    fold (o_I (-1) 1 (Xn X0 f f' n));
      apply (in_open_xn a0 b0 c X0  (-1) 1  f f' A0_b0_c pr
	Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps n); intuition 
    | subst s0; assumption
    | subst k; rewrite <- H0; field_Z]).

assert (Hf0': f'(Xn X0 f f' n ) <> 0).
  clear kanto_correct.
  abstract (destruct (lemma_princ a0 b0 c X0 (-1) 1  f f' A0_b0_c pr
    Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps n); intuition).

assert   (Hdif_f'2 : (fun x =>  1/Z2R b*f'((IDR k + x)/Z2R b)) r  <> 0).
  clear kanto_correct.
  abstract (rewrite H0; apply prod_neq_R0; [ZR0compat|trivial]).

assert (Habs_a2 : Rabs (/ (fun x =>  1/Z2R b*f'((IDR k + x)/Z2R b)) r) 
   <= Z2R b*An a0  n).
  clear kanto_correct.
  abstract (rewrite H0; 
  replace (/ (1 / Z2R b * f' (Xn X0 f f' n))) with ( Z2R b  / f' (Xn X0 f f' n));
    [| field_Z; assumption];
  unfold Rdiv;rewrite Rabs_mult; 
  rewrite Rabs_right ;[|ZR0compat];
    apply Rmult_le_compat_l;[ZR0compat|];
  destruct (lemma_princ a0 b0 c X0 (-1) 1  f f' A0_b0_c  pr
    Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps n); intuition).

assert (Habs_b2 : Rabs ((fun x => f ((IDR k + x)/Z2R b)) r / 
  (fun x =>  (1/Z2R b)*f'((IDR k + x)/Z2R b)) r)  <= Z2R b*(Bn a0 b0 c n)).
  clear kanto_correct.
  abstract ( replace (f ((IDR k + r) / Z2R b) / ((1 / Z2R b) * f' ((IDR k + r) / Z2R b))) with 
    (Z2R b * (f ((IDR k + r) / Z2R b) /  (f' ((IDR k + r) / Z2R b)))) ;
    [| field_Z; rewrite H0; trivial];
  rewrite Rabs_mult; rewrite Rabs_right; [|ZR0compat];
  apply Rmult_le_compat_l; [ZR0compat|];
  rewrite H0;
  destruct (lemma_princ a0 b0 c X0  (-1) 1  f f' A0_b0_c pr
    Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps n); intuition).

assert (Hreq: r = Z2R b * Xn X0 f f' n - IDR k) by 
  (clear kanto_correct; abstract (rewrite <- H0; field_IDR)).

assert (Hb_eps2 : included (c_disc r (2 *  (Z2R b * Bn a0 b0 c n) )) (c_I (-1) 1)).
  clear kanto_correct.
  abstract (destruct (lemma_princ a0 b0 c X0  (-1) 1  f f' A0_b0_c pr
    Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps n) as [_[Hlp _] ];
  intros t Ht;
  unfold c_I;
  rewrite rewrite_c_disc in Ht;
  elim Ht; intros Ht1 Ht2;
  split; [
    apply Rle_trans with  ( r - 2 * (Z2R b * Bn a0 b0 c n) );[|assumption];
    rewrite Hreq;
    replace (Z2R b * Xn X0 f f' n - IDR k - 2 * (Z2R b * Bn a0 b0 c n)) with
      ( (Xn X0 f f' n + - 2 * Bn a0 b0 c n) * Z2R b - IDR k) by field_IDR;
    assert (Henbn:= leq_en_bn a0 b0 c X0 (-1) 1 f f' A0_b0_c 
      Hlim_f' Hincl  Habs_a Habs_b n);
    assert (Hbn:= Bn_ge_0 a0 b0 c X0 (-1) 1 f f' 
      Hlim_f' Hincl  Habs_a Habs_b n);
    assert ( -1 <=  (Xn X0 f f' n + - 2 * Bn a0 b0 c n) * Z2R b - IDR k <= 1);
      [|intuition];
    apply (make_digit_add (EXn g EX0 n) k s); try assumption; [
      apply sym_eq; assumption

      | fold (c_I (-1) 1 (Xn X0 f f' n + -2 * Bn a0 b0 c n ));
      apply Hlp;
      rewrite rewrite_c_disc;
      split; [unfold Rminus|]; apply Rplus_le_compat_l; [
        fourier
	| apply Rle_trans with R0; fourier]

      |rewrite <-Rabs_le_le_eq;
      rewrite Rabs_mult;
      rewrite (Rabs_right (2 * Z2R b * Z2R b)); [| ZR0compat];
      rewrite Ropp_mult_distr_l_reverse; rewrite Rabs_Ropp;
      rewrite Rabs_right; [| ZR0compat];
      apply Rmult_le_reg_l with (Rpower 2 (INR n - 1));
        [apply Rpow_pos; fourier|];
	
	apply Rle_trans with ( b0 * (2 * Z2R b * Z2R b));[
  	  rewrite <-Rmult_assoc;
	  apply Rmult_le_compat_r; [ZR0compat|];
	  replace (Rpower 2 (INR n - 1) * (2 * Bn a0 b0 c n)) with
	    (2^n *  Bn a0 b0 c n); [
 	      assert (Hxx:= rel_bn2 a0 b0 c X0 (-1) 1  f f' A0_b0_c 
		Hlim_f' Hincl Habs_a Habs_b n);
	      apply Rmult_le_reg_l with ( /2^n);[
	        apply Rinv_0_lt_compat; apply pow2_diff0
	        | replace (/ 2 ^ n * (2 ^ n * Bn a0 b0 c n)) with 
		  (Bn a0 b0 c n) ;[assumption| field; apply Rgt_not_eq; apply pow2_diff0]]
	      | rewrite <-Rmult_assoc;
	      ring_simplify;
	      pattern Rmult at -3 ; rewrite Rmult_comm;
	      rewrite Rmult_comm;
	      rewrite Rmult_assoc;
	      apply Rmult_eq_compat_l;
	      rewrite <-Rpower_pow; [|fourier];
	      pattern 2 at 2; rewrite <-(Rpower_1 2); [|fourier];
	      rewrite <-Rpower_plus;
	      replace (1 + (INR n - 1)) with (INR n); [trivial| ring] ]

	  | pattern Rmult at -4; rewrite Rmult_comm;
	  apply Rge_le; assumption] ]

    | apply Rle_trans with  ( r + 2 * (Z2R b * Bn a0 b0 c n) );[assumption|];
    rewrite Hreq;
    replace (Z2R b * Xn X0 f f' n - IDR k + 2 * (Z2R b * Bn a0 b0 c n)) with
      ( (Xn X0 f f' n +  2 * Bn a0 b0 c n) * Z2R b - IDR k) by field_IDR;
    assert (Henbn:= leq_en_bn a0 b0 c X0 (-1) 1 f f' A0_b0_c 
      Hlim_f' Hincl  Habs_a Habs_b n);
    assert (Hbn:= Bn_ge_0 a0 b0 c X0 (-1) 1 f f' 
      Hlim_f' Hincl  Habs_a Habs_b n);
    assert ( -1 <=  (Xn X0 f f' n + 2 * Bn a0 b0 c n) * Z2R b - IDR k <= 1);
      [|intuition];
    apply (make_digit_add (EXn g EX0 n) k s); try assumption;[
      apply sym_eq; assumption

      | apply Hlp;
      rewrite rewrite_c_disc;
      split; [unfold Rminus|]; apply Rplus_le_compat_l;[
        apply Rle_trans with R0; fourier | fourier]

      | rewrite <-Rabs_le_le_eq;
      rewrite Rabs_mult;
      repeat (rewrite Rabs_right; [| ZR0compat]);
      apply Rmult_le_reg_l with (Rpower 2 (INR n - 1)); 
	[apply Rpow_pos; fourier| ];
      apply Rle_trans with ( b0 * (2 * Z2R b * Z2R b));[
        rewrite <-Rmult_assoc;
	apply Rmult_le_compat_r; [ZR0compat| ];
	replace (Rpower 2 (INR n - 1) * (2 * Bn a0 b0 c n)) with
	  (2^n *  Bn a0 b0 c n);[
	  assert (Hxx:= rel_bn2 a0 b0 c X0 (-1) 1  f f' A0_b0_c 
	    Hlim_f' Hincl Habs_a Habs_b n);
	  apply Rmult_le_reg_l with ( /2^n);[
	    apply Rinv_0_lt_compat; apply pow2_diff0
	    | replace (/ 2 ^ n * (2 ^ n * Bn a0 b0 c n)) with 
	      (Bn a0 b0 c n) ;[assumption| field; apply Rgt_not_eq; apply pow2_diff0]]

	  | rewrite <-Rmult_assoc;
	  ring_simplify;
	  pattern Rmult at -3 ; rewrite Rmult_comm;
	  rewrite Rmult_comm;
	  rewrite Rmult_assoc;
	  apply Rmult_eq_compat_l;
	  rewrite <-Rpower_pow; [|fourier];
	  pattern 2 at 2; rewrite <-(Rpower_1 2); [|fourier];
	  rewrite <-Rpower_plus;
	  replace (1 + (INR n - 1)) with (INR n); [trivial| ring]]


        | pattern Rmult at -4; rewrite Rmult_comm;
	apply Rge_le; assumption] ] ]).

assert (Hrelbn:=rel_bn2 a0 b0 c X0 (-1) 1  f f' 
  A0_b0_c Hlim_f' Hincl Habs_a Habs_b n).

assert (R1 :  xs _ _ _ _ (-1) 1 _ _ A0_b0_c2 pr2 Hder_f2 Hcont_f'2 Hlim_f'2 Hrin
    Hdif_f'2 Habs_a2 Habs_b2 Hb_eps2 =  xs0 * Z2R b - IDR k).
  clear kanto_correct.
abstract (apply kantoro_unic2; [
  subst r;
    unfold c_disc;
      replace (Z2R b * Xn X0 f f' n - IDR k - (xs0 * Z2R b - IDR k))
	with ( Z2R b * (Xn X0 f f' n - xs0)) by field_Z;
	rewrite Rabs_mult; rewrite Rabs_right; [|ZR0compat];
	  rewrite <-Rmult_assoc; rewrite (Rmult_comm 2 );
	    rewrite Rmult_assoc; apply Rmult_le_compat_l; [ZR0compat|];
	      rewrite Rabs_minus_sym;
		apply unic_aux with (-1) 1 pr; try assumption

  | replace ((IDR k + (xs0 * Z2R b - IDR k)) / Z2R b) with xs0 by field_Z];
elim (conv_root  a0 b0 c X0  (-1) 1  f f' A0_b0_c pr
    Hder_f Hcont_f' Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps xs0 Hcv1);
intros; assumption).

assert (U4 : Un_cv
      (Xn r (fun x : R => f ((IDR k + x) / Z2R b))
         (fun x : R => 1 / Z2R b * f' ((IDR k + x) / Z2R b))) ( xs0 * Z2R b - IDR k) /\
    c_disc r (2 * (Z2R b * Bn a0 b0 c n)) ( xs0 * Z2R b - IDR k) /\ 
     f ((IDR k + ( xs0 * Z2R b - IDR k)) / Z2R b) = 0).
  clear kanto_correct.
  abstract (apply (sig_proj _ 
  (fun xs =>  (Un_cv
    (Xn r (fun x : R => f ((IDR k + x) / Z2R b))
      (fun x : R => 1 / Z2R b * f' ((IDR k + x) / Z2R b))) xs /\
    c_disc r (2 * (Z2R b * Bn a0 b0 c n)) xs /\ f ((IDR k + xs) / Z2R b) = 0)) 
  (kantoro_exists_b (Z2R b * An a0 n) (Z2R b * Bn a0 b0 c n) (c/Z2R b^2)
    r (-1) 1 (fun x => f ((IDR k + x)/Z2R b)) (fun x => 1/(Z2R b)*f'((IDR k + x)/Z2R b))
    A0_b0_c2 pr2 Hder_f2 Hcont_f'2 Hlim_f'2 Hrin Hdif_f'2 Habs_a2 Habs_b2 Hb_eps2));
  rewrite <-R1; unfold xs; reflexivity).

destruct U4 as [U4 [U5 U6]].

apply (kanto_correct 
(* new g   *) (fun x : Stream digit => multb (g (k :: x))) 
(* new EX0 *) s
(* new f   *) (fun x => f ((IDR k + x)/Z2R b))
(* new f'  *) (fun x => 1/(Z2R b)*f'((IDR k + x)/Z2R b))
(* new X0  *) r
(* new a0  *) (Z2R b * An a0 n) 
(* new b0  *) (Z2R b * Bn a0 b0 c n)
(* new c  *) (c/Z2R b^2)) with pr2; clear  kanto_correct; try assumption.

abstract (apply Rle_ge;
  apply Rle_trans with (b0 * (2 * Z2R b * Z2R b) );
    [|fourier];
  apply Rmult_le_compat_r;[ZR0compat|];
  apply Rle_trans with (Z2R b* (/2^n*b0)); 
    [ apply Rmult_le_compat_l; [ZR0compat| apply (rel_bn2 a0 b0 c X0 (-1) 1  f f' 
    A0_b0_c Hlim_f' Hincl Habs_a Habs_b n)] |];
  rewrite <-Rmult_assoc; pattern b0 at 2; replace b0 with (1*b0) by ring;
  apply Rmult_le_compat_r; [apply (H_b0 b0 X0 f f' Habs_b)| ];
  apply Rmult_le_reg_l with (2^n); [apply pow2_diff0|ring_simplify];
  rewrite Rinv_r_simpl_m; [assumption|apply Rgt_not_eq; apply pow2_diff0]).

(intros x vx Hxn_in_disc Hxvx Hd0 Hin1;
elim (RIneq.Rmult_neq_0_reg _ _ Hd0); intros Hvf1 Hvf2;
assert (Heq2:
  (f ((IDR k + vx) / Z2R b) / (1 / Z2R b * f' ((IDR k + vx) / Z2R b)))
  = (Z2R b *(f ((IDR k + vx) / Z2R b) / ( f' ((IDR k + vx) / Z2R b)))) ) 
   by (  field_Z; assumption);
rewrite Heq2;
apply multb_correct; [|rewrite  <- Heq2; assumption];
apply Hgeq).
   constructor; try assumption;
    apply (Srepresents_interval x); assumption.
assert (Hdpos:= Hdif_pos  a0 b0 c X0 (-1) 1  f f' A0_b0_c  pr
    Hder_f Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps n 0%nat).
simpl in Hdpos;
rewrite <-H0 in Hdpos;
rewrite Rabs_minus_sym in Hdpos;
fold (c_disc X0 (2 * b0 - 2 * b0 / 2 ^ n) ((IDR k + r) / Z2R b)) in Hdpos.
(*rewrite c_disc_c in Hxvx.*)
rewrite rewrite_c_disc in Hxvx, Hdpos |-*.

elim Hxvx; elim Hdpos; intros He1 He2 He3 He4; split.
apply Rle_trans with ((IDR k +( r - 2 * (Z2R b * Bn a0 b0 c n) )) / Z2R b).
replace ((IDR k + (r - 2 * (Z2R b * Bn a0 b0 c n))) / Z2R b) with
  ((IDR k + r) / Z2R b - 2 * Bn a0 b0 c n) by field_Z.
apply Rplus_le_reg_l with (2 * Bn a0 b0 c n); ring_simplify; 
 unfold Rminus ; rewrite Rplus_assoc; rewrite Rplus_comm.
apply Rle_trans with (X0 - (2 * b0 - 2 * b0 / 2 ^ n)); [|assumption].
unfold Rminus; rewrite Rplus_assoc;  apply Rplus_le_compat_l; ring_simplify.
apply  Rplus_le_compat_l.
unfold Rdiv; rewrite Rmult_assoc; apply Rmult_le_compat_l; try fourier.
fold (Rdiv b0 (2^n) ).
apply Bn_desc_2 with X0 (-1) 1 f f'; try assumption.
unfold Rdiv; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat ; ZR0compat.
apply Rplus_le_compat_l; assumption.
apply Rle_trans with ((IDR k +( r + 2 * (Z2R b * Bn a0 b0 c n) )) / Z2R b).
unfold Rdiv; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat ; ZR0compat.
apply Rplus_le_compat_l; assumption.
replace ((IDR k + (r + 2 * (Z2R b * Bn a0 b0 c n))) / Z2R b) with
  ((IDR k + r) / Z2R b + 2 * Bn a0 b0 c n) by field_Z.
apply Rplus_le_reg_l with (-(2 * Bn a0 b0 c n)); ring_simplify; 
 unfold Rminus ; rewrite Rplus_assoc; rewrite (Rplus_comm (-2*Bn a0 b0 c n)).
apply Rle_trans with (X0 + (2 * b0 - 2 * b0 / 2 ^ n)); [assumption|].
unfold Rminus; rewrite Rplus_assoc;  apply Rplus_le_compat_l; ring_simplify.
unfold Rminus; apply  Rplus_le_compat_l.
apply Ropp_le_contravar; unfold Rdiv; 
  rewrite Rmult_assoc; apply Rmult_le_compat_l; try fourier.
fold (Rdiv b0 (2^n) ).
apply Bn_desc_2 with X0 (-1) 1 f f'; try assumption.

assumption.

rabs_intros;
rewrite Heq2 in Hin1. 
rewrite Rabs_mult in Hin1;
rewrite Rabs_right in Hin1; [|ZR0compat].
apply Rmult_le_reg_l with ((Z2R b)); [ZR0compat|].
apply Rle_trans with 1; [assumption|];
ring_simplify; R_to_Z.
Qed.


Theorem kanto_correct' : 
  forall g EX0 f f' X0 a0 b0 c n
    (A0_b0_c: 2*a0*b0*c <=1 )
    (pr: forall t:R, c_I (-1) 1 t -> derivable_pt f t)
    (Hder_f: forall (t:R)(H: c_I (-1) 1 t), derive_pt f t (pr t H)=f' t) 
    (Hcont_f':forall y, (c_I (-1) 1 y) -> continuity_pt f' y) 
    (Hlim_f':forall y1 y2:R, c_I (-1) 1 y1 -> c_I (-1) 1 y2 -> 
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2))) 
    (Hincl: -1 < X0 < 1) 
    (Hdif_f':f' X0 <> 0 ) 
    (Habs_a: Rabs (/(f' X0)) <= a0) 
    (Habs_b:Rabs (f X0 / (f' X0)) <= b0 ) 
    ( Hb_eps: included (c_disc X0 (2*b0)) (c_I (-1) 1)),
    (Z2R b - 2) * Rpower 2 (INR n-1) >= b0 * (2* Z2R b * Z2R b) ->  
    (Z2R  b)  <=    2^n ->
    (forall x vx, Srepresents x vx ->(c_disc X0 (2*b0) vx) ->
 f' vx <>0 -> -1 <= f vx / f' vx <= 1 ->
      Srepresents (g x) (f vx/ f' vx)) -> 
    Srepresents EX0 X0 ->
    Srepresents (exact_newton g EX0 n) (xs a0 b0 c X0 (-1) 1 f f' A0_b0_c  pr 
      Hder_f Hcont_f' Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps).
Proof.
intros g EX0 f f' X0 a0 b0 c n A0_b0_c  pr 
  Hder_f Hcont_f'   Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps
  Hspeed Hnb  Hgeq HrepX0 .

set (xs0 := (xs a0 b0 c X0 (-1) 1 f f' A0_b0_c pr Hder_f Hcont_f' Hlim_f' Hincl
        Hdif_f' Habs_a Habs_b Hb_eps)).
assert (U1 : Un_cv (Xn X0 f f') xs0 /\ c_disc X0 (2 * b0) xs0 /\ f xs0 = 0).
apply (sig_proj _  
  (fun xs1 =>  Un_cv (Xn X0 f f') xs1 /\ c_disc X0 (2 * b0) xs1 /\ f xs1 = 0)
  (kantoro_exists_b a0 b0 c X0 (-1) 1 f f'
    A0_b0_c pr Hder_f Hcont_f' Hlim_f' Hincl Hdif_f' Habs_a Habs_b Hb_eps)).
subst xs0; unfold xs; reflexivity.
destruct U1 as [U1 [U2 U3]].
apply (kanto_correct g EX0 f f' X0 a0 b0 c n 
    (xs a0 b0 c X0 (-1) 1 f f' A0_b0_c pr Hder_f Hcont_f' Hlim_f' Hincl
        Hdif_f' Habs_a Habs_b Hb_eps) pr); try assumption.
Qed.


End NEWTON.


(*Application to the square root*)

Module SQRT(Nums : NUMS).

Module Newton_correct := NEWTON(Nums).
Export Newton_correct.

Section sqrt_correct.
(* Square root*)
Variable vy : R.
Variable  n : nat.
Variable y EX0: Stream digit.
Variable X0 : R.

Definition b0 := Rabs (((Z2R b ^2) * X0^2 - vy )/ (2*(Z2R b ^2) * X0)).
Definition g_sqrt y x := (x/ (1+1)%D - y * ((Sinverse (2) x) / (1+1)%D))%S.
Definition f_sqrt y x := (Z2R b ) ^2 * x ^ 2 - y.
Definition f'_sqrt x := 2*(Z2R b)^2 * x.

Hypothesis Hin_disc:  included (c_disc X0 (2*b0)) (c_I (/Z2R b^2) 1).
Hypothesis H_X0: /Z2R b <= X0 < 1.                                                                                                                                                                           
Hypothesis Hn: Z2R b <= 2 ^ n.
Hypothesis Hn2: (Z2R b - 2) * Rpower 2 (INR n-1) >= b0 * (2* Z2R b * Z2R b).
Hypothesis Hny:  / (Z2R b)^2 <= vy <= 1.
Hypothesis Hrepy: Srepresents y vy.
Hypothesis HrepX0: Srepresents EX0 X0.


Theorem g_f_f'2:
  forall x vx,
    0 <= vx ->
    Srepresents x vx ->
    (c_disc X0 (2*b0) vx) ->
    f'_sqrt vx <> 0 -> 
    -1 <= f_sqrt vy vx / f'_sqrt vx <= 1 ->
(* 1 <= vx * Z2R b ^ 2 \/ vx * Z2R b ^ 2 <= -1 ->*)
    Srepresents (g_sqrt y x) ((f_sqrt vy vx)/ (f'_sqrt vx)).
Proof.
intros x vx Hvx Hrepx Hdisc Hdif Hineq.
 assert (Hbpos := base_positive). 
unfold g_sqrt, f_sqrt, f'_sqrt.


assert (Hone: 1 <= vx * Z2R b ^ 2).
assert (Hin:= Hin_disc vx Hdisc).
elim Hin; intros I1 I2.
apply Rmult_le_reg_l with (/(Z2R b^2)).
apply Rinv_0_lt_compat; simpl pow; ZR0compat.
replace (/ Z2R b ^ 2 * (vx * Z2R b ^ 2)) with vx;
ring_simplify; [assumption|field; ZR0compat].

replace (((Z2R b) ^2* vx ^ 2 - vy) / (2 * (Z2R b) ^2* vx)) with 
  (vx / IDR 2%D - vy * ((1 / ( (Z2R b) ^2* vx))/ IDR 2)%D).

apply Ssub_correct.
apply div_digit_correct.
simpl_IDZ; omega.
simpl_IDZ; omega.
assumption.
apply Smult_correct.
assumption.
apply div_digit_correct.
simpl_IDZ; omega.
simpl_IDZ; omega.

apply Sinverse_correct.
assumption.
left; assumption.

assert (U1 :=  Srepresents_interval _ _ Hrepx).
assert (U2 :=  Srepresents_interval _ _ Hrepy).
simpl_IDR.
rabs_intros.
pattern 1%R at 6; replace 1%R with (1/2+(1*1)/2) by field.
unfold Rminus; apply Rabs_add.
  rewrite Rabs_div; [| ZR0compat]. 
  rewrite (Rabs_right 2); [|ZR0compat].
  apply Rdiv_le_compat_r; ZR0compat.

  rewrite Rabs_Ropp.
  replace (vy * (1 / (Z2R b ^ 2 * vx) / 2)) with 
    ((vy * (1 / (Z2R b ^ 2 * vx))) / 2) by (unfold Rdiv; ring).
  rewrite Rabs_div; [| ZR0compat]. 
  rewrite (Rabs_right 2); [|ZR0compat].
  apply Rdiv_le_compat_r; [ZR0compat |].
  rewrite Rabs_mult.
  apply Rmult_le_compat; assumption || ZR0compat.
  rabs_elim; rewrite Rmult_comm.
  apply Sinverse_tech_lemma.
left; assumption.

field_IDR.
ZR0compat.
apply Rlt_le_trans with (1/Z2R b^2); [ZR0compat|].
apply Rle_div_r; ZR0compat.
Qed.

Definition Ssqrt (x : Stream digit) := 
  exact_newton (g_sqrt x) (EX0)%S n%nat.

Lemma pr_f_sqrt vx:
forall x, c_I (-1) 1 x -> derivable_pt (f_sqrt vx) x.
Proof.
intros; unfold f_sqrt; reg.
Qed.


Theorem sqrt_correct : 
  forall vx, 0 <= vx ->
    vx^2 =  vy ->
    Srepresents (Ssqrt y) (vx/Z2R b).
Proof.
assert (Hbpos:= base_positive).
intros vx Hvxpos Heq.
unfold Ssqrt.
assert (Hdif: 0 < X0 <1).
split; elim H_X0; intros; [apply Rlt_le_trans with (/Z2R b); [ZR0compat|trivial]|trivial].
assert (Hf': f'_sqrt X0 > 0).
unfold f'_sqrt;  ZR0compat; dec_tac; assumption.
assert (H0: forall (t : R) (H : c_I (-1) 1 t),
derive_pt (f_sqrt vy) t (pr_f_sqrt vy t H) = f'_sqrt t).
intros.
unfold f_sqrt; unfold pow; unfold f'_sqrt; reg.
assert (H1:forall y0 : R, c_I (-1) 1 y0 -> continuity_pt f'_sqrt y0).
intros; unfold f'_sqrt; reg.
assert (H2: forall y1 y2 : R,
c_I (-1) 1 y1 ->
c_I (-1) 1 y2 ->
Rabs (f'_sqrt y1 - f'_sqrt y2) <= 2 * Z2R b ^ 2 * Rabs (y1 - y2)).
intros; unfold f'_sqrt.
replace (2 * Z2R b ^ 2 * y1 - 2 * Z2R b ^ 2 * y2) with (2 * Z2R b ^ 2 * (y1 -  y2)) by ring.
rewrite Rabs_mult.
apply Rmult_le_compat_r; [apply Rabs_pos|rewrite Rabs_right; [right;trivial|ZR0compat]].
assert (H3:2 * / (2 * Z2R b ^ 2 * X0) * b0 * (2 * Z2R b ^ 2) <= 1).
replace (2 * ( / (2 * Z2R b ^ 2 * X0)) * b0 * (2 * Z2R b ^ 2) ) with
 (2 * (1 / X0) * b0  ); [| field; split; dec_tac;  ZR0compat].
unfold b0.
 rewrite Rabs_div; [|dec_tac;ZR0compat].
pattern Rabs at 1; rewrite Rabs_right; [|dec_tac; ZR0compat].
replace (2 * (1 / X0) * (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy) / (2 * Z2R b ^ 2 * X0)) ) with
  ( (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy) / ( Z2R b ^ 2 * X0^2)) ); 
   [|field; split; dec_tac;ZR0compat].
apply Rmult_le_reg_l with ((Z2R b ^ 2 * X0 ^ 2) );[dec_tac; ZR0compat|].
 replace (Z2R b ^ 2 * X0 ^ 2 * (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy) / (Z2R b ^ 2 * X0 ^ 2)))
 with (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy)); [|field_Z; dec_tac; apply Rgt_not_eq; trivial].
rewrite Rabs_right.
rewrite Rmult_1_r; 
 pattern (Z2R b ^ 2 * X0 ^ 2 ) at 2; rewrite <-(Rplus_0_r (Z2R b ^ 2 * X0 ^ 2 )). 
unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_0; apply Rle_trans 
  with (/Z2R b^2); 
[left; ZR0compat|dec_tac; assumption].
apply Rle_ge; apply Rle_trans with (1-vy); [dec_tac; fourier|].
unfold Rminus; apply Rplus_le_compat_r.
replace 1 with (Z2R b ^ 2 * (/Z2R b) ^ 2) by field_Z.
apply Rmult_le_compat_l; [ZR0compat|].
apply pow_incr; split; [left; ZR0compat|elim H_X0; intros; assumption].
(assert (H4: -1 < X0 < 1)).
dec_tac; split;  fourier.
assert (H5: f'_sqrt X0 <> 0).
apply Rgt_not_eq; assumption.
assert (H6: Rabs (/ f'_sqrt X0) <= / (2 * Z2R b ^ 2 * X0)).
rewrite Rabs_Rinv; [|apply Rgt_not_eq; assumption]; rewrite Rabs_right;
  [|left;assumption].
unfold f'_sqrt; right; trivial.
assert (H7: Rabs (f_sqrt vy X0 / f'_sqrt X0) <= b0).
unfold b0; unfold f_sqrt; unfold f'_sqrt; right; trivial.
assert (H8: included (c_disc X0 (2 * b0)) (c_I (-1) 1)).
apply included_trans with (c_I (/ Z2R b ^ 2) 1); [trivial|].
intros x U; elim U; intros U1 U2; split; [|trivial].
apply Rle_trans with 0; [fourier|apply Rle_trans with (/Z2R b^2); 
  [left|];ZR0compat].
assert (H9: forall (x : Stream digit) (vx0 : R),
Srepresents x vx0 ->
c_disc X0 (2 * b0) vx0 ->
f'_sqrt vx0 <> 0 ->
-1 <= f_sqrt vy vx0 / f'_sqrt vx0 <= 1 ->
Srepresents (g_sqrt y x) (f_sqrt vy vx0 / f'_sqrt vx0)). 
intros x vx0 Hq0 Hd Hq1 Hq3; apply g_f_f'2; try assumption.
assert (Hd1:= Hin_disc vx0 Hd).
elim Hd1; intros U I.
apply Rle_trans with (/Z2R b^2 ); [left; ZR0compat|assumption].

apply kanto_correct  with (f_sqrt vy ) (f'_sqrt) X0 (/(2*Z2R b^2 * X0)) 
  b0 (2*Z2R b^2)  (pr_f_sqrt vy); try assumption.

elim ( kantoro_exists_b _ _ _ X0 _ _ _ _ H3 (pr_f_sqrt vy) H0 H1 H2 H4 
  H5 H6 H7 H8  ); intros xs [Hxscv [Hxsd Hxs0]].
assert (root: f_sqrt vy (vx / Z2R b) = 0).
unfold f_sqrt; rewrite <-Heq; try field_Z .
unfold f_sqrt in Hxs0, root.

replace (Z2R b ^ 2 * xs ^ 2 ) with (Rsqr (Z2R b * xs) ) in Hxs0;[|unfold Rsqr; field].
replace (Z2R b ^ 2 * (vx / Z2R b) ^ 2 ) with (Rsqr (Z2R b * (vx / Z2R b)) )
 in root;[|unfold Rsqr;  field_Z].
assert (haux: Rsqr (Z2R b * xs) = vy ).
apply (Rplus_eq_reg_l (-vy)); ring_simplify; rewrite Rplus_comm; assumption.
assert (au: Rsqr (Z2R b * (vx / Z2R b)) = vy).
apply (Rplus_eq_reg_l (-vy)); ring_simplify; rewrite Rplus_comm; assumption.
assert (E:xs = vx/Z2R b).
apply Rsqr_inj.
  elim (Hin_disc xs).
  intros Z1 Z2.
  apply Rle_trans with (/Z2R b^2); [left;ZR0compat|assumption].
  assumption.
  ZR0compat.
  apply Rmult_eq_reg_l with (Rsqr (Z2R b)).
  repeat rewrite <-Rsqr_mult.
  rewrite haux; rewrite au; trivial.
  apply Rgt_not_eq; apply Rlt_gt. 
  apply Rsqr_pos_lt;  ZR0compat.

rewrite <- E; assumption.
Qed.


Theorem sqrt_correct' : 
    Srepresents (Ssqrt y) ((sqrt vy)/(Z2R b)).
Proof.
assert (Hbpos:= base_positive).
assert (Hy0: 0 <= vy).
elim Hny;  intros Z1 Z2.
 apply Rle_trans with (/Z2R b^2); [left;ZR0compat|assumption].
assert (H2: (sqrt vy) ^ 2 = vy).
unfold  pow.
rewrite Rmult_1_r.
 apply sqrt_sqrt; try assumption.
exact ( sqrt_correct (sqrt vy) (sqrt_positivity vy Hy0) H2 ).
Qed.


End sqrt_correct.


End SQRT.