Require Export pi_correct.
Require Import series_utils.

Module INV_CORRECT(Nums : NUMS).

Module Pi_correct := PI_CORRECT(Nums).
Export Pi_correct.


Theorem pow_rabs_1 : 
  forall (x : R) (n : nat), Rabs x <= 1 -> Rabs (x ^ n) <= 1.
Proof.
intros x n Hx.
induction n; simpl.
rewrite Rabs_right; ZR0compat.

replace 1 with (1*1) by ring.
rewrite Rabs_mult.
apply Rmult_le_compat; ZR0compat.
Qed.

Theorem Rmult_pow : 
  forall x y n, 
    x ^n * y ^n = (x*y)^n.
Proof.
intros x y n; induction n; simpl.
ring.
rewrite <- IHn; ring.
Qed.

Theorem Rabs_neq: 
  forall x y : R, 0 < y -> Rabs x <> y -> x <> y.
intros x y Hypos.
elim (Rle_or_lt 0 x); intros G1.
rewrite Rabs_right; [tauto| ZR0compat].
intros _.
apply Rlt_not_eq.
apply Rlt_trans with 0; assumption.
Qed.

Theorem Rabs_minus:
  forall w x y z : R, Rabs w <= y -> Rabs x <= z -> Rabs (w - x) <= y + z.
Proof.
intros w x y z Hwy Hxz.
unfold Rminus.
apply Rabs_add.
assumption.
rewrite Rabs_Ropp; assumption.
Qed.

Theorem Rneq_le_lt : 
  forall x y, 
    x <> y -> x <= y -> x < y.
Proof.
intros x y Hneq Hle.
elim Hle; intros G1.
assumption.
absurd (x=y); assumption.
Qed.

Theorem inv_step_correct : 
  forall x vx y y' k m, 
    (2 <= IBZ k)%Z ->
    0 <= IBR y ->
    IBR y' = IBR y * 2* Z2R b * Z2R b ->
    Rabs (vx * IBR k) <= 1 -> 
    IBR y * Rabs vx^m * (2* Z2R b) <=  (Z2R b -4)* IBR k * (1 - vx)  ->
    Srepresents x vx -> 
    forall p z xn sx n,
      IBR y * Rabs vx^(n+p) * (2* Z2R b * Z2R b) <= (Z2R b -4) *IBR k * (1 - vx)  ->
      IBR z = IBR k ^ n * (Z2R b - 4) ->
      (m <= n)%nat ->
      Srepresents xn (IBR y * vx ^n / IBR k) -> 
      Srepresents sx (IBR y * (vx^m - vx ^ n) / (IBR k * (1-vx))) ->
      let (p0, sx') := inv_step y' z k x xn sx p in
        let (z', xn') := p0 in
          exists n', 
            (n <= n')%nat /\
            IBR z' = IBR k ^ n' * (Z2R b - 4) /\
            IBR y * Rabs vx^n' * (2* Z2R b * Z2R b) <=  (Z2R b -4)  * IBR k * (1 - vx) /\
          Srepresents xn' (IBR y * vx ^n' / IBR k) /\
          Srepresents sx' (IBR y * (vx^m - vx ^ n') / ( IBR k * (1-vx))).
Proof.           
assert (Hbpos:=base_positive).
intros x vx y y' k m Hkpos Hypos Hy' Hvxk Hboundsm Hrepx p; induction p; simpl;
intros z xn sx n Hboundsp Hz Hlemn Hrepxn Hrepsx.

exists n.
replace (n+0)%nat with n in Hboundsp by ring.
intuition.

assert (U0 : 2 <= IBR k) by IBR_to_Z.

simpl_IBZ; Zbool G1.
exists n.
split;[ omega|].
split; [assumption|].
split; [| intuition].
apply Rle_trans with ((Rabs vx ^ n * IBR k ^n) *  ((Z2R b - 4) * IBR k * (1 - vx) )).
  replace (IBR y * Rabs vx ^ n * (2 * Z2R b * Z2R b)) with 
    (Rabs vx ^ n * (IBR y *  2 * Z2R b * Z2R b)) by ring.
  replace ((Rabs vx ^ n * IBR k ^n) *  ( (Z2R b - 4) *IBR k * (1 - vx) )) with 
    (Rabs vx ^ n * (IBR k ^n *  (IBR k * (1 - vx) * (Z2R b - 4)))) by ring.
  apply Rmult_le_compat_l; [ZR0compat|].
  apply Rle_trans with (IBR z).
  rewrite <- Hy'; IBR_to_Z; assumption.
  rewrite Hz.
  apply Rmult_le_compat_l; [ZR0compat|].
  pattern (Z2R b - 4) at 1; replace (Z2R b - 4) with (1*(Z2R b - 4)) by ring.
  apply Rmult_le_compat_r;[R_to_Z|].
  pattern 1 at 1; replace 1 with (2+ -1) by ring.
  replace (IBR k * (1 - vx)) with (IBR k + - (vx * IBR k)) by ring.
  rabs_elim; dec_tac; fourier.

  pattern ( (Z2R b - 4) * IBR k * (1 - vx) ) at 2; replace 
     ((Z2R b - 4)* IBR k * (1 - vx) ) with (1* ((Z2R b - 4)* IBR k * (1 - vx))) by ring.
  apply Rmult_le_compat_r; [R_to_Z|].
  ZR0compat; [ R_to_Z|].
  assert (U2 := Srepresents_interval _ _ Hrepx).
  dec_tac; fourier.
  replace ( Rabs vx ^ n * IBR k ^ n) with (Rabs ((vx  * IBR k) ^ n)).
  apply pow_rabs_1; assumption.
  
  rewrite <- RPow_abs.
  rewrite Rabs_mult.
  rewrite (Rabs_right (IBR k)); [| ZR0compat].
  rewrite Rmult_pow; reflexivity.


assert (Hboundsp2 : IBR y * Rabs vx ^ (S n + p) * (2 * Z2R b * Z2R b) <=
            (Z2R b - 4)*IBR k * (1 - vx)).
  replace (S n+p)%nat with (n +S p)%nat by omega; assumption.

assert (Hz2: (IBR (z * k)%B = IBR k ^ S n * (Z2R b - 4))). 
  simpl; simpl_IBR; fold_IBR; rewrite Hz; ring.

assert (Hlemn2 : (m <= S n)%nat) by omega.

assert (Hrepxn2 : Srepresents (xn * x)%S (IBR y * vx ^ S n / IBR k)).
  simpl.
  replace (IBR y * (vx * vx ^ n) / IBR k) with 
    ((IBR y * (vx ^ n) / IBR k)* vx) by (unfold Rdiv; ring).
  apply Smult_correct; assumption.

assert (U1 : vx <> 1).
  apply Rabs_neq;[auto with real|].
  apply Rlt_not_eq.
  apply Rle_lt_trans with (1/IBR k).
  apply Rle_div_l; [IBR_to_Z|].
  apply Rle_trans with (2:= Hvxk); right.
  rewrite Rabs_mult; rewrite (Rabs_right (IBR k)); 
    [reflexivity| ZR0compat; IBR_to_Z].
  apply Rlt_div_r; IDR_to_Z.

destruct (Srepresents_interval _ _ Hrepx) as [U2 U3].

assert (U4 : IBR k * (1-vx) <> 0 ).
  apply prod_neq_R0; [ZR0compat|].
  auto with real.

assert (U5: 0 < 1 - vx).
  assert (vx < 1); [| fourier].
  apply Rneq_le_lt; assumption.

assert (Hrepsx2 : Srepresents (xn + sx)%S
  (IBR y * (vx^m - vx ^ (S n)) / (IBR k * (1 - vx)))).
  clear IHp.
  simpl.
  replace (IBR y * (vx^m - vx * vx ^ n) / (IBR k * (1 - vx))) with 
    (IBR y * vx ^ n / IBR k + IBR y * (vx^m - vx ^ n ) / (IBR k * (1 - vx))) 
    by (  field; split; ZR0compat).
  apply Sadd_correct; try assumption.

  rabs_intros.
  
  replace ( IBR y * vx ^ n / IBR k + 
    IBR y * (vx ^ m  - vx ^ n) / (IBR k *(1 - vx)))  with 
  (IBR y * vx ^ m * (1-vx ^(n+1-m)) / (IBR k * (1 - vx))).
  
  rewrite Rabs_div;[| assumption].
  do 3 rewrite Rabs_mult.
  
  apply Rle_trans with  (Rabs (IBR y) * Rabs (vx ^ m) * 2 /
    (Rabs (IBR k) * Rabs (1 - vx))).
  apply Rdiv_le_compat_r; [ZR0compat|].
  apply Rmult_le_compat_l; [ZR0compat|].
  apply Rabs_minus.
  rewrite Rabs_right; ZR0compat.
  apply pow_rabs_1.
  rabs_elim; intuition.
  
  rewrite (Rabs_right (IBR k)); [|ZR0compat].
  rewrite (Rabs_right (IBR y)); [|ZR0compat; IBR_to_Z].
  rewrite (Rabs_right (1-vx)); [|ZR0compat].
  rewrite <- RPow_abs.
  
  apply Rle_trans with ((Z2R b - 4)/ Z2R b).
  apply Rdiv_mult_compat.
  ZR0compat.
  ZR0compat.
  rewrite Rmult_assoc.
  rewrite <- (Rmult_assoc (Z2R b - 4)).
  assumption.
  apply Rle_div_r; [ZR0compat | R_to_Z].
  
  replace (IBR y * vx ^ m * (1 - vx ^ (n + 1 - m))) with 
    (IBR y * (vx ^m - vx ^(S n))).
  replace (vx ^ m * (1 - vx ^ (n - m))) with 
    (vx^m - vx^n).
  simpl; field.
  split; [auto with real|ZR0compat].
  pattern n at 1; replace n with (m + (n-m))%nat by omega.
  rewrite pow_add; ring.
  replace (S n)%nat with (m+ (n+1-m))%nat by omega.
  rewrite pow_add; ring.
  
generalize (IHp (z * k)%B (xn * x)%S (xn + sx)%S (S n) 
  Hboundsp2 Hz2 Hlemn2 Hrepxn2 Hrepsx2).

let_simpl; intros aux sx' U6.
let_simpl; intros z' xn' U7; subst aux.
intros [n' U8].

exists n'.
split; [omega| intuition].
Qed.

Theorem inv_series_correct : 
  forall x vx k nmax, 
    Srepresents x vx ->
    (2 <= IBZ k)%Z ->
    Rabs (vx * IBR k) <= 1 -> 
    Z2R b * Rabs vx^nmax <= 1 -> 
    forall y y' n r vr z xn, 
      IBR y * Rabs vx ^ n * (2 * Z2R b) <= (Z2R b - 4) * IBR k * (1 - vx) ->
      IBR z = IBR k ^ n * (Z2R b - 4) -> 
      Srepresents r vr ->
      Srepresents xn (IBR y * vx^n / IBR k) ->
      - (Z2R b + 2) <= 2 * Z2R b * vr <= Z2R b + 2 ->
      0 <= IBR y ->
      IBR y' = IBR y * 2 * Z2R b *Z2R b ->
      Srepresents (inv_series y'  z k nmax x xn r) (vr + IBR y*vx^n / (IBR k*(1-vx))).
Proof.
assert (Hbpos:= base_positive).
intros x vx k nmax  Hrepx Hkpos Hvxk Hnmax.

cofix.
intros y y' n r vr z xn Hnbounds Hz Hrepr Hrepxn Hrbounds Hypos Hy'.
str_unfold (inv_series y' z k nmax x xn r).
simpl.
fold_decomp.

assert (G1 : vx <> 1).
  clear inv_series_correct;
  abstract(apply Rabs_neq;[auto with real|];
  apply Rlt_not_eq;
  apply Rle_lt_trans with (1/IBR k); 
    [|   apply Rlt_div_r; IBR_to_Z];
  apply Rle_div_l; [IBR_to_Z|];
  apply Rle_trans with (2:= Hvxk); right;
  rewrite Rabs_mult; rewrite (Rabs_right (IBR k)); 
    [reflexivity| ZR0compat; IBR_to_Z]).
  
destruct (Srepresents_interval _ _ Hrepx) as [G2 G3].

assert (G4 : IBR k * (1-vx) <> 0 ).
  clear inv_series_correct;
  abstract(apply prod_neq_R0; [IBR_to_Z| auto with real]).

assert (G5: 0 < 1 - vx).
  clear inv_series_correct;
  abstract(assert (vx < 1); [| fourier];
  apply Rneq_le_lt; assumption).

assert (U1 : (n<=n)%nat) by omega.

assert (HrepSzero: Srepresents Szero  (IBR y * (vx ^ n - vx ^ n) / (IBR k * (1 - vx)))).
  clear inv_series_correct;
  abstract(const_rep; unfold Rdiv; ring).

assert (Hnmaxbounds: IBR y * Rabs vx^(n+nmax) * (2* Z2R b * Z2R b) 
  <= (Z2R b -4) *IBR k * (1 - vx)).
  clear inv_series_correct;
  abstract( apply Rle_trans with (2:= Hnbounds);
  replace (IBR y * Rabs vx ^ (n + nmax) * (2 * Z2R b * Z2R b)) with 
    (IBR y * Rabs vx ^ n * (Z2R b * Rabs vx ^ nmax * (2  * Z2R b)))
    by (rewrite pow_add; ring);
  apply Rmult_le_compat_l; [ZR0compat|];
  pattern (2 * Z2R b) at 2; replace (2 * Z2R b) with (1*(2 * Z2R b)) by ring;
  apply Rmult_le_compat_r; ZR0compat).

generalize (inv_step_correct x vx y y' k n Hkpos Hypos Hy' Hvxk Hnbounds Hrepx
  nmax z xn Szero n Hnmaxbounds Hz U1 Hrepxn HrepSzero).

let_simpl; intros aux sx' U2.
let_simpl; intros z' xn' U3.
intros [n' [U4 [U5 [U6 [U7 U8]]]]].

replace (vr + IBR y * vx ^ n / (IBR k * (1 - vx)) ) with 
  (vr + (IBR y * (vx ^ n - vx ^ n') / (IBR k * (1 - vx))) + 
    IBR y * vx ^ n' / (IBR k * (1 - vx))) by (unfold Rdiv; ring).

assert (- (Z2R b * Z2R b - 3 * Z2R b - 2) <=
   2 * Z2R b * Z2R b * (IBR y * (vx ^ n - vx ^ n') / (IBR k * (1 - vx))) <=
   Z2R b * Z2R b - 3 * Z2R b - 2).
  clear inv_series_correct;
  abstract(rabs_intros;
  rewrite Rabs_mult;
  rewrite Rabs_right; [| ZR0compat];
  apply splitted_series_bound_rabs with (IBR y *  vx ^ n'/ ( IBR k * (1 - vx)));[
    replace (IBR y * (vx ^ n - vx ^ n') / (IBR k * (1 - vx)) +
      IBR y * vx ^ n' / (IBR k * (1 - vx))) with 
    (IBR y * vx ^ n / (IBR k * (1 - vx))) by (unfold Rdiv; ring);
    replace (2 * Z2R b * Rabs (IBR y * vx ^ n / (IBR k * (1 - vx)))) with 
      (IBR y * Rabs vx ^ n * (2 * Z2R b) / (IBR k * (1 - vx)));[
      apply Rle_div_r;[ZR0compat; IBR_to_Z|];
      apply Rle_trans with (1:=Hnbounds); right; ring|];
      rewrite Rabs_div;[|ZR0compat; IBR_to_Z];
    repeat rewrite Rabs_mult;
    rewrite RPow_abs;
    rewrite (Rabs_right (IBR k)); [| ZR0compat; IBR_to_Z];
    rewrite (Rabs_right (IBR y)); [| ZR0compat; IBR_to_Z];
    rewrite (Rabs_right (1-vx)); [unfold Rdiv; ring| ZR0compat]|];
    
    replace (2 * Z2R b * Z2R b * Rabs (IBR y * vx ^ n' / (IBR k * (1 - vx)))) with 
      (IBR y * Rabs vx ^ n' * (2 * Z2R b * Z2R b) / (IBR k * (1 - vx)));[
      apply Rle_trans with (Z2R b -4); [| ZR0compat];
      apply Rle_div_r;[ZR0compat; IBR_to_Z|];
      apply Rle_trans with (1:=U6); right; ring|];
    rewrite Rabs_div;[|ZR0compat; IBR_to_Z];
    repeat rewrite Rabs_mult;
    rewrite RPow_abs;
    rewrite (Rabs_right (IBR k)); [| ZR0compat; IBR_to_Z];
    rewrite (Rabs_right (IBR y)); [| ZR0compat; IBR_to_Z];
    rewrite (Rabs_right (1-vx)); [unfold Rdiv; ring| ZR0compat]).

md_tac2 d r' vr' U10 U11 inv_series_correct.
2:ZR0compat.
destruct U11 as [U11 [U12 U13]].

assert (U14: -1 <= Z2R b * (IBR y * vx ^ n' / IBR k) <= 1).
  clear inv_series_correct;
  abstract(rabs_intros;
  rewrite Rabs_mult;
  rewrite Rabs_div;[|ZR0compat; IBR_to_Z];
  rewrite Rabs_mult;
  rewrite <- RPow_abs;
  rewrite (Rabs_right (Z2R b)); [| ZR0compat];
  rewrite (Rabs_right (IBR y)); [| ZR0compat; IBR_to_Z];
  rewrite (Rabs_right (IBR k)); [| ZR0compat; IBR_to_Z];
  apply Rmult_le_reg_l with (2*Z2R b); [ZR0compat|];
  replace (2 * Z2R b * (Z2R b * (IBR y * Rabs vx ^ n' / IBR k))) with 
    ((IBR y * Rabs vx ^ n' *(2 * Z2R b * Z2R b))/ IBR k) by (unfold Rdiv; ring);
  apply Rle_div_r;[IBR_to_Z|];
  apply Rle_trans with (1:=U6);
  replace (2 * Z2R b * 1 * IBR k) with 
    (Z2R b * IBR k * 2) by ring;
  apply Rmult_le_compat; ZR0compat; IBR_to_Z;
  apply Zmult_le_compat_r; ZR0compat).

constructor; try assumption.

replace (IBR y * vx ^ n' / (IBR k * (1 - vx)) * Z2R b) with 
  (IBR (y*Bbase)%B* vx ^ n' / (IBR k * (1 - vx)))
  by (unfold Rdiv; simpl_IBR;fold_IBR; ring).

apply inv_series_correct; clear inv_series_correct; try assumption.

replace (IBR (y * Bbase)%B * Rabs vx ^ n' * (2 * Z2R b)) with 
  (IBR y * Rabs vx ^ n' * (2 * Z2R b * Z2R b)) ;[ |
    unfold Rdiv; simpl_IBR;fold_IBR; ring].
assumption.

replace (IBR (y * Bbase)%B * vx ^ n' / IBR k) with 
  (Z2R b *  (IBR y * vx ^ n' / IBR k)) by (simpl_IBR; unfold Rdiv; ring).
apply multb_correct.
assumption.
assumption. (* from the cut*)
simpl_IBR; fold_IBR; ZR0compat.
simpl_IBR; fold_IBR; rewrite Hy'; ring.

clear inv_series_correct.
rabs_intros.
pattern 1 at 3; replace 1 with 
  (( Z2R b + 2) / ( 2 * Z2R b) + ( Z2R b - 2) / ( 2 * Z2R b)) by field_Z.
apply Rabs_add.
  apply Rle_div_l;[ ZR0compat|].
  rewrite Rmult_comm.
  rewrite <- (Rabs_right (2 * Z2R b)); [| ZR0compat].
  rewrite <- Rabs_mult; assumption.
 
  replace ( Rabs (IBR y * vx ^ n' / (IBR k * (1 - vx)) * Z2R b))
    with ((IBR y * Rabs vx ^ n' * Z2R b) / (IBR k * (1 - vx))).
  apply Rdiv_mult_compat.
  ZR0compat.
  ZR0compat; IBR_to_Z.
  replace (IBR y * Rabs vx ^ n' * Z2R b * (2 * Z2R b))  with 
    (IBR y * Rabs vx ^ n' * (2 * Z2R b * Z2R b)) by ring.
  apply Rle_trans with (1:=U6).
  replace ((Z2R b - 4) * IBR k * (1 - vx)) with 
    ((Z2R b - 4) * (IBR k * (1 - vx))) by ring.
  apply Rmult_le_compat_r; ZR0compat.
  IBR_to_Z.

  rewrite Rabs_mult.
  rewrite Rabs_div; [| ZR0compat; IBR_to_Z].
  repeat rewrite Rabs_mult.
  rewrite (Rabs_right (IBR y));[| ZR0compat; IBR_to_Z].
  rewrite (Rabs_right (IBR k));[| ZR0compat; IBR_to_Z].
  rewrite (Rabs_right (Z2R b));[| ZR0compat].
  rewrite (Rabs_right (1-vx));[| ZR0compat].
  rewrite <- RPow_abs; unfold Rdiv; ring.
Qed.

Theorem log_aux_correct :
  forall k, 
    (2 <= IDZ k)%Z  -> 
    forall m n kn,
      (IDZ kn < b ->
	b < IDZ k^(Z_of_nat (n+m)%nat+1) -> 
	IDZ kn = IDZ k ^ (Z_of_nat n) ->
	IDZ k^(Z_of_nat (n+log_aux m k kn)%nat) < b
	<= IDZ k^(Z_of_nat (n+log_aux m k kn)+1))%Z.
Proof.
intros k Hk m; induction m.
simpl.
intros n kn U1 U2 U3; subst.
rewrite <- plus_n_O in *.
rewrite <- U3; intuition.

intros n kn Hknb Hbknm Hkn. 
simpl; unfold Dle_bool; compare_gt_tac G1.

assert (U1: (forall n m, n+ S m = (n+1) +m)%nat) by (intros; omega).
rewrite U1; apply IHm.
simpl_IDZ; omega.
rewrite <- U1; assumption.
replace (Z_of_nat (n+1)) with (Z_of_nat n +1)%Z; 
  [|rewrite inj_plus; ring].
simpl_IDZ; rewrite Zpower_exp; ZR0compat.
rewrite Zpower_1_r; rewrite <- Hkn; ring.

rewrite <- plus_n_O.
rewrite Zpower_exp; ZR0compat.
rewrite Zpower_1_r; rewrite Hkn in *.
intuition; rewrite Zmult_comm; omega.
Qed.

Theorem log_correct : 
  forall k, 
    (2 <= IDZ k)%Z  -> 
    (IDZ k^(Z_of_nat (log k)) < b <= IDZ k^(Z_of_nat (log k)+1))%Z.
Proof.
pose (Hbpos := base_positive).
intros k Hk.
replace (log k) with (0+log k)%nat by ring.
unfold log.
apply log_aux_correct;
  assumption || (simpl; simpl_IDZ; omega) || idtac.
rewrite plus_0_l.
simpl_IDZ.
rewrite Z_of_nat_Zabs_nat.
assert (U1 : (0 < b)%Z) by omega.
generalize (N_digits_correct (b) U1).
intros [_ U2].
apply Zlt_le_trans with (1:= U2).
apply Zpower_incr; ZR0compat.
apply ZERO_le_N_digits.
apply ZERO_le_N_digits.
Qed.

Theorem Sinv_init_tech_lemma : 
  forall k, 
    (2 <= k ->
    5 <= b ->
    2*b <= (b-4)*((k-1)*k*k*k*k))%Z.
Proof.    
intros k U1 U2.
apply Zle_trans with ((b - 4) * (1* 2 * 2* 2*2))%Z.
omega.
apply Zmult_le_compat_l; [|omega].
repeat (apply Zmult_le_compat; try omega).
Qed.

Theorem Rabs_pow_1 : 
  forall x n,
    Rabs x <= 1 ->
    x ^ n <= 1.
Proof.
intros x n Hrabs.
apply Rle_trans with (Rabs (x^n)).
apply RRle_abs.
apply pow_rabs_1; assumption.
Qed.

Theorem inv_init_correct : 
  forall x k vx,
    (2 <= IDZ k <= b)%Z -> 
    Rabs (vx * IDR k) <= 1 ->
    Srepresents x vx ->
    Srepresents (inv_init x k) (1 / (IDR k * ( 1 - vx ) )).
Proof.
assert (Hbpos:= base_positive).
intros x k vx [H2lek Hkleb] Hxkbounds Hrepx.
unfold inv_init.

assert (U1 :  Srepresents (Sone / k)%S (1 / IDR k)).
  apply div_digit_correct; omega || apply rep_Sone.

assert (U2 : (2 <= IBZ (big_of_digit k))%Z).
  simpl_IBZ; assumption.

assert (U3: Rabs (vx * IBR (big_of_digit k)) <= 1).
  simpl_IBR; fold_IDR; assumption.

assert (U4 : Z2R b * Rabs vx ^ (log k + 1) <= 1).
  apply Rle_trans with (Z2R b / IDR k ^ (log k+1)).
    replace (Z2R b / IDR k ^ (log k+1)) with 
      (Z2R b *(1/ IDR k ^ (log k+1))) by (unfold Rdiv; ring).
    apply Rmult_le_compat_l; [ZR0compat|].
    apply Rle_div_l; [ZR0compat; IDR_to_Z|].
    rewrite Rmult_pow.
    rewrite <- (Rabs_right (IDR k)); [| ZR0compat; IDR_to_Z].
    rewrite <- Rabs_mult; rewrite RPow_abs; rabs_elim.
    apply pow_1_1; assumption.
    
    apply Rle_div_r; [ZR0compat; IDR_to_Z|].
    replace (1 * IDR k ^ (log k + 1)) with 
      (Z2R (IDZ k ^ (Z_of_nat (log k) + 1))%Z).
    assert (U0 := log_correct _ H2lek).
    R_to_Z;  intuition.
    IDR_to_Z; rewrite inj_plus.
    simpl Z_of_nat; ring.
    
assert(U5 : (Z2R b - 4) * (IDR k - 1) <= (Z2R b - 4) * IDR k * (1 - vx)).
  rewrite Rmult_assoc; apply Rmult_le_compat_l; [R_to_Z|].
  replace (IDR k * (1 - vx)) with (IDR k - vx * IDR k) by ring.
  unfold Rminus; apply Rplus_le_compat_l.
  apply Ropp_le_contravar.
  apply Rle_trans with (2:= Hxkbounds).
  apply RRle_abs.

assert (U6 : 1-vx <> 0).
  assert (vx <> 1 ); [| auto with real].
  apply Rabs_neq;[auto with real|].
  apply Rlt_not_eq.
  apply Rle_lt_trans with (1/IDR k).
  apply Rle_div_l; [IDR_to_Z|].
  apply Rle_trans with (2:= Hxkbounds); right.
  rewrite Rabs_mult; rewrite (Rabs_right (IDR k)); 
    [reflexivity| ZR0compat; IDR_to_Z].
  apply Rlt_div_r; IDR_to_Z.

assert (U7 :  - (Z2R b + 2) <= 2 * Z2R b * 0 <= Z2R b + 2).
split; R_to_Z.

assert (U8 :  0 <= IBR 1%B).
IBR_to_Z.

assert (U9 : Srepresents Szero 0).
apply rep_Szero.

assert (U10:  -1 <= 1 / (IDR k * (1 - vx)) <= 1).
assert (U0: IDR k * (1 - vx) <> 0).
  apply prod_neq_R0; [IDR_to_Z| assumption].
rabs_intros.
rewrite Rabs_div; [|assumption].
apply Rle_div_r; [apply Rabs_pos_lt; assumption|].
rewrite Rabs_R1; rewrite Rmult_1_l.
replace (IDR k * (1 - vx)) with (IDR k - vx*IDR k) by ring.
assert (U01 : 1<= IDR k - vx * IDR k); 
  [|rewrite Rabs_right; fourier].
replace 1 with (2- 1) by ring.
unfold Rminus; apply Rplus_le_compat;[IDR_to_Z|].
apply Ropp_le_contravar.
apply Rle_trans with (2:= Hxkbounds).
apply RRle_abs.

assert (U11 : Srepresents (x / k)%S (vx / IDR k)).
  apply div_digit_correct; try omega; assumption.

assert (U12: Srepresents (x * x / k)%S (vx ^ 2 / IDR k)).
  apply div_digit_correct; try omega.
  replace (vx^2) with (vx*vx) by ring.
  apply Smult_correct; assumption.
  
assert (U13: Srepresents (x * x * x / k)%S (vx ^ 3 / IDR k)).
  apply div_digit_correct; try omega.
  replace (vx^3) with (vx*vx*vx) by ring.
  repeat apply Smult_correct; assumption.

assert (V1 : Rabs vx <= 1 / 2).
  apply Rle_div_l;[fourier|].
  apply Rle_trans with (2:= Hxkbounds).
  rewrite Rabs_mult.
  apply Rmult_le_compat_l; [ZR0compat|].
  rewrite Rabs_right; ZR0compat; IDR_to_Z.

assert (V2 : Rabs (vx^2) <= 1 / 4).
  replace (vx^2) with (vx*vx) by ring.
  rewrite Rabs_mult.
  replace (1/4) with (1/2 * (1/2)) by field.
  apply Rmult_le_compat; ZR0compat.

assert (V3 : Rabs (vx^3) <= 1 / 8).
  replace (vx^3) with (vx^2*vx) by ring.
  rewrite Rabs_mult.
  replace (1/8) with (1/4 * (1/2)) by field.
  apply Rmult_le_compat; ZR0compat.

assert (V4 : -1 <= (1 + vx) / IDR k <= 1).
   rabs_intros.
   rewrite Rabs_div; [|IDR_to_Z].
   apply Rle_div_r; [ZR0compat; IDR_to_Z|].
   apply Rle_trans with 2; [| rewrite Rabs_right; ZR0compat; IDR_to_Z].
   replace 2 with (1+1) by ring.
   apply Rabs_add; [rewrite Rabs_R1|]; fourier.

assert (V5 : -1 <= (1 + vx+vx^2) / IDR k <= 1).
   rabs_intros.
   rewrite Rabs_div; [|IDR_to_Z].
   apply Rle_div_r; [ZR0compat; IDR_to_Z|].
   apply Rle_trans with 2; [| rewrite Rabs_right; ZR0compat; IDR_to_Z].
   replace 2 with (1+1/2+1/2) by field.
   repeat (apply Rabs_add; [| fourier]).
   rewrite Rabs_R1; fourier.

assert (V6 : -1 <= (1 + vx+vx^2+vx^3) / IDR k <= 1).
   rabs_intros.
   rewrite Rabs_div; [|IDR_to_Z].
   apply Rle_div_r; [ZR0compat; IDR_to_Z|].
   apply Rle_trans with 2; [| rewrite Rabs_right; ZR0compat; IDR_to_Z].
   replace 2 with (1+1/2+1/4+1/4) by field.
   repeat (apply Rabs_add; [| fourier]).
   rewrite Rabs_R1; fourier.

assert (U14 : Srepresents (Sone / k + x / k)%S (1 / IDR k + vx / IDR k)).
  apply Sadd_correct; try assumption.
  replace (1 / IDR k + vx / IDR k) with 
    ((1 + vx) / IDR k) by (unfold Rdiv; ring).
  assumption.

assert (U15: Srepresents (Sone / k + x / k + x * x / k)%S
     (1 / IDR k + vx / IDR k + vx ^ 2 / IDR k)).
  apply Sadd_correct; try assumption.
  replace (1 / IDR k + vx / IDR k+vx^2/IDR k) with 
    ((1 + vx+vx^2) / IDR k) by (unfold Rdiv; ring).
  assumption.

assert (U16: Srepresents (Sone / k + x / k + x * x / k + x * x * x / k)%S
     (1 / IDR k + vx / IDR k + vx ^ 2 / IDR k + vx ^ 3 / IDR k)).
  apply Sadd_correct; try assumption.
  replace (1 / IDR k + vx / IDR k+vx^2/IDR k+vx^3/IDR k) with 
    ((1 + vx+vx^2+vx^3) / IDR k) by (unfold Rdiv; ring).
  assumption.

simpl_IBZ; Zbool G1.

replace (1 / (IDR k * (1 - vx))) with
  (0 + IBR 1%B*vx^0 / (IBR (big_of_digit k) * (1-vx)));
  [| simpl_IBR; unfold Rdiv; ring].

apply inv_series_correct; try assumption.

simpl; simpl_IBR; fold_IDR.
apply Rle_trans with (2:= U5).
IDR_to_Z.
repeat rewrite Zmult_1_l.
apply Zle_trans with (1:=G1); apply Zeq_le; ring.
  
simpl; IBR_to_Z.

replace (IBR 1%B * vx ^ 0 / IBR (big_of_digit k)) with 
  (1 / IDR k); [assumption |simpl_IBR; simpl; unfold Rdiv; ring].

simpl_IBR; ring.

clear G1.
Zbool G1.

assert (R1 : 1 / (IDR k * (1 - vx)) =
  1 / IDR k + (0 + IBR 1%B*vx^1 / (IBR (big_of_digit k) * (1-vx))))
    by (field_IBR; assumption).
rewrite R1.

apply Sadd_correct;[assumption| |rewrite <- R1; assumption].

apply inv_series_correct; try assumption.

simpl; simpl_IBR; fold_IDR.
apply Rle_trans with (2:= U5).
apply Rle_trans with ((2 * Z2R b)/ IDR k).
  apply Rle_div_l;[IDR_to_Z|].
  replace (1 * (Rabs vx * 1) * (2 * Z2R b) * IDR k) with 
    (Rabs vx * IDR k * (2 * Z2R b)) by ring. 
  pattern (2 * Z2R b) at 2; replace (2 * Z2R b) with
    (1*(2 * Z2R b)) by ring.
  apply Rmult_le_compat_r; [ZR0compat|].
  apply Rle_trans with (2:= Hxkbounds); right.
  rewrite Rabs_mult.
  rewrite (Rabs_right (IDR k)); [reflexivity| ZR0compat; IDR_to_Z].
  
  apply Rle_div_r; IDR_to_Z.
  apply Zle_trans with (1:=G1); apply Zeq_le; ring.
  
simpl; simpl_IBR; ring.

replace (IBR 1%B * vx ^ 1 / IBR (big_of_digit k)) with 
  (vx / IDR k); [ assumption |simpl_IBR; simpl; unfold Rdiv; ring].

simpl_IBR; ring.

clear G1.
Zbool G1.

assert (R1 : 1 / (IDR k * (1 - vx)) =
  (1 / IDR k + vx/IDR k) +(0 + IBR 1%B*vx^2 / (IBR (big_of_digit k) * (1-vx))))
    by (field_IBR; assumption).
rewrite R1.

apply Sadd_correct; [assumption| | rewrite <- R1; assumption].

apply inv_series_correct; try assumption.

simpl; simpl_IBR; fold_IDR.
apply Rle_trans with (2:= U5).
apply Rle_trans with ((2 * Z2R b)/ (IDR k*IDR k)).
  apply Rle_div_l;[ZR0compat; IDR_to_Z|].
  replace (1 * (Rabs vx * (Rabs vx * 1)) * (2 * Z2R b) * (IDR k*IDR k)) with 
    ((Rabs vx * IDR k)^2 * (2 * Z2R b)) by ring. 
  pattern (2 * Z2R b) at 2; replace (2 * Z2R b) with
    (1*(2 * Z2R b)) by ring.
  apply Rmult_le_compat_r; [ZR0compat|].
  apply Rabs_pow_1.
  apply Rle_trans with (2:= Hxkbounds); right.
  do 2 rewrite Rabs_mult.
  rewrite Rabs_Rabsolu; reflexivity.
 
  apply Rle_div_r; [ZR0compat|]; IDR_to_Z.
  apply Zle_trans with (1:=G1); apply Zeq_le; ring.
  
simpl; simpl_IBR; ring.

replace (IBR 1%B * vx ^ 2 / IBR (big_of_digit k)) with 
  (vx^2 / IDR k); [ assumption |simpl_IBR; simpl; unfold Rdiv; ring].

simpl_IBR; ring.

clear G1.
Zbool G1.

assert (R1 : 1 / (IDR k * (1 - vx)) =
  (1 / IDR k + vx/IDR k+vx^2/IDR k) +(0 + IBR 1%B*vx^3 / (IBR (big_of_digit k) * (1-vx))))
    by (field_IBR; assumption).
rewrite R1.

apply Sadd_correct; [assumption| | rewrite <- R1; assumption].

apply inv_series_correct; try assumption.

simpl; simpl_IBR; fold_IDR.
apply Rle_trans with (2:= U5).
apply Rle_trans with ((2 * Z2R b)/ (IDR k*IDR k*IDR k)).
  apply Rle_div_l;[ZR0compat; IDR_to_Z|].
  replace (1 * (Rabs vx * (Rabs vx * (Rabs vx * 1))) *
    (2 * Z2R b) * (IDR k * IDR k * IDR k)) with 
    ((Rabs vx * IDR k)^3 * (2 * Z2R b)) by ring. 
  pattern (2 * Z2R b) at 2; replace (2 * Z2R b) with
    (1*(2 * Z2R b)) by ring.
  apply Rmult_le_compat_r; [ZR0compat|].
  apply Rabs_pow_1.
  apply Rle_trans with (2:= Hxkbounds); right.
  do 2 rewrite Rabs_mult.
  rewrite Rabs_Rabsolu; reflexivity.
 
  apply Rle_div_r; [ZR0compat|]; IDR_to_Z.
  apply Zle_trans with (1:=G1); apply Zeq_le; ring.
  
simpl; simpl_IBR; ring.

replace (IBR 1%B * vx ^ 3 / IBR (big_of_digit k)) with 
  (vx^3 / IDR k); [ |simpl_IBR; simpl; unfold Rdiv; ring].
apply div_digit_correct; try omega.
replace (vx ^3) with (vx*vx*vx) by ring.
repeat apply Smult_correct; assumption.

simpl_IBR; ring.

clear G1.

assert (G1 : (2*b <= (b-4)*((IDZ k-1)*(IDZ k) * (IDZ k) *(IDZ k)*(IDZ k)))%Z).
  apply Sinv_init_tech_lemma; omega.

assert (R1 : 1 / (IDR k * (1 - vx)) =
  (1 / IDR k + vx/IDR k+vx^2/IDR k + vx^3/IDR k) +
  (0 + IBR 1%B*vx^4 / (IBR (big_of_digit k) * (1-vx))))
    by (field_IBR; assumption).
rewrite R1.

apply Sadd_correct; [assumption| | rewrite <- R1; assumption].

apply inv_series_correct; try assumption.

simpl; simpl_IBR; fold_IDR.
apply Rle_trans with (2:= U5).
apply Rle_trans with ((2 * Z2R b)/ (IDR k*IDR k*IDR k*IDR k)).
  apply Rle_div_l;[ZR0compat; IDR_to_Z|].
  replace (1 * (Rabs vx * (Rabs vx * (Rabs vx * (Rabs vx * 1)))) *
    (2 * Z2R b) * (IDR k * IDR k * IDR k * IDR k)) with 
    ((Rabs vx * IDR k)^4 * (2 * Z2R b)) by ring. 
  pattern (2 * Z2R b) at 2; replace (2 * Z2R b) with
    (1*(2 * Z2R b)) by ring.
  apply Rmult_le_compat_r; [ZR0compat|].
  apply Rabs_pow_1.
  apply Rle_trans with (2:= Hxkbounds); right.
  do 2 rewrite Rabs_mult.
  rewrite Rabs_Rabsolu; reflexivity.
 
  apply Rle_div_r; [ZR0compat|]; IDR_to_Z.
  apply Zle_trans with (1:=G1); apply Zeq_le; ring.
  
simpl; simpl_IBR; ring.

replace (IBR 1%B * vx ^ 4 / IBR (big_of_digit k)) with 
  (vx^4 / IDR k); [ |simpl_IBR; simpl; unfold Rdiv; ring].
apply div_digit_correct; try omega.
replace (vx ^4) with (vx*vx*vx*vx) by ring.
repeat apply Smult_correct; assumption.

simpl_IBR; ring.
Qed.

Theorem Ropp_le : 
  forall x y, 
    -x <= - y <= x -> 
    -x <=  y <= x.
Proof.
intros x y [Hxy Hyx].
split; fourier.
Qed.

Theorem Sinverse_tech_lemma:
  forall x,
    1 <= x \/ x <= -1 ->
    -1 <= 1/ x <= 1.
Proof.
assert (U1 : forall x,  1 <= x -> -1 <= 1/ x <= 1).
  intros x Hx.
  split.
    apply Rle_trans with 0 ;[fourier|].
    apply Rmult_le_reg_l with x; [fourier|].
    replace (x * 0) with 0 by ring.
    replace (x * (1 / x)) with 1; [fourier|field; ZR0compat].

    apply Rmult_le_reg_l with x; [fourier|].
    replace (x * 1) with x by ring.
    replace (x * (1 / x)) with 1; [assumption|field; ZR0compat].

intros x; destruct 1 as [Hx | Hx].
apply U1; assumption.
apply Ropp_le.
replace (- (1 / x)) with (1 / (-x));[|field].
apply U1; fourier.
assert (x < 0) by fourier.
auto with real.
Qed.

Theorem Sinverse_correct : 
  forall n x vx,
    Srepresents x vx -> 
    1 <= vx*Z2R b^n \/ vx*Z2R b^n <= -1 ->
    Srepresents (Sinverse n x) (1 / (Z2R b^n * vx)).
Proof.
intros n; induction n; simpl.
intros x vx Hrepx Hvx.
rewrite Rmult_1_r in Hvx.
assert (U1: -1 <= vx <= 1).
  apply Srepresents_interval with (1:=Hrepx). 

elim Hvx; intros G1.
  assert (vx = 1) by intuition; subst.
  replace (1/(1*1)) with 1 by field; assumption.

  assert (vx = -1) by intuition; subst.
  replace (1/(1*-1)) with (-1) by field; assumption.

intros x vx Hrepx Hvx.
assert (Hbpos := base_positive).
assert (U1 : 0 < Z2R b) by IDR_to_Z.
destruct x as [k x']; inversion Hrepx; subst.

replace (Z2R b * Z2R b ^ n * ((IDR k + r) / Z2R b)) with 
  (Z2R b ^ n * (IDR k + r)) by field_Z.
compare_tac G1.

(*case k=1*)
replace (IDR k) with 1 in * by IDR_to_Z; clear G1.
destruct x' as [k' x'']; inversion_to_variable x''; subst.
compare_tac G1.

(*case k'=0 (and k=1) *)
assert (R1 : IDR k' = 0) by IDR_to_Z.
assert (R2 : 1/(Z2R b ^ n * (1 + (IDR k' + r0) / Z2R b)) = 
  Z2R b * ((1/ (IDR Dbase * (1 - r0 / IDR  (-Dbase)%D)))/ (Z2R b ^ n))).
  rewrite R1; field_IDR.
    apply Rlt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac; fourier.
    ZR0compat.
    apply Rgt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac; fourier.
rewrite R2.
apply multb_correct.
apply str_add_zero_correct.
apply inv_init_correct.
simpl_IDZ; omega.
replace ((r0 / IDR (- Dbase)%D)* IDR Dbase) with (-r0) by field_IDR.
rabs_elim; dec_tac; split; fourier.
apply div_digit_correct; assumption || simpl_IDZ; omega.
rewrite <- R2.
apply Sinverse_tech_lemma.
match goal with 
  id : 1 <= ?a \/ ?a <= -1 |- 1 <= ?c \/ ?c <= -1 => 
    replace c with a; [assumption | try field_IDR]
end.

(*case k'<0 (and k=1)  : k::k'::x = 0::k'+b::x*)
replace (1 + (IDR k' + r0) / Z2R b) with 
  ((IDR (k'+Dbase)%D + r0)/Z2R b) by field_IDR.
apply IHn.
constructor; assumption || simpl_IDZ; omega.
match goal with 
  id : 1 <= ?a \/ ?a <= -1 |- 1 <= ?c \/ ?c <= -1 => 
    replace c with a; [assumption | try field_IDR]
end.

(*case k'>0 (and k=1)  : k::k'::x = 2::k'-b::x*)
assert (U2 : (0 <= Z2R (IDZ k')))by R_to_Z.
assert (R1 : 1/(Z2R b ^ n * (1 + (IDR k' + r0) / Z2R b)) = 
   (1/ (IDR 2%D * (1 - ((IDR (k'-Dbase)%D + r0) / Z2R b)/ IDR (-(2))%D))) / (Z2R b ^ n)).
  field_IDR.
    apply Rlt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac; fourier.
    ZR0compat.
    apply Rgt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac;fourier.
rewrite R1.
apply str_add_zero_correct.
apply inv_init_correct.
  simpl_IDZ; omega.
  replace ( ((IDR (k' - Dbase)%D + r0) / Z2R b / IDR (- (2))%D)*IDR 2%D) with 
  (- (IDR (k' - Dbase)%D + r0) / Z2R b) by field_IDR.
  rabs_elim; apply Ropp_le.
  replace ( - (- (IDR (k' - Dbase)%D + r0) / Z2R b)) with 
  ((IDR k' - Z2R b + r0)/Z2R b) by field_IDR.
  apply Rle_div_1; [R_to_Z|].
  split.
    assert (1 <=  IDR k') by IDR_to_Z; dec_tac; fourier.
    assert (IDR k' <= Z2R b-1) by IDR_to_Z; dec_tac; fourier.
  apply div_digit_correct; try (simpl_IDZ; omega).
  constructor; assumption || simpl_IDZ; omega.

(*case k < 1*)
compare_tac G2.

(*case k = -1*)
replace (IDR k) with (-1) in * by IDR_to_Z; clear G1 G2.
destruct x' as [k' x'']; inversion_to_variable x''; subst.
compare_tac G1.

(*case k'=0 (and k=-1) *)
assert (R1 : IDR k' = 0) by IDR_to_Z.
assert (R2 : 1/(Z2R b ^ n * (- 1 + (IDR k' + r0) / Z2R b)) = 
  Z2R b * (-(1/ (IDR Dbase * (1 - r0 / IDR  Dbase)))/ (Z2R b ^ n))).
  rewrite R1; field_IDR.
    apply Rgt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac; fourier.
    ZR0compat.
    apply Rlt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac; fourier.
rewrite R2.
apply multb_correct.
apply str_add_zero_correct.
apply Sopp_correct.
apply inv_init_correct.
simpl_IDZ; omega.
replace ((r0 / IDR Dbase)* IDR Dbase ) with r0 by field_IDR.
rabs_elim; assumption.
apply div_digit_correct; assumption || simpl_IDZ; omega.
rewrite <- R2.
apply Sinverse_tech_lemma.
match goal with 
  id : 1 <= ?a \/ ?a <= -1 |- 1 <= ?c \/ ?c <= -1 => 
    replace c with a; [assumption | try field_IDR]
end.

(*case k'<0 (and k=-1) *)
assert (U2 : (0 <= - Z2R (IDZ k')))by R_to_Z.
assert (R1 : 1/(Z2R b ^ n * (-1 + (IDR k' + r0) / Z2R b)) = 
   -(1/ (IDR 2%D * (1 - ((IDR (k'+Dbase)%D + r0) / Z2R b)/ IDR 2%D))) / (Z2R b ^ n)).
  field_IDR.
    apply Rgt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac; fourier.
    ZR0compat.
    apply Rlt_not_eq; assert (1 < Z2R b) by R_to_Z; dec_tac;fourier.
rewrite R1.
apply str_add_zero_correct.
apply Sopp_correct.
apply inv_init_correct.
  simpl_IDZ; omega.
  replace (((IDR (k' + Dbase)%D + r0) / Z2R b / IDR 2%D)* IDR 2%D ) with 
  ((IDR k' + Z2R b + r0) / Z2R b) by field_IDR.
  rabs_elim; apply Rle_div_1; [R_to_Z|].
  split.
    assert (-Z2R b +1 <=  IDR k') by IDR_to_Z; dec_tac; fourier.
    assert (IDR k' <= -1) by IDR_to_Z; dec_tac; fourier.
  apply div_digit_correct; try (simpl_IDZ; omega).
  constructor; assumption || simpl_IDZ; omega.

(* case k' >0 (and k=-1)*)
replace (-1 + (IDR k' + r0) / Z2R b) with 
  ((IDR (k' - Dbase)%D + r0)/Z2R b) by field_IDR.
apply IHn.
constructor; assumption || simpl_IDZ; omega.
match goal with 
  id : 1 <= ?a \/ ?a <= -1 |- 1 <= ?c \/ ?c <= -1 => 
    replace c with a; [assumption | try field_IDR]
end.

(*case k <= -2*)
clear G1.
assert (U2 : IDR k <= - 2) by IDR_to_Z.
assert (R1 : 1 / (Z2R b ^ n * (IDR k + r)) = 
  (- (1 /  (IDR (-k)%D * (1 - r/IDR (-k)%D))))/ (Z2R b ^ n)).
  field_IDR; fold_IDR.
    apply Rgt_not_eq; dec_tac; fourier.
    ZR0compat.
    apply Rlt_not_eq; dec_tac;fourier.
rewrite R1.
apply str_add_zero_correct.
apply Sopp_correct.
apply inv_init_correct.
  simpl_IDZ; omega.
  replace ( (r / IDR (-k)%D) * IDR (-k)%D ) with r by field_IDR.
  rabs_elim; assumption.
  apply div_digit_correct; assumption || simpl_IDZ; omega.

(*case k=0*)
replace (IDR k) with 0 in * by IDR_to_Z.
replace (0 + r) with r in * by ring.
replace ( r / Z2R b * (Z2R b * Z2R b ^ n)) with 
  (r *  Z2R b ^ n) in Hvx by field_Z.
apply IHn; assumption.

(*case k >= 2*)
assert (U2 : 2 <= IDR k) by IDR_to_Z.
assert (R1 : 1 / (Z2R b ^ n * (IDR k + r)) = 
  (1 /  (IDR k * (1 - r/IDR (-k)%D)))/ (Z2R b ^ n)).
  field_IDR; fold_IDR.
    apply Rlt_not_eq; dec_tac; fourier.
    ZR0compat.
    apply Rgt_not_eq; dec_tac;fourier.
rewrite R1.
apply str_add_zero_correct.
apply inv_init_correct.
  simpl_IDZ; omega.
  replace ( (r / IDR (-k)%D) * IDR k) with (-r) by field_IDR.
  rabs_elim; apply Ropp_le; replace (- - r) with r by ring; assumption.
  apply div_digit_correct; assumption || simpl_IDZ; omega.
Qed.

Theorem inverse_tech_lemma : 
  forall x y z, 
    1 <= x *Rabs (y*z) -> 
    z <> 0.
Proof.
intros x y z Hxyz.
apply Rmult_neq_0 with y.
assert (U1 : forall x, Rabs x <> 0 -> x <> 0).
intros w Hw.
elim (Rle_or_lt 0 w); intros G1.
rewrite <- (Rabs_right w); [assumption| ZR0compat].
auto with real.
apply U1.
replace (z*y) with (y * z) by ring.
apply Rmult_neq_0 with x.
replace (Rabs (y * z) * x ) with (x * Rabs (y * z)) by ring.
apply Rlt_neq_0; fourier.
Qed.

Theorem inverse_tech_lemma2 : 
  forall r e n,
    1  <= Z2R b ^ n * Rabs (Z2R b ^ e * r) -> 
    1 <= r * Z2R b ^ (n + e) \/ r * Z2R b ^ (n + e) <= -1.
Proof.
assert (Hbpos:= base_positive).
intros r e n H1.
rewrite pow_add.
elim (Rle_or_lt 0 r); intros G1.
left; apply Rle_trans with (1:=H1); right.
rewrite Rabs_right; [ring| ZR0compat].
right; apply Ropp_le_cancel.
replace (- - 1) with 1 by ring.
apply Rle_trans with (1:=H1); right.
rewrite Rabs_mult.
rewrite (Rabs_left r); [| auto].
rewrite Rabs_right; [ring|ZR0compat].
Qed.

Theorem inverse_correct : 
  forall x n vx, 
    1 / Z2R b ^ n <=  Rabs vx ->
    represents x vx ->
    represents (inverse x n) (1/vx).
Proof.
assert (Hbpos:= base_positive).
intros x n vx Hvx Hrep.
unfold inverse.
inversion Hrep; subst.
assert (U1: 1 <= Z2R b ^ n * Rabs (Z2R b ^ e * r)).
replace 1 with ( Z2R b ^ n * (1/  Z2R b ^ n)); [| field; ZR0compat].
apply Rmult_le_compat_l; ZR0compat.

replace (Z2R b ^ e * r * Z2R b ^ n) with (r * Z2R b ^ (n+e)) in Hvx 
  by  (rewrite pow_add; ring).
replace (1 / (Z2R b ^ e * r)) with 
  (Z2R b ^ n * (1 / (Z2R b ^ (n + e) * r))).
constructor.
apply Sinverse_correct; try assumption.

apply inverse_tech_lemma2; assumption.

rewrite pow_add; field.
repeat split.
apply  inverse_tech_lemma with (1:=U1).
ZR0compat.
ZR0compat.
Qed.

Theorem inverse_correct2 : 
  forall x n vx, 
    1 <= Z2R b ^ n * Rabs vx ->
    represents x vx ->
    represents (inverse x n) (1/vx).
Proof.
assert (Hbpos:= base_positive).
intros x n vx Hvx Hrep.
unfold inverse.
inversion Hrep; subst.
replace (Z2R b ^ e * r * Z2R b ^ n) with (r * Z2R b ^ (n+e)) in Hvx 
  by  (rewrite pow_add; ring).
replace (1 / (Z2R b ^ e * r)) with 
  (Z2R b ^ n * (1 / (Z2R b ^ (n + e) * r))).
constructor.
apply Sinverse_correct; try assumption.

apply inverse_tech_lemma2; assumption.

rewrite pow_add; field.
repeat split.
apply  inverse_tech_lemma with (1:=Hvx).
ZR0compat.
ZR0compat.
Qed.

End INV_CORRECT.
