Require Export Reals.
Require Export Fourier.
Require Import Ranalysis.
Require Export Field.
Open Local Scope R_scope.


Lemma Rlt_0_mult_reg:
forall r1 r2, r1 * r2 <> 0 -> r1 <>0 /\ r2 <> 0.
Proof.
intros r1 r2 H; split.
elim ( (Req_dec r1 0)) ; intros Hr1; 
  [rewrite Hr1 in H; rewrite Rmult_0_l in H; elim (H (refl_equal 0))| assumption] .
elim ( (Req_dec r2 0)) ; intros Hr1; 
  [rewrite Hr1 in H; rewrite Rmult_0_r in H; elim (H (refl_equal 0))| assumption] .
Qed.

Lemma Rle_Rabs:
forall (r1 r2:R), 0<= r1->
r1<=r2->
Rabs r1 <= Rabs r2.
Proof.
intros.
assert (Rabs r1=r1).
apply Rabs_pos_eq; auto.
assert (Rabs r2 = r2).
apply Rabs_pos_eq; auto.
apply Rle_trans with r1; assumption.
rewrite H1.
rewrite H2.
trivial.
Qed.

Lemma pos_div2 : 
  forall x, 0 < x -> 0 < x/2.
Proof.
intros;fourier.
Qed.

Hint Resolve pos_div2.

Lemma Rabs_def2_le :
forall x a: R, 
Rabs x <= a <->
-a<=x<=a.
Proof.
intros x a.
split; [intros Hip| intros [H1 H2]].
split; split_Rabs; fourier.
split_Rabs; fourier.
Qed.

Lemma dif_0 :
forall x:R, Rabs (1-x)<=1/2 -> x > 0.
Proof.
intros.
assert (H1: -(1/2) <= (1-x) <= 1/2).
destruct (Rabs_def2_le (1-x) (1/2)).
apply (H0 H).
assert (H3:1/2<=x).
elim H1.
intros; fourier.
fourier.
Qed.

Lemma split_fr:
forall (r1 r2:R), r2<>0 ->
Rabs (r1/r2) = Rabs r1/Rabs r2.
Proof.
intros r1 r2 Hdif.
unfold Rdiv.
rewrite Rabs_mult. 
rewrite Rabs_Rinv; auto.
Qed.

Lemma help_le:
forall (r1 r2:R), 0<r1*r2 -> r1<= r2 -> /r2 <= /r1.
Proof.
intros r1 r2 H H0.
unfold Rle in H0 |-*.
elim H0;intros Hc.
left.
apply (Rinv_lt_contravar r1 r2 H Hc).
right.
rewrite Hc; auto.
Qed.

Lemma op_nr_r:
forall (r1 r2:R),
r1-r2=r1->
r2=0.
Proof.
intros.
apply Rle_antisym; fourier.
Qed.


Lemma Rnot_gt_le:
forall r1 r2, (~(r1>r2))->r1<=r2.
intros r1 r2; generalize (Rtotal_order r1 r2); unfold Rle in |- *; tauto.
Qed.

Lemma Rnot_lt_ge:
forall r1 r2, (~(r1<r2))->r1>=r2.
intros r1 r2; generalize (Rtotal_order r1 r2); unfold Rge in |- *; tauto.
Qed.

Lemma Rmult_neq_0_reg2 : forall r1 r2, r1 * r2 <> 0 -> r1 <> 0 /\ r2 <> 0.
Proof.
intros r1 r2 H; split; red in |- *; intro; apply H; auto with real.
Qed.

(*definition and properties of closed and open intervals and discs*) 
Definition c_I (a b x:R) : Prop:= a<=x<=b.
Definition o_I (a b x:R) : Prop:= a<x<b.
Definition c_disc (x delta y: R) : Prop :=
(Rabs (x -y))<=delta.

Lemma inta:
forall a b, a<=b -> c_I a b a.
Proof.
intros a b H; split; fourier.
Qed. 

Lemma intb:
forall a b, a<=b -> c_I a b b.
Proof.
intros a b H; split; fourier.
Qed. 


Lemma lung_interv:
forall a b x dist,
Rabs (b-a)<=dist->
(c_I a b x) ->
Rabs (b-x)<=dist.
Proof.
intros a b x dist Hba [Hia Hib].
apply Rle_trans with (Rabs(b-a)).
apply Rle_Rabs; fourier.
fourier.
Qed.

Lemma lung_interv2:
forall a b x dist,
Rabs (b-a)<=dist->
(c_I a b x) ->
Rabs (x-a)<=dist.
Proof.
intros a b x dist Hba [Hia Hib].
apply Rle_trans with (Rabs(b-a)).
apply Rle_Rabs; fourier.
fourier.
Qed.

Lemma inc_neigh:
forall (x y r:R), (Rabs (x- y)) <= (r/2)
-> included  (c_disc y (r/2)) (c_disc x r).
Proof.
intros x y r HRabs2 x0 Hdisc .
unfold c_disc in Hdisc |-*.

apply Rle_trans with (Rabs(x-y)+Rabs(y-x0)).
replace (x-x0) with (x-y+(y-x0)) by ring.
apply (Rabs_triang (x-y) (y-x0)).
replace r with ((r/2)+ (r/2)).
 fourier.
field. 
Qed.

Hint Resolve inc_neigh.

Lemma in_open_closed:
forall a b, included (o_I a b ) (c_I a b).
Proof.
intros a b t H; unfold o_I in *; unfold c_I in *; elim H;
intros; split; fourier.
Qed.

Lemma open_closed_int:
forall a b c d, included (c_I a b) (c_I c d) ->
included (o_I a b) (o_I c d).
Proof.
intros a b c d Hin.
intros t Ht.
elim (Rle_dec a b); intros Hab.
assert (H1 : c <= a).
elim (Hin a (inta a b Hab)); tauto.
assert (H2: b <= d).
elim (Hin b (intb a b Hab)); tauto.
elim Ht; intros.
unfold o_I, c_I in *; split; 
  [apply Rle_lt_trans with a|apply Rlt_le_trans with b]; assumption.
assert (Hcon: a < b).
elim Ht ; intros;  apply Rlt_trans with t; assumption.
elim Hab; fourier.
Qed.


Lemma rewrite_c_disc:
forall x e y, 
(c_disc x e y <-> x-e<=y<=x+e).
intros x e y .
split; intros Hip.
unfold c_disc in Hip.

rewrite Rabs_def2_le in Hip .
elim Hip; intros Hi1 Hi2.
split; fourier.
unfold c_disc.

rewrite Rabs_def2_le.
elim Hip; intros Hin1 Hin2.
split; fourier.
Qed.

Lemma center_in_disc:
forall c r, 0<=r -> c_disc c r c.
Proof.
intros c r H.
rewrite rewrite_c_disc; split.
pattern c at 2; replace c with (c-0) by ring.
unfold Rminus; apply Rplus_le_compat_l;
apply Ropp_le_contravar; trivial .
pattern c at 1; replace c with (c+0) by ring.
apply Rplus_le_compat_l; trivial.
Qed.

Lemma edgel_in_disc:
forall c r, 0<=r-> c_disc c r (c - r).
Proof.
intros c r Hr.
rewrite rewrite_c_disc.
split; try fourier.
Qed.

Lemma edger_in_disc:
forall c r, 0<=r-> c_disc c r (c + r).
Proof.
intros c r Hr.
rewrite rewrite_c_disc.
split; try fourier.
Qed.

Lemma c_disc_c:
forall c r t, c_disc c r t <->  - r <=  t - c <= r.
intros c r t ; split; intro H; rewrite rewrite_c_disc in *;
elim H ; intros H1 H2; split; try fourier.
Qed. 


Lemma c_disc_el:
forall c r t, c_disc c r t <->  - r + t <=  c <= r + t .
intros c r t ; split; intro H; rewrite rewrite_c_disc in *;
elim H ; intros H1 H2; split; try fourier.
Qed. 

Lemma incl_c_disc:
forall c r1 r2, r1<=r2->
included (c_disc c r1 ) (c_disc c r2).
Proof.
intros c r1 r2 Hr t Hr1.
rewrite c_disc_c in *.
elim Hr1; intros Hr11 Hr12; split; try fourier.
Qed.

Lemma disc_translation_pos:
forall c r1 d, 0 <= d ->
forall x, c_disc c r1 x -> c_disc c (r1+d) (x+d).
Proof.
intros c r d x H0 Hx.
rewrite c_disc_c in *.
elim Hx; intros; split;  ring_simplify; try fourier.
Qed.

(*
Lemma disc_scale_pos:
forall c r1 d, d <> 0 ->
forall x, c_disc c r1 x -> c_disc c (r1/d) (x/d).
Proof.
intros c r d H0 x Hx.
rewrite c_disc_c in *.
elim Hx; intros; split; unfold Rdiv;  try fourier; trivial.

Qed.
*)

Lemma disc_interval:
 forall c r a b, 0 < r -> included (c_disc c r) (c_I a b) ->
   a < c < b.
Proof.
intros c r a b Hr Hin.
assert (h1: c_I a b ( c - r/2)).
apply Hin.
unfold c_disc; replace (c - (c- r/2)) with (r/2) by ring.
rewrite Rabs_right; fourier.
unfold c_I in h1; elim h1; clear h1; intros h11 _.
split; [fourier|].
assert (h1: c_I a b ( c + r/2)).
apply Hin.
unfold c_disc; replace (c - (c + r/2)) with (- (r/2)) by ring.
rewrite Rabs_left; fourier.
unfold c_I in h1; elim h1; clear h1; intros _ h22; fourier.
Qed.





(*borrowed and new proofs to apply archimede's principle*)
Theorem INR_Zabs_nat_eq :
  forall (z:Z), (0 <= z)%Z -> (IZR z = INR(Zabs_nat z))%R.
unfold Zabs_nat.
intros z; case z.
simpl.
intros; auto.
unfold IZR.
intros; auto.
compute.
intros x H'; elim H'; auto.
Qed.


Lemma pow2_larger :
  forall n, (INR n <= 2^n).
induction n.
simpl; fourier.
replace (S n) with (1+n)%nat.
rewrite pow_add; rewrite pow_1.
assert (Hlem: (forall a b, 1<= b -> a<=b -> 1+a<= 2*b)%R).
intros; fourier.
rewrite plus_INR; simpl (INR 1).
apply Hlem; auto.
apply pow_R1_Rle;fourier.
omega.
Qed.

Lemma pow_half_smaller :
 forall r, (0 < r) -> (1/(2^(Zabs_nat(up (1/r)))) <= r).
intros r Hpos.
elim (archimed (1/r)); intros Harc1 Harc2.
apply (Rmult_le_reg_l (2^(Zabs_nat (up (1/r))))).
apply pow_lt.
fourier.
replace (2 ^ Zabs_nat (up (1 / r)) * (1 / 2 ^ Zabs_nat (up (1 / r))))%R
 with 1%R.
apply Rle_trans with ((INR (Zabs_nat (up (1/r))))*r)%R.
rewrite <- INR_Zabs_nat_eq.
pattern 1%R at 1; replace 1%R with (1/r * r)%R ;[|field; auto with real].
apply Rmult_le_compat_r.
fourier.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
apply le_IZR.
apply Rle_trans with (1/r)%R.
replace (1/r)%R with (/r)%R.
simpl IZR.
assert (0 < / r)%R.
apply Rinv_0_lt_compat; auto.
unfold Rle; left; assumption.
field.
apply Rgt_not_eq; auto.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
apply Rmult_le_compat_r.
fourier.
apply pow2_larger.
field.
apply pow_nonzero.
discrR.
Qed.

Lemma archim:
forall v, 0 < v -> {n:nat| (1/(2^n) <= v)%R}.
intros v Hvpos.
exists (Zabs_nat (up (1/v)))%nat.
destruct (archimed (1/v)) as [Hb1 _].
apply pow_half_smaller.
assumption.
Qed.

Lemma archim_lt:
(forall v, 0 < v -> {n:nat| 1/(2^n) <= v/2}).
intros v Hvpos.
exists (Zabs_nat (up (1/v))+1)%nat.
destruct (archimed (1/v)) as [Hb1 _].
rewrite pow_add.
replace (1 / (2 ^ Zabs_nat (up (1 / v))*2^1))%R
 with ((1 / (2 ^ Zabs_nat (up (1 / v))))*/2)%R.
replace (v/2)%R with (v*/2)%R.
apply Rmult_le_compat_r.
assert (0 < /2)%R.
apply Rinv_0_lt_compat;auto.
fourier.
fourier.
apply pow_half_smaller.
fourier.
field.
replace (2^1)%R with 2%R.
field.
assert (2^Zabs_nat(up(1*/v)) <> 0)%R.
apply pow_nonzero;auto.
discrR.
simpl;assumption.
simpl;field.
Qed.

Lemma archim_lt2:
forall v, 0 < v -> {n:nat| (1/(2^n) < v)%R}.
intros v Hvpos.
assert (Hhalf: {n:nat| 1/(2^n) <= v/2}).
apply archim_lt; assumption.
elim Hhalf; intros n Hineq.
exists n.
assert (v/2<v).
fourier.
unfold Rle in Hineq; elim Hineq; intros.
apply Rlt_trans with (v/2); assumption.
rewrite H0.
assumption.
Qed.


Lemma pow2_diff0:
forall n:nat,
2^n>0.
induction n; simpl; fourier.
Qed.

Lemma archim2_le:
forall v k, 0 < v ->0<=k-> {n:nat| (k/(2^n) < v)%R}.
intros v k Hvpos Hkpos.
case (Rle_lt_or_eq_dec _ _ Hkpos); [intros Heq |intros Hneq].
assert ({ n:nat| 1/2^n<v/k}).
apply (archim_lt2 (v/k)).
unfold Rdiv.
apply Rlt_mult_inv_pos; assumption.
elim H; intros N Hprop.
exists N.
replace (k/2^N) with (k*(1/2^N)) ;[|field; apply Rgt_not_eq; apply pow2_diff0].
replace v with (k*(v/k)).
apply (Rmult_lt_compat_l k (1/2^N) (v/k)).
fourier.
assumption.
field.
apply Rgt_not_eq.
fourier.
auto with real.
exists 0%nat.
simpl.
rewrite <-Hneq.
fourier.
Qed.


Section Kantorovitch.

(*variables we need in the theorem*)

Variable (a0 b0 c:R).
Variable X0:R.
(*Variable eps:R.*)


(*define f f' as section variables*)
Variable a:R.
Variable b:R.
Variable f:R->R.
Variable f':R->R.

Hypothesis A0_b0_c: 2*a0*b0*c<=1.
(*Hypothesis about_eps:0<eps.*)
Hypothesis pr: forall t:R, c_I a b t -> derivable_pt f t.
Hypothesis Hder_f: forall (t:R)(H:c_I a b t), derive_pt f t (pr t H)=f' t. 
Hypothesis Hcont_f':forall y, (c_I a b y) -> continuity_pt f' y.
Hypothesis Hlim_f':forall y1 y2:R, c_I a b y1 -> c_I a b y2 ->
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2)).
Hypothesis Hincl: a< X0 < b.
Hypothesis Hdif_f':f' X0 <> 0 .
Hypothesis Habs_a: Rabs (/(f' X0)) <= a0.
Hypothesis Habs_b:Rabs (f X0/(f' X0)) <= b0 .
Hypothesis Hb_eps: included (c_disc X0 (2*b0)) (c_I a b).

Definition eps := 2*b0.

Fixpoint An (n:nat) {struct n} : R :=
match n with
|0=>a0
| S n => 2*(An n)
end.

Fixpoint Bn (n:nat) : R:=
match n with
|0=>b0
|S n => (An n)*(Bn n)*(Bn n)*c
end.


Fixpoint En (n:nat): R :=
match n with
|0 => eps
|S n => (En n)/2
end.


Fixpoint Xn (n:nat) {struct n}: R :=
match n with 
| 0 => X0
|S n => (Xn n) - (f (Xn n))/(f' (Xn n))
end.

(*properties for these variables*)
Lemma H_a0:0<=a0.
Proof.
apply Rle_trans with (Rabs (/(f' X0))); 
[apply Rabs_pos|trivial].
Qed.

Lemma H_b0:0<=b0.
Proof.
apply Rle_trans with (Rabs (f X0/(f' X0)));
[apply Rabs_pos|trivial].
Qed.

Lemma H_c:0<=c.
Proof.
assert (Hab: a < b).
elim Hincl ; intros Ha Hb; fourier.

assert (Hab1: a <= b) by fourier.
assert (H1:= Hlim_f' a b (inta a b Hab1) (intb a b Hab1)).
apply ( Rmult_le_reg_l ) with (Rabs ( a - b)).
apply Rabs_pos_lt; apply Rlt_not_eq; fourier.
ring_simplify; rewrite Rmult_comm; apply Rle_trans
with (Rabs ( f' a - f' b)); [apply Rabs_pos| trivial].
Qed.




Hint Resolve H_a0 H_b0 H_c.





Lemma aux1:
forall r1 r2, 0<=r1 -> 0<=r2-> 0<=r1*r2.
Proof.
intros r1 r2 Hr1 Hr2.
replace 0 with (0*r2) by ring.
auto with real.
Qed.

Lemma eps_pos:
forall n, 0<=En n.
Proof.
induction n.
simpl; unfold eps; apply aux1;
[fourier|auto].
simpl; fourier.
Qed.



Lemma greater0: 0<=2*a0*b0*c.
Proof. 
repeat apply aux1; auto.
fourier.
Qed.

Hint Resolve greater0.

Lemma An_ge_0:
forall n, 0<=An n.
Proof.
induction n.
simpl; auto.
simpl; apply aux1; [fourier|auto].
Qed.

Hint Resolve An_ge_0.

Lemma Bn_ge_0:
forall n, 0<=Bn n.
Proof.
induction n.
simpl; auto.
simpl; repeat apply aux1; auto.
Qed.

Lemma An_bn_c0:
forall (n:nat) , 0<=2*(An n)*(Bn n)*c.
Proof.
induction n.
simpl; auto.
simpl.
replace (2 * (2 * An n) * (An n * Bn n * Bn n * c) * c) with ((2*An n*Bn n * c)*(2*An n*Bn n * c));
 [idtac | ring].
fold (Rsqr (2*An n*Bn n*c)).
auto with real.
Qed.

Lemma An_bn_c1:
forall (n:nat) , 2*(An n)*(Bn n)*c<=1.
Proof.
induction n.
simpl; auto.
simpl.
replace (2 * (2 * An n) * (An n * Bn n * Bn n * c) * c) with ((2*An n*Bn n * c)*(2*An n*Bn n * c));
 [idtac | ring].
fold (Rsqr (2*An n*Bn n*c)).
pattern 1 at 3;rewrite <- Rsqr_1.
apply Rsqr_incr_1.
auto.
apply An_bn_c0.
fourier.
Qed.


Lemma An_bn_c12:
forall (n:nat) , (An n)*(Bn n)*Bn n*c<=1/2*Bn n.
Proof.
intros.
assert ( 2*(An n)*(Bn n)*c<=1).
apply An_bn_c1.
assert  ((An n)*(Bn n)*c<=1/2).
unfold Rdiv.
replace (An n*Bn n*c) with ((1/2)*(2*(An n)*(Bn n)*c)) by field.
replace (1*/2) with (1/2*1) by field.
apply (Rmult_le_compat_l (1/2) (2*(An n)*(Bn n)*c) 1); fourier .
replace ((An n)*(Bn n)*Bn n*c) with (An n*Bn n*c*Bn n) by ring.
apply Rmult_le_compat_r ; [|assumption].
apply Bn_ge_0;assumption.
Qed.

Hint Resolve An_bn_c12.

Lemma Bn_desc_2:
forall n:nat,
Bn n<=b0/2^n.
Proof.
induction n.
simpl; fourier.
simpl.
unfold Rdiv in IHn.
assert (1/2* Bn n<=b0*/(2*2^n)).
replace (b0*/(2*2^n)) with (1/2*(b0/2^n)).
apply Rmult_le_compat_l.
fourier.
assumption.
field.
apply Rgt_not_eq.
apply pow2_diff0.
apply Rle_trans with (1/2*Bn n); auto.
Qed.


(*proofs about limit of sequences*)
Theorem Heine1:
forall (f:R->R) (Un:nat->R) (x:R),
continuity_pt f x -> Un_cv Un x -> Un_cv (fun i:nat=>f (Un i)) (f x).
Proof. 
intros F Un x H H0 eps0 He.
destruct (H eps0 He) as [delta [H1 H2]] .
destruct (H0 delta H1) as [N H3] .
exists N.
intros n Hn.
assert (Heq_dec : Un n = x \/ Un n <> x) by
 (elim (Rtotal_order x (Un n)); auto with real).
destruct Heq_dec as [Heq | Hneq].
rewrite Heq; unfold R_dist.
replace (F x-F x) with 0 by ring;  rewrite Rabs_R0; auto.
apply (H2 (Un n)); unfold no_cond; split; [split;auto | exact (H3 n Hn)].
Qed.

Lemma disc_incl_disc:
forall Un,
(forall n, included (c_disc (Un (S n)) (En (S n))) 
(c_disc (Un n) (En n))) ->
(forall n p, included (c_disc (Un (n+p)%nat) (En (n+p)%nat)) 
(c_disc (Un n) (En n))).
intros.
induction p.
replace (n+0)%nat with n by omega.
unfold included.
intros; auto.
replace (n+S p)%nat with (S (n+p)%nat) by omega.
assert (included (c_disc (Un (S (n+p))) (En (S (n+p))))
(c_disc (Un (n+p)%nat) (En (n+p)%nat))).
apply (H (n+p)%nat).
apply included_trans with (c_disc  (Un (n+p)%nat) (En (n+p))); auto.
Qed.

Lemma disc_incl_pct:
forall Un, 
(forall n, included (c_disc (Un (S n)) (En (S n))) 
(c_disc (Un n) (En n)))->
(forall n p, (c_disc (Un n) (En n) (Un (n+p)%nat))).
Proof.
intros.
assert (included (c_disc (Un (n+p)%nat) (En (n+p)%nat)) (c_disc (Un n) (En n))).
apply (disc_incl_disc Un H n p).
unfold included in H0.
apply (H0 (Un (n+p)%nat)).
unfold c_disc.
rewrite Rminus_diag_eq.
rewrite Rabs_R0.
unfold Rle; apply eps_pos.
trivial.
Qed.

Lemma lim_conv_up:
forall (un:nat->R) (a l:R),
(forall n, un n <=a)->
Un_cv un l ->
l<=a.
Proof.
intros un t l H H0.
cut (~(l>t)).
intros H1.
apply Rnot_gt_le; auto.
unfold not; intros H1.
unfold Un_cv in H0.
assert (l-t>0) by fourier.
assert (Hcon: exists N, forall n, (n>=N)%nat ->R_dist (un n) l <l-t) by
apply (H0 (l-t) H2).
elim Hcon; intros N Hlim.
assert (Hdist:R_dist (un N) l<l-t).
apply (Hlim N); omega.
unfold R_dist in Hdist.
assert (Ha: un N <= t) by apply (H N).
rewrite Rabs_minus_sym in Hdist.
rewrite Rabs_right in Hdist; [|fourier].
assert (un N>t); fourier.
Qed.

Lemma lim_conv_low:
forall (un:nat->R) (a l:R),
(forall n, a<=un n)->
Un_cv un l ->
a<=l.
Proof.
intros un t l H H0.
cut (~(l<t)).
intros H1.
apply Rnot_gt_le; auto.
unfold not; intros.
unfold Un_cv in H0.
assert (t-l>0).
 fourier.
assert (Hcon: exists N, forall n, (n>=N)%nat ->R_dist (un n) l <t-l).
apply (H0 (t-l) H2).
elim Hcon; intros N Hlim.
assert (Hdist:R_dist (un N) l<t-l).
apply (Hlim N).
omega.
unfold R_dist in Hdist.
assert (Ha: t<=un N).
apply (H N).
rewrite Rabs_right in Hdist; [|fourier].
assert (un N<t); fourier.
Qed.

Lemma lim_in_interv:
forall (un:nat->R) (a b l:R),
(forall n, a<=un n <=b)->
Un_cv un l ->
a<=l<=b.
Proof.
intros un t y l H H0.
split.
apply lim_conv_low with un; auto.
intros n.
destruct (H n); assumption.
apply lim_conv_up with un; auto.
intros n.
destruct (H n); assumption.
Qed.

Lemma lim_seq_monot:
forall (xn yn:nat->R) (x y:R), 
(forall n, xn n<= yn n)->
(Un_cv xn x)->
(Un_cv yn y)->
x<=y.
intros.
assert (0<=y-x).
assert (forall n, 0<=yn n- xn n).
intros.
elim (H n);intros; fourier.
assert (Un_cv (fun i:nat => yn i - xn i) (y - x)).
apply CV_minus; assumption.
apply lim_conv_low with (fun  i:nat => yn i - xn i);
assumption.
fourier.
Qed.

Lemma cont_lim_0:
forall (f:R->R) (yn xn:nat->R) (a b y:R),
(forall t:R, c_I a b t -> continuity_pt f t)->
(Un_cv xn 0)->
(forall n, a<=yn n<=b)->
(Un_cv yn y)->
(Un_cv (fun i:nat=>(f (yn i))*xn i) 0).
intros f0 yn xn a1 b1 y H H0 H1 H2.
assert (H4: Un_cv (fun i:nat=> f0 (yn i)) (f0 y)).
apply (Heine1 ).
assert (H3:c_I a1 b1 y).
unfold c_I.
apply lim_in_interv with yn; auto.
apply (H y H3).
assumption.
replace 0 with (f0 y*0) by ring.
apply CV_mult;assumption.
Qed.

Lemma same_lim:
forall (xn:nat->R) (x:R),
(Un_cv xn x)->
(Un_cv (fun i:nat => xn (S i)) x).
intros.
unfold Un_cv in H |-*.
intros.
elim (H eps0 H0); intros N Hxn.
exists (N-1)%nat.
intros.
apply (Hxn (S n)).
omega.
Qed.

Lemma lim_minus_0:
forall (xn:nat->R) (x:R),
(Un_cv xn x)->
(Un_cv (fun i:nat => xn i - xn (S i)) 0).
intros.
replace 0 with (x-x) by ring.
apply CV_minus.
assumption.
apply same_lim;assumption.
Qed.

Lemma lim_prod_minus:
forall (f:R->R) (yn:nat->R) (a b y:R),
(forall t:R, c_I a b t -> continuity_pt f t)->
(forall n, a<=yn n<=b)->
(Un_cv yn y)->
(Un_cv (fun i:nat=>(f (yn i))*(yn i - yn (S i))) 0).
intros.
apply (cont_lim_0 f0 yn (fun i:nat => yn i - yn (S i)) a1 b1 y); auto.
apply lim_minus_0 with y; assumption.
Qed.

(*more proofs for sequences*)
Lemma Cauchy_crit2:
forall Un,
Cauchy_crit Un <->
(forall eps0:R,
eps0>0->
exists N:nat,
(forall n p:nat,
(n>=N)%nat->R_dist (Un n) (Un (n+p)%nat) < eps0)).
intros.
split; intros.
unfold Cauchy_crit in H.
elim (H eps0 H0); intros N Hchy.
exists N.
intros.
apply (Hchy n (n+p)%nat).
assumption.
omega.
unfold Cauchy_crit.
intros.
elim (H eps0 H0); intros N Hcrit.
exists N.
intros.
elim (le_or_lt n m); intros.
replace m with (n+(m-n))%nat by omega.
apply (Hcrit n (m-n)%nat).
assumption.
rewrite R_dist_sym.
replace n with (m+(n-m))%nat by omega.
apply (Hcrit m (n-m)%nat).
assumption.
Qed.

Lemma transf_en:
forall n:nat,
En n = eps/2^n.
Proof.
induction n.
simpl;field.
simpl En.
rewrite IHn.
simpl; field.
apply Rgt_not_eq.
apply pow2_diff0.
Qed.

Lemma pow2_monot:
forall (n :nat), eps/2^(S n)<=eps/2^n.
Proof.
assert (He:= eps_pos).
induction n .
simpl.
assert (He1: 0<=eps).
replace eps with (En 0); [|simpl; trivial];
apply He.
fourier.
simpl in IHn |-*.
replace (eps/(2*(2*2^n))) with (1/2*(eps/(2*2^n))) ;[|field; auto;apply Rgt_not_eq; apply pow2_diff0].
pattern (eps/(2*2^n)) at 2;replace (eps/(2*2^n)) with (1/2*(eps/2^n));[| field; apply Rgt_not_eq; apply pow2_diff0].
apply (Rmult_le_compat_l (1/2) (eps/(2*2^n)) (eps/2^n)).
fourier.
assumption.
Qed.

Lemma pow2_monot2:
forall (n p:nat), 
eps/2^(n+p)<=eps/2^n.
intros n p.
induction p.
replace (n+0)%nat with n by omega.
unfold Rle; right; trivial.
replace (n+S p )%nat with (S (n+p)%nat) by omega.
assert (IH2: eps/2^S (n+p)%nat<=eps/2^(n+p)%nat).
apply pow2_monot.
apply Rle_trans with (eps/2^(n+p)%nat); assumption.
Qed.

Lemma pow2_monot3:
forall (n N:nat), (n>=N)%nat->
eps/2^n<=eps/2^N.
intros.
replace n with (N+(n-N))%nat by omega.
apply (pow2_monot2 N (n-N)%nat).
Qed.

Lemma cauch_seq:
forall (Un:nat->R) ,
(forall (n:nat), (n<>0)%nat->included (c_disc (Un n) (En n)) 
(c_disc (Un (pred n)) (En (pred n))))
->
Cauchy_crit Un.
intros Un H.
rewrite Cauchy_crit2.
intros eps0 H0.
assert (Hincl0:forall n p, (c_disc (Un n) (En n) (Un (n+p)%nat))).
apply disc_incl_pct.
intros n.
apply (H (S n)).
omega.
assert (Harch: {n:nat| (eps/(2^n) < eps0)}).
apply archim2_le; [fourier|unfold eps; apply aux1;[fourier|auto]].
elim Harch; intros N Heps.
exists N; intros n p H1.
assert (Hnp:c_disc (Un n) (En n) (Un (n+p)%nat)).
apply (Hincl0 n p).
rewrite transf_en in Hnp.
unfold c_disc in Hnp.
assert (eps/2^n<=eps/2^N).
apply (pow2_monot3 n N); assumption.
unfold R_dist.
assert (eps/2^n<eps0).
unfold Rle in H2; elim H2; intros H3.
apply Rlt_trans with (eps/2^N); assumption.
rewrite H3; assumption.
unfold Rle in Hnp; elim Hnp; intros.
apply Rlt_trans with (eps/2^n); assumption.
rewrite H4; assumption.
Qed.

Lemma aux_bn_cv:
forall n,
Bn (S n) <= b0/2^n.
Proof.
induction n.
simpl.
unfold Rdiv.
rewrite Rinv_1.
replace (a0*b0*b0*c) with (b0*(a0*b0*c)) by ring.
apply Rmult_le_compat_l.
auto;fourier.
apply Rle_trans with (2*(a0*b0*c)).
pattern (a0*b0*c) at 1; replace (a0*b0*c) with (1*(a0*b0*c)) by ring.
apply Rmult_le_compat_r.
repeat apply aux1 ; auto. 
fourier.
replace (2*(a0*b0*c)) with (2*a0*b0*c) by ring.
assumption.
replace (Bn (S (S n))) with (Bn (S n)*(An (S n)*Bn (S n)*c)); [|simpl;ring].
replace (2^(S n)) with (2*2^n).
replace (b0/(2*2^n)) with ((b0/2^n)*(1/2)).
apply Rmult_le_compat.
apply Bn_ge_0.
replace (An (S n)*Bn (S n)*c) with (1/2*(2*An (S n)*Bn (S n)*c)) by field.
replace 0 with (1/2*0) by field .
apply Rmult_le_compat_l.
fourier.
apply An_bn_c0.
assumption.
apply Rmult_le_reg_l with 2.
fourier.
repeat rewrite <-Rmult_assoc.
replace (2*(1/2)) with 1 by field.
apply An_bn_c1.
unfold Rdiv; field.
apply Rgt_not_eq; apply pow2_diff0.
simpl.
trivial.
Qed.

Lemma Bn_monot:
forall n, Bn (S n)<=Bn n.
Proof.
induction n.
simpl.
replace (a0*b0*b0*c) with (b0*(a0*b0*c)) by ring.
pattern b0 at 3; replace b0 with (b0*1) by ring.
apply Rmult_le_compat_l ; [auto|].
apply Rle_trans with (2*(a0*b0*c)).
pattern (a0*b0*c) at 1; replace (a0*b0*c) with (1*(a0*b0*c)) by ring.
apply Rmult_le_compat_r.
repeat apply aux1; auto.
fourier.
replace (2*(a0*b0*c)) with (2*a0*b0*c) by ring.
assumption.
replace (Bn (S (S n))) with (Bn (S n)*(An (S n)*Bn (S n)*c)); [|simpl;ring].
pattern (Bn (S n)) at 3; replace (Bn (S n)) with (Bn (S n)*1) by ring.
apply Rmult_le_compat_l.
apply Bn_ge_0.
apply Rle_trans with (2*(An (S n)*Bn (S n)*c)).
pattern (An (S n)*Bn (S n)*c) at 1; replace (An (S n)*Bn (S n)*c) with (1*(An (S n)*Bn (S n)*c)) by ring.
apply Rmult_le_compat_r; [|fourier].
do 2 (apply aux1; auto).
apply Bn_ge_0.
replace (2*(An (S n)*Bn (S n)*c)) with (2*An (S n)*Bn (S n)*c) by ring.
apply An_bn_c1.
Qed.

Lemma Bn_monot2:
forall (n p:nat), Bn (n+(S p))%nat <= Bn n.
Proof.
intros n p.
induction p.
replace (n+1)%nat with (S n) by omega.
apply Bn_monot.
replace (n+S (S p) )%nat with (S (n+(S p))%nat) by omega.
apply Rle_trans with (Bn (n+(S p)))%nat.
apply Bn_monot.
assumption.
Qed.

Lemma Bn_monot3:
forall (n p:nat), ( (S p)<=n)%nat-> Bn n<=Bn  p.
Proof.
intros n p Hlt.
replace n with (p+S(n- (S p)))%nat by omega.
apply Bn_monot2.
Qed.


Lemma Bn_conv_0:
Un_cv Bn 0.
Proof.
red.
intros.
assert (Hb0:  {n:nat| (b0/(2^n) < eps0)}).
apply archim2_le;[  fourier| auto].
elim Hb0; intros N Hlt.
exists (S (S N)).
intros n Hineq.
apply Rle_lt_trans with (b0/2^N).
unfold R_dist.
replace (Bn n - 0) with (Bn n) by ring.
rewrite Rabs_pos_eq; [| apply Bn_ge_0].
apply Rle_trans with (Bn (S N)).
apply (Bn_monot3 n (S N)).
omega.
apply aux_bn_cv.
assumption.
Qed.

(*lemmas needed to prove the troublesome inequality*)

Lemma derive_decreasing_interv_var :
  forall (a b:R) (f:R -> R) (pr:forall t, a <= t <= b -> derivable_pt f t),
    (forall (t:R)(H:a <= t <= b), derive_pt f t (pr t H)<=0) ->
    forall x y:R, a <= x <= b -> a <= y <= b -> x < y -> f x >= f y.
intros a1 b1 f1 pr0 Hd x y Hinx Hiny Hxy.
assert (Hin1 : forall t, x < t < y -> a1 <= t <= b1).
intros t Hint; split; destruct Hint; destruct Hinx; destruct Hiny; fourier.
pose (pr1:= fun t (H : x < t < y) => pr0 t (Hin1 t H)).
assert (Hmvt := MVT f1 id x y pr1 (fun t h => derivable_pt_id t)).
destruct Hmvt as [z [Hin Heq]]; auto.
intros t Hint; apply derivable_continuous_pt; apply pr0.
split; destruct Hint; destruct Hinx; destruct Hiny; fourier.
intros; apply derivable_continuous_pt; apply derivable_id.
rewrite derive_pt_id in Heq.
unfold id in Heq; rewrite Rmult_1_r in Heq.
assert (Hfy : f1 y - f1 x <= 0).
rewrite <- Heq.
assert ( 0 <= - ((y - x)*derive_pt f1 z (pr1 z Hin))).
rewrite <- Ropp_mult_distr_r_reverse.
apply Rmult_le_pos.
fourier.
assert (derive_pt f1 z (pr1 z Hin)<= 0).
apply (Hd z (Hin1 z Hin)).
fourier.
fourier.
fourier.
Qed.

Lemma derive_increasing_interv_var :
  forall (a b:R) (f:R -> R) (pr:forall t, a <= t <= b -> derivable_pt f t),
    (forall (t:R)(H:a <= t <= b), 0<=derive_pt f t (pr t H)) ->
    forall x y:R, a <= x <= b -> a <= y <= b -> x < y -> f x <= f y.
intros a1 b1 f1 pr0 Hd x y Hinx Hiny Hxy.
assert (Hin1 : forall t, x < t < y -> a1 <= t <= b1).
intros t Hint; split; destruct Hint; destruct Hinx; destruct Hiny; fourier.
pose (pr1:= fun t (H : x < t < y) => pr0 t (Hin1 t H)).
assert (Hmvt := MVT f1 id x y pr1 (fun t h => derivable_pt_id t)).
destruct Hmvt as [z [Hin Heq]]; auto.
intros t Hint; apply derivable_continuous_pt; apply pr0.
split; destruct Hint; destruct Hinx; destruct Hiny; fourier.
intros; apply derivable_continuous_pt; apply derivable_id.
rewrite derive_pt_id in Heq.
unfold id in Heq; rewrite Rmult_1_r in Heq.
assert (Hfy : 0<=f1 y - f1 x ).
rewrite <- Heq.
assert (0<=derive_pt f1 z (pr1 z Hin)).
apply (Hd z (Hin1 z Hin)).
apply Rmult_le_pos.
fourier.
assumption.
fourier.
Qed.


Lemma derive_decreasing_interv_var2 :
  forall (a b:R) (f:R -> R) (pr:forall t, c_I a b t -> derivable_pt f t),
    (forall (t:R)(H: c_I a b t), derive_pt f t (pr t H)<=0) ->
    forall x y:R, c_I a b x -> c_I a b y -> x < y -> f x >= f y.
Proof.
intros a1 b1 f1 pr1 Hder x y Hinx Hiny Hineq.
assert (pr':forall t:R, a1<= t<=b1-> derivable_pt f1 t).
intros t Hab.
fold (c_I a1 b1 t) in Hab.
apply (pr1 t Hab).
apply derive_decreasing_interv_var with a1 b1 pr'.
intros t H.
fold (c_I a1 b1 t) in H.
assert (Heq: derive_pt f1 t (pr1 t H) = derive_pt f1 t (pr' t H)) by apply pr_nu.
rewrite <-Heq.
apply (Hder t H).
unfold c_I in Hinx; exact Hinx.
unfold c_I in Hiny; exact Hiny.
assumption.
Qed.

Lemma derive_increasing_interv_var2 :
  forall (a b:R) (f:R -> R) (pr:forall t, c_I a b t -> derivable_pt f t),
    (forall (t:R)(H: c_I a b t), 0<=derive_pt f t (pr t H)) ->
    forall x y:R, c_I a b x -> c_I a b y -> x < y -> f x <= f y.
Proof.
intros a1 b1 f1 pr1 Hder x y Hinx Hiny Hineq.
assert (pr':forall t:R, a1<= t<=b1-> derivable_pt f1 t).
intros t Hab.
fold (c_I a1 b1 t) in Hab.
apply (pr1 t Hab).
apply derive_increasing_interv_var with a1 b1 pr'.
intros t H.
fold (c_I a1 b1 t) in H.
assert (Heq: derive_pt f1 t (pr1 t H) = derive_pt f1 t (pr' t H)) by apply pr_nu.
rewrite <-Heq.
apply (Hder t H).
unfold c_I in Hinx; exact Hinx.
unfold c_I in Hiny; exact Hiny.
assumption.
Qed.

Lemma equiv_cond_f':
(forall (y1 y2:R) (H1: c_I a b y1) (H2: c_I a b y2),
y2<y1->
f' y1 - f' y2 <= c * (y1-y2)).
Proof.
intros y1 y2 H1 H2 Hineq.
assert (Habs: Rabs (f' y1 -f' y2)<=c*Rabs(y1-y2)) by apply (Hlim_f' y1 y2 H1 H2).
rewrite (Rabs_right (y1-y2)) in Habs; [|fourier].

rewrite Rabs_def2_le in Habs.
elim Habs; intros Hmin Hplus; assumption.
Qed.

Lemma equiv_cond_f'2:
(forall (y1 y2:R) (H1: c_I a b y1) (H2: c_I a b y2),
y2<y1->
-(c*(y1-y2))<=f' y1 - f' y2 ).
Proof.
intros y1 y2 H1 H2 Hineq.
assert (Habs: Rabs (f' y1 -f' y2)<=c*Rabs(y1-y2)) by apply (Hlim_f' y1 y2 H1 H2).
rewrite (Rabs_right (y1-y2)) in Habs; [|fourier].

rewrite Rabs_def2_le in Habs.
elim Habs; intros Hmin Hplus; assumption.
Qed.


Lemma Hxinb:
forall x x0,
c_I a b x0->
c_I x0 b x ->
c_I a b x. 
Proof.
intros x x0 [Ha0 Hb0] [Hx0 Hb].
split; fourier.
Qed.

Lemma Hxina:
forall x x0,
c_I a b x0->
c_I a x0 x ->
c_I a b x. 
Proof.
intros x x0 [Ha0 Hb0] [Hx0 Hb].
split; fourier.
Qed.

Definition g_fun (x0 x:R):R:=
f x -  f x0 - (f' x0)*x+(f' x0)*x0 - 1/2*c*Rsqr(x-x0).

Lemma pr_gfun:
forall (x0 x:R) (Hx0:c_I a b x0) (Hx: c_I a b x) ,
derivable_pt (g_fun x0) x.
intros.
unfold g_fun.
reg.
apply (pr x Hx).
Qed.

Lemma der_gfun:
forall (x0 x:R) (H0:c_I a b x0) (H:c_I a b x),
derive_pt (g_fun x0) x (pr_gfun x0 _ H0 H) = 
derive_pt f x (pr x H) - f' x0 - c*(x-x0).
intros.
unfold g_fun.
reg.
assert(Huni:derive_pt f x H1 = derive_pt f x (pr x H)).
apply pr_nu.
rewrite Huni.
field.
apply (pr x H).
Qed. 

Lemma gfun_x0_is_0:
forall  x0 ,
g_fun  x0 x0 = 0.
Proof.
intros.
unfold g_fun.
unfold Rsqr.
field.
Qed.

Lemma der_g_pos_neg:
forall (x0:R) (Hx0:c_I a b x0),
(forall (x:R) (Hxb: c_I x0 b x),
derive_pt (g_fun x0) x (pr_gfun x0 _ Hx0 (Hxinb x x0 Hx0 Hxb)) <=0) /\
(forall (x:R) (Hxa: c_I a x0 x),
0<=derive_pt (g_fun x0) x (pr_gfun x0 _ Hx0 (Hxina x x0 Hx0 Hxa))).
Proof.

intros x0 Hx0.
assert (Hf': (forall (y1 y2:R) (H1: c_I a b y1) (H2: c_I a b y2),
y2<y1->
f' y1 - f' y2 <= c * (y1-y2))) by apply equiv_cond_f'.
split; intros.
rewrite der_gfun.
assert (H: a<=x<=b) by apply (Hxinb x x0 Hx0 Hxb).
assert (Heqff': derive_pt f x (pr x H) = f' x) by apply (Hder_f x H).
assert (Heq: derive_pt f x (pr x (Hxinb x x0 Hx0 Hxb)) = derive_pt f x (pr x H)) by apply pr_nu.
rewrite Heq.
rewrite Heqff'.
elim Hxb; intros Hexx0 _.
elim Hexx0; intro H1.
assert (f' x - f' x0 <= c*(x-x0)).
apply (Hf' x x0 H Hx0 H1).
fourier.
rewrite H1.
replace (f' x - f' x -c*(x-x)) with 0 by ring; fourier.
rewrite der_gfun.
assert (H: a<=x<=b) by apply (Hxina x x0 Hx0 Hxa).
assert (Heqff': derive_pt f x (pr x H) = f' x) by apply (Hder_f x H).
assert (Heq: derive_pt f x (pr x (Hxina x x0 Hx0 Hxa)) = derive_pt f x (pr x H)) by apply pr_nu.
rewrite Heq.
rewrite Heqff'.
elim Hxa; intros _ Hexx0 .
elim Hexx0; intro H1.
assert ( f' x0 - f' x <=c*(x0-x)).
apply (Hf' x0 x Hx0 H H1).
replace (c*(x0-x)) with (-(c*(x-x0))) in H0 by ring.
fourier.
rewrite <-H1.
replace (f' x - f' x -c*(x-x)) with 0 by ring; fourier.
Qed.

Lemma rle_contr_rew:
forall r, 0<=-r <-> r<=0.
Proof.
intros.
split; intros; fourier.
Qed.

Lemma g_monot_int:
forall (x0:R) (Hx0:c_I a b x0),
(forall x y:R, c_I x0 b x -> c_I x0 b y -> x <= y -> 
(g_fun x0) x >= (g_fun  x0) y)/\
(forall x y:R, c_I a x0 x -> c_I a x0 y -> x <= y -> 
(g_fun x0) x <= (g_fun x0) y).
Proof.
intros x0 Hx0.
assert (Hder: (forall (x:R) (Hxb: c_I x0 b x),
derive_pt (g_fun x0) x (pr_gfun x0 _ Hx0 (Hxinb x x0 Hx0 Hxb)) <=0) /\
(forall (x:R) (Hxa: c_I a x0 x),
0<=derive_pt (g_fun x0) x (pr_gfun x0 _ Hx0  (Hxina x x0 Hx0 Hxa)))) 
   by apply der_g_pos_neg.
elim Hder; intros Hderb Hdera.
split; intros x y Hcx Hcy Hineq.
unfold Rle in Hineq; elim Hineq; intros.
assert (prg: forall t, c_I x0 b t -> derivable_pt (g_fun x0) t).
intros t Hx0b.
apply (pr_gfun x0 t Hx0 (Hxinb t x0 Hx0 Hx0b)).
apply derive_decreasing_interv_var2 with x0 b prg.
intros t H0.
assert (Heq: derive_pt (g_fun x0) t (pr_gfun x0 t Hx0 (Hxinb t x0 Hx0 H0))
 = derive_pt (g_fun x0) t (prg t H0)) by apply pr_nu.
rewrite <-Heq.
apply (Hderb t H0).
assumption.
assumption.
assumption.
rewrite H.
unfold Rle; right; trivial.
unfold Rle in Hineq; elim Hineq; intros.
assert (prg: forall t, c_I a x0 t -> derivable_pt (g_fun x0) t).
intros t Hax0.
apply (pr_gfun x0 t Hx0 (Hxina t x0 Hx0 Hax0)).
apply derive_increasing_interv_var2 with a x0 prg.
intros t H0.
assert (Heq: derive_pt (g_fun x0) t (pr_gfun x0 t Hx0 (Hxina t x0 Hx0 H0))
 = derive_pt (g_fun x0) t (prg t H0)) by apply pr_nu.
rewrite <-Heq.
apply (Hdera t H0).
assumption.
assumption.
assumption.
rewrite H.
unfold Rle; right; trivial.
Qed.

Lemma g_less_0:
forall (x0:R) (Hx0:c_I a b x0),
(forall x, c_I a b x ->g_fun x0 x<=0).
Proof.
intros x0 Hx0 x Hab.
assert (Hmon: (forall x y:R, c_I x0 b x -> c_I x0 b y -> x <= y -> 
(g_fun x0) x >= (g_fun  x0) y)/\
(forall x y:R, c_I a x0 x -> c_I a x0 y -> x <= y -> 
(g_fun x0) x <= (g_fun x0) y)) by apply (g_monot_int x0 Hx0).
elim Hmon; intros Hdes Hcres.
elim (Rtotal_order x x0); intros Horder.
rewrite <-(gfun_x0_is_0 x0).
unfold c_I in Hx0, Hab |-*.
elim Hx0; intros.
elim Hab; intros.
apply (Hcres x x0).
split; fourier.
split; fourier.
fourier.
elim Horder; intros Ho1.
unfold Rle; right.
rewrite Ho1.
apply (gfun_x0_is_0 x0).
rewrite <-(gfun_x0_is_0 x0).
unfold c_I in Hx0, Hab |-*.
elim Hx0; intros.
elim Hab; intros.
assert (Haux: g_fun x0 x0>=g_fun x0 x).
apply (Hdes x0 x).
split; fourier.
split; fourier.
fourier.
fourier.
Qed.

Definition g_fun2 (x0 x:R):R:=
f x -  f x0 - (f' x0)*x+(f' x0)*x0 + 1/2*c*Rsqr(x-x0).

Lemma pr_gfun2:
forall (x0 x:R) (Hx0:c_I a b x0) (Hx: c_I a b x) ,
derivable_pt (g_fun2 x0) x.
intros.
unfold g_fun2.
reg.
apply (pr x Hx).
Qed.

Lemma der_gfun2:
forall (x0 x:R) (H0:c_I a b x0) (H:c_I a b x),
derive_pt (g_fun2 x0) x (pr_gfun2 x0 _ H0 H) = 
derive_pt f x (pr x H) - f' x0 + c*(x-x0).
intros.
unfold g_fun2.
reg.
assert(Huni:derive_pt f x H1 = derive_pt f x (pr x H)).
apply pr_nu.
rewrite Huni.
field.
apply (pr x H).
Qed. 

Lemma gfun2_x0_is_0:
forall  x0 ,
g_fun2  x0 x0 = 0.
Proof.
intros.
unfold g_fun2.
unfold Rsqr.
field.
Qed.

Lemma der_g_pos_neg2:
forall (x0:R) (Hx0:c_I a b x0),
(forall (x:R) (Hxb: c_I x0 b x),
0<=derive_pt (g_fun2 x0) x (pr_gfun2 x0 _ Hx0 (Hxinb x x0 Hx0 Hxb))) /\
(forall (x:R) (Hxa: c_I a x0 x),
derive_pt (g_fun2 x0) x (pr_gfun2 x0 _ Hx0 (Hxina x x0 Hx0 Hxa))<=0).
Proof.

intros x0 Hx0.
assert (Hf'2: (forall (y1 y2:R) (H1: c_I a b y1) (H2: c_I a b y2),
y2<y1->
-(c * (y1-y2)) <= f' y1 - f' y2)) by apply equiv_cond_f'2.
split; intros.
rewrite der_gfun2.
assert (H: a<=x<=b) by apply (Hxinb x x0 Hx0 Hxb).
assert (Heqff': derive_pt f x (pr x H) = f' x) by apply (Hder_f x H).
assert (Heq: derive_pt f x (pr x (Hxinb x x0 Hx0 Hxb)) = derive_pt f x (pr x H)) by apply pr_nu.
rewrite Heq.
rewrite Heqff'.


elim Hxb; intros Hexx0 _.
elim Hexx0; intro H1.
assert ( - (c*(x-x0))<=f' x - f' x0).
apply (Hf'2 x x0 H Hx0 H1).
fourier.
rewrite H1.
replace (f' x - f' x +c*(x-x)) with 0 by ring; fourier.


rewrite der_gfun2.
assert (H: a<=x<=b) by apply (Hxina x x0 Hx0 Hxa).
assert (Heqff': derive_pt f x (pr x H) = f' x) by apply (Hder_f x H).
assert (Heq: derive_pt f x (pr x (Hxina x x0 Hx0 Hxa)) = derive_pt f x (pr x H)) by apply pr_nu.
rewrite Heq.
rewrite Heqff'.
elim Hxa; intros _ Hexx0 .
elim Hexx0; intro H1.
assert (-(c*(x0-x))<= f' x0 - f' x).
apply (Hf'2 x0 x Hx0 H H1).
replace (-(c*(x0-x))) with (c*(x-x0)) in H0 by ring.
fourier.
rewrite <-H1.
replace (f' x - f' x +c*(x-x)) with 0 by ring; fourier.
Qed.

Lemma g_monot_int2:
forall (x0:R) (Hx0:c_I a b x0),
(forall x y:R, c_I x0 b x -> c_I x0 b y -> x <= y -> 
(g_fun2 x0) x <= (g_fun2  x0) y)/\
(forall x y:R, c_I a x0 x -> c_I a x0 y -> x <= y -> 
(g_fun2 x0) x >= (g_fun2 x0) y).
Proof.
intros x0 Hx0.
assert (Hder: (forall (x:R) (Hxb: c_I x0 b x),
0<=derive_pt (g_fun2 x0) x (pr_gfun2 x0 _ Hx0 (Hxinb x x0 Hx0 Hxb)) ) /\
(forall (x:R) (Hxa: c_I a x0 x),
derive_pt (g_fun2 x0) x (pr_gfun2 x0 _ Hx0  (Hxina x x0 Hx0 Hxa))<=0)) 
   by apply der_g_pos_neg2.
elim Hder; intros Hderb Hdera.
split; intros x y Hcx Hcy Hineq.
unfold Rle in Hineq; elim Hineq; intros.
assert (prg: forall t, c_I x0 b t -> derivable_pt (g_fun2 x0) t).
intros t Hx0b.
apply (pr_gfun2 x0 t Hx0 (Hxinb t x0 Hx0 Hx0b)).
apply derive_increasing_interv_var2 with x0 b prg.
intros t H0.
assert (Heq: derive_pt (g_fun2 x0) t (pr_gfun2 x0 t Hx0 (Hxinb t x0 Hx0 H0))
 = derive_pt (g_fun2 x0) t (prg t H0)) by apply pr_nu.
rewrite <-Heq.
apply (Hderb t H0).
assumption.
assumption.
assumption.
rewrite H.
unfold Rle; right; trivial.
unfold Rle in Hineq; elim Hineq; intros.
assert (prg: forall t, c_I a x0 t -> derivable_pt (g_fun2 x0) t).
intros t Hax0.
apply (pr_gfun2 x0 t Hx0 (Hxina t x0 Hx0 Hax0)).
apply derive_decreasing_interv_var2 with a x0 prg.
intros t H0.
assert (Heq: derive_pt (g_fun2 x0) t (pr_gfun2 x0 t Hx0 (Hxina t x0 Hx0 H0))
 = derive_pt (g_fun2 x0) t (prg t H0)) by apply pr_nu.
rewrite <-Heq.
apply (Hdera t H0).
assumption.
assumption.
assumption.
rewrite H.
unfold Rle; right; trivial.
Qed.

Lemma g2_greater_0:
forall (x0:R) (Hx0:c_I a b x0),
(forall x, c_I a b x ->0<=g_fun2 x0 x).
Proof.
intros x0 Hx0 x Hab.
assert (Hmon: (forall x y:R, c_I x0 b x -> c_I x0 b y -> x <= y -> 
(g_fun2 x0) x <= (g_fun2  x0) y)/\
(forall x y:R, c_I a x0 x -> c_I a x0 y -> x <= y -> 
(g_fun2 x0) x >= (g_fun2 x0) y)) by apply (g_monot_int2 x0 Hx0).
elim Hmon; intros Hdes Hcres.
elim (Rtotal_order x x0); intros Horder.
rewrite <-(gfun2_x0_is_0 x0).
unfold c_I in Hx0, Hab |-*.
elim Hx0; intros.
elim Hab; intros.
assert (Haux: g_fun2 x0 x>=g_fun2 x0 x0).
apply (Hcres x x0).
split; fourier.
split; fourier.
fourier.
fourier.
elim Horder; intros Ho1.
unfold Rle; right.
rewrite Ho1.
rewrite (gfun2_x0_is_0 x0); trivial.
unfold c_I in Hx0, Hab |-*.
elim Hx0; intros.
elim Hab; intros.
rewrite <-(gfun2_x0_is_0 x0).
apply (Hdes x0 x).
split; fourier.
split; fourier.
fourier.
Qed.

Lemma lemma_sec:
(forall x y, c_I a b x ->c_I a b y->
-(1/2*c*Rsqr(x-y))<=f x -  f y - (f' y)*x+(f' y)*y <=1/2*c*Rsqr(x-y)).
Proof.
intros x y Hxin Hyin.
split.
assert (Hg: 0<=f x - f y - f' y * x + f' y * y +1/2 * c * Rsqr (x-y)).
fold (g_fun2 y x). 
apply g2_greater_0 ; try assumption.
fourier.
assert (Hg: f x - f y - f' y * x + f' y * y -1/2 * c * Rsqr (x-y)<=0).
fold (g_fun y x). 
apply g_less_0; try assumption.
fourier.
Qed.

Lemma lemma_third:
(forall x y, c_I a b x ->c_I a b y->
Rabs (f x -  f y - (f' y)*x+(f' y)*y )<=1/2*c*Rsqr(x-y)).
Proof.
intros.
assert (Hsec:-(1/2*c*Rsqr(x-y))<=f x -  f y - (f' y)*x+(f' y)*y <=1/2*c*Rsqr(x-y)).
apply lemma_sec; assumption.
rewrite Rabs_def2_le; assumption.
Qed.

Lemma Rinv_Rmult_le:
forall (r1 r2 r3:R), 0<r1->0<r2->0<r3->
/r1*r2>=/r3->r1*/r2<=r3.
Proof.
intros r1 r2 r3 Hi1 Hi2 Hi3 Hineq.
replace (r1*/r2) with (/(/r1*r2)) ;[|field; auto].
elim Hineq; intro H'.
unfold Rle ; left.
replace r3 with (//r3); [|apply Rinv_involutive; apply Rgt_not_eq; auto].
apply Rinv_lt_contravar.
repeat apply Rmult_lt_0_compat; auto with real.
unfold Rgt in H'.
assumption.
unfold Rle; right.
rewrite H'; auto with real.
split; apply Rgt_not_eq; assumption.
Qed.


Definition mu n := 2*An n*Bn n*c.

Lemma rel_mu:
forall n, mu (S n) = (mu n)^2.
Proof.
induction n; unfold mu;  simpl; ring.
Qed.

Fixpoint Npow (r:nat) (n:nat) {struct n} : nat :=
  match n with
    | O => 1%nat
    | S n =>  (r  * (Npow r n))%nat
  end.




Lemma rel_mu0_aux:
forall p, mu p =  (mu 0) ^ (Npow 2 p).
Proof.
induction p; [unfold mu; simpl; ring|].
rewrite rel_mu; rewrite IHp ;
rewrite <- pow_mult.
assert (Heq: (Npow 2 p * 2)%nat = Npow 2 (S p)).
  simpl; omega.
rewrite Heq; trivial.
Qed.

Lemma eq_pow:
forall r n, INR (Npow r n)  = (INR r) ^ n.
Proof.
intros r n.
induction n; simpl; [trivial
   |rewrite mult_INR; rewrite IHn; trivial].
Qed.

Lemma Npow_mon1:
forall (x n:nat) , (1 <= x)%nat -> (Npow x  n <= Npow x (S n))%nat.
Proof.
intros x n Hx; induction n; [simpl; omega|].
simpl in IHn |-*.
apply mult_le_compat_l; assumption .
Qed.



Lemma Npow_monot:
forall x m n, (1 <= x)%nat -> 
(Npow x (m) <= Npow x (m+n))%nat.
Proof.
intros x m n Hx .
induction n.
replace (m+0)%nat with m;omega.
apply le_trans with (Npow x (m+n)).
trivial.
replace (m+(S n))%nat 
  with (S (m+n)) by omega; apply Npow_mon1;trivial.
Qed.


Infix Local "^R" := Rpower (at level 30, right associativity) : R_scope.

Lemma rel_bn1:
forall n, Bn (S n) = /2*(mu 0)^(Npow 2 n)*Bn n.
Proof.
intro n; simpl;
rewrite <- rel_mu0_aux; unfold mu; field.
Qed.

Lemma rel_bn : 
forall n, Bn n = /2^n * (mu 0)^((Npow 2 n)-1) * b0 .
Proof.
induction n; [simpl;field|].
rewrite rel_bn1; rewrite IHn; simpl.
field_simplify.
rewrite <- pow_add; rewrite Rmult_comm.
assert (Heq: (Npow 2 n + (Npow 2 n - 1))%nat  =
             (Npow 2 n + (Npow 2 n + 0) -1)%nat) by omega.
rewrite Heq.
 trivial.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.

Lemma pow_le1:
forall x n, 0 <=x <=1 -> 0 <= x^n <= 1.
Proof.
intros x n [H0 H1].
induction n; simpl; [split; [fourier|right; trivial]| elim IHn; intros IH1 IH2].
split; [apply aux1; assumption|
replace 1 with (1*1) by ring; apply Rmult_le_compat; assumption].
Qed.


Lemma pow_le1_le:
forall x m n, 0 <= x <= 1 -> (0 <= n <= m)%nat -> x^m <=x^n.
Proof.
intros x m n [H0 H1] [Hm Hmn].
replace m with (m - n + n)%nat by omega.
  rewrite pow_add.
  pattern (x ^ n) at 2 in |- *; replace (x ^ n) with (1 * x ^ n);
    auto with real.
  apply Rminus_le.
  repeat rewrite (fun y:R => Rmult_comm y (x ^ n));
    rewrite <- Rmult_minus_distr_l.
  replace 0 with (x ^ n * 0); auto with real.
  apply Rmult_le_compat_l; auto with real.
   apply pow_le; auto with real.
apply Rle_minus.
exact (proj2 (pow_le1 x (m-n) (conj H0 H1))).
Qed.

Lemma Npow_gt1:
forall x n, (0 <x)%nat -> (1 <= Npow x n)%nat .
Proof.
intros x n Hx; induction n; simpl; [omega|].
apply le_trans with (Npow x n); [trivial|].
elim (mult_O_le (Npow x n) x) ; intros H0; [|trivial] .
  assert (H2:x<>O) by omega; elim H2; exact H0.
Qed.

Lemma minus_le_compat:
forall m n p, (m <= n)%nat -> (m-p <= n - p)%nat.
Proof.
intros m n p H.
elim H.
omega.
intros m0 H1 H2.
apply le_trans with (m0 - p)%nat; omega.
Qed.



Lemma Rpow_pos:
forall a x, 0 < a -> 0 < a ^R x.
Proof.
intros x y Ha; unfold Rpower; apply exp_pos.
Qed.


Lemma rel_bn2: forall n,
Bn n <= /2^n * b0.
Proof.
induction n.
simpl; fourier.
rewrite rel_bn1.
simpl .
replace (/ (2 * 2 ^ n) * b0) with (/2 * (/2^n * b0));
[| field; apply Rgt_not_eq; apply pow2_diff0].
assert (H2: 0 <= mu 0 ^ Npow 2 n <=1).
apply pow_le1; unfold mu; simpl; split; [apply greater0|trivial].
apply Rmult_le_compat.
apply aux1; [fourier| intuition].
apply Bn_ge_0.
pattern (/2) at 2; replace (/2) with (/2*1) by field.
apply Rmult_le_compat_l; [fourier|intuition].
assumption.
Qed.



Lemma leq_en_bn:
forall n, Bn n <= En n/2.
Proof.
induction n; [simpl; unfold eps; right; field|].
rewrite rel_bn1.
simpl En.
unfold Rdiv at 1; rewrite (Rmult_comm (En n/2)).
rewrite Rmult_assoc.
apply Rmult_le_compat_l; [fourier|].
apply Rle_trans with (Bn n); try fourier.
pattern (Bn n) at 2; replace (Bn n) with (1* Bn n) by ring.
apply Rmult_le_compat_r; [apply Bn_ge_0|].
assert (0<=mu 0^Npow 2 n <= 1); [|intuition].
apply pow_le1; unfold mu; simpl; split; [apply greater0|trivial].
Qed.


Theorem lemma_princ:
forall n:nat,
(*0<En n/\*)
(n<>0%nat-> included 
   (c_disc (Xn n) (En n))
   (c_disc (Xn (pred n)) (En (pred n)))) /\
included (c_disc (Xn n) (En n)) (c_I a b)/\
(c_I a b (Xn n))/\
f' (Xn n) <>0/\
Rabs (/(f' (Xn n))) <= An n/\
Rabs ((f (Xn n))/(f' (Xn n))) <= Bn n/\
Bn n<= (En n)/2.
Proof.
induction n.
split.
intro Hdif;elim Hdif; trivial.
split; simpl; try assumption.
split; [split; elim Hincl; intros; fourier|].
repeat (split; try assumption).
unfold eps; right; field.
(*induction step*)
destruct IHn as [IHincl1 [IHincl2 [IHincl3 [IHf' [IHa [IHb1 IHb2]]]]]].
assert (Hxn: (Xn n) - (Xn (S n)) = f (Xn n)/f' (Xn n)).
simpl; field;assumption. 
(*split.

simpl; auto.*)
assert (Haux: included (c_disc (Xn (S n)) (En (S n)))
(c_disc (Xn (pred (S n))) (En (pred (S n))))).
simpl (pred (S n)).
simpl (En (S n) ).
apply inc_neigh.
rewrite Hxn.
apply Rle_trans with (Bn n); auto.
split.
intro Hdif; assumption.
assert (Haux1: included (c_disc (Xn (S n)) (En (S n))) (c_I a b)).
simpl (pred (S n)) in Haux.
apply included_trans with (c_disc (Xn n) (En n)); auto.
split; auto.
assert (Haux2: c_I a b (Xn (S n))).
assert (c_disc (Xn (S n)) (En (S n)) (Xn (S n))).
unfold c_disc.
replace ((Xn (S n))- (Xn (S n))) with 0 by ring.
rewrite Rabs_R0; apply eps_pos. 
auto.
split; auto.
(*proving f'(xn)<>0)*)
assert (Haux3: Rabs (1- (f' (Xn (S n))/f' (Xn n))) <= 1/2).
replace (1- (f' (Xn (S n))/f' (Xn n))) with ((f' (Xn n)- (f' (Xn (S n))))/f' (Xn n)); [idtac|field;auto].
unfold Rdiv at 1.
rewrite Rabs_mult.
apply Rle_trans with (Rabs (f' (Xn n) - f' (Xn (S n))) * An n).
apply Rmult_le_compat_l ;[apply Rabs_pos|assumption].
apply Rle_trans with ((c*Rabs ((Xn n)-(Xn (S n))))*An n).
apply Rmult_le_compat_r.
apply An_ge_0.
apply (Hlim_f' (Xn n) (Xn (S n)) IHincl3 Haux2).
rewrite Hxn.
replace (c*Rabs(f (Xn n)/f' (Xn n)) *An n) with 
         (Rabs(f (Xn n)/f' (Xn n)) *An n*c) by ring.
apply Rle_trans with (Bn n*An n*c).
repeat apply Rmult_le_compat_r.
auto.
apply An_ge_0.
assumption.
replace (1/2) with (1/2*1) by field.
replace (Bn n*An n*c) with ((1/2)*(2*(An n)*(Bn n)*c)) by field.
apply (Rmult_le_compat_l (1/2) (2*(An n)*(Bn n)*c) 1) .
fourier.
apply An_bn_c1.
assert (Haux4: (f' (Xn (S n)))/(f' (Xn n)) <> 0).
apply Rgt_not_eq.
apply (dif_0 (f' (Xn (S n))/f' (Xn n)) Haux3).
unfold Rdiv in Haux4.
assert (Hf'dif:f' (Xn (S n))<>0).
destruct (Rmult_neq_0_reg2 (f' (Xn (S n))) (/f' (Xn n)) Haux4); assumption.
split. (*here i get f'<>0)*)
trivial.
(*done prouving f'(xn+1)<>0*)
(* goal |1/f'(xn+1)| <=an+1*)
assert (Ha: Rabs(/f' (Xn (S n))) <= An (S n)).
replace (/f' (Xn (S n))) with ((/f' (Xn n))*(f' (Xn n)/f' (Xn (S n)))); [|field; auto].
rewrite Rabs_mult.
apply Rle_trans with (An n*2).
apply Rmult_le_compat.
apply Rabs_pos.
apply Rabs_pos.
assumption.
assert (Haux51: 1-Rabs ((f' (Xn (S n)))/(f' (Xn n)))
                          <=Rabs (1- (f' (Xn (S n))/f' (Xn n)))).
pattern 1 at 1; rewrite <-Rabs_R1.
apply Rabs_triang_inv .
assert (Haux52: 1-Rabs ((f' (Xn (S n)))/(f' (Xn n)))<=1/2).
fourier.
assert (Haux53:Rabs ((f' (Xn (S n)))/(f' (Xn n))) >= 1/2).
fourier.
unfold Rdiv in Haux53 |-*.
rewrite Rabs_mult in Haux53|-*.
rewrite Rabs_Rinv in Haux53 |-*.
apply Rinv_Rmult_le.
apply Rabs_pos_lt; assumption.
apply Rabs_pos_lt; assumption.
fourier.
rewrite Rmult_comm in Haux53.
replace (1*/2) with (/2) in Haux53 by field.
assumption.
assumption.
assumption.
simpl; fourier.
(*for the goal*)
split;[assumption|].
(*the tricky ineq*)
assert (Haux5: Rabs (f (Xn (S n))) <= 1/2*c*Bn n*Bn n).
apply Rle_trans with (1/2*c*Rsqr (Xn (S n) - Xn n)).
replace (f (Xn (S n))) with 
( f (Xn (S n)) -f (Xn n) - f' (Xn n)*Xn (S n) + f'(Xn n)* Xn n) ;[| simpl;field; auto].
apply lemma_third; assumption.
replace (1/2*c*Bn n*Bn n) with (1/2*c*(Bn n*Bn n)) by ring. 
fold (Rsqr (Bn n)).
apply (Rmult_le_compat_l).
apply aux1; [fourier|auto].
rewrite Rsqr_neg_minus.
rewrite Hxn.
rewrite Rsqr_abs.
apply Rsqr_incr_1.
assumption.
apply Rabs_pos.
apply Bn_ge_0.
split.
unfold Rdiv.
rewrite Rabs_mult.
simpl Bn.
simpl An in Ha.
replace (An n*Bn n*Bn n*c) with (1/2*c*Bn n*Bn n*(2*An n)) by field.
apply  (Rmult_le_compat (Rabs (f(Xn (S n)))) (1/2*c*Bn n*Bn n) 
(Rabs (/f' (Xn (S n)))) (2*An n)).
apply Rabs_pos.
apply Rabs_pos.
assumption.
assumption.
simpl.
assert (2*An n*Bn n*c*Bn n<=En n/2).
replace (En n/2) with (1*(En n/2)) by field.
apply  (Rmult_le_compat (2*An n*Bn n*c) 1 (Bn n) (En n/2)).
apply An_bn_c0.
apply Bn_ge_0.
apply An_bn_c1.
assumption.
replace (An n*Bn n*Bn n*c) with (1/2*(2*An n* Bn n*c*Bn n)) by field.
replace (En n/2/2) with (1/2*(En n/2)) by field.
apply Rmult_le_compat_l.
fourier.
assumption.
(*done with my property on the sequence*)
Qed.

Lemma rabs_is_0:
forall r, Rabs r = 0 -> r = 0.
Proof.
intros r H.
case (Rcase_abs r ); intros Hc.
assert (j1:= Rlt_not_eq r 0 Hc).
elim ( Rabs_no_R0 r j1 H).
elim Hc; intros Hc1.
assert (j1:= Rgt_not_eq r 0 Hc1).
elim ( Rabs_no_R0 r j1 H).
assumption.
Qed.

Lemma in_open_xn:
forall n, (o_I a b (Xn n)).
Proof.
induction n.
unfold o_I; simpl; assumption. 
destruct (lemma_princ ( n)) as [H1[H2[H3[H4[H5[H6 H7]]]]]].
simpl pred in H1.
elim (Rle_dec (Bn ( n)) 0); intros Hbn.
assert (Hbn1:= Bn_ge_0 ( n)).
assert (Hbn3: Bn ( n) = 0) by (apply Rle_antisym;fourier).
rewrite Hbn3 in H6.
assert (H61: Rabs (f (Xn ( n)) / f' (Xn ( n))) = 0).
apply Rle_antisym;[fourier| apply Rabs_pos].
simpl Xn.
replace ((Xn n - f (Xn n) / f' (Xn n))) with (Xn n); try assumption.
assert (J1:= rabs_is_0 _ H61).
rewrite J1; ring.
assert (Hgt:=  Rnot_le_lt _ _ Hbn).
destruct (lemma_princ (S n)) as [H1s[H2s[H3s[H4s[H5s[H6s H7s]]]]]].
simpl pred in H1s.
assert (Hin2: included (c_disc (Xn (S n)) (Bn ( n))) (c_disc (Xn (S n)) (En (S n)))).
apply incl_c_disc.
simpl En; fourier.
assert (Hin3:= Hin2 (Xn (S n) - Bn ( n)) (edgel_in_disc _ _ (Bn_ge_0 ( n)))).
assert (Hin4:= H2s _ Hin3).
assert (Hin5:= Hin2 (Xn (S n) + Bn ( n)) (edger_in_disc _ _ (Bn_ge_0 ( n)))).
assert (Hin6:= H2s _ Hin5).

unfold c_I in Hin4; unfold c_I in Hin6.
elim Hin4; intros Hin41 Hin42.
elim Hin6; intros Hin61 Hin62.
split.
apply Rlt_le_trans with (a + Bn n); fourier.
apply Rle_lt_trans with (b - Bn n); fourier.
Qed.


Theorem kantoro_exists:
{ xs:R & Un_cv Xn xs /\ c_disc X0 eps xs /\ f xs = 0}. 
Proof.
assert (Hcauchy: Cauchy_crit Xn).
assert (Hdisc:(forall (n:nat), (n<>0)%nat->included (c_disc (Xn n) (En n)) 
(c_disc (Xn (pred n)) (En (pred n))))).
intro n.
destruct (lemma_princ n) as  [Hd [_]]; intros.
apply (Hd H0).
apply cauch_seq; assumption.
elim (R_complete Xn Hcauchy); intros xs Hcv.
exists xs.
assert (Hindisc: c_disc X0 eps xs).
assert (Hdisc:(forall (n:nat), (n<>0)%nat->included (c_disc (Xn n) (En n)) 
(c_disc (Xn (pred n)) (En (pred n))))).
intros n Hdif.
destruct (lemma_princ n) as  [Hd [_]]; intros.
apply (Hd Hdif).
assert (Hxn: forall p, (c_disc (Xn 0) (En 0) (Xn p))).
intro p.
replace p with (0+p)%nat by omega.
apply disc_incl_pct.
intros n; apply (Hdisc (S n));omega.
assert (forall p, c_disc X0 eps (Xn p)).
intros p; apply (Hxn p).
assert (forall p, X0-eps<=(Xn p)<=X0+eps).
intros p.
rewrite <-rewrite_c_disc.
exact (H p).
rewrite rewrite_c_disc.
apply lim_in_interv with Xn ;assumption.
split; [assumption|].
split; [assumption|].
assert (Hxn: forall n, 
f (Xn n) = f' (Xn n)*((Xn n)-(Xn (S n)))).
intros n.
simpl;field.
destruct (lemma_princ n) as [_ [_[_[Hf'[_]]]]]; intros; assumption.
assert (Hconv:Un_cv (fun i:nat=> f' (Xn i)*((Xn i)-(Xn (S i)))) 0).
apply lim_prod_minus with a b xs.
assumption.
intros n.
destruct (lemma_princ n) as  [_[_[Hab2 [_]]]]; intros.
unfold c_I in Hab2; assumption.
assumption.
assert (Un_cv (fun i:nat =>f (Xn i)) 0).
unfold Un_cv in Hconv |-*.
intros eps0 He0.
elim (Hconv eps0 He0); intros N Habs.
exists N.
intros n Hneq.
rewrite (Hxn n).
exact (Habs n Hneq).
assert (Hcv2:Un_cv (fun i:nat => f (Xn i)) (f xs)).
apply Heine1.
apply derivable_continuous_pt.
assert (Hin: c_I a b xs).
destruct (lemma_princ 0%nat) as  [_[Hd [_]]]; intros.
simpl in Hd.
apply (Hd xs); assumption.
exact (pr xs Hin).
assumption.
apply UL_sequence with (fun i:nat=>f (Xn i)); assumption.
Qed.

Lemma conv_root:
 forall xs:R , Un_cv Xn xs  -> c_disc X0 eps xs /\ f xs = 0. 
Proof.
assert (Hcauchy: Cauchy_crit Xn).
assert (Hdisc:(forall (n:nat), (n<>0)%nat->included (c_disc (Xn n) (En n)) 
(c_disc (Xn (pred n)) (En (pred n))))).
intro n.
destruct (lemma_princ n) as  [Hd [_]]; intros.
apply (Hd H0).
apply cauch_seq; assumption.
elim (R_complete Xn Hcauchy); intros xs Hcv.
intros xs1 Hcv1.
assert (Hul:= UL_sequence  Xn xs xs1 Hcv Hcv1).
rewrite <-Hul.

assert (Hindisc: c_disc X0 eps xs).
assert (Hdisc:(forall (n:nat), (n<>0)%nat->included (c_disc (Xn n) (En n)) 
(c_disc (Xn (pred n)) (En (pred n))))).
intros n Hdif.
destruct (lemma_princ n) as  [Hd [_]]; intros.
apply (Hd Hdif).
assert (Hxn: forall p, (c_disc (Xn 0) (En 0) (Xn p))).
intro p.
replace p with (0+p)%nat by omega.
apply disc_incl_pct.
intros n; apply (Hdisc (S n));omega.
assert (forall p, c_disc X0 eps (Xn p)).
intros p; apply (Hxn p).
assert (forall p, X0-eps<=(Xn p)<=X0+eps).
intros p.
rewrite <-rewrite_c_disc.
exact (H p).
rewrite rewrite_c_disc.
apply lim_in_interv with Xn ;assumption.
split; [assumption|].
assert (Hxn: forall n, 
f (Xn n) = f' (Xn n)*((Xn n)-(Xn (S n)))).
intros n.
simpl;field.
destruct (lemma_princ n) as [_ [_[_[Hf'[_]]]]]; intros; assumption.
assert (Hconv:Un_cv (fun i:nat=> f' (Xn i)*((Xn i)-(Xn (S i)))) 0).
apply lim_prod_minus with a b xs.
assumption.
intros n.
destruct (lemma_princ n) as  [_[_[Hab2 [_]]]]; intros.
unfold c_I in Hab2; assumption.
assumption.
assert (Un_cv (fun i:nat =>f (Xn i)) 0).
unfold Un_cv in Hconv |-*.
intros eps0 He0.
elim (Hconv eps0 He0); intros N Habs.
exists N.
intros n Hneq.
rewrite (Hxn n).
exact (Habs n Hneq).
assert (Hcv2:Un_cv (fun i:nat => f (Xn i)) (f xs)).
apply Heine1.
apply derivable_continuous_pt.
assert (Hin: c_I a b xs).
destruct (lemma_princ 0%nat) as  [_[Hd [_]]]; intros.
simpl in Hd.
apply (Hd xs); assumption.
exact (pr xs Hin).
assumption.
apply UL_sequence with (fun i:nat=>f (Xn i)); assumption.
Qed.

(*
Lemma bn_in_disc:
forall un,
(forall n, Rabs (un (S n) - un n)<=Bn n) ->
(forall n, c_disc (un 0%nat) (2*b0 - (2* Bn n)) (un n)).
Proof.
intros un Hn.
assert (Hdif_pos: forall n, Rabs (un n - un 0%nat)<=2*b0 - (2*b0)/2^n).
induction n.
replace (un 0%nat- un 0%nat) with 0 by ring.
rewrite Rabs_R0;simpl.
replace (2*b0 - 2*b0/1) with 0; [|field; auto]; fourier. 
simpl.
replace (un (S n) - un 0%nat) with (un (S n) - un n +(un n - un 0%nat))
by ring.
assert (Rabs (un (S n) - un n)<=b0/2^n).
apply Rle_trans with (Bn n).
apply (Hn n).
apply Bn_desc_2.
apply Rle_trans with (Rabs (un (S n) - un n) +Rabs (un n - un 0%nat)).
apply (Rabs_triang (un (S n) - un n) (un n - un 0%nat)).
replace (2*b0 - 2*b0/(2*2^n)) with (b0/2^n + (2*b0 - 2*b0/2^n)).
apply Rplus_le_compat; assumption.
unfold Rdiv.
field.
apply Rgt_not_eq.
apply pow2_diff0.
intros m.
unfold c_disc.
rewrite Rabs_minus_sym.
apply Rle_trans with (2*b0-2*b0/2^( m)).
apply (Hdif_pos ( m)).
simpl.
unfold Rdiv.
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.

apply Rle_mult_inv_pos.
auto.
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.


replace (2*b0-2*b0*/(2*2^m)) with (2*b0 + (-2*b0*/(2*2^m))); [|field;auto].
pattern (2*b0) at 2; replace (2*b0) with (2*b0+0) by ring.
apply Rplus_le_compat_l.
replace (-2*b0*/(2*2^m)) with (-(b0*/2^m)); [|field;auto].
replace 0 with (-0) by ring.
apply Ropp_le_contravar.
apply Rle_mult_inv_pos.
auto.
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.
*)

Lemma not_series:
forall un:nat->R,
(forall n, Rabs (un (S n) - un n)<=Bn n) ->
(forall n, c_disc (un 0%nat) (2*b0) (un n)).
Proof.
intros un  Hineq m.
induction m.
unfold c_disc.
replace (un 0%nat- un 0%nat) with 0 by ring.
rewrite Rabs_R0; apply aux1; auto with real.
assert (Hdif_pos: forall n, Rabs (un n - un 0%nat)<=2*b0 - (2*b0)/2^n).
induction n.
replace (un 0%nat- un 0%nat) with 0 by ring.
rewrite Rabs_R0;simpl.
replace (2*b0 - 2*b0/1) with 0; [|field; auto]; fourier. 
simpl.
replace (un (S n) - un 0%nat) with (un (S n) - un n +(un n - un 0%nat))
by ring.
assert (Rabs (un (S n) - un n)<=b0/2^n).
apply Rle_trans with (Bn n).
apply (Hineq n).
apply Bn_desc_2.
apply Rle_trans with (Rabs (un (S n) - un n) +Rabs (un n - un 0%nat)).
apply (Rabs_triang (un (S n) - un n) (un n - un 0%nat)).
replace (2*b0 - 2*b0/(2*2^n)) with (b0/2^n + (2*b0 - 2*b0/2^n)).
apply Rplus_le_compat; assumption.
unfold Rdiv.
field.
apply Rgt_not_eq.
apply pow2_diff0.
unfold c_disc.
rewrite Rabs_minus_sym.
apply Rle_trans with (2*b0-2*b0/2^(S m)).
apply (Hdif_pos (S m)).
simpl.
unfold Rdiv.
replace (2*b0-2*b0*/(2*2^m)) with (2*b0 + (-2*b0*/(2*2^m))); [|field;auto].
pattern (2*b0) at 2; replace (2*b0) with (2*b0+0) by ring.
apply Rplus_le_compat_l.
replace (-2*b0*/(2*2^m)) with (-(b0*/2^m)); [|field;auto].
replace 0 with (-0) by ring.
apply Ropp_le_contravar.
apply Rle_mult_inv_pos.
auto.
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.

Lemma xn_in_disc :
 forall n, c_disc X0 (2*b0) (Xn n).
Proof.
replace X0 with (Xn 0) ; [|simpl; trivial].
apply not_series.
intro n.
destruct (lemma_princ n) as [_[_[_[_[_ [Hb _]]]]]].
rewrite Rabs_minus_sym.
replace (Xn n - Xn (S n)) with (f ((Xn n))/f' (Xn n));[assumption|simpl;field].
destruct (lemma_princ n) as  [_ [_[_[Hf'[_]]]]]; intros; assumption.
Qed.


Theorem kantoro_exists_b:
{ xs:R & Un_cv Xn xs /\ c_disc X0 (2*b0) xs /\ f xs = 0}. 
Proof.
elim kantoro_exists; intros xs Hlim.
destruct Hlim as [Hl1 [Hl2 Hl3]].
exists xs.
split; [assumption|].
split; [|assumption].
assert (Hxn: forall n, c_disc (Xn 0%nat) (2*b0) (Xn n)).
apply not_series.
intro n.
destruct (lemma_princ n) as [_[_[_[_[_ [Hb _]]]]]].
rewrite Rabs_minus_sym.
replace (Xn n - Xn (S n)) with (f ((Xn n))/f' (Xn n));[assumption|simpl;field].
destruct (lemma_princ n) as  [_ [_[_[Hf'[_]]]]]; intros; assumption.
rewrite rewrite_c_disc.
apply lim_in_interv with Xn .
intro n.
assert (Hseq: c_disc (Xn 0) (2*b0) (Xn n)) by apply (Hxn n).
simpl in Hseq.
rewrite rewrite_c_disc in Hseq.
assumption.
assumption.
Qed.

Definition xs := projT1 kantoro_exists_b.


Lemma uniq_aux1:
forall xs2:R,
  (*c_disc X0 (2*b0) xs2->*)
  c_I a b xs2 ->
  f xs2 = 0 ->
 forall n, Rabs (xs2-Xn (S n)) <= An n * 1/2 *  c * Rsqr(xs2-Xn n).
Proof. 
intros xs2 Hdom  Hrad.
induction n.
(*unfold c_disc in Hdisc2.*)
(*rewrite Rabs_minus_sym.*)
(*assert (Hf'_not_zero : f' (Xn n) <> 0).*)
(*destruct (lemma_princ n) as [_ [_ [_[Hf' _]]]]; intros; assumption.*)

assert (Hr_seq2: Xn 1 - xs2 = (/(f' (Xn 0)))*
(f xs2 - f (Xn 0)-f' (Xn 0)*xs2 + f' (Xn 0) * Xn 0)). 
rewrite Hrad; simpl; field; assumption.
(*destruct (lemma_princ n) as  [_ [_[_[Hf' _]]]]; intros; assumption.*)
rewrite Rabs_minus_sym.
rewrite Hr_seq2.
rewrite Rabs_mult.
simpl.
unfold Rdiv ; repeat rewrite Rmult_assoc; apply Rmult_le_compat.
apply Rabs_pos.
apply Rabs_pos.
assumption.
repeat rewrite <-Rmult_assoc; fold (Rdiv 1 2); apply lemma_third.
assumption.
unfold c_I; split; elim Hincl; intros; fourier.
assert (Hf'_not_zero : forall n, f' (Xn ( n)) <> 0).
intro n0; destruct (lemma_princ ( n0)) as [_ [_ [_[Hf' _]]]]; intros; assumption.
assert (Hr_seq2: forall n, Xn ((S n)) - xs2 = (/(f' (Xn ( n))))*
(f xs2 - f (Xn ( n))-f' (Xn ( n))*xs2 + f' (Xn n) * Xn ( n))). 
intro n0; rewrite Hrad; simpl. field.
apply Hf'_not_zero.
(*destruct (lemma_princ n) as  [_ [_[_[Hf' _]]]]; intros; assumption.*)
rewrite Rabs_minus_sym.
rewrite Hr_seq2.
rewrite Rabs_mult.
unfold Rdiv; repeat rewrite Rmult_assoc.
apply Rmult_le_compat.
apply Rabs_pos.
apply Rabs_pos.
destruct (lemma_princ (S n) ) as [_ [_ [_[_[Hf' _]]]]].
assumption.
repeat rewrite <-Rmult_assoc; fold (Rdiv 1 2); apply lemma_third.
assumption.
apply Hb_eps; apply xn_in_disc.
Qed.


Lemma unic_aux:
forall xs2:R,
  c_disc X0 (2*b0) xs2->
  f xs2 = 0 ->
 forall n, Rabs (xs2-Xn n) <= 2*Bn n.
Proof.
intros xs2 Hdisc2 Hrad.
induction n.
simpl.
unfold c_disc in Hdisc2.
rewrite Rabs_minus_sym; assumption.
assert (Hf'_not_zero : f' (Xn n) <> 0).
destruct (lemma_princ n) as [_ [_ [_[Hf' _]]]]; intros; assumption.
assert (Hr_seq2: Xn (S n) - xs2 = (1/(f' (Xn n)))*
(f xs2 - f (Xn n)-f' (Xn n)*(xs2-Xn n))). 
rewrite Hrad; simpl; field; assumption.
(*destruct (lemma_princ n) as  [_ [_[_[Hf' _]]]]; intros; assumption.*)
rewrite Rabs_minus_sym.
rewrite Hr_seq2.
rewrite Rabs_mult.
simpl.
replace (2*(An n* Bn n*Bn n*c)) with (An n*(2*Bn n*Bn n*c)) by ring.
apply Rmult_le_compat.
apply Rabs_pos.
apply Rabs_pos.
destruct (lemma_princ n) as [_ [_ [_[_[Hf' _]]]]].
unfold Rdiv; replace (1*/f' (Xn n)) with (/f' (Xn n));[|field];
assumption.

apply Rle_trans with (1/2*c*Rsqr(xs2-Xn n)).
replace ((f xs2 - f (Xn n)-f' (Xn n)*(xs2-Xn n))) with 
(f xs2 - f (Xn n)-f' (Xn n)*xs2+ f' (Xn n)*Xn n) by field.
apply lemma_third.
rewrite rewrite_c_disc in Hdisc2.
unfold included in Hincl.
apply (Hb_eps xs2).
rewrite rewrite_c_disc.
elim Hdisc2; intros.
split;fourier.
destruct (lemma_princ n) as  [_ [_[Hk _]]].
assumption.
assert (Hsqr: Rsqr (xs2 - Xn n) <=2*Bn n*(2*Bn n)).
fold (Rsqr (2*Bn n)).
apply Rsqr_le_abs_1 .
replace (Rabs(2*Bn n)) with (2*Bn n).
assumption.
rewrite Rabs_right.
trivial.
apply Rle_ge.
apply aux1; [fourier|apply Bn_ge_0].
replace (2*Bn n*Bn n*c) with (1/2*c*(2*Bn n*(2*Bn n)));[| field]. 
apply Rmult_le_compat_l.
apply aux1; [fourier|auto].
fourier.
Qed.


Theorem kantoro_unic:
forall xs2:R,
  c_disc X0 (2*b0) xs2->
  f xs2 = 0 ->
  Un_cv Xn xs2. 
Proof.
intros xs2 Hdisc2 Hrad.
assert (Hseq:= unic_aux xs2 Hdisc2 Hrad).
red.
intros delta Hdpos.
assert (HBn_cv: Un_cv Bn 0); [apply Bn_conv_0|].
unfold Un_cv in HBn_cv.
assert (Hd2: delta/2>0).
fourier.
elim (HBn_cv (delta/2) Hd2); intros N Hbcv.
exists N.
intros n Hineq.
unfold R_dist.
rewrite Rabs_minus_sym.
apply Rle_lt_trans with (2*Bn n).
apply (Hseq n).
assert (Hbn_d: R_dist (Bn n) 0<delta/2).
apply (Hbcv n Hineq).
unfold R_dist in Hbn_d.
replace (Bn n-0) with (Bn n) in Hbn_d by ring.
rewrite Rabs_right in Hbn_d.
fourier.
apply Rle_ge; apply Bn_ge_0.
Qed.

Theorem kantoro_unic2:
forall xs2:R,
  c_disc X0 (2*b0) xs2->
  f xs2 = 0 ->
  xs = xs2. 
Proof.
intros xs2 H H1.
apply UL_sequence with Xn;
[elim (projT2 kantoro_exists_b); intros H2 _;
unfold xs; assumption
|apply kantoro_unic; assumption].
Qed.

Theorem kantoro_unic3:
forall xs2:R,
Un_cv Xn xs2 ->
  xs = xs2. 
Proof.
intros xs2 H.
apply UL_sequence with Xn;
[elim (projT2 kantoro_exists_b); intros H2 _;
unfold xs; assumption
|assumption]. 
Qed.

(*
Lemma conv_crit:
forall an bn l,
Un_cv bn 0 ->
Rabs (
*)
Lemma kantoro_uniq_extended:
forall xs2,
0 < mu 0 -> 
included (c_disc X0 (2/mu 0 * b0)) (c_I a b) ->
c_disc X0 (2/mu 0 * b0) xs2 ->
f xs2 = 0 ->
xs = xs2.
Proof.
intros xs2 Hmu Hdom Hdisc Hroot.
assert (Hmu0: forall n, 0 < mu n).
induction n; simpl; [assumption| rewrite rel_mu; unfold pow; apply Rmult_lt_0_compat;
  ring_simplify; assumption].
assert (Habc0: 0 < a0 /\ 0 < b0 /\ 0 < c).
assert (H00:= Rgt_not_eq _ _ Hmu).
 elim (Rlt_0_mult_reg _ _ H00) ; intros  H02 H01 .
 elim ( Rlt_0_mult_reg _ _ H02) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
simpl in H04, H05.
split.
elim H_a0; intro Ha00; [trivial|rewrite Ha00 in H05; elim (H05 (refl_equal a0))].
split.
elim H_b0; intro Ha00; [trivial|rewrite Ha00 in H04; elim (H04 (refl_equal b0))].
elim H_c; intro Ha00; [trivial|rewrite Ha00 in H01; elim (H01 (refl_equal c))].
assert ( Hind: forall n, Rabs (xs2 - Xn n) <= 2 / mu n * Bn n).
induction n.
rewrite Rabs_minus_sym; simpl; unfold c_disc in Hdisc; trivial.
apply Rle_trans with (An n * 1/2 *  c * Rsqr (xs2 - Xn n)).
apply uniq_aux1; [apply Hdom; assumption|assumption].
apply Rle_trans with (An n * 1 / 2 * c * Rsqr (2 / mu n * Bn n)).
apply Rmult_le_compat_l.
unfold Rdiv; repeat apply aux1; [apply An_ge_0|fourier |fourier |apply H_c].
apply Rsqr_le_abs_1; apply Rle_trans with ( 2 / mu n * Bn n ); [assumption| apply RRle_abs].
right; unfold Rsqr; unfold mu; simpl; field.
assert (H00:= Rgt_not_eq _ _ Hmu).
 elim (Rlt_0_mult_reg _ _ H00) ; intros  _ H01 .
clear H00; assert (H00:= Rgt_not_eq _ _ (Hmu0 n)).
 elim (Rlt_0_mult_reg _ _ H00) ; intros  H02 _.
 elim ( Rlt_0_mult_reg _ _ H02) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat split; assumption.
assert (Hbmu: forall n, Bn n / mu n = /2 ^ n * (b0/ mu 0)).
induction n; [simpl; field; apply Rgt_not_eq; apply Hmu|].
replace (/2^ (S n)) with (/2 * / 2 ^ n); [|simpl pow; field; apply Rgt_not_eq; apply pow2_diff0].
rewrite Rmult_assoc; rewrite <- IHn.
replace (Bn (S n)) with (/2 * mu n * Bn n).
rewrite rel_mu; unfold pow; field; apply Rgt_not_eq; unfold Rgt; apply Hmu0.
rewrite rel_bn1 ; rewrite rel_mu0_aux; trivial.
apply kantoro_unic3.
assert (Hscn: forall n, (0<n)%nat -> 2*/2^n = /2^ (n-1)).
intros n Hn; induction n.
assert (H : ~ (0 < 0)%nat) by omega; elim (H Hn).
(*elim (le_lt_or_eq 0 n).
 intros H2.*)
simpl.
replace (n-0)%nat with n by omega.
field; apply Rgt_not_eq; apply pow2_diff0.

assert (Hind2:  forall n : nat, (0 < n)%nat -> Rabs (xs2 - Xn ( n)) <= / 2 ^ (n - 1)  * (b0 / mu 0)).
intros n Hn; apply Rle_trans with (2 / mu ( n) * Bn ( n)); [apply Hind|].
replace (2 / mu (n ) * Bn ( n) ) with (2* ( Bn ( n) / mu ( n) )) ;
  [| field; apply Rgt_not_eq;unfold Rgt; apply Hmu0].
rewrite Hbmu; simpl; right.
rewrite <-(Hscn n Hn). 
field; split; apply Rgt_not_eq; [apply Hmu|apply pow2_diff0].
intros e He; unfold R_dist.
assert (Har: 0 < e * (mu 0 / b0) ).
unfold Rdiv; apply Rmult_lt_0_compat.
apply He.
apply Rmult_lt_0_compat.
assumption.
apply Rinv_0_lt_compat.
elim Habc0; intros _ Habc1; elim Habc1; intros; assumption.
elim (archim_lt2 _ Har); intros N H; exists (S N); intros n Hn.
rewrite Rabs_minus_sym.
elim (le_lt_or_eq 0 n);[| | omega].
 intro Hn0.
 apply Rle_lt_trans with (/ 2 ^ (n - 1) * (b0 / mu 0)).
apply Hind2; assumption.
apply Rle_lt_trans with (/ 2 ^ (N) * (b0 / mu 0) ).
apply Rmult_le_compat_r.
unfold Rdiv; apply aux1; [apply H_b0|left; apply Rinv_0_lt_compat; assumption].
apply Rle_Rinv; [ apply pow2_diff0|apply pow2_diff0|apply Rle_pow; [fourier|omega]].
apply Rmult_lt_reg_l with (mu 0/ b0).
unfold Rdiv; apply Rmult_lt_0_compat; 
 [ assumption|apply Rinv_0_lt_compat; elim Habc0; intros _ H0 ; elim H0; intros; assumption ].
replace (mu 0 / b0 * (/ 2 ^ N * (b0 / mu 0))) with (1/2^N).
rewrite Rmult_comm; assumption.
field; split; [apply Rgt_not_eq; apply Hmu|split; apply Rgt_not_eq; [apply pow2_diff0|
  elim Habc0; intros _ H0 ; elim H0; intros; assumption]].
intro Hn0; subst n.
assert (Hct: ~ (0 >= S N)%nat) by omega; elim Hct; trivial.
Qed.


Lemma Hdif_pos: forall n p, Rabs (Xn (p + n) - Xn p)<=2*Bn p - (2*Bn p)/2^n.
Proof.
intros n p; induction n.
simpl; rewrite plus_0_r; unfold Rminus at 1; rewrite Rplus_opp_r; 
rewrite Rabs_R0; unfold Rdiv; rewrite Rinv_1; ring_simplify; right; trivial.
simpl.
replace (Xn (p + S n) - Xn p) with (Xn (p + S n) - Xn (p+n) + (Xn (p + n)- Xn p)) 
by ring.
assert (H0 : Rabs (Xn (p + (S n)) - Xn ( p+ n)) <= Bn p / 2^n).
apply Rle_trans with (Bn (p+n)).
replace (p + S n)%nat with (S ( p+ n)) by omega.
rewrite Rabs_minus_sym.
destruct (lemma_princ (p+n)) as [_[_[_[_[_ [Hineq _]]]]]].
replace (Xn  (p + n) - Xn (S (p + n))) with (f (Xn (p + n)) / f' (Xn (p + n)));
[assumption|simpl; field].
destruct (lemma_princ (p+n)) as [_ [_[_[Hf'[_]]]]]; intros; assumption.
repeat rewrite rel_bn.
field_simplify; [|split; apply Rgt_not_eq; apply pow2_diff0|
   apply Rgt_not_eq; apply pow2_diff0].
rewrite Rmult_comm; rewrite <-pow_add.
unfold Rdiv; repeat rewrite Rmult_assoc.
apply Rmult_le_compat_l; [auto|].
apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; apply pow2_diff0.
apply pow_le1_le.
split; simpl; [apply greater0| trivial].
split.
assert (H02: (0<2)%nat) by omega.
assert (H1:= Npow_gt1 2 p H02).
omega.
apply minus_le_compat.
apply Npow_monot; omega.
apply Rle_trans with (Rabs (Xn (p + S n) - Xn (p + n)) 
  + Rabs(Xn (p + n) - Xn p)); [apply Rabs_triang|].
replace (2*Bn p - 2*Bn p/(2*2^n)) with (Bn p/2^n + (2*Bn p - 2*Bn p/2^n)).
apply Rplus_le_compat; assumption.
unfold Rdiv; field; apply Rgt_not_eq; apply pow2_diff0.
Qed.

End Kantorovitch.

