Require Import Reals.
Require Import Fourier.
Require Import kanto_exists_uniq.

Section Kprop.

Variable (a0 b0 c:R).
Variable X0:R.
(*Variable eps:R.*)

(*define f f' as section variables*)
Variable a:R.
Variable b:R.
Variable f:R->R.
Variable f':R->R.

Hypothesis pr: forall t:R, c_I a b t -> derivable_pt f t.
Hypothesis Hder_f: forall (t:R)(H:c_I a b t), derive_pt f t (pr t H)=f' t. 
Hypothesis Hcont_f':forall y, (c_I a b y) -> continuity_pt f' y.
Hypothesis Hlim_f':forall y1 y2:R, c_I a b y1 -> c_I a b y2 ->
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2)).
Hypothesis Hincl: a< X0 < b.
Hypothesis Hdif_f':f' X0 <> 0 .
Hypothesis Habs_a: Rabs (/(f' X0)) <= a0.
Hypothesis Habs_b:Rabs (f X0/(f' X0)) <= b0 .

Variable x0':R.

Hypothesis Hmu: 0 < mu a0 b0 c 0 < 1.
Hypothesis Hmu3: included (c_disc X0 (2/mu a0 b0 c  0 * b0)) (c_I a b).
Hypothesis  Hdom: Rabs (x0' - X0) <= (1-mu a0 b0 c  0)/(2*mu a0 b0 c 0)*b0.


Lemma  A0_b0_c: 2*a0*b0*c<=1.
Proof.
elim Hmu; intros Hmu0 Hmu1.
unfold mu in Hmu1; simpl in Hmu1; left; assumption.
Qed.

Lemma Hb_eps: included (c_disc X0 (2*b0)) (c_I a b).
Proof.
apply included_trans with (c_disc X0 (2 / mu a0 b0 c 0 * b0)); [|assumption].
apply incl_c_disc.
unfold Rdiv.
apply Rmult_le_compat_r.
apply H_b0 with X0 f f'; assumption.
pattern 2 at 1; replace 2 with (2*/1) by field.
apply Rmult_le_compat_l; [fourier|].
apply Rle_Rinv; elim Hmu; intros ; fourier.
Qed.



Theorem conv_stability:
let xstar := (xs a0 b0 c X0 a b f f' A0_b0_c pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b Hb_eps) in
      Un_cv (Xn x0' f f' ) xstar.
Proof.
intros xstar .
destruct Hmu as [Hmu0 Hmu1].
assert (Hmu2: (c_I a b x0')).
apply Hmu3; unfold c_disc; apply Rle_trans with
   ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0) * b0).
rewrite Rabs_minus_sym; assumption.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
unfold Rdiv.
rewrite Rinv_mult_distr; [|apply Rgt_not_eq; fourier|apply Rgt_not_eq; assumption].
 rewrite <-Rmult_assoc; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; assumption.
apply Rle_trans with (1*/2); [|fourier].
apply Rmult_le_compat_r; [fourier|].
pattern 1 at 2; replace 1 with (1-0) by ring; unfold Rminus;
   apply Rplus_le_compat_l; fourier.
assert (h1: Rabs (1 - f'(x0') / f' (X0)) <= (1 - mu a0 b0 c 0)/4 ).

replace ((1 - mu a0 b0 c 0)/4) with (a0*c*((1- mu a0 b0 c 0)/(2*mu a0 b0 c 0) * b0)). 
apply Rle_trans with (a0*c*Rabs (x0' - X0)).
apply Rle_trans with (Rabs (/f'(X0)) * Rabs (f'(x0' ) - f'(X0))).
rewrite <- Rabs_mult; rewrite Rabs_minus_sym; right; f_equal; field; exact Hdif_f'.
rewrite Rmult_assoc; apply Rmult_le_compat; 
  [apply Rabs_pos| apply Rabs_pos| assumption |].
  apply Hlim_f'; [assumption|apply Hb_eps; apply center_in_disc]. 
  unfold Rdiv; repeat apply aux1.
   fourier.
   apply H_b0 with X0 f f'; assumption.
   (*rewrite (Rmult_assoc). (a0*c)  ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0)) b0).*)
apply Rmult_le_compat_l; [apply aux1; 
  [apply H_a0 with X0 f'|apply H_c with X0 a b f']; assumption|exact Hdom].
  unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*; field.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat split; assumption.
assert (h11: Rabs (1 - f'(x0') / f' (X0)) <= 1/4).
apply Rle_trans with ((1 - mu a0 b0 c 0) / 4); [assumption|].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
  pattern 1 at 2; replace 1 with (1 - 0) by ring.
  unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar; left; assumption.
assert (h2: f' (x0') <> 0 /\ /f' X0 <>0).
apply Rlt_0_mult_reg .
fold (Rdiv (f' x0') (f' X0)); apply Rgt_not_eq; apply dif_0; apply Rle_trans with (1/4);
  [assumption|fourier].
elim h2; clear h2; intros h2 h0'.
assert (h3: Rabs (f' (X0)/ f' (x0')) <= 4 /(3 + mu  a0 b0 c 0)).
replace (f' X0 / f' x0') with (1 / (1 - (1 - f' (x0') / f' X0))); [|field; split; assumption].
unfold Rdiv at 1; rewrite Rabs_mult; rewrite Rabs_Rinv; 
  [|replace (1 - (1 - f' x0' / f' X0)) with
  (f' x0' / f' X0) by ring; unfold Rdiv; apply prod_neq_R0; assumption ].
rewrite Rabs_right; [|fourier].
apply Rle_trans with (1*/ (1 - Rabs (1 - f' x0' / f' X0))).
apply  Rmult_le_compat_l; [fourier| ].
assert (hx: (0 < 1 - Rabs (1 - f' x0' / f' X0) <= Rabs (1 - (1 - f' x0' / f' X0)))).
split.
apply Rgt_minus; apply Rgt_ge_trans with (1/4); [fourier|apply Rle_ge; apply h11].
pattern 1 at 1; replace 1 with (Rabs 1); [apply Rabs_triang_inv| apply Rabs_right; fourier].
apply Rle_Rinv; elim hx; intros hq hw; 
 [|apply Rlt_le_trans with ( 1 - Rabs (1 - f' x0' / f' X0))|];assumption.
assert (hx: 0 < (1 - (1 - mu a0 b0 c 0) / 4) <= (1 - Rabs (1 - f' x0' / f' X0)) ).
split.
field_simplify; unfold Rdiv; rewrite Rinv_1; rewrite Rmult_0_l .
apply Rmult_lt_0_compat; [|fourier].
 apply Rlt_trans with 3; [fourier|].
  pattern 3 at 1; replace 3 with (0+3) by ring; apply Rplus_lt_compat_r; assumption.
unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar .
unfold Rminus in h1; apply h1.
elim hx; intros hx1 hx2; clear hx.
apply Rle_trans with (1*/ (1 - (1 - mu a0 b0 c 0)/4)).
apply Rmult_le_compat_l; [fourier|].
apply Rle_Rinv; [|apply Rlt_le_trans with ( 1 - (1 - mu a0 b0 c 0) / 4)|]; assumption.
assert (hf: 3 + mu a0 b0 c 0 <> 0).
apply Rgt_not_eq; unfold Rgt; apply Rlt_trans with 3; [fourier|].
  pattern 3 at 1; replace 3 with (3+0) by ring; apply Rplus_lt_compat_l;  assumption .
field_simplify; [right; trivial| |].
assumption.
replace (4 - (1 - mu a0 b0 c 0) ) with (3+ mu a0 b0 c 0 ) by ring; assumption.
assert (h4: Rabs (/f'(x0')) <= 4 / (3+mu a0 b0 c 0) * a0).
replace (/f' x0') with (f' X0 / f' x0' * / f' X0);  [|field; split; assumption].
rewrite Rabs_mult.
apply Rmult_le_compat; [apply Rabs_pos| apply Rabs_pos| assumption|assumption].
assert (h5: Rabs (f (x0') / f' (X0)) <= (3+mu a0 b0 c 0)^2/ (16 * mu a0 b0 c 0)*b0).
replace (f x0' / f' X0) with (/f' (X0)* (f x0' - f X0 - f' X0 *(x0' - X0)) +
   f X0 / f' X0 + (x0' - X0));
  [|field; assumption].
apply Rle_trans with (Rabs (/ f' X0 * (f x0' - f X0 - f' X0 * (x0' - X0)) +
   f X0 / f' X0)  + Rabs (x0' - X0)).
apply Rabs_triang.
apply Rle_trans with (Rabs (/ f' X0 * (f x0' - f X0 - f' X0 * (x0' - X0))) +
   Rabs (f X0 / f' X0) +Rabs (x0' - X0)).
apply Rplus_le_compat_r.
apply Rabs_triang.
rewrite Rabs_mult.
apply Rle_trans with (a0 * (1/2 * c * Rsqr (x0' - X0) ) + b0 + Rabs (x0' - X0)).
repeat apply Rplus_le_compat.
apply Rmult_le_compat; [apply Rabs_pos|apply Rabs_pos|assumption |].
replace (f x0' - f X0 - f' X0 * (x0' - X0)) with (f x0' - f X0 - f' X0 * x0' + f' X0* X0) by ring.
apply lemma_third with a b pr; try  assumption.
apply Hb_eps; apply center_in_disc.
apply aux1; [fourier|apply H_b0 with X0 f f']; assumption.
assumption.
right; trivial.
apply Rle_trans with ((1/4 * mu a0 b0 c 0 * /b0) *
   ((Rsqr ( (1 - mu a0 b0 c 0) / (2 *(mu a0 b0 c 0)) * b0))) 
   + b0 + (1 - mu a0 b0 c 0)/ (2 * mu a0 b0 c 0) * b0 ).
apply Rplus_le_compat; [|assumption].
apply Rplus_le_compat_r.
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat.
apply aux1.
apply aux1.
 apply H_a0 with X0 f'; assumption.
fourier.
apply H_c with X0 a b f'; assumption.
apply Rle_0_sqr.
unfold mu; simpl; right. field.
unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros; assumption .
apply Rsqr_le_abs_1.
pattern Rabs at 1; rewrite Rabs_right; 
 [|apply Rle_ge; apply Rle_trans with (Rabs (x0' - X0)); [apply Rabs_pos|]]; assumption.
right; unfold Rsqr; field.
split; [apply Rgt_not_eq; unfold Rgt; assumption|].
unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros; assumption .
assert (h6: Rabs (f x0' / f' x0') <= (3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0).
replace (f x0' / f' x0') with ((f' X0 / f' x0') * (f x0' / f' X0)); [|field; split; assumption] .
rewrite Rabs_mult.
replace ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0) with
  (4 / (3 + mu a0 b0 c 0) * ((3 + mu a0 b0 c 0) ^ 2 / (16 * mu a0 b0 c 0) * b0)).
apply Rmult_le_compat; [apply Rabs_pos|apply Rabs_pos|assumption| assumption].
field; split; apply Rgt_not_eq; unfold Rgt; 
  [|apply Rlt_trans with 3; [fourier|]; pattern 3 at 1 ; replace 3 with (3+0) by ring; 
  apply Rplus_lt_compat_l ] ; assumption.
set (a' :=  4 / (3 + mu a0 b0 c 0) * a0).
set (b' :=   (3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0).
assert (Hx1: included (c_disc x0' (2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0)))
  (c_disc X0 (2 / mu a0 b0 c 0 * b0))).
intros x Hdisc.
unfold c_disc in Hdisc |-*.
replace (X0 - x) with (X0 - x0' + (x0' - x)) by ring.
apply Rle_trans with (Rabs (X0 - x0') + Rabs (x0' - x)); [apply Rabs_triang|].
apply Rle_trans with ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0) * b0+
  2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0) ).
apply Rplus_le_compat; [rewrite Rabs_minus_sym; assumption|assumption]. 
right; field; apply Rgt_not_eq; unfold Rgt; assumption.
assert (hx: included (c_disc x0' (2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0)))
  (c_I a b)).
apply included_trans with (c_disc X0 (2 / mu a0 b0 c 0 * b0)); assumption.
assert (h7: 
{xs : R &  Un_cv (Xn x0' f f') xs /\ c_disc x0' (2 * b') xs /\ f xs = 0}).
apply kantoro_exists_b with a' c a b pr; unfold a'; unfold b'; try assumption.
right; unfold mu; simpl; field.
 unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat split; try assumption.
apply Rgt_not_eq; unfold Rgt; apply Rlt_trans with 3; [fourier|];
   pattern 3 at 1 ; replace 3 with (3+0) by ring; apply Rplus_lt_compat_l ; assumption.
apply disc_interval with (2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0)); 
  [|assumption].
unfold Rdiv; repeat apply Rmult_lt_0_compat; [fourier| | |].
apply Rlt_trans with 3; [fourier|];
   pattern 3 at 1 ; replace 3 with (3+0) by ring; apply Rplus_lt_compat_l ; assumption.
apply Rinv_0_lt_compat; apply Rmult_lt_0_compat; [fourier| assumption].
unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros _ Hb.
assert (hb0: 0 <= b0).
apply H_b0 with X0 f f'; assumption.
elim hb0; intro hb0'; [assumption| rewrite <- hb0' in Hb; elim (Hb (refl_equal 0))].
elim h7; intros xs2 Hxs2.
assert (h: xstar = xs2).
unfold xstar; apply kantoro_uniq_extended.
assumption.
assumption.
apply (Hx1 xs2); elim Hxs2; intros _ H1; 
  elim H1; intros H2 _; unfold b' in H2; assumption.  
 elim Hxs2; intros _ H1; elim H1; intros; assumption.
rewrite h; elim Hxs2; intros; assumption.
Qed.


Definition a' :=  4 / (3 + mu a0 b0 c 0) * a0.
Definition b' :=   (3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0.


Lemma disc_x0': included (c_disc x0' (2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0)))
  (c_disc X0 (2 / mu a0 b0 c 0 * b0)).
Proof.
intros x Hdisc.
unfold c_disc in Hdisc |-*.
replace (X0 - x) with (X0 - x0' + (x0' - x)) by ring.
apply Rle_trans with (Rabs (X0 - x0') + Rabs (x0' - x)); [apply Rabs_triang|].
apply Rle_trans with ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0) * b0+
  2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0) ).
apply Rplus_le_compat; [rewrite Rabs_minus_sym; assumption|assumption]. 
right; field; apply Rgt_not_eq; unfold Rgt; elim Hmu; intros; assumption.
Qed.

Lemma domain_x0': included (c_disc x0' (2 * b'))
  (c_I a b).
Proof.
unfold b'.
apply included_trans with (c_disc X0 (2 / mu a0 b0 c 0 * b0));
[apply disc_x0'|]; assumption.
Qed.

Lemma A0_b0_c': 2* a' * b' * c <= 1.
 Proof.
right; unfold a'; unfold b'; unfold mu; simpl; field.
 unfold mu in Hmu |-*;  simpl in Hmu |-*; elim Hmu; 
  clear Hmu; intros Hmu0 Hmu1.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat split; try assumption.
apply Rgt_not_eq; unfold Rgt; apply Rlt_trans with 3; [fourier|];
   pattern 3 at 1 ; replace 3 with (3+0) by ring; apply Rplus_lt_compat_l ; assumption.
Qed.


Lemma Hincl': a < x0' < b.
Proof.
apply disc_interval with (2 * ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0)); 
  [|apply domain_x0'].
unfold Rdiv; repeat apply Rmult_lt_0_compat; [fourier| | |].
apply Rlt_trans with 3; [fourier|];
   pattern 3 at 1 ; replace 3 with (3+0) by ring; apply Rplus_lt_compat_l;
     elim Hmu; intros; assumption.
apply Rinv_0_lt_compat; apply Rmult_lt_0_compat; [fourier|elim Hmu; tauto ].
unfold mu in Hmu |-*;  simpl in Hmu |-*; elim Hmu; 
  clear Hmu; intros Hmu0 Hmu1.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros _ Hb.
assert (hb0: 0 <= b0).
apply H_b0 with X0 f f'; assumption.
elim hb0; intro hb0'; [assumption| rewrite <- hb0' in Hb; elim (Hb (refl_equal 0))].
Qed.


Lemma Hdif_x0': f' x0' <> 0.
Proof.
elim Hmu; intros Hmu0 Hmu1.
assert (Hmu2: (c_I a b x0')).
apply Hmu3; unfold c_disc; apply Rle_trans with
   ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0) * b0).
rewrite Rabs_minus_sym; assumption.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
unfold Rdiv.
rewrite Rinv_mult_distr; [|apply Rgt_not_eq; fourier|apply Rgt_not_eq; assumption].
 rewrite <-Rmult_assoc; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; assumption.
apply Rle_trans with (1*/2); [|fourier].
apply Rmult_le_compat_r; [fourier|].
pattern 1 at 2; replace 1 with (1-0) by ring; unfold Rminus; 
  apply Rplus_le_compat_l; fourier.
assert (h1: Rabs (1 - f'(x0') / f' (X0)) <= (1 - mu a0 b0 c 0)/4 ).
replace ((1 - mu a0 b0 c 0)/4) with (a0*c*((1- mu a0 b0 c 0)/(2*mu a0 b0 c 0) * b0)). 
apply Rle_trans with (a0*c*Rabs (x0' - X0)).
apply Rle_trans with (Rabs (/f'(X0)) * Rabs (f'(x0' ) - f'(X0))).
rewrite <- Rabs_mult; rewrite Rabs_minus_sym; right; f_equal; field; exact Hdif_f'.
rewrite Rmult_assoc; apply Rmult_le_compat; 
  [apply Rabs_pos| apply Rabs_pos| assumption |].
  apply Hlim_f'; [assumption|apply Hb_eps; apply center_in_disc]. 
  unfold Rdiv; repeat apply aux1.
   fourier.
   apply H_b0 with X0 f f'; assumption.
apply Rmult_le_compat_l; [apply aux1; 
  [apply H_a0 with X0 f'|apply H_c with X0 a b f']; assumption|exact Hdom].
  unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*; field.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat split; assumption.
assert (h11: Rabs (1 - f'(x0') / f' (X0)) <= 1/4).
apply Rle_trans with ((1 - mu a0 b0 c 0) / 4); [assumption|].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
  pattern 1 at 2; replace 1 with (1 - 0) by ring.
  unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar; left; assumption.
assert (h2: f' (x0') <> 0 /\ /f' X0 <>0).
apply Rlt_0_mult_reg .
fold (Rdiv (f' x0') (f' X0)); apply Rgt_not_eq; apply dif_0; apply Rle_trans with (1/4);
  [assumption|fourier].
elim h2; clear h2; intros h2 h0'.
assumption.
Qed.


Lemma haux_abs_a':
   Rabs (f' (X0)/ f' (x0')) <= 4 /(3 + mu  a0 b0 c 0).
Proof.
elim Hmu; intros Hmu0 Hmu1.
assert (Hmu2: (c_I a b x0')).
apply Hmu3; unfold c_disc; apply Rle_trans with
   ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0) * b0).
rewrite Rabs_minus_sym; assumption.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
unfold Rdiv.
rewrite Rinv_mult_distr; [|apply Rgt_not_eq; fourier|apply Rgt_not_eq; assumption].
 rewrite <-Rmult_assoc; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; assumption.
apply Rle_trans with (1*/2); [|fourier].
apply Rmult_le_compat_r; [fourier|].
pattern 1 at 2; replace 1 with (1-0) by ring; unfold Rminus;
   apply Rplus_le_compat_l; fourier.
assert (h1: Rabs (1 - f'(x0') / f' (X0)) <= (1 - mu a0 b0 c 0)/4 ).
replace ((1 - mu a0 b0 c 0)/4) with (a0*c*((1- mu a0 b0 c 0)/(2*mu a0 b0 c 0) * b0)). 
apply Rle_trans with (a0*c*Rabs (x0' - X0)).
apply Rle_trans with (Rabs (/f'(X0)) * Rabs (f'(x0' ) - f'(X0))).
rewrite <- Rabs_mult; rewrite Rabs_minus_sym; right; f_equal; field; exact Hdif_f'.
rewrite Rmult_assoc; apply Rmult_le_compat; 
  [apply Rabs_pos| apply Rabs_pos| assumption |].
  apply Hlim_f'; [assumption|apply Hb_eps; apply center_in_disc]. 
  unfold Rdiv; repeat apply aux1.
   fourier.
   apply H_b0 with X0 f f'; assumption.
 apply Rmult_le_compat_l; [apply aux1; 
  [apply H_a0 with X0 f'|apply H_c with X0 a b f']; assumption|exact Hdom].
  unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*; field.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat split; assumption.
assert (h11: Rabs (1 - f'(x0') / f' (X0)) <= 1/4).
apply Rle_trans with ((1 - mu a0 b0 c 0) / 4); [assumption|].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
  pattern 1 at 2; replace 1 with (1 - 0) by ring.
  unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar; left; assumption.
assert (h2: f' (x0') <> 0 /\ /f' X0 <>0).
apply Rlt_0_mult_reg .
fold (Rdiv (f' x0') (f' X0)); apply Rgt_not_eq; apply dif_0; apply Rle_trans with (1/4);
  [assumption|fourier].
elim h2; clear h2; intros h2 h0'.
replace (f' X0 / f' x0') with (1 / (1 - (1 - f' (x0') / f' X0)));
   [|field; split; [apply Hdif_x0'| apply Hdif_f']].
unfold Rdiv at 1; rewrite Rabs_mult; rewrite Rabs_Rinv;
[|replace (1 - (1 - f' x0' / f' X0)) with
  (f' x0' / f' X0) by ring; unfold Rdiv; apply prod_neq_R0; assumption ].
rewrite Rabs_right; [|fourier].
apply Rle_trans with (1*/ (1 - Rabs (1 - f' x0' / f' X0))).
apply  Rmult_le_compat_l; [fourier| ].
assert (hx: (0 < 1 - Rabs (1 - f' x0' / f' X0) <= Rabs (1 - (1 - f' x0' / f' X0)))).
split.
apply Rgt_minus; apply Rgt_ge_trans with (1/4); [fourier|apply Rle_ge ;apply h11].
pattern 1 at 1; replace 1 with (Rabs 1); [apply Rabs_triang_inv| apply Rabs_right; fourier].
apply Rle_Rinv; elim hx; intros hq hw; 
 [|apply Rlt_le_trans with ( 1 - Rabs (1 - f' x0' / f' X0))|];assumption.
assert (hx: 0 < (1 - (1 - mu a0 b0 c 0) / 4) <= (1 - Rabs (1 - f' x0' / f' X0)) ).
split.
field_simplify; unfold Rdiv; rewrite Rinv_1; rewrite Rmult_0_l .
apply Rmult_lt_0_compat; [|fourier].
 apply Rlt_trans with 3; [fourier|].
  pattern 3 at 1; replace 3 with (0+3) by ring; apply Rplus_lt_compat_r; assumption.
unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar .
unfold Rminus in h1; apply h1.
elim hx; intros hx1 hx2; clear hx.
apply Rle_trans with (1*/ (1 - (1 - mu a0 b0 c 0)/4)).
apply Rmult_le_compat_l; [fourier|].
apply Rle_Rinv; [|apply Rlt_le_trans with ( 1 - (1 - mu a0 b0 c 0) / 4)|]; assumption.
assert (hf: 3 + mu a0 b0 c 0 <> 0).
apply Rgt_not_eq; unfold Rgt; apply Rlt_trans with 3; [fourier|].
  pattern 3 at 1; replace 3 with (3+0) by ring; apply Rplus_lt_compat_l;  assumption .
field_simplify; [right; trivial| |].
assumption.
replace (4 - (1 - mu a0 b0 c 0) ) with (3+ mu a0 b0 c 0 ) by ring; assumption.
Qed.

Lemma H_in_disc:  (c_I a b x0').
Proof.
apply Hmu3; unfold c_disc; apply Rle_trans with
   ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0) * b0).
rewrite Rabs_minus_sym; assumption.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
unfold Rdiv.
elim Hmu;  intros Hmu0 Hmu1.
rewrite Rinv_mult_distr; [|apply Rgt_not_eq; fourier|apply Rgt_not_eq; assumption].
 rewrite <-Rmult_assoc; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; assumption.
apply Rle_trans with (1*/2); [|fourier].
apply Rmult_le_compat_r; [fourier|].
pattern 1 at 2; replace 1 with (1-0) by ring; unfold Rminus; 
  apply Rplus_le_compat_l; fourier.
Qed.


Lemma Habs_a':  Rabs (/f'(x0')) <= a'.
Proof.
 elim Hmu; intros Hmu0 Hmu1.
unfold a'.
replace (/f' x0') with (f' X0 / f' x0' * / f' X0);  [|field; split; [apply Hdif_x0'| apply Hdif_f']].
rewrite Rabs_mult.
apply Rmult_le_compat; [apply Rabs_pos| apply Rabs_pos|
  apply haux_abs_a' |assumption].
Qed.

Lemma Habs_b':
  Rabs (f x0' / f' x0') <= (3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0.
Proof.
elim Hmu; intros Hmu0 Hmu1.
assert (h5: Rabs (f (x0') / f' (X0)) <= (3+mu a0 b0 c 0)^2/ (16 * mu a0 b0 c 0)*b0).
replace (f x0' / f' X0) with 
  (/f' (X0)* (f x0' - f X0 - f' X0 *(x0' - X0)) + f X0 / f' X0 + (x0' - X0));
  [|field; assumption].
apply Rle_trans with (Rabs (/ f' X0 * (f x0' - f X0 - f' X0 * (x0' - X0)) +
   f X0 / f' X0)  + Rabs (x0' - X0)).
apply Rabs_triang.
apply Rle_trans with (Rabs (/ f' X0 * (f x0' - f X0 - f' X0 * (x0' - X0))) +
   Rabs (f X0 / f' X0) +Rabs (x0' - X0)).
apply Rplus_le_compat_r.
apply Rabs_triang.
rewrite Rabs_mult.
apply Rle_trans with (a0 * (1/2 * c * Rsqr (x0' - X0) ) + b0 + Rabs (x0' - X0)).
repeat apply Rplus_le_compat.
apply Rmult_le_compat; [apply Rabs_pos|apply Rabs_pos|assumption |].
replace (f x0' - f X0 - f' X0 * (x0' - X0)) with (f x0' - f X0 - f' X0 * x0' + f' X0* X0) by ring.
apply lemma_third with a b pr; try  assumption.
apply H_in_disc.
apply Hb_eps; apply center_in_disc.
(*unfold Rdiv; repeat apply aux1; [fourier|left; apply Rinv_0_lt_compat; 
  exact Hmu0|apply  H_b0 with X0 f f'] .
assumption.
apply Hb_eps; apply center_in_disc;*) apply aux1; [fourier|apply H_b0 with X0 f f'];
   assumption.
assumption.
right; trivial.
apply Rle_trans with ((1/4 * mu a0 b0 c 0 * /b0) *
   ((Rsqr ( (1 - mu a0 b0 c 0) / (2 *(mu a0 b0 c 0)) * b0))) 
   + b0 + (1 - mu a0 b0 c 0)/ (2 * mu a0 b0 c 0) * b0 ).
apply Rplus_le_compat; [|assumption].
apply Rplus_le_compat_r.
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat.
apply aux1.
apply aux1.
apply H_a0 with X0 f'; assumption.
fourier.
apply H_c with X0 a b f'; assumption.
apply Rle_0_sqr.
unfold mu; simpl; right. field.
unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros; assumption .
apply Rsqr_le_abs_1.
pattern Rabs at 1; rewrite Rabs_right; 
 [|apply Rle_ge; apply Rle_trans with (Rabs (x0' - X0)); [apply Rabs_pos|]]; assumption.
right; unfold Rsqr; field.
split; [apply Rgt_not_eq; unfold Rgt; assumption|].
unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros; assumption .

replace (f x0' / f' x0') with ((f' X0 / f' x0') * (f x0' / f' X0));
   [|field; split; [apply Hdif_x0'|apply Hdif_f']] .
rewrite Rabs_mult.
replace ((3 + mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0) with
  (4 / (3 + mu a0 b0 c 0) * ((3 + mu a0 b0 c 0) ^ 2 / (16 * mu a0 b0 c 0) * b0)).
apply Rmult_le_compat; 
  [apply Rabs_pos|apply Rabs_pos|apply haux_abs_a' | assumption ].
field; split; apply Rgt_not_eq; unfold Rgt; 
  [|apply Rlt_trans with 3; [fourier|]; pattern 3 at 1 ; replace 3 with (3+0) by ring; 
  apply Rplus_lt_compat_l ] ; assumption.
Qed.


Theorem conv_stability2:
let xstar := (xs a0 b0 c X0 a b f f' A0_b0_c pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b Hb_eps) in
  let xstar2 := (xs a' b' c x0' a b f f' A0_b0_c' pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl' Hdif_x0' Habs_a' Habs_b' domain_x0') in 
   xstar = xstar2.
Proof.
intros xstar xstar2.
apply UL_sequence with (Xn x0' f f').
apply conv_stability; assumption.
unfold xstar2; unfold xs; case kantoro_exists_b; simpl; tauto.
Qed.

End Kprop.


Section Approx.

Variable (a0 b0 c:R).
Variable X0:R.


(*define f f' as section variables*)
Variable a:R.
Variable b:R.
Variable f:R->R.
Variable f':R->R.

Hypothesis pr: forall t:R, c_I a b t -> derivable_pt f t.
Hypothesis Hder_f: forall (t:R)(H:c_I a b t), derive_pt f t (pr t H)=f' t. 
Hypothesis Hcont_f':forall y, (c_I a b y) -> continuity_pt f' y.
Hypothesis Hlim_f':forall y1 y2:R, c_I a b y1 -> c_I a b y2 ->
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2)).
Hypothesis Hincl: a< X0 < b.
Hypothesis Hdif_f':f' X0 <> 0 .
Hypothesis Habs_a: Rabs (/(f' X0)) <= a0.
Hypothesis Habs_b:Rabs (f X0/(f' X0)) <= b0 .


Variable x0':R.
Hypothesis Hmu0 : 0 < mu a0 b0 c 0 < 1.
Hypothesis Hmudisc: included (c_disc X0 (2/mu a0 b0 c  0 * b0)) (c_I a b).
Hypothesis Hdom: Rabs (x0' - X0) <= (1-mu a0 b0 c  0)/(4*mu a0 b0 c 0)*b0.

Definition a0' :=  8 /(7 + mu a0 b0 c 0) * a0.
Definition b0' := let mu0 := mu a0 b0 c 0 in
  ((mu0 ^2 + 46*mu0 + 17) / (8 * (7 + mu0)* mu0) * b0).

Lemma a0_gt_0:
 0 < a0.
Proof.
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
 unfold mu in Hmu0; simpl in Hmu0.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
assert (ha0 := H_a0 a0 X0 f' Habs_a).
elim ha0; intro ha1; [assumption| elim H05; subst; trivial].
Qed.


Lemma b0_gt_0:
 0 < b0.
Proof.
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
 unfold mu in Hmu0; simpl in Hmu0.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
assert (ha0 := H_b0 b0 X0 f f' Habs_b).
elim ha0; intro ha1; [assumption| elim H04; subst; trivial].
Qed.

Lemma c_gt_0:
 0 < c.
Proof.
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
 unfold mu in Hmu0; simpl in Hmu0.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
assert (ha0 := H_c c X0 a b f' Hlim_f' Hincl).
elim ha0; intro ha1; [assumption| elim H02; subst; trivial].
Qed.

Lemma ap_disc_x0': included (c_disc x0' (2 * b0'))
  (c_disc X0 (2 / mu a0 b0 c 0 * b0)).
Proof.
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
intros x Hdisc.
unfold c_disc in Hdisc |-*.
replace (X0 - x) with (X0 - x0' + (x0' - x)) by ring.
apply Rle_trans with (Rabs (X0 - x0') + Rabs (x0' - x)); [apply Rabs_triang|].
apply Rle_trans with ( (1-mu a0 b0 c  0)/(4*mu a0 b0 c 0)*b0 + 2 * b0').
apply Rplus_le_compat; [rewrite Rabs_minus_sym|]; assumption.
unfold b0'.
 set (mu0 := mu a0 b0 c 0) in *.
replace ((1 - mu0) / (4 * mu0) * b0 +
2 * ((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0)*mu0) * b0)) with
(( (1 - mu0) / (4 * mu0)  +
2 * ((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0)*mu0))) * b0) by ring.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
replace ( (1 - mu0) / (4 * mu0) +
2 * ((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0) * mu0)) ) with 
((( (1 - mu0)* (7+ mu0) + (mu0 ^ 2 + 46 * mu0 + 17)) / (4 * (7 + mu0))) / mu0);
[|field; split; apply Rgt_not_eq; fourier].
unfold Rdiv; apply Rmult_le_compat_r; [left; apply Rinv_0_lt_compat; assumption|].
apply Rmult_le_reg_l with (4 * (7 + mu0)); [apply Rmult_lt_0_compat; fourier|].
rewrite <-Rmult_assoc; rewrite Rinv_r_simpl_m; 
  [|apply Rgt_not_eq; unfold Rgt; apply Rmult_lt_0_compat; fourier].
ring_simplify.
replace (40*mu0 + 24) with (8 * mu0 + (32 * mu0 + 24)) by ring.
apply Rplus_le_compat_l.
apply Rplus_le_reg_l with (- 24); ring_simplify.
pattern 32 at 2; replace 32 with (32*1) by ring.
apply Rmult_le_compat_l; fourier.
Qed.

Lemma ap_domain_x0': included (c_disc x0' (2 * b0'))
  (c_I a b).
Proof.
unfold b0'.
apply included_trans with (c_disc X0 (2 / mu a0 b0 c 0 * b0));
[apply ap_disc_x0'|]; assumption.
Qed.

Lemma ap_gt_A0_b0_c': 0 < 2* a0' * b0' * c < 1.
 Proof.
unfold a0'; unfold b0'.
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
set  (mu0 := mu a0 b0 c 0) in *.

split.

unfold Rdiv; repeat apply Rmult_lt_0_compat; try fourier.
apply Rinv_0_lt_compat; fourier.
apply a0_gt_0.
apply Rlt_trans with (mu0 ^ 2); 
  [unfold pow;  apply  Rmult_lt_0_compat|]; fourier.
apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat.
apply Rmult_lt_0_compat.
  fourier.
fourier.
assumption.
apply b0_gt_0.
apply c_gt_0.
replace (2 * (8 / (7 + mu0) * a0) *
((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0) * mu0) * b0) * c) with
((2 * a0 * b0 * c / mu0)  * (8 / (7 + mu0) ) *
((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0) ) ) ) ;
  [|field; split; apply Rgt_not_eq; fourier].
replace (2 * a0 * b0 * c / mu0) with 1; [ring_simplify|].
replace (8 / (7 + mu0) * ((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0))))
with (((mu0 ^ 2 + 46 * mu0 + 17) / ((7 + mu0) ^ 2))); 
  [|field; apply Rgt_not_eq; fourier].
apply Rmult_lt_reg_l with ((7+mu0)^2); 
  [unfold pow; repeat apply Rmult_lt_0_compat; fourier |].
unfold Rdiv; rewrite <-Rmult_assoc; rewrite Rinv_r_simpl_m; 
 [|apply Rgt_not_eq; unfold Rgt; unfold pow; repeat apply Rmult_lt_0_compat; fourier].
unfold pow; ring_simplify.
repeat rewrite Rplus_assoc; apply Rplus_lt_compat_l.
replace (46*mu0 + 17) with (14*mu0 + (32*mu0 + 17)) by ring.
apply Rplus_lt_compat_l.
replace 49 with (32 + 17) by ring; apply Rplus_lt_compat_r.
pattern 32 at 2; replace 32 with (32*1) by ring.
apply Rmult_lt_compat_l; fourier.
 unfold mu0; unfold mu; simpl; field.
repeat split; apply Rgt_not_eq; [apply c_gt_0 | apply b0_gt_0|apply a0_gt_0].
Qed.
 
Lemma ap_A0_b0_c': 2* a0' * b0' * c <= 1.
 Proof.
left; elim ap_gt_A0_b0_c'; tauto.
Qed.

Lemma ap_Hincl': a < x0' < b.
Proof.
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
set  (mu0 := mu a0 b0 c 0) in *.
apply disc_interval with (2 * b0'); 
  [|apply ap_domain_x0'].
unfold b0'; unfold Rdiv; fold mu0;  repeat apply Rmult_lt_0_compat; [fourier| | |].
apply Rlt_trans with (mu0 ^ 2); [unfold pow; rewrite Rmult_1_r;
apply Rmult_lt_0_compat|]; fourier.
apply Rinv_0_lt_compat; unfold pow; apply Rmult_lt_0_compat.
apply Rmult_lt_0_compat.
fourier.
fourier.
assumption.
apply b0_gt_0.
Qed.

Lemma ap_Hdif_x0': f' x0' <> 0.
Proof.
assert (hmu:= Hmu0).
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
set  (mu0 := mu a0 b0 c 0) in *.
assert (Hmu2: (c_I a b x0')).
apply Hmudisc; unfold c_disc.
 apply Rle_trans with ((1 - mu0) / (4 * mu0) * b0).
rewrite Rabs_minus_sym; assumption.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
unfold Rdiv.
rewrite Rinv_mult_distr; [|apply Rgt_not_eq; fourier|apply Rgt_not_eq; assumption].
 rewrite <-Rmult_assoc; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; assumption.
apply Rle_trans with (1*/4); [|fourier].
apply Rmult_le_compat_r; [fourier|].
pattern 1 at 2; replace 1 with (1-0) by ring; unfold Rminus; 
  apply Rplus_le_compat_l; fourier.
assert (h1: Rabs (1 - f'(x0') / f' (X0)) <= (1 - mu a0 b0 c 0)/8 ).

replace ((1 - mu a0 b0 c 0)/8) with (a0*c*((1- mu a0 b0 c 0)/(4*mu a0 b0 c 0) * b0)). 
apply Rle_trans with (a0*c*Rabs (x0' - X0)).
apply Rle_trans with (Rabs (/f'(X0)) * Rabs (f'(x0' ) - f'(X0))).
rewrite <- Rabs_mult; rewrite Rabs_minus_sym; right; f_equal; field; exact Hdif_f'.
rewrite Rmult_assoc; apply Rmult_le_compat; 
  [apply Rabs_pos| apply Rabs_pos| assumption |].
  apply Hlim_f'; [assumption|].
assert ( Hb_eps1:  included (c_disc X0 (2 * b0)) (c_I a b)).
apply Hb_eps with a0 c f f'; trivial. 
 apply Hb_eps1; apply center_in_disc.
  unfold Rdiv; repeat apply aux1.
   fourier.
   apply H_b0 with X0 f f'; assumption.
   (*rewrite (Rmult_assoc). (a0*c)  ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0)) b0).*)
apply Rmult_le_compat_l; [apply aux1; 
  [apply H_a0 with X0 f'|apply H_c with X0 a b f']; assumption|exact Hdom].
  unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*; field.
repeat split; apply Rgt_not_eq; [apply c_gt_0 | apply b0_gt_0|apply a0_gt_0].
assert (h11: Rabs (1 - f'(x0') / f' (X0)) <= 1/8).
apply Rle_trans with ((1 - mu a0 b0 c 0) / 8); [assumption|].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
  pattern 1 at 2; replace 1 with (1 - 0) by ring.
  unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar; left; assumption.
assert (h2: f' (x0') <> 0 /\ /f' X0 <>0).
apply Rlt_0_mult_reg .
fold (Rdiv (f' x0') (f' X0)); apply Rgt_not_eq; apply dif_0; apply Rle_trans with (1/8);
  [assumption|fourier].
elim h2; clear h2; intros h2 h0'.
assumption.
Qed.


Lemma ap_haux_abs_a':
  Rabs (f' (X0)/ f' (x0')) <= 8 /(7 + mu a0 b0 c 0).
Proof.
assert (hmu:= Hmu0).
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
set (mu0 := mu a0 b0 c 0) in *.
assert (Hmu2: (c_I a b x0')).
apply Hmudisc; unfold c_disc; apply Rle_trans with ((1 - mu0) / (4 * mu0) * b0).
rewrite Rabs_minus_sym; assumption.
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption|].
unfold Rdiv.
rewrite Rinv_mult_distr; [|apply Rgt_not_eq; fourier|apply Rgt_not_eq; assumption].
 rewrite <-Rmult_assoc; apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; assumption.
apply Rle_trans with (1*/4); [|fourier].
apply Rmult_le_compat_r; [fourier|].
pattern 1 at 2; replace 1 with (1-0) by ring; 
  unfold Rminus; apply Rplus_le_compat_l; fourier.
assert (h1: Rabs (1 - f'(x0') / f' (X0)) <= (1 - mu0)/8 ).
replace ((1 - mu0)/8) with (a0*c*((1- mu0)/(4*mu0) * b0)). 
apply Rle_trans with (a0*c*Rabs (x0' - X0)).
apply Rle_trans with (Rabs (/f'(X0)) * Rabs (f'(x0' ) - f'(X0))).
rewrite <- Rabs_mult; rewrite Rabs_minus_sym; right; f_equal; field; exact Hdif_f'.
rewrite Rmult_assoc; apply Rmult_le_compat; 
  [apply Rabs_pos| apply Rabs_pos| assumption |].
assert ( Hb_eps1:  included (c_disc X0 (2 * b0)) (c_I a b)).
apply Hb_eps with a0 c f f'; trivial. 
  apply Hlim_f'; [assumption|apply Hb_eps1; apply center_in_disc]. 
  unfold Rdiv; repeat apply aux1.
   fourier.
   apply H_b0 with X0 f f'; assumption.
   (*rewrite (Rmult_assoc). (a0*c)  ((1 - mu a0 b0 c 0) / (2 * mu a0 b0 c 0)) b0).*)
apply Rmult_le_compat_l; [apply aux1; 
  [apply H_a0 with X0 f'|apply H_c with X0 a b f']; assumption|exact Hdom].

  unfold mu0 in *; unfold mu in * ; simpl in *; field.
repeat split; apply Rgt_not_eq; [apply c_gt_0 | apply b0_gt_0|apply a0_gt_0].
assert (h11: Rabs (1 - f'(x0') / f' (X0)) <= 1/8).
apply Rle_trans with ((1 - mu0) / 8); [assumption|].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
  pattern 1 at 2; replace 1 with (1 - 0) by ring.
  unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar; left; assumption.
assert (h2: f' (x0') <> 0 /\ /f' X0 <>0).
apply Rlt_0_mult_reg .
fold (Rdiv (f' x0') (f' X0)); apply Rgt_not_eq; apply dif_0; apply Rle_trans with (1/8);
  [assumption|fourier].
elim h2; clear h2; intros h2 h0'.

replace (f' X0 / f' x0') with (1 / (1 - (1 - f' (x0') / f' X0)));
   [|field; split; [apply ap_Hdif_x0'| apply Hdif_f']].
unfold Rdiv at 1; rewrite Rabs_mult; rewrite Rabs_Rinv;
[|replace (1 - (1 - f' x0' / f' X0)) with
  (f' x0' / f' X0) by ring; unfold Rdiv; apply prod_neq_R0; assumption ].
rewrite Rabs_right; [|fourier].
apply Rle_trans with (1*/ (1 - Rabs (1 - f' x0' / f' X0))).
apply  Rmult_le_compat_l; [fourier| ].
assert (hx: (0 < 1 - Rabs (1 - f' x0' / f' X0) <= Rabs (1 - (1 - f' x0' / f' X0)))).
split.
apply Rgt_minus; apply Rgt_ge_trans with (1/8); [fourier|apply Rle_ge ;apply h11].
pattern 1 at 1; replace 1 with (Rabs 1); [apply Rabs_triang_inv| apply Rabs_right; fourier].
apply Rle_Rinv; elim hx; intros hq hw; 
 [|apply Rlt_le_trans with ( 1 - Rabs (1 - f' x0' / f' X0))|];assumption.
assert (hx: 0 < (1 - (1 - mu0) / 8) <= (1 - Rabs (1 - f' x0' / f' X0)) ).
split.
field_simplify; unfold Rdiv; rewrite Rinv_1; rewrite Rmult_0_l .
apply Rmult_lt_0_compat; [|fourier].
 fourier.

unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_contravar .
unfold Rminus in h1; apply h1.
elim hx; intros hx1 hx2; clear hx.
apply Rle_trans with (1*/ (1 - (1 - mu0)/8)).
apply Rmult_le_compat_l; [fourier|].
apply Rle_Rinv; [|apply Rlt_le_trans with ( 1 - (1 - mu0) / 8)|]; assumption.
field_simplify ; [right; trivial| |]; apply Rgt_not_eq; fourier.
Qed.


Lemma ap_H_in_disc:  (c_I a b x0').
Proof.
unfold c_I; elim (ap_Hincl'); intros; split; intuition.
Qed.


Lemma ap_Habs_a':  Rabs (/f'(x0')) <= a0'.
Proof.
 elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
unfold a0'.
replace (/f' x0') with (f' X0 / f' x0' * / f' X0);  
  [|field; split; [apply ap_Hdif_x0'| apply Hdif_f']].
rewrite Rabs_mult.
apply Rmult_le_compat; 
  [apply Rabs_pos| apply Rabs_pos|apply ap_haux_abs_a' |assumption].
Qed.

Lemma ap_Habs_b':
  Rabs (f x0' / f' x0') <= b0'.
Proof.
assert (hmu:= Hmu0).
elim Hmu0; clear Hmu0; intros Hmu0 Hmu1.
unfold b0'.
set (mu0 := mu a0 b0 c 0) in *.
assert (h5: Rabs (f (x0') / f' (X0)) <= (mu0 ^2 + 46 * mu0 + 17) / (64 * mu0)*b0).
replace (f x0' / f' X0) with (/f' (X0)*
   (f x0' - f X0 - f' X0 *(x0' - X0)) + f X0 / f' X0 + (x0' - X0));
  [|field; assumption].
apply Rle_trans with (Rabs (/ f' X0 * (f x0' - f X0 - f' X0 * (x0' - X0)) + 
  f X0 / f' X0)  + Rabs (x0' - X0)).
apply Rabs_triang.
apply Rle_trans with (Rabs (/ f' X0 * (f x0' - f X0 - f' X0 * (x0' - X0))) + 
  Rabs (f X0 / f' X0) +Rabs (x0' - X0)).
apply Rplus_le_compat_r.
apply Rabs_triang.
rewrite Rabs_mult.
apply Rle_trans with (a0 * (1/2 * c * Rsqr (x0' - X0) ) + b0 + Rabs (x0' - X0)).
repeat apply Rplus_le_compat.
apply Rmult_le_compat; [apply Rabs_pos|apply Rabs_pos|assumption |].
replace (f x0' - f X0 - f' X0 * (x0' - X0)) with (f x0' - f X0 - f' X0 * x0' + f' X0* X0) by ring.
apply lemma_third with a b pr; try  assumption.
apply ap_H_in_disc.
assert ( Hb_eps1:  included (c_disc X0 (2 * b0)) (c_I a b)).
apply Hb_eps with a0 c f f'; trivial. 
apply Hb_eps1; apply center_in_disc.
(*unfold Rdiv; repeat apply aux1; [fourier|left; apply Rinv_0_lt_compat; 
  exact Hmu0|apply  H_b0 with X0 f f'] .
assumption.
apply Hb_eps; apply center_in_disc;*) 
  apply aux1; [fourier|apply H_b0 with X0 f f']; assumption.
assumption.
right; trivial.
apply Rle_trans with ((1/4 * mu0 * /b0) * ((Rsqr ( (1 - mu0) / (4 *(mu0)) * b0))) 
   + b0 + (1 - mu0)/ (4 * mu0) * b0 ).
apply Rplus_le_compat; [|assumption].
apply Rplus_le_compat_r.
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat.
apply aux1.
apply aux1.
apply H_a0 with X0 f'; assumption.
fourier.
apply H_c with X0 a b f'; assumption.
apply Rle_0_sqr.
unfold mu0; unfold mu; simpl; right. field.
apply Rgt_not_eq;  apply b0_gt_0.
apply Rsqr_le_abs_1.
pattern Rabs at 1; rewrite Rabs_right; 
 [|apply Rle_ge; apply Rle_trans with (Rabs (x0' - X0)); [apply Rabs_pos|]]; assumption.
right; unfold Rsqr; field.
split; [apply Rgt_not_eq; unfold Rgt; assumption|].
unfold mu in Hmu0 |-*;  simpl in Hmu0 |-*.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 _.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros; assumption .
replace (f x0' / f' x0') with ((f' X0 / f' x0') * (f x0' / f' X0)); [|field; split; 
  [apply ap_Hdif_x0'|apply Hdif_f']] .
rewrite Rabs_mult.
replace ((mu0 ^ 2 + 46 * mu0 + 17) / (8 * (7 + mu0) * mu0) * b0) with
( (8 / (7 + mu0)) * ((mu0 ^ 2 + 46 * mu0 + 17) / (64 * mu0) * b0)); 
  [|field; split; apply Rgt_not_eq; fourier].
apply Rmult_le_compat;
   [apply Rabs_pos|apply Rabs_pos|apply ap_haux_abs_a' | assumption ].
Qed.


Theorem ap_conv_stability2:
let xstar := (xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmu0) pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b 
                     (Hb_eps a0 b0 c X0 a b f f' Habs_b Hmu0 Hmudisc)) in
  let xstar2 := (xs a0' b0' c x0' a b f f' ap_A0_b0_c' pr Hder_f Hcont_f' Hlim_f' 
                          ap_Hincl' ap_Hdif_x0' ap_Habs_a' ap_Habs_b' ap_domain_x0') in 
   xstar = xstar2.
Proof.
intros xstar xstar2.
elim Hmu0;  intros Hmu00 Hmu1.
apply UL_sequence with (Xn x0' f f').
unfold xstar.

 apply conv_stability.
apply Rle_trans with ((1 - mu a0 b0 c 0) / (4 * mu a0 b0 c 0) * b0); 
  [assumption|].
unfold Rdiv;  apply Rmult_le_compat_r.
apply H_b0 with X0 f f' ; assumption.
apply Rmult_le_compat_l; [fourier|].
apply Rle_Rinv; fourier.
unfold xstar2; unfold xs; case kantoro_exists_b; simpl; tauto.
Qed.


Lemma conv_stability3:  
    {xstar | Un_cv (Xn x0' f f' ) xstar }.
Proof.
exists (xs a0' b0' c x0' a b f f' ap_A0_b0_c' pr Hder_f Hcont_f' Hlim_f' 
                          ap_Hincl' ap_Hdif_x0' ap_Habs_a' ap_Habs_b' ap_domain_x0').
 unfold xs; case kantoro_exists_b; simpl; tauto.
Qed.



End Approx.


Section ApproxSequence.


Variable (a0 b0 c:R).
Variable X0:R.

(*define f f' as section variables*)
Variable a:R.
Variable b:R.
Variable f:R->R.
Variable f':R->R.

Hypothesis pr: forall t:R, c_I a b t -> derivable_pt f t.
Hypothesis Hder_f: forall (t:R)(H:c_I a b t), derive_pt f t (pr t H)=f' t. 
Hypothesis Hcont_f':forall y, (c_I a b y) -> continuity_pt f' y.
Hypothesis Hlim_f':forall y1 y2:R, c_I a b y1 -> c_I a b y2 ->
      Rabs (f' y1 - f' y2) <= c * (Rabs (y1-y2)).
Hypothesis Hincl: a < X0 < b.
Hypothesis Hdif_f':f' X0 <> 0 .
Hypothesis Habs_a: Rabs (/(f' X0)) <= a0.
Hypothesis Habs_b:Rabs (f X0/(f' X0)) <= b0 .


(*we model the approximation that come in the computation with machine reals*)
(*it takes the real to approximate and a nat which corresponds to the precision *)
(*at which we are approximating*)

Variable aprox : R  -> nat -> R.

Fixpoint Yn n {struct n}: nat -> R :=
  match n with 
    |0%nat  => Xn X0 f f'
    |S n' => Xn (aprox (Yn n' 1) n) f f'
  end.

Definition Zn n := Yn n 0.

Fixpoint Aseq n {struct n}:  R :=
  match n with 
    |0%nat => a0 
    |S n' => a0' (An (Aseq n') 1%nat) (Bn (Aseq n') (Bseq n') c 1%nat) c  
end
  with 
Bseq n {struct n} : R :=
  match n with 
    |O =>  b0 
    |S n' => b0' (An (Aseq n') 1%nat) (Bn (Aseq n') (Bseq n') c 1%nat) c  
end.

Fixpoint Tn n {struct n} : R :=
match n with
|O => X0 
|S n' => let tn := Tn n' in (aprox (tn - f tn / f' tn) n)
end.

Lemma yn_as_xn:
forall n, Yn n 1 = Xn (Yn n 0) f f' 1.
Proof.
intros n; case n; simpl; reflexivity.
Qed. 

Lemma eq_tn_zn:
forall n, Zn n = Tn n.
Proof.
induction n; unfold Zn;  simpl; [trivial|].
rewrite yn_as_xn; simpl; unfold Zn in IHn; rewrite IHn; trivial.
Qed.

Hypothesis aprox_in_ab:
forall x n, o_I a b x  -> o_I a b (aprox  x n).

Definition Round n :=  /3^n * ((1 - (mu a0 b0 c 0)^2)/ (8* (mu a0 b0 c 0))*b0) .

Hypothesis Haprox:
forall n, 
Rabs ((aprox (Xn (Yn n 0) f f' 1) (S n)) - Xn (Yn n 0) f f' 1) <= Round n.

Hypothesis mu0_in_bound:  1/2 <= mu a0 b0 c 0 < 1. 



Hypothesis Hbla3: included (c_disc X0 (3*b0)) (c_I a b).

Lemma H2b0: included (c_disc X0 (2*b0)) (c_I a b).
Proof.
apply included_trans with (2:=Hbla3).
apply incl_c_disc.
apply Rmult_le_compat_r.
apply H_b0 with X0 f f'; assumption.
fourier.
Qed.

Lemma Hmuineq: 0 < mu a0 b0 c 0 < 1.
Proof.
elim mu0_in_bound; intros; split; fourier.
Qed.


(* rabs (aprox r n) - r <= 1/Z2R b ^ n  if it models truncate; but it should also be 
correlated with the number of digits that can be certified at each iteration*)

Lemma mult_neg: forall r1 r2, r1 <= 0 -> 0 <= r2 -> r1*r2 <= 0.
intros; replace 0 with (r1*0) by ring; apply Rmult_le_compat_neg_l; assumption.
Qed.

Lemma mu_bound:
 forall x, 1/2 <= x <= 1-> 
  x * ((x ^ 4 + 46 * x^2 + 17) / (8 * (7 + x^2) * x^2)) <= 1.
Proof.
intros x Hx.
elim Hx; intros.
assert (H1:  0 < 8 * (7 + x ^ 2) * x ^ 2).
apply Rmult_lt_0_compat; [apply Rmult_lt_0_compat|].
fourier.
replace 0 with (0 + 0) by ring.
apply Rplus_lt_compat; [fourier|apply pow_lt; fourier].
apply pow_lt; fourier.
apply Rmult_le_reg_l with ((8 * (7 + x ^ 2) * x ^ 2)); [assumption|].
replace (8 * (7 + x ^ 2) * x ^ 2 *
(x * ((x ^ 4 + 46 * x ^ 2 + 17) / (8 * (7 + x ^ 2) * x ^ 2))) ) with 
(x * ((x ^ 4 + 46 * x ^ 2 + 17) )) ; [|field].
apply Rplus_le_reg_l with (- 8 * (7 + x ^ 2) * x ^ 2 * 1); ring_simplify.
replace (x ^ 5 - 8 * x ^ 4 + 46 * x ^ 3 - 56 * x ^ 2 + 17 * x) with
 ((x - 1)*x*(x^3 - 7*x^2 + 39*x - 17)); [|ring].
apply mult_neg; [apply mult_neg; fourier|].
set (h := fun  x =>   x ^ 3 - 7 * x ^ 2 + 39 * x - 17).
assert (H2:= derive_increasing_interv_var2 0 1 h).
assert (H3:  forall t : R, c_I 0 1 t -> derivable_pt h t).
intros; unfold h; reg.
assert (H4: forall (t : R) (H5 : c_I 0 1 t), derive_pt h t (H3 t H5) = 
  3*t^2 - 14*t +39).
intros; unfold h; simpl; reg.
assert  (H5: (forall (t : R) (H5 : c_I 0 1 t), 0 <= derive_pt h t (H3 t H5))) .
intros t Ht; rewrite H4.
unfold Rminus; rewrite Rplus_assoc.
replace 0 with (0+0) by ring.
apply Rplus_le_compat.
apply aux1; [fourier|apply pow_le; unfold c_I in Ht; intuition].
replace (-(14*t)+39) with (14 * (1-t) + 25) by ring.
replace 0 with (0+0) by ring.
apply Rplus_le_compat; unfold c_I in Ht; elim Ht; intros;
[apply aux1 | ];  fourier.
assert (H6 := H2 H3 H5).
fold (h x).
apply Rle_trans with (h (1/2)).
unfold h; simpl.
field_simplify; fourier.
elim H; intro H'.
apply H6.
unfold c_I; split; fourier.
unfold c_I; split; [apply Rle_trans with (2:=H)|]; fourier.
assumption.
rewrite H'; right; trivial.
split; apply Rgt_not_eq; unfold Rgt; [fourier|].
replace 0 with (0 + 0) by ring.
apply Rplus_lt_compat; [fourier|apply pow_lt; fourier].
Qed.




Lemma Xn_add_terms:
forall r m n, Xn r f f' (m+n)%nat = Xn (Xn r f f' m) f f' n.
Proof.
intros r m n; induction n.
simpl; replace (m+0)%nat with m; trivial.
replace (m + S n)%nat with (S m + n)%nat by omega.
simpl; rewrite IHn; trivial.
Qed.


Lemma strict_mu_n:
forall n, 0 < mu (Aseq n) (Bseq n) c 0 < 1.
Proof.
intros  n; induction n.
destruct Hmuineq as [Hmu0 Hmu1].
simpl; split; assumption.
simpl.
assert (H1 : 0 < mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0 < 1).
replace (mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0) with
  (mu (Aseq n) (Bseq n) c 1); [|unfold mu; simpl; trivial].
set (mun := mu (Aseq n) (Bseq n) c 0).
fold mun in  IHn.
set (mun1 := mu (Aseq n) (Bseq n) c 1 ).
assert (Hmun: mun1 = mun^2).
unfold mun; unfold mun1; unfold mu; simpl; ring.
rewrite Hmun.
elim IHn; intros IH1 IH2.
split.
apply pow_lt; elim IHn; intuition.
unfold pow;  replace 1 with (1*1) by ring; apply Rmult_gt_0_lt_compat; ring_simplify.
trivial.
apply Rlt_0_1.
assumption.
assumption.
unfold mu; unfold a0'; unfold b0'; simpl.
set (mu1' := mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0).
fold mu1' in H1; elim H1; intros H11 H12.
replace (2 * (8 / (7 + mu1') * (2 * Aseq n)) *
((mu1' * (mu1' * 1) + 46 * mu1' + 17) / (8 * (7 + mu1') * mu1') *
 (Aseq n * Bseq n * Bseq n * c)) * c) with 
(2 * (8 / (7 + mu1') * (2 * Aseq n)) *
((mu1' * (mu1' * 1) + 46 * mu1' + 17) / 
  (8 * (7 + mu1') * (2* (2 * Aseq n) * (Aseq n * Bseq n * Bseq n * c) * c )) *
 (Aseq n * Bseq n * Bseq n * c)) * c); [|unfold mu1'; unfold mu; simpl; trivial] .
replace 
(2 * (8 / (7 + mu1') * (2 * Aseq n)) *
((mu1' * (mu1' * 1) + 46 * mu1' + 17) /
 (8 * (7 + mu1') * (2 * (2 * Aseq n) * (Aseq n * Bseq n * Bseq n * c) * c)) *
 (Aseq n * Bseq n * Bseq n * c)) * c) with
( (mu1' * (mu1' * 1) + 46 * mu1' + 17) / ( (7 + mu1')^2 )); [|field].
split.

apply Rmult_lt_reg_l with ((7+mu1')^2); 
  [unfold pow; repeat apply Rmult_lt_0_compat; unfold mu1' in *; fourier |].

unfold Rdiv; rewrite <-Rmult_assoc; rewrite Rinv_r_simpl_m;
 [|apply Rgt_not_eq; unfold Rgt; unfold pow; repeat apply Rmult_lt_0_compat; 
  unfold mu1' in *; fourier].
unfold pow; ring_simplify.

replace 0 with (0+0) by ring; rewrite Rplus_assoc.
apply Rplus_lt_compat.
apply pow_lt; assumption.
replace 0 with (0+0) by ring.
apply Rplus_lt_compat; [|unfold mu1' in *; fourier].
apply Rmult_lt_0_compat; unfold mu1' in *; fourier.


apply Rmult_lt_reg_l with ((7+mu1')^2); 
  [unfold pow; repeat apply Rmult_lt_0_compat; unfold mu1' in *; fourier |].

unfold Rdiv; rewrite <-Rmult_assoc; rewrite Rinv_r_simpl_m;
 [|apply Rgt_not_eq; unfold Rgt; unfold pow; repeat apply Rmult_lt_0_compat; 
  unfold mu1' in *; fourier].
unfold pow; ring_simplify.
repeat rewrite Rplus_assoc.
apply Rplus_lt_compat_l.

replace (46*mu1' + 17) with (14*mu1' + (32*mu1' + 17)) by ring.
apply Rplus_lt_compat_l.
replace 49 with (32 + 17) by ring; apply Rplus_lt_compat_r.
pattern 32 at 2; replace 32 with (32*1) by ring.
apply Rmult_lt_compat_l; unfold mu1' in *; fourier.

 unfold mu in IHn; simpl in IHn; elim IHn; intros IH1 IH2.
 assert (H00:= Rgt_not_eq _ _ IH1).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
do 3 (split;  try assumption).
apply Rgt_not_eq; unfold mu1' in *; fourier.
Qed.


Lemma mu_monot:
forall n,
mu (Aseq n) (Bseq n) c 0 <= mu (Aseq (S n)) (Bseq (S n)) c 0.
Proof.
destruct Hmuineq as [Hmu0 Hmu1].
intros  n.
induction n.
simpl; unfold a0'; unfold  b0'.
unfold mu at 2; unfold An; unfold Bn; simpl.
replace (2 * (8 / (7 + mu (2 * a0) (a0 * b0 * b0 * c) c 0) * (2 * a0)) *
((mu (2 * a0) (a0 * b0 * b0 * c) c 0 *
  (mu (2 * a0) (a0 * b0 * b0 * c) c 0 * 1) +
  46 * mu (2 * a0) (a0 * b0 * b0 * c) c 0 + 17) /
 (8 * (7 + mu (2 * a0) (a0 * b0 * b0 * c) c 0) *
  mu (2 * a0) (a0 * b0 * b0 * c) c 0) * (a0 * b0 * b0 * c)) * c) with

(2 * (8 / (7 + mu (2 * a0) (a0 * b0 * b0 * c) c 0) * (2 * a0)) *
((mu (2 * a0) (a0 * b0 * b0 * c) c 0 *
  (mu (2 * a0) (a0 * b0 * b0 * c) c 0 * 1) +
  46 * mu (2 * a0) (a0 * b0 * b0 * c) c 0 + 17) /
 (8 * (7 + mu (2 * a0) (a0 * b0 * b0 * c) c 0) *
  (2 * (2 * a0) * (a0 * b0 * b0 * c) * c )) * (a0 * b0 * b0 * c)) * c); [|unfold mu; trivial].
set (mu1' := mu (2 * a0) (a0 * b0 * b0 * c) c 0).
replace (2 * (8 / (7 + mu1') * (2 * a0)) *
((mu1' * (mu1' * 1) + 46 * mu1' + 17) /
 (8 * (7 + mu1') * (2 * (2 * a0) * (a0 * b0 * b0 * c) * c)) *
 (a0 * b0 * b0 * c)) * c) with 
( ((mu1' * (mu1' * 1) + 46 * mu1' + 17) /
 ( (7 + mu1')^2))).
set (mu0 := mu a0 b0 c 0).
replace mu1' with (mu0^2).
apply Rmult_le_reg_l with ((7 + mu0 ^ 2) ^ 2).
apply pow_lt; replace 0 with (0+0) by ring; 
  apply Rplus_lt_le_compat; [fourier|apply pow_le].
unfold mu0; unfold mu; simpl ; apply greater0 with X0 a b f f'; assumption .
replace ( (7 + mu0 ^ 2) ^ 2 *
((mu0 ^ 2 * (mu0 ^ 2 * 1) + 46 * mu0 ^ 2 + 17) / ( (7 + mu0 ^ 2) ^ 2))) with
((mu0 ^ 2 * (mu0 ^ 2 * 1) + 46 * mu0 ^ 2 + 17)); [|field].
ring_simplify.

apply Rplus_le_reg_l with (- mu0 ^ 4 - 46 * mu0 ^ 2 - 17); ring_simplify.
replace (mu0 ^ 5 - mu0 ^ 4 + 14 * mu0 ^ 3 - 46 * mu0 ^ 2 + 49 * mu0 - 17) with
( (mu0-1)^3*(mu0^2+2*mu0+17)) by ring.
assert (haux: forall r1 r2, r1 <= 0 -> 0 <= r2 -> r1*r2 <= 0).
intros; replace 0 with (r1*0) by ring; apply Rmult_le_compat_neg_l; assumption.
apply haux.
replace ((mu0 - 1) ^ 3 ) with ((mu0 - 1) * ((mu0 -1) ^ 2)) by ring.
apply haux; [unfold mu0; fourier|].
replace 2%nat with (2*1)%nat by omega; rewrite  pow_Rsqr; ring_simplify.
 apply Rle_0_sqr.
replace 0 with (0+0) by ring; apply Rplus_le_compat.
replace 0 with (0+0) by ring; apply Rplus_le_compat.
simpl pow; apply aux1; unfold mu0; fourier.
unfold mu0; fourier.
fourier.
apply Rgt_not_eq; unfold Rgt.
 replace 0 with (0+0) by ring; apply Rplus_lt_le_compat; [fourier|apply pow_le].
unfold mu0; unfold mu; simpl ; apply greater0 with X0 a b f f'; assumption .
unfold mu0; unfold mu1'; unfold mu; simpl; ring.
field.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
do 3 (split;  try assumption).
set (mu0 := mu a0 b0 c 0).
replace mu1' with (mu0^2); [|unfold mu0; unfold mu1'; unfold mu; simpl; ring].
apply Rgt_not_eq; unfold Rgt.
replace 0 with (0+0) by ring; apply Rplus_lt_le_compat; [fourier|apply pow_le].
unfold mu0; unfold mu; simpl ; apply greater0 with X0 a b f f'; assumption .
set (m := S n).
unfold mu; unfold  An; unfold Bn.
change (Aseq (S m)) with (a0' (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c).
change (Bseq (S m)) with (b0' (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c).
unfold a0'; unfold b0'.
change
((mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0 ^ 2 +
  46 * mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0 + 17) /
 (8 * (7 + mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0) *
  mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0) *
 Bn (Aseq m) (Bseq m) c 1) with
(((mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0 ^ 2 +
  46 * mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0 + 17) /
 (8 * (7 + mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0) *
  (2* (An (Aseq m) 1) * (Bn (Aseq m) (Bseq m) c 1) * c )) *
 Bn (Aseq m) (Bseq m) c 1)).
set (mu1' :=mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0).
assert (Hstr:  0 < mu (Aseq m) (Bseq m) c 0 < 1).
apply strict_mu_n ; split; assumption.

replace (2 * (8 / (7 + mu1') * An (Aseq m) 1) *
((mu1' ^ 2 + 46 * mu1' + 17) /
 (8 * (7 + mu1') * (2 * An (Aseq m) 1 * Bn (Aseq m) (Bseq m) c 1 * c)) *
 Bn (Aseq m) (Bseq m) c 1) * c) with ((mu1' ^ 2 + 46 * mu1' + 17) /
 ( (7 + mu1')^2)).
set (mu0:=2 * Aseq m * Bseq m * c).
replace (mu1') with (mu0^2); [|unfold mu1'; unfold mu0; simpl; unfold mu; simpl; ring].
apply Rmult_le_reg_l with ((7 + mu0 ^ 2) ^ 2).
apply pow_lt; replace 0 with (0+0) by ring; apply Rplus_lt_le_compat; 
  [fourier|apply pow_le].
unfold mu0; elim Hstr; intuition.
replace ((7 + mu0 ^ 2) ^ 2 *
   (((mu0 ^ 2) ^ 2 + 46 * mu0 ^ 2 + 17) / (7 + mu0 ^ 2) ^ 2)) with
((mu0 ^ 2 * (mu0 ^ 2 ) + 46 * mu0 ^ 2 + 17)); [|field].
ring_simplify.
apply Rplus_le_reg_l with (- mu0 ^ 4 - 46 * mu0 ^ 2 - 17); ring_simplify.
replace (mu0 ^ 5 - mu0 ^ 4 + 14 * mu0 ^ 3 - 46 * mu0 ^ 2 + 49 * mu0 - 17) with
( (mu0-1)^3*(mu0^2+2*mu0+17)) by ring.
assert (Haux2: 0 < mu0 < 1).
unfold mu0; unfold  mu in Hstr ; unfold m; simpl in Hstr |-*; assumption.
elim Haux2; intros H1 H2.
assert (haux: forall r1 r2, r1 <= 0 -> 0 <= r2 -> r1*r2 <= 0).
intros; replace 0 with (r1*0) by ring; apply Rmult_le_compat_neg_l; assumption.
apply haux.
replace ((mu0 - 1) ^ 3 ) with ((mu0 - 1) * ((mu0 -1) ^ 2)) by ring.
apply haux; [ fourier|].
replace 2%nat with (2*1)%nat by omega; rewrite  pow_Rsqr; ring_simplify.
 apply Rle_0_sqr.
replace 0 with (0+0) by ring; apply Rplus_le_compat.
replace 0 with (0+0) by ring; apply Rplus_le_compat.
simpl pow; apply aux1; fourier.
 fourier.
fourier.
apply Rgt_not_eq; unfold Rgt.
 replace 0 with (0+0) by ring; apply Rplus_lt_le_compat; [fourier|apply pow_le].
intuition.
field.
split.
 assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros _ H02; assumption.

split. 
unfold Bn; unfold An.
elim Hstr; intros Hs1 Hs2.
assert (H00:= Rgt_not_eq _ _ Hs1).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
repeat apply prod_neq_R0; trivial.

split.
unfold Bn; unfold An.
elim Hstr; intros Hs1 Hs2.
assert (H00:= Rgt_not_eq _ _ Hs1).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
trivial.

replace mu1' with ((mu (Aseq m) (Bseq m) c 0)^2);
 [| unfold mu1'; unfold mu; simpl; ring].
apply Rgt_not_eq; unfold Rgt.
replace 0 with (0+0) by ring; apply Rplus_lt_le_compat; [fourier|apply pow_le].
intuition.
Qed.

(*We do not need an induction on n for the proof; the reasoning
   is the same in both cases
and the ind hyp. is not used *)

Lemma mu0_le_mu:
forall n, mu a0 b0 c 0 <= mu (Aseq n) (Bseq n) c 0.
Proof.
intros n; induction n.
simpl; right; trivial.
apply Rle_trans with (mu (Aseq n) (Bseq n) c 0); 
  [assumption| apply mu_monot; assumption].
Qed.

(* need proofs for quadratic equation*)



Lemma aseq_bseq_pos:
forall n, 0 < Aseq n /\ 0 < Bseq n .
Proof.
intros n.
elim Hmuineq; intros Hmu0 Hmu1.
unfold mu in Hmu0; simpl in Hmu0 |-*.
assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
induction n.
split.
assert (Ha:= H_a0 a0 X0 f' Habs_a).
elim Ha; intro Ha1; [assumption|elim H05; rewrite Ha1; trivial ]. 
assert (Ha:= H_b0 b0 X0 f f' Habs_b).
elim Ha; intro Ha1; [assumption|elim H04; rewrite Ha1; trivial ].
elim IHn; intros IHa IHb.
simpl; unfold a0', b0'.
replace ((mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0 )) with
 ((mu (Aseq n) (Bseq n) c 0)^2); [|unfold mu; simpl; ring].
assert (H:= strict_mu_n n).
(*set (mu0 := (mu (Aseq n) (Bseq n) c 0)).
fold mu0 in H.*)
split.
unfold Rdiv; do 2 apply Rmult_lt_0_compat.
fourier.
apply Rinv_0_lt_compat.
replace 0 with (0+0) by ring.
apply Rplus_lt_compat.
fourier.
apply pow_lt; intuition.
fourier.
assumption.
unfold Rdiv; do 2 apply Rmult_lt_0_compat.
replace 0 with (0+0+0) by ring.
apply Rplus_lt_compat.
apply Rplus_lt_compat.
rewrite <-pow_mult.
apply pow_lt; intuition.
apply Rmult_lt_0_compat.
fourier.
apply pow_lt; intuition.
fourier.
apply Rinv_0_lt_compat.
 apply Rmult_lt_0_compat.
apply Rmult_lt_0_compat.
fourier.
replace 0 with (0+0) by ring.
apply Rplus_lt_compat.
fourier.
apply pow_lt; intuition.
apply pow_lt; intuition.
repeat apply Rmult_lt_0_compat; assumption.
assert (Ha:= H_c c X0 a b f'  Hlim_f' Hincl).
elim Ha; intro Ha1; [assumption|elim H02; rewrite Ha1; trivial ].
Qed. 



Definition Rn n :=
  (1 - mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0)/
       (4 * mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0) *
       (Aseq n * Bseq n * Bseq n * c).


Lemma speed_conv_approx:
forall n, 
/ 3 * Rn n <= Rn (S n).
 Proof.
intro n.
unfold Rn.
set (as0 := 2 * Aseq n).
set (bs0 := (Aseq n * Bseq n * Bseq n * c)).
simpl.
fold as0; fold bs0.
set (mu0 := mu as0 bs0 c 0).
set (mus1' := mu (2 * a0' as0 bs0 c) 
  (a0' as0 bs0 c * b0' as0 bs0 c * b0' as0 bs0 c * c) c 0).
set (bs1' := (a0' as0 bs0 c * b0' as0 bs0 c * b0' as0 bs0 c * c)).
set (mus1 := mu (a0' as0 bs0 c) (b0' as0 bs0 c ) c 0).
assert (H1: mus1' = mus1^2).
unfold mus1'; unfold mus1; unfold mu; simpl;ring.
rewrite H1.
assert (H5: mu0 = (mu (Aseq n) (Bseq n) c 0)^2).
unfold mu0; unfold mu; unfold as0; unfold bs0; simpl; ring.
assert (H6 : 0 < mu0 <1).
assert (hh:=  strict_mu_n n).
rewrite H5; split; [ apply pow_lt; intuition|].
replace 1 with (1*1) by ring; simpl; apply Rmult_le_0_lt_compat;
 elim hh; intros; fourier.
destruct H6 as [H7 H8].
assert (H2 : bs1' = /2*mus1*(b0' as0 bs0 c)).
unfold bs1'; unfold mus1; unfold mu; simpl; field.
rewrite H2.
unfold b0'; fold mu0.
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
assert (HH := aseq_bseq_pos n).
unfold bs0; repeat apply aux1; [| | |apply H_c with X0 a b  f']; intuition.
assert (H3: mus1 = (mu0^2+46*mu0 + 17) / ((7+mu0)^2)).
unfold mus1;   unfold a0'; unfold b0' ; fold mu0; unfold mu; simpl.
replace (2 * (8 / (7 + mu0) * as0) *
((mu0 * (mu0 * 1) + 46 * mu0 + 17) / (8 * (7 + mu0) * mu0) * bs0) * c) with
 ((2*as0*bs0*c)*(8* / (7 + mu0))* 
  ((mu0 * (mu0 * 1) + 46 * mu0 + 17) / (8*(7 + mu0) * mu0))); [|field].
replace (2*as0*bs0*c) with mu0; [|unfold mu0; unfold mu; simpl; trivial]; field.
split; apply Rgt_not_eq; unfold Rgt; fourier.
split; apply Rgt_not_eq; unfold Rgt; fourier.
replace (1+4*2*2) with 17 by ring.
rewrite H3.
assert (Haux1 : 0 < 7 + mu0).
fourier.
assert (Haux2: 0 < (mu0 ^ 2 + 46 * mu0 + 17)).
replace 0 with (0+0) by ring; rewrite Rplus_assoc; 
  apply Rplus_lt_compat; [apply pow_lt|]; fourier.
set (up := (mu0 ^ 2 + 46 * mu0 + 17)).
set (low :=(7 + mu0) ).
replace ((1 - (up / low ^ 2) ^ 2) / (4 * (up / low ^ 2) ^ 2) * / 2 * (up / low ^ 2) *
(up / (4*2 * low * mu0))) with ((1 - (up / low ^ 2) ^ 2) / (4 * (up / low ^ 2))  * / 2  *
(up / (8 * low * mu0))); [|field].
replace ((1 - (up / low ^ 2) ^ 2) / (4 * (up / low ^ 2)) * / 2 *
(up / (8 * low * mu0))) with ((1 - (up / low ^ 2) ^ 2) / (4 * ( / low ^ 2)) * / 2 *
( / (8 * low * mu0))); [|field].
replace ((1 - (up / low ^ 2) ^ 2) / (4 * / low ^ 2) * / 2 * / (8 * low * mu0)) with 
( (1 - (up / low ^ 2) ^ 2) / (4 * / low ) * / 2 * / (8 * mu0)); [|field].
replace ((1 - (up / low ^ 2) ^ 2) / (4 * / low) * / 2 * / (8 * mu0)) with 
( /3 * 3  * (1 - (up / low ^ 2) ^ 2) /  / low *  /16 * / (4 * mu0)); [|field].
unfold Rdiv.
repeat rewrite Rmult_assoc. 
apply Rmult_le_compat_l; [fourier|].
repeat rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; fourier.
replace (4*2*2) with 16 by ring.
replace (3*(1 - (up * / low ^ 2) ^ 2) * / / low *  / 16) with 
(3* (1 - (up * / low ^ 2) ^ 2) *  low * / 16 ); [|field].
replace (3* (1 - (up * / low ^ 2) ^ 2) * low * / 16) with 
(3*(low^4-up^2) / low^3 */16); [|field].
unfold up; unfold low.
replace (3*((7 + mu0) ^ 4 - (mu0 ^ 2 + 46 * mu0 + 17) ^ 2) / (7 + mu0) ^ 3 * / 16) with
 (((1-mu0) * (3*32 * (2*mu0^2+60*mu0+66) / (7 + mu0) ^ 3 * / 16))); [|field].
pattern (1-mu0) at 1; replace (1-mu0) with ((1-mu0)*1) by ring.
apply Rmult_le_compat_l; [fourier|].
replace (3*32 * (2 * mu0 ^ 2 + 60 * mu0 + 66) / (7 + mu0) ^ 3 * / 16) with
(3 * 2* (2 * mu0 ^ 2 + 60 * mu0 + 66) / (7 + mu0) ^ 3); [|field].
apply Rmult_le_reg_l with ((7+mu0)^3).
apply pow_lt; assumption.
replace ((7 + mu0) ^ 3 * (3*2 * (2 * mu0 ^ 2 + 60 * mu0 + 66) / (7 + mu0) ^ 3)) with 
( (6 * (2 * mu0 ^ 2 + 60 * mu0 + 66))); [|field].
apply Rplus_le_reg_l with (-(7 + mu0) ^ 3 * 1).
ring_simplify.
replace (-mu0^3- 9 * mu0 ^ 2 + 213 * mu0 + 53) with 
  ( (- mu0^3 +1) + (-9*mu0^2+9*1) + 213*mu0 + 43) by ring.
apply Rplus_le_le_0_compat; [|fourier].
apply Rplus_le_le_0_compat; [|fourier].
apply Rplus_le_le_0_compat.
apply Rplus_le_reg_l with (mu0^3); ring_simplify.
replace 1 with (1^3) by ring.
apply pow_incr; split; fourier.
apply Rplus_le_reg_l with (9*mu0^2); ring_simplify.
pattern 9 at 2; replace 9 with (9*1) by ring; 
  apply Rmult_le_compat_l; [|replace 1 with (1^2) by ring; apply pow_incr; split]; fourier.
apply Rgt_not_eq; assumption.
apply Rgt_not_eq; assumption.
apply Rgt_not_eq; assumption.
apply Rgt_not_eq; assumption.
apply Rgt_not_eq; assumption.
split; apply Rgt_not_eq; assumption.
split; apply Rgt_not_eq; assumption.
repeat split; apply Rgt_not_eq; assumption.
repeat split; apply Rgt_not_eq; assumption.
Qed.



Lemma bound_Rn0:
Rn 0 <= 3/16*b0.
Proof.
elim Hmuineq; intros H0 H2.
unfold Rn; simpl.
replace (mu (2 * a0) (a0 * b0 * b0 * c) c 0) with 
  (mu a0 b0 c 0 ^2); [|unfold mu; simpl; ring].
replace (   (a0 * b0 * b0 * c)) with (/2*mu a0 b0 c 0 * b0); 
  [|unfold mu; simpl; field].
apply Rle_trans with ((1 - 1/4) / (4 * 1/4) * (/ 2 * 1/2 * b0)).
replace ((1 - mu a0 b0 c 0 ^ 2) / (4 * mu a0 b0 c 0 ^ 2) * (/ 2 * mu a0 b0 c 0 * b0))
with ((1 - mu a0 b0 c 0 ^ 2) / (4 * mu a0 b0 c 0 ) * (/ 2 * b0)); 
  [|field; apply Rgt_not_eq; intuition].
unfold Rdiv; repeat rewrite <-Rmult_assoc.
apply Rmult_le_compat_r.
apply H_b0 with   X0 f f'; assumption.
apply Rmult_le_compat_r; [fourier|].
repeat rewrite Rmult_assoc.
apply Rmult_le_compat.
assert (H1: mu a0 b0 c 0 ^ 2 <= 1).
replace 1 with (1*1) by ring; unfold pow; 
  apply Rmult_le_compat; fourier.
fourier.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat; fourier.
unfold Rminus; apply Rplus_le_compat; [right; trivial|].
apply Ropp_le_contravar.
elim mu0_in_bound; intros G1 G2.
replace (1*/4) with ((1/2) * (1/2)) by field; unfold pow; apply Rmult_le_compat;
 fourier.
replace ( /(2 * (2 * (1 * / 4))) * (/ 2 * 1)) with (/2); [|field].
apply Rle_Rinv; [fourier| apply Rmult_lt_0_compat; fourier|].
pattern 2 at 1; replace 2 with (2*1) by ring.
apply Rmult_le_compat_l; [fourier|].
apply Rmult_le_reg_l with (/2); [fourier| field_simplify].
elim mu0_in_bound; intros G1 G2.
replace ( mu a0 b0 c 0 / 1) with (mu a0 b0 c 0) by field; assumption.
repeat rewrite <-Rmult_assoc.
apply Rmult_le_compat_r.
apply H_b0 with   X0 f f'; assumption.
right; field.
Qed.



Lemma Hapr:
forall n, 
Rabs ((aprox (Xn (Yn n 0) f f' 1) (S n)) - Xn (Yn n 0) f f' 1) <=
(1 - mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0) /
(4 * mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0) *
(Aseq n * Bseq n * Bseq n * c).
Proof.
intros n; fold (Rn n).
assert (Hmuineq := Hmuineq).
apply Rle_trans with (Round n).
apply Haprox.
induction n.
unfold Round; unfold Rn; simpl.
replace (mu (2 * a0) (a0 * b0 * b0 * c) c 0) with (mu a0 b0 c 0 ^2); 
  [|unfold mu; simpl; ring].
replace (a0 * b0 * b0 * c) with (1/2* mu a0 b0 c 0 * b0); [|unfold mu; simpl; field].
replace ((1 - mu a0 b0 c 0 ^ 2) / (4 * mu a0 b0 c 0 ^ 2) * (1 / 2 * mu a0 b0 c 0 * b0))
with ((1 - mu a0 b0 c 0 ^ 2) / (8 * mu a0 b0 c 0 ) * b0); 
  [|field; apply Rgt_not_eq; intuition].
right; field; apply Rgt_not_eq; intuition.
apply Rle_trans with (/ 3 * Rn n ); [|apply speed_conv_approx].
unfold Round in IHn |-*; simpl.
replace (/ (3 * 3 ^ n)) with (/3* (/3^n)); [|field].
rewrite Rmult_assoc.
apply Rmult_le_compat_l; [fourier|apply IHn].
apply Rgt_not_eq; unfold Rgt;  apply pow_lt; fourier.
Qed.

Lemma disc_in_disc:
forall c1 c2 r, 
included (c_disc c2 r) (c_disc c1 (r+ (Rabs (c2 - c1)))).
Proof.
intros c1 c2 r x H2.
unfold c_disc in H2 |-*.
apply Rplus_le_reg_l with (- Rabs (c2 - c1)); ring_simplify.
rewrite Rplus_comm; apply Rle_trans with (Rabs (c1-x + (c2 - c1))).
rewrite <-(Rabs_Ropp (c2 - c1)).
assert (H:= Rabs_triang_inv (c1 - x) (-(c2 - c1))); 
  unfold Rminus in H |-*; rewrite Ropp_involutive in H;
assumption.
replace (c1 - x + (c2 - c1)) with (c2 - x) by ring; assumption.
Qed.

Lemma bseq_monot:
forall n, Bseq (S n) <= /2*Bseq n.
Proof.
intros n.
assert (Hmuineq := Hmuineq).
induction n.
replace (Bseq 1 ) with (b0' (An (Aseq 0) 1%nat) (Bn (Aseq 0) (Bseq 0) c 1%nat) c);
   [|simpl; trivial].
pattern (Bseq 0) at 2; replace (Bseq 0) with (b0 ) by (simpl; trivial).
unfold b0' .
set (mu0 :=mu a0 b0 c 0).
simpl Aseq; simpl Bseq.
set (mu1 := mu (An a0 1) (Bn a0 b0 c 1) c 0).
assert (Heq: mu1 = mu0^2).
unfold mu1; unfold mu0; unfold mu; simpl; ring.
assert (Hmu00: 0 < mu0).
unfold mu0; intuition.
rewrite rel_bn1 with (a0) (b0) c  ( X0) a b 0%nat; [|assumption].
simpl Npow.
simpl Bn; fold mu0.
ring_simplify.
(*pattern b0 at 2; replace b0 with (1*b0) by ring. *)
apply Rmult_le_compat_r.
apply H_b0 with (X0 ) f f'; assumption.
rewrite Rmult_comm; rewrite <-Rmult_assoc.
pattern (/2) at 2; replace (/2) with (1*/2) by ring.
apply Rmult_le_compat_r; [fourier|].
rewrite Heq; replace ((mu0^2)^2) with (mu0^4) ;[|unfold pow; simpl; ring].
apply mu_bound.
unfold mu0; elim mu0_in_bound; intros; split; fourier.
set (m:= S n).
replace (Bseq (S m) ) with (b0' (An (Aseq m) 1%nat) (Bn (Aseq m) (Bseq m) c 1%nat) c);
   [|simpl; trivial].
unfold b0' .
set (mu0 :=mu (Aseq m) (Bseq m) c 0).
set (mu1 := mu (An (Aseq m) 1) (Bn (Aseq m) (Bseq m) c 1) c 0).
assert (Heq: mu1 = mu0^2).
unfold mu1; unfold mu0; unfold mu; simpl; ring.
elim (strict_mu_n m); intros Hmu00 Hmu01.
fold mu0 in Hmu00 ; fold mu0 in Hmu01.
rewrite rel_bn1 with (Aseq m) (Bseq m) c  ( X0) a b 0%nat; [|assumption].
simpl Npow.
replace (Bn (Aseq m) (Bseq m) c 0) with (Bseq m); [|unfold Bn; simpl; trivial].
fold mu0.
ring_simplify.
(*pattern b0 at 2; replace b0 with (1*b0) by ring. *)
apply Rmult_le_compat_r.
assert (Hab:=aseq_bseq_pos m); intuition.
rewrite Rmult_comm; rewrite <-Rmult_assoc.
pattern (/2) at 2; replace (/2) with (1*/2) by ring.
apply Rmult_le_compat_r; [fourier|].
rewrite Heq; replace ((mu0^2)^2) with (mu0^4) ;[|unfold pow; simpl; ring].
apply mu_bound.
unfold mu0 in Hmu00, Hmu01 |-*; split; [|fourier].
apply Rle_trans with (mu a0 b0 c 0); 
  [elim mu0_in_bound; intros; assumption| unfold mu0; apply mu0_le_mu].
Qed.


Lemma aux_bseq_cv:
forall n,
Bseq  n <= b0/2^n.
Proof.
induction n.
simpl; fourier.
apply Rle_trans with (/2*Bseq n).
apply bseq_monot.
 simpl.
replace (b0/(2*2^n)) with (/2*(b0/2^n)); [|field; apply Rgt_not_eq; apply pow2_diff0].
apply Rmult_le_compat_l; [ fourier | assumption].
Qed.


Lemma Bseq_conv_0:
Un_cv Bseq 0.
Proof.
red.
intros eps0 Heps0.
assert (Hb0:  {n:nat| (b0/(2^n) < eps0)}).
apply archim2_le;[  fourier| auto].
apply H_b0 with X0 f f'; assumption.
elim Hb0;  intros N Hlt.
exists (S ( N)).
intros n HineqN.
apply Rle_lt_trans with (b0/2^n).
unfold R_dist.
replace (Bseq n - 0) with (Bseq n) by ring.
rewrite Rabs_pos_eq.
apply aux_bseq_cv.
assert (Hnn:= aseq_bseq_pos n); intuition.
apply Rle_lt_trans with ( b0 / 2 ^ N); [|assumption].
unfold Rdiv; apply Rmult_le_compat_l.
apply H_b0 with X0 f f'; assumption.
apply Rle_Rinv; [| | apply Rle_pow; [fourier|omega]]; apply pow2_diff0.
Qed.

Lemma round_eq:
forall n, Round n = /3^n*Round 0.
Proof.
intros n; induction n; [field|].
destruct Hmuineq as [Hmu0 Hmu1].
assert (H1: Round (S n) = /3* Round n).
unfold Round; simpl; field; split; apply Rgt_not_eq; 
  [| unfold Rgt; apply pow_lt]; fourier.
rewrite H1; rewrite IHn; simpl; field; 
  apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
Qed.

Lemma my_nat_ind:
forall P : nat -> Prop,
  P 0%nat -> 
  (forall n, (forall m: nat, (m <=n)%nat ->P m) -> P (S n)) -> 
    forall n : nat, P n.
Proof.
intros P H0 H.
intros n.
assert (H1: (forall m : nat, (m <= n)%nat -> P m) /\ P n).
induction n.
split.
intros m Hm; assert (heq: m = 0%nat) by omega.
rewrite heq; assumption.
assumption.
elim IHn; intros IH1 IH2.
split.
intros m Hm.
elim (le_lt_eq_dec m (S n) Hm); intros Hord.
apply IH1; omega.
rewrite Hord; apply H; assumption.
apply H; assumption.
elim H1; intros; assumption.
Qed.

Lemma  kanto_aprox: forall n,
0 < 2 * Aseq n * Bseq n * c < 1 /\
o_I a b (Yn n 0) /\
f' (Yn n 0) <> 0 /\
Rabs (/ f' (Yn n 0)) <= Aseq n /\
Rabs (f (Yn n 0) / f' (Yn n 0)) <= Bseq n /\
included (c_disc (Yn n 0) (2 * Bseq n)) (c_I a b) /\
(forall n0, (n0 <=n)%nat ->
   Rabs (Tn  (S n) - Tn 0) <= sum_f_R0  (fun n => /3^n) n * Round 0 
  + sum_f_R0 (fun n => /2^n) n * b0) /\
(included
  (c_disc (Xn (Yn n 0) f f' 1)
     (2 / mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0 *
      (Aseq n * Bseq n * Bseq n * c))) (c_disc X0 (3*b0))) .
Proof.
intros n.
assert (Hmuineq := Hmuineq).
induction n using my_nat_ind.
unfold mu in Hmuineq; simpl in Hmuineq |-*.
elim Hmuineq; elim Hincl; intros H1 H2 Hmu0 Hmu1.
split; [intuition|].
split; [assumption|].
split; [assumption|].
split; [assumption|].
split; [assumption|].
split; [apply H2b0|].
split.
intros m Hm.
change ((X0 - f X0 / f' X0)) with ( (Xn (Yn 0 0) f f' 1)).
replace (aprox (Xn (Yn 0 0) f f' 1) 1 - X0) with
  ((aprox (Xn (Yn 0 0) f f' 1) 1 - (Xn (Yn 0 0) f f' 1)) + ((Xn (Yn 0 0) f f' 1) - X0)) by ring.
apply Rle_trans with (Rabs (aprox (Xn (Yn 0 0) f f' 1) 1 - (Xn (Yn 0 0) f f' 1)) 
  + Rabs ((Xn (Yn 0 0) f f' 1) - X0)); [apply Rabs_triang|].
replace (/1) with 1 by field; ring_simplify.
apply Rplus_le_compat.
apply Haprox.
simpl;replace (X0 - f X0 / f' X0 - X0) with (- (f X0 / f' X0)) by ring;
rewrite Rabs_Ropp; assumption.

apply included_trans with 
  (c_disc X0 ( (2 / mu (2 * a0) (a0 * b0 * b0 * c) c 0 * (a0 * b0 * b0 * c)) +
  Rabs ( (X0 - f X0 / f' X0) - X0))); [apply disc_in_disc|].
apply incl_c_disc.
replace (X0 - f X0 / f' X0 - X0) with (- (f X0 / f' X0)) by ring.
rewrite Rabs_Ropp.
replace (2 / mu (2 * a0) (a0 * b0 * b0 * c) c 0 * (a0 * b0 * b0 * c)) with
  (/ mu a0 b0 c 0 * b0) ; [|unfold mu; simpl; field].
replace (3* b0) with (2*b0 + b0) by ring.
apply Rplus_le_compat; [|assumption].
apply Rmult_le_compat_r; [apply H_b0 with X0 f f'; assumption |].
replace 2 with (/ (1 / 2)) by field.
destruct mu0_in_bound as [Hmu12 Hmu11].
apply Rle_Rinv; [fourier|unfold mu; simpl; assumption | assumption].
unfold mu in Hmu0; simpl in Hmu0 |-*.
assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
do 3 (split; try assumption).
(*induction step*)
assert (Hleq: forall n, (n <= n)%nat).
intros n0; omega.
assert (IHn := H n (Hleq n)).

assert (H1 : 0 < mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0 < 1).
replace (mu (2 * Aseq n) (Aseq n * Bseq n * Bseq n * c) c 0) with
  (mu (Aseq n) (Bseq n) c 1); [|unfold mu; simpl; trivial].
rewrite rel_mu; elim IHn; intros JH1 JH2; unfold mu in *; 
elim JH1; intros; unfold pow. 
split; simpl;
 [apply Rmult_lt_0_compat |
 pattern 1 at 6; replace 1 with (1*1) by ring; apply Rmult_gt_0_lt_compat ]; fourier.
assert (H2: Yn n 1 = Xn (Yn n 0) f f' 1).
destruct n;  simpl; trivial.
destruct IHn as  [IH1 [IH2 [IH3 [IH4 [IH5 [IH6 [IH7 IH8]] ]]]]].
assert (IH0: 2 * Aseq n * Bseq n * c <= 1).
elim IH1; intros; intuition.
destruct (lemma_princ (Aseq n) (Bseq n) c (Yn n 0) a b f f' IH0 pr Hder_f  Hlim_f'
  IH2 IH3 IH4 IH5 IH6 1)
as   [lp1 [lp2[ lp3[ lp4[lp5[lp6 lp7]]]]]].
split.
simpl .
unfold mu; simpl. 
 apply ap_gt_A0_b0_c' with (Xn (Yn n 0) f f' 1) a b f f' 
   (aprox (Xn (Yn n 0) f f' 1) (S n)); try assumption.
fold (o_I a b (Xn (Yn n 0) f f' 1)); 
  apply in_open_xn with (Aseq n) (Bseq n) c pr; assumption .
(*apply Hbla2.*)
apply included_trans with (c_disc X0 (3*b0)); assumption.
apply Hapr.
split.
simpl.
rewrite yn_as_xn.
unfold o_I; apply aprox_in_ab.
apply  in_open_xn with (Aseq n) (Bseq n) c pr; try assumption.
assert (Hdif_aux: f' (Yn (S n) 0) <> 0).
simpl.
rewrite H2.
apply ap_Hdif_x0' with 
 (An (Aseq n) 1) (Bn (Aseq n) (Bseq n) c 1) c (Xn (Yn n 0) f f' 1) a b f;
  try assumption.
fold (o_I a b (Xn (Yn n 0) f f' 1)); 
  apply in_open_xn with (Aseq n) (Bseq n) c pr; assumption .
apply included_trans with (c_disc X0 (3*b0)); assumption.
unfold An; unfold Bn; simpl An; simpl Bn.
apply Hapr.
(*apply Hbla2.
unfold An; unfold Bn; simpl An; simpl Bn.*)
split; [assumption|].
split.
simpl.
rewrite H2.
apply ap_Habs_a' with  (Xn (Yn n 0) f f' 1) a b f; try assumption.
fold (o_I a b (Xn (Yn n 0) f f' 1)); 
  apply in_open_xn with (Aseq n) (Bseq n) c pr; assumption .
unfold An; unfold Bn; simpl An; simpl Bn.
apply included_trans with (c_disc X0 (3*b0)); assumption.
apply Hapr.
assert (Hb_aux: Rabs (f (Yn (S n) 0) / f' (Yn (S n) 0)) <= Bseq (S n)).
simpl.
rewrite H2.
apply  ap_Habs_b' with (Xn (Yn n 0) f f' 1) a b pr; try assumption.
fold (o_I a b (Xn (Yn n 0) f f' 1)); 
  apply in_open_xn with (Aseq n) (Bseq n) c pr; assumption .
unfold An; unfold Bn; simpl An; simpl Bn.
apply included_trans with (c_disc X0 (3*b0)); assumption.
apply Hapr.
split; [assumption|].
split.
simpl.
rewrite H2.
apply ap_domain_x0' with (Xn (Yn n 0) f f' 1) f f'; try assumption.
apply included_trans with (c_disc X0 (3*b0)); assumption.
apply Hapr.
assert (dif_approx: forall m:nat , (m <= S  n)%nat ->
 Rabs (Tn (S m) - Tn m) <=/3^m* Round 0 + /2^m * b0).
intros m Hm.
 simpl.
assert (H21:= Haprox m); simpl in H21.
replace (aprox (Tn m  - f (Tn m) / f' (Tn m)) (S m) - Tn m ) with 
  ( (aprox (Tn m - f (Tn m) / f' (Tn m)) (S m) -  (Tn m - f (Tn m) / f' (Tn m))) 
  +( (Tn m - f (Tn m) / f' (Tn m)) - Tn m)) by ring.
apply Rle_trans with 
 (Rabs ( aprox (Tn m - f (Tn m) / f' (Tn m)) (S m) -  (Tn m - f (Tn m) / f' (Tn m))) +
 Rabs ((Tn m - f (Tn m) / f' (Tn m)) - Tn m)); [apply Rabs_triang|].
apply Rplus_le_compat;
rewrite <-eq_tn_zn; unfold Zn.
rewrite <-round_eq; assumption.

replace (Yn m 0 - f (Yn m 0) / f' (Yn m 0) - Yn m 0) with (-(f (Yn m 0) / f' (Yn m 0))) by ring.
rewrite Rabs_Ropp.
elim (le_lt_eq_dec m (S n) Hm); intros Hord.
assert (Hm' : (m <= n)%nat) by omega.
destruct (H m Hm') as  [IHm1 [IHm2 [IHm3 [IHm4 [IHm5 [IHm6 [IHm7 IHm8]] ]]]]].
apply Rle_trans with (1:=IHm5).
 replace (/ 2 ^ m * b0) with (b0/2^m) ;[| field; apply Rgt_not_eq; apply pow2_diff0].
apply aux_bseq_cv.
rewrite <-Hord in Hb_aux; apply Rle_trans with (1:= Hb_aux).
 replace (/ 2 ^ m * b0) with (b0/2^m) ;[| field; apply Rgt_not_eq; apply pow2_diff0].
apply aux_bseq_cv.
assert (dif_approx2: (forall n0 : nat,
 (n0 <= S n)%nat ->
 Rabs (Tn (S (S n)) - Tn 0) <=
 sum_f_R0 (fun n1 : nat => / 3 ^ n1) (S n) * Round 0 +
 sum_f_R0 (fun n1 : nat => / 2 ^ n1) (S n) * b0) ).

intros m Hm.
assert (Heq: forall n, Tn (S n) - Tn 0 = sum_f_R0 (fun  n => (Tn (S n) - Tn n)) n).
intros m0; induction m0; [simpl; auto|].
rewrite tech5; rewrite <-IHm0; ring.
rewrite Heq.
apply Rle_trans with (sum_f_R0 (fun i : nat => Rabs ( Tn (S i) - Tn i)) (S n)).
apply Rsum_abs.
rewrite Rmult_comm.
rewrite (Rmult_comm (sum_f_R0 (fun n0 : nat => / 2 ^ n0) (S n) )).
repeat rewrite scal_sum.
rewrite <-(plus_sum  ).
apply sum_Rle.
intros m1 Hm1.
apply dif_approx; assumption.
split; [assumption|].
assert ( dif_approx3:
 Rabs ((Xn (Yn (S n) 0) f f' 1)  - Tn 0) <=
  sum_f_R0  (fun n => /3^n) n * Round 0 
  + sum_f_R0 (fun n => /2^n) (S n) * b0).


replace ((Xn (Yn (S n) 0) f f' 1)  - Tn 0) with
  ( ((Xn (Yn (S n) 0) f f' 1) - Tn (S n)) + (Tn (S n) - Tn 0)) by ring.
apply Rle_trans with (Rabs (Xn (Yn (S n) 0) f f' 1 - Tn (S n)) +
 Rabs( (Tn (S n) - Tn 0))); [apply Rabs_triang|].
rewrite tech5; rewrite Rmult_plus_distr_r;
rewrite <-Rplus_assoc.
 pattern Rplus at 1; rewrite Rplus_comm .
apply Rplus_le_compat; [|apply IH7 with n; omega].
rewrite <-eq_tn_zn; unfold Zn.
apply Rle_trans with (Bseq (S n)); 
  [|replace (/2^S n * b0) with (b0/2^S n) ; 
    [|field; apply Rgt_not_eq; apply pow2_diff0]; apply aux_bseq_cv].
change (Xn (Yn (S n) 0) f f' 1) with (Yn (S n) 0 - f (Yn (S n) 0) / f' (Yn (S n) 0)).

replace (Yn (S n) 0 - f (Yn (S n) 0) / f' (Yn (S n) 0) - Yn (S n) 0) with 
  (- (f (Yn (S n) 0) / f' (Yn (S n) 0)) ); [|field; assumption].
rewrite Rabs_Ropp; assumption.
set (p:=S n).
replace (mu (2 * Aseq p) (Aseq p * Bseq p * Bseq p * c) c 0) with
  (mu (Aseq p) (Bseq p) c 0^2); [|unfold mu; simpl; ring].
replace (Aseq p * Bseq p * Bseq p * c) with (/2* mu (Aseq p) (Bseq p) c 0 *Bseq p);
 [|unfold mu; simpl; field].
replace (2 / mu (Aseq p) (Bseq p) c 0 ^ 2 *
      (/ 2 * mu (Aseq p) (Bseq p) c 0 * Bseq p)) with
  ( / mu (Aseq p) (Bseq p) c 0 * Bseq p);
  [|field; apply Rgt_not_eq; unfold Rgt; elim (strict_mu_n p); intros ;assumption].
apply included_trans with (c_disc (Xn (Yn p 0) f f' 1) (2*Bseq p)).
apply incl_c_disc.
apply Rmult_le_compat. 
left; apply Rinv_0_lt_compat; elim (strict_mu_n p); intros; assumption.
elim ( aseq_bseq_pos p); intros; left; assumption.
replace 2 with (/ (1/ 2)) by field. 
apply Rle_Rinv; [fourier | elim (strict_mu_n p); intros; assumption| ].
destruct mu0_in_bound as [Hmu12 _].
apply Rle_trans with (1:= Hmu12).
apply  mu0_le_mu.
right; trivial.
apply included_trans with (c_disc (Tn 0) (2*Bseq p + Rabs (Xn (Yn p 0) f f' 1 - Tn 0 ))).
apply disc_in_disc.
(*apply included_trans with((c_disc X0 (3 * b0)) ); [|assumption].*)
simpl (Tn 0) at 1.
apply incl_c_disc.
(*destruct p.
simpl.
replace ((X0 - f X0 / f' X0 - X0)) with (- (f X0 / f' X0)) by ring.
replace (3*b0) with (2*b0 + b0) by ring.
apply Rplus_le_compat_l.
rewrite Rabs_Ropp; assumption.*)
unfold p.
apply Rle_trans with (2*Bseq (S n) + (sum_f_R0  (fun n => /3^n) n * Round 0 
  + sum_f_R0 (fun n => /2^n) (S n) * b0));
[apply Rplus_le_compat_l; apply dif_approx3|].
rewrite(
  sum_eq (fun n0 : nat => / 3 ^ n0) (fun n0 : nat => (/ 3) ^ n0) n).
rewrite(
  sum_eq (fun n0 : nat => / 2 ^ n0) (fun n0 : nat => (/ 2) ^ n0) (S n)).
repeat rewrite tech3.
assert (H4: forall n, 0 <= 3 - / 3 ^ n).
intros m; induction m; simpl; [fourier|].
apply Rle_trans with (1:= IHm).
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.
rewrite Rinv_mult_distr.
pattern (/3^m) at 2; replace (/3^m) with (1* /3^m) by ring.
apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; apply pow_lt.
fourier.
fourier.
apply Rgt_not_eq; fourier.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
assert (Heq2: (1 - (/ 3) ^ S n) / (1 - / 3) = (3 - /3^n) / 2).
assert (H3: forall n , ((/3)^n)  = /3^ n).
intros m; induction m. 
simpl; field.
simpl; rewrite IHm; field; apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
rewrite H3.
replace (1-/3) with (2/3) by field.
replace ((1 - / 3 ^ S n) / (2 / 3) ) with (((1 - / 3 ^ S n) / 2)* (/ / 3) ); [|field].
rewrite Rinv_involutive.
replace (/3^n) with (3/3^(S n)); [|simpl; field].
field.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
apply Rgt_not_eq; fourier.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.


apply Rle_trans with (2 * (b0 /2^ (S n)) +
((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 +
 (1 - (/ 2) ^ S (S n)) / (1 - / 2) * b0)).
apply Rplus_le_compat_r.
apply Rmult_le_compat_l; [fourier|apply aux_bseq_cv].
assert  (2 * (b0 / 2 ^ S n) +
((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 +
 (1 - (/ 2) ^ S (S n)) / (1 - / 2) * b0) = 
  ((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0)+ (2 + /2^(S n) )*b0).
rewrite Rplus_comm.
rewrite Rplus_assoc.
apply Rplus_eq_compat_l.
replace ((1 - (/ 2) ^ S (S n)) / (1 - / 2) * b0 + 2 * (b0 / 2 ^ S n) ) with
(((1 - (/ 2) ^ S (S n)) / (1 - / 2) + 2 * (/ 2 ^ S n))*b0) ; [|field].
assert (H3: forall n , ((/2)^n)  = /2^ n) .
intros m; induction m. 
simpl; field.
simpl; rewrite IHm; field; apply Rgt_not_eq; apply pow2_diff0.
rewrite H3.
simpl.
replace (1-/2) with (/2) by field.
replace (((1 - / (2 * (2 * 2 ^ n))) / / 2)) 
 with ((1 - / (2 * (2 * 2 ^ n))) * 2); [|field].
field.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
rewrite H0.
replace ((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 + (2 + / 2 ^ S n) * b0) with
 (2*b0 + ((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 + ( / 2 ^ S n) * b0)) by ring.
replace (3* b0 ) with (2*b0 + b0) by ring.
apply Rplus_le_compat_l.
apply Rle_trans with ((1 - (/ 3) ^ S n) / (1 - / 3) * (3/16*b0) + / 2 ^ S n * b0).
apply Rplus_le_compat; [|right; trivial].
apply Rmult_le_compat_l.

rewrite Heq2.
unfold Rdiv; apply aux1.

apply H4.
fourier.
assert (G1: Round 0 = Rn 0).
unfold Round, Rn, mu; simpl; field.
elim Hmuineq; intros Hmu0 Hmu1; unfold mu in Hmu0; simpl in Hmu0.
assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
do 3 (split;   try assumption).
rewrite G1.
apply bound_Rn0.
rewrite Heq2.
apply Rle_trans with ((3 - 0) / 2 * (3 / 16 * b0) + / 2 ^ S 0 * b0) .
apply Rplus_le_compat.
apply Rmult_le_compat.
unfold Rdiv; apply aux1; [apply H4|fourier].
apply aux1; [fourier|apply H_b0 with X0 f f'; assumption].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.
rewrite Rinv_pow; [apply pow_le|apply Rgt_not_eq]; fourier.
right; trivial.
apply Rmult_le_compat.
rewrite Rinv_pow; [apply pow_le|apply Rgt_not_eq]; fourier.
apply H_b0 with X0 f f'; assumption.
apply Rle_Rinv; 
  [apply pow_lt; fourier|apply pow_lt; fourier |apply Rle_pow; [fourier|omega]].
right; trivial.
replace ((3 - 0) / 2 * (3 / 16 * b0) + / 2 ^ 1 * b0) with (50/64*b0); [|field].
pattern b0 at 2; replace b0 with (1*b0) by ring.
apply Rmult_le_compat_r.
apply H_b0 with X0 f f'; assumption.
fourier.
apply Rlt_not_eq; fourier.
apply Rlt_not_eq; fourier.
intros; apply Rinv_pow; apply Rgt_not_eq; fourier.
intros; apply Rinv_pow; apply Rgt_not_eq; fourier.
Qed.


Lemma xn_aprox_conv:
let xstar := (xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmuineq)
       pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b 
     (H2b0 ) ) in
forall n , Un_cv (Yn n) xstar /\ Rabs (Yn n 0 - xstar) <=2*Bseq n .
Proof.
 intros xstar  p.

induction p.
simpl.
unfold xstar; unfold xs. 
case kantoro_exists_b; simpl; intros xst Hxst;
 elim Hxst; intros; intuition.

destruct IHp as [IHp1 IHp2]; 
simpl Yn.
destruct (kanto_aprox p) as [IH1 [IH2 [ IH3 [ IH4 [ IH5  [IH6 [IH7 IH8]]]]]]].
assert (Hx: 2 * Aseq p * Bseq p * c <= 1).
elim IH1; intros; intuition.
assert (Hun_cv:  Un_cv (Xn (Yn p 1) f f')  xstar).
destruct p.
unfold Yn.
unfold Yn in IHp1.

intros e He; 
  elim (IHp1 e He); intros N HN; exists N; intros n Hn.
rewrite <-Xn_add_terms.
apply HN; omega.

simpl Yn in IHp1.
replace (Yn (S p) 1) with ((Xn (aprox (Yn p 1) (S p)) f f') 1%nat); [|simpl; trivial].
intros e He; elim (IHp1 e He); intros N HN; exists N; intros n Hn.
replace (Xn (Xn (aprox (Yn p 1) (S p)) f f' 1) f f' n) with 
  (Xn (aprox (Yn p 1) (S p)) f f' (1 + n)%nat);
  [apply HN; omega|(apply Xn_add_terms)].


rewrite yn_as_xn.
assert (Hx1: a < Xn (Yn p 0) f f' 1  < b).
fold (o_I a b (Xn (Yn p 0) f f' 1 )).
 apply in_open_xn with (Aseq p) (Bseq p)  c pr; try assumption.

destruct  (lemma_princ (Aseq p) (Bseq p) c (Yn p 0) a b f f' Hx pr Hder_f  Hlim_f' IH2
    IH3 IH4 IH5 IH6 1%nat  ) as [Hlp1 [Hlp2 [Hlp3 [Hlp4 [Hlp5 [Hlp6 Hlp7]]]]]].

assert (Han: 2 * An (Aseq p) 1 * Bn (Aseq p) (Bseq p) c 1 * c <= 1).
apply  An_bn_c1 with (Yn p 0) a b f f'; assumption.

assert (Hlp8: included (c_disc (Xn (Yn p 0) f f' 1) (2 * Bn (Aseq p) (Bseq p) c 1)) (c_I a b)).
apply included_trans with (2:= Hlp2).
apply incl_c_disc.
apply Rmult_le_reg_l with (/2); [fourier|].
replace (/2*(2*Bn (Aseq p) (Bseq p) c 1)) with (Bn (Aseq p) (Bseq p) c 1) by field.
replace (/ 2 * En (Bseq p) 1) with ( En (Bseq p) 1/2) by field.
apply leq_en_bn with (Yn p 0) a b f f'; assumption.

assert  (Hmu01 : 0 < mu (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c 0 < 1).
unfold mu; simpl.
replace (2 * (2 * Aseq p) * (Aseq p * Bseq p * Bseq p * c) * c ) with 
  ((2 * Aseq p * Bseq p * c )*(2 * Aseq p * Bseq p * c)) by ring.
generalize IH1; generalize ( 2 * Aseq p * Bseq p * c ); intros r Hr; elim Hr; intros ; split.
replace 0 with (0*r) by ring; apply Rmult_lt_compat_r; assumption.
replace 1 with (1*1) by ring; apply Rmult_gt_0_lt_compat; fourier.
assert (Hmudisc : included
                       (c_disc (Xn (Yn p 0) f f' 1)
                          (2 /
                           mu (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c 0 *
                           Bn (Aseq p) (Bseq p) c 1)) (c_I a b)).
simpl  An; simpl  Bn.
apply included_trans with (c_disc X0 (3*b0)); assumption.

assert   (Hdom : Rabs (aprox (Yn p 1) (S p) - Xn (Yn p 0) f f' 1) <=
                  (1 - mu (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c 0) /
                  (2 * mu (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c 0) *
                  Bn (Aseq p) (Bseq p) c 1).
simpl An in Hmu01; simpl Bn in Hmu01.
simpl An; simpl Bn; rewrite yn_as_xn.
apply Rle_trans with (1:= Hapr p).
apply Rmult_le_compat_r.
replace ((Aseq p * Bseq p * Bseq p * c) ) with (Bn  (Aseq p) (Bseq p) c 1);
  [|simpl; trivial].
apply Bn_ge_0 with  (Yn p 0) a b f f'; assumption.
unfold Rdiv;  apply Rmult_le_compat_l.
 elim Hmu01; intros; fourier.
apply Rle_Rinv; elim Hmu01; intros.
apply Rmult_lt_0_compat; fourier.
apply Rmult_lt_0_compat; fourier.
apply Rmult_le_compat_r; fourier.

assert (Hstab:= conv_stability2
 (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c ((Xn (Yn p 0) f f' 1)) 
  a b f f' pr Hder_f Hcont_f' Hlim_f' Hx1
    Hlp4 Hlp5 Hlp6   (aprox (Yn p 1) (S p))  Hmu01 Hmudisc Hdom).
assert (Heq2: (Xn (Yn p 0) f f' 1) =(Yn p 1)).
destruct p; simpl; trivial.
assert (Hapr1:= Hapr p).
rewrite Heq2 in Hapr1.
pattern (Yn p 1) at 2 in Hapr1; rewrite <-Heq2 in Hapr1.

assert (Hstab2:=  ap_conv_stability2  
  (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c ((Xn (Yn p 0) f f' 1)) 
  a b f f' pr Hder_f Hcont_f' Hlim_f' Hx1
    Hlp4 Hlp5 Hlp6  (aprox (Yn p 1) (S p))  Hmu01 Hmudisc Hapr1).

assert (Hcv1: Un_cv (Xn (aprox (Xn (Yn p 0) f f' 1) (S p)) f f')
  (xs (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c
             (Xn (Yn p 0) f f' 1) a b f f'
             (A0_b0_c (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c Hmu01) pr
             Hder_f Hcont_f' Hlim_f' Hx1 Hlp4 Hlp5 Hlp6
             (Hb_eps (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c
                (Xn (Yn p 0) f f' 1) a b f f' Hlp6 Hmu01 Hmudisc))).
apply conv_stability.
replace (aprox (Xn (Yn p 0) f f' 1) (S p)) with (aprox (Yn p 1) (S p));
 [| rewrite yn_as_xn; trivial].
assumption.

assert (Hcv2: Un_cv (Xn (Xn (Yn p 0) f f' 1)  f f')
  (xs (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c
            (Xn (Yn p 0) f f' 1) a b f f'
            (A0_b0_c (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c Hmu01) pr
            Hder_f Hcont_f' Hlim_f' Hx1 Hlp4 Hlp5 Hlp6
            (Hb_eps (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c
               (Xn (Yn p 0) f f' 1) a b f f' Hlp6 Hmu01 Hmudisc))).
unfold xs; case kantoro_exists_b; simpl; intuition.
assert (Heq: xstar  = 
  (xs (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c
            (Xn (Yn p 0) f f' 1) a b f f'
            (A0_b0_c (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c Hmu01) pr
            Hder_f Hcont_f' Hlim_f' Hx1 Hlp4 Hlp5 Hlp6
            (Hb_eps (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c
               (Xn (Yn p 0) f f' 1) a b f f' Hlp6 Hmu01 Hmudisc))).
apply UL_sequence with ((Xn (Xn (Yn p 0) f f' 1) f f')).
rewrite <- yn_as_xn; assumption.
assumption.
split; rewrite Heq.
assumption.
rewrite Hstab2.
replace (Bseq (S p)) with (b0' (An (Aseq p) 1) (Bn (Aseq p) (Bseq p) c 1) c).
unfold xs;  case kantoro_exists_b; simpl.
intros x [ _ [H _]]; unfold c_disc in H.
rewrite <-Heq2 in H; simpl in H.
assumption.
simpl; trivial.
Qed.

Lemma zn_bound:
let xstar := (xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmuineq)
         pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b H2b0) in
forall n, Rabs (Zn n - xstar) <= 2* Bseq n.
Proof.
intros xstar n.
assert (HH:=xn_aprox_conv n).
fold xstar in HH.
unfold Zn; intuition.
Qed.

Lemma zn_conv:
let xstar := (xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmuineq)
    pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b H2b0) in
Un_cv Zn xstar .
Proof.
intros xstar.
assert (HH:= zn_bound).
intros e He.
assert (JJ:= Bseq_conv_0).
unfold Un_cv in JJ.
assert (Hd2: e/2>0).
fourier.
elim (JJ (e/2) Hd2); intros N Hbcv.
exists N.
intros n Hineq2.
unfold R_dist.
apply Rle_lt_trans with (2*Bseq n).
apply (HH n).
assert (Hbseq := Hbcv n Hineq2).
unfold R_dist in Hbseq; rewrite Rminus_0_r in Hbseq.
replace (Bseq n) with (Rabs (Bseq n)).
fourier.
apply Rabs_right; unfold Rge.
assert (KK:= aseq_bseq_pos n); intuition.
Qed.

Lemma tn_conv:
let xstar := (xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmuineq) 
    pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b H2b0) in
Un_cv Tn xstar .
Proof.
intro xstar; red.
intros e He; elim (zn_conv e He);
 intros N; exists N; intros;
rewrite <- (eq_tn_zn ); auto.
Qed.

Lemma conv_tn_speed:
forall n,
let xstar := (xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmuineq) 
    pr Hder_f Hcont_f' Hlim_f' Hincl Hdif_f' Habs_a Habs_b H2b0) in
Rabs (Tn n - xstar) <= / Rpower 2 (INR n - 1) * b0.
Proof.
intros n xstar.
destruct (xn_aprox_conv n) as [_ H].
fold xstar in H.
induction n; [simpl in H |-*|].
replace ( / Rpower 2 (0 - 1) ) with 2; [assumption|].
rewrite <-Rpower_Ropp.
replace (-(0-1)) with 1 by ring.
rewrite Rpower_1; auto with real.
apply Rle_trans with ( 2 * Bseq (S n)).
rewrite <-eq_tn_zn; unfold Zn; assumption.
apply Rle_trans with (2* (b0/2^(S n))).
apply Rmult_le_compat_l; [fourier|].
apply aux_bseq_cv.
right.
replace (INR (S n) - 1) with (INR n).
rewrite (Rpower_pow); simpl pow.
 field; apply Rgt_not_eq; apply pow2_diff0.
fourier.
rewrite S_INR; ring.
Qed.



Lemma eq_lim_tn_xn:
forall xs0, Un_cv Tn xs0 -> Un_cv (Xn X0 f f') xs0.
Proof.
intros xs0 Hcv.
set (xstar := xs a0 b0 c X0 a b f f' (A0_b0_c a0 b0 c Hmuineq)
      pr Hder_f Hcont_f' Hlim_f' 
                                   Hincl Hdif_f' Habs_a Habs_b H2b0).
assert (Hcv2 := tn_conv).
fold xstar in Hcv2.
assert (H2:  xs0 = xstar).
apply UL_sequence with Tn; assumption.
rewrite H2.
unfold xstar; unfold xs; case kantoro_exists_b; simpl; intuition.
Qed.

Lemma dif_approx: 
forall n, Rabs (Tn (S n) - Tn n) <=/3^n* Round 0 + /2^n * b0.
Proof.
intros n.
 simpl.
assert (H2:= Haprox n); simpl in H2.
replace (aprox (Tn n  - f (Tn n) / f' (Tn n)) (S n) - Tn n ) with 
  ( (aprox (Tn n - f (Tn n) / f' (Tn n)) (S n) -  (Tn n - f (Tn n) / f' (Tn n))) 
  +( (Tn n - f (Tn n) / f' (Tn n)) - Tn n)) by ring.
apply Rle_trans with 
 (Rabs ( aprox (Tn n - f (Tn n) / f' (Tn n)) (S n) -  (Tn n - f (Tn n) / f' (Tn n))) +
 Rabs ((Tn n - f (Tn n) / f' (Tn n)) - Tn n)); [apply Rabs_triang|].
apply Rplus_le_compat;
rewrite <-eq_tn_zn; unfold Zn.
rewrite <-round_eq; assumption.
destruct (kanto_aprox n) as [IH1 [IH2 [ IH3 [ IH4 [ IH5  IH6]]]]].
replace (Yn n 0 - f (Yn n 0) / f' (Yn n 0) - Yn n 0) with (-(f (Yn n 0) / f' (Yn n 0))) by ring.
rewrite Rabs_Ropp.
apply Rle_trans with (1:=IH5).
 replace (/ 2 ^ n * b0) with (b0/2^n) ;[| field; apply Rgt_not_eq; apply pow2_diff0].
apply aux_bseq_cv.
Qed.

Lemma dif_approx2:
forall n, Rabs (Tn  (S n) - Tn 0) <= sum_f_R0  (fun n => /3^n) n * Round 0 
  + sum_f_R0 (fun n => /2^n) n * b0.
Proof.
intros n.
assert (Heq: forall n, Tn (S n) - Tn 0 = sum_f_R0 (fun  n => (Tn (S n) - Tn n)) n).
intros m; induction m; [simpl; auto|].
rewrite tech5; rewrite <-IHm; ring.
rewrite Heq.
apply Rle_trans with (sum_f_R0 (fun i : nat => Rabs ( Tn (S i) - Tn i)) n);
  [apply Rsum_abs|].
rewrite Rmult_comm.
rewrite (Rmult_comm (sum_f_R0 (fun n0 : nat => / 2 ^ n0) n )).
repeat rewrite scal_sum.
rewrite <-(plus_sum  ).
apply sum_Rle; intros m Hm.
apply dif_approx.
Qed.

Lemma dif_approx3:
forall n, Rabs ((Xn (Yn (S n) 0) f f' 1)  - Tn 0) <=
  sum_f_R0  (fun n => /3^n) n * Round 0 
  + sum_f_R0 (fun n => /2^n) (S n) * b0.
Proof.
intros n.
replace ((Xn (Yn (S n) 0) f f' 1)  - Tn 0) with
  ( ((Xn (Yn (S n) 0) f f' 1) - Tn (S n)) + (Tn (S n) - Tn 0)) by ring.
apply Rle_trans with (Rabs (Xn (Yn (S n) 0) f f' 1 - Tn (S n)) +
 Rabs( (Tn (S n) - Tn 0))); [apply Rabs_triang|].
rewrite tech5; rewrite Rmult_plus_distr_r;
rewrite <-Rplus_assoc.
 pattern Rplus at 1; rewrite Rplus_comm .
apply Rplus_le_compat; [|apply dif_approx2].
rewrite <-eq_tn_zn; unfold Zn.
apply Rle_trans with (Bseq (S n)); 
  [|replace (/2^S n * b0) with (b0/2^S n) ; 
    [|field; apply Rgt_not_eq; apply pow2_diff0]; apply aux_bseq_cv].
change (Xn (Yn (S n) 0) f f' 1) with (Yn (S n) 0 - f (Yn (S n) 0) / f' (Yn (S n) 0)).
destruct (kanto_aprox (S n)) as [ _ [_ [H1 [_ [H2 _]]]]].
replace (Yn (S n) 0 - f (Yn (S n) 0) / f' (Yn (S n) 0) - Yn (S n) 0) with 
  (- (f (Yn (S n) 0) / f' (Yn (S n) 0)) ); [|field; assumption].
rewrite Rabs_Ropp; assumption.
Qed.


Lemma elem_in_c_disc:
forall n, included (c_disc (Tn n) (2* Bseq n)) (c_disc X0 (3*b0)).
Proof.
intros n.
destruct (kanto_aprox n) as [_ [_ [_ [_ [_ [_ [_ H]]]]]]]. induction n.
simpl; apply incl_c_disc; apply Rmult_le_compat_r;
  [apply H_b0 with   X0 f f'; assumption|fourier].
apply included_trans with (c_disc X0 ((2* Bseq (S n)) + Rabs (Tn (S n) - X0))).
apply disc_in_disc.
apply incl_c_disc.
apply Rle_trans with (2 * (b0/2^ (S n)) + (sum_f_R0  (fun n => /3^n) n * Round 0 
  + sum_f_R0 (fun n => /2^n) n * b0)).
apply Rplus_le_compat.
apply Rmult_le_compat_l; [fourier| apply aux_bseq_cv].
change X0 with (Tn 0); apply dif_approx2.


rewrite(
  sum_eq (fun n0 : nat => / 3 ^ n0) (fun n0 : nat => (/ 3) ^ n0) n).
rewrite(
  sum_eq (fun n0 : nat => / 2 ^ n0) (fun n0 : nat => (/ 2) ^ n0) ( n)).
repeat rewrite tech3.
assert (H4: forall n, 0 <= 3 - / 3 ^ n).
intros m; induction m; simpl; [fourier|].
apply Rle_trans with (1:= IHm).
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.
rewrite Rinv_mult_distr.
pattern (/3^m) at 2; replace (/3^m) with (1* /3^m) by ring.
apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; apply pow_lt.
fourier.
fourier.
apply Rgt_not_eq; fourier.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
assert (Heq2: (1 - (/ 3) ^ S n) / (1 - / 3) = (3 - /3^n) / 2).
assert (H3: forall n , ((/3)^n)  = /3^ n).
intros m; induction m. 
simpl; field.
simpl; rewrite IHm; field; apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
rewrite H3.
replace (1-/3) with (2/3) by field.
replace ((1 - / 3 ^ S n) / (2 / 3) ) with (((1 - / 3 ^ S n) / 2)* (/ / 3) ); [|field].
rewrite Rinv_involutive.
replace (/3^n) with (3/3^(S n)); [|simpl; field].
field.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier.
apply Rgt_not_eq; fourier.
apply Rgt_not_eq; unfold Rgt; apply pow_lt; fourier. 
apply Rle_trans with (2 * (b0 /2^ (S n)) +
((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 +
 (1 - (/ 2) ^ S ( n)) / (1 - / 2) * b0)).
apply Rplus_le_compat_r.
apply Rmult_le_compat_l; [fourier|right; trivial].
assert  (2 * (b0 / 2 ^ S n) +
((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 +
 (1 - (/ 2) ^ S ( n)) / (1 - / 2) * b0) = 
  ((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0)+ (2 )*b0).
rewrite Rplus_comm.
rewrite Rplus_assoc.
apply Rplus_eq_compat_l.
replace ((1 - (/ 2) ^ (S n)) / (1 - / 2) * b0 + 2 * (b0 / 2 ^ n) ) with
(((1 - (/ 2) ^ (S n)) / (1 - / 2) + 2 * (/ 2 ^ n))*b0) ; [|field].
assert (H3: forall n , ((/2)^n)  = /2^ n) .
intros m; induction m. 
simpl; field.
simpl; rewrite IHm; field; apply Rgt_not_eq; apply pow2_diff0.
rewrite H3.
simpl.
replace (1-/2) with (/2) by field.
replace ((1 - / (2 * 2 ^ n)) / / 2)
 with ((1 - / (2 * 2 ^ n)) * 2); [|field].
field.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
rewrite H0.
replace ((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0 + 2 * b0) with
 (2*b0 + ((1 - (/ 3) ^ S n) / (1 - / 3) * Round 0)) by ring.
replace (3* b0 ) with (2*b0 + b0) by ring.
apply Rplus_le_compat_l.
apply Rle_trans with ((1 - (/ 3) ^ S n) / (1 - / 3) * (3/16*b0) ).

apply Rmult_le_compat_l.

rewrite Heq2.
unfold Rdiv; apply aux1.

apply H4.
fourier.
assert (G1: Round 0 = Rn 0).
unfold Round, Rn, mu; simpl; field.
elim Hmuineq; intros Hmu0 Hmu1; unfold mu in Hmu0; simpl in Hmu0.
assert (H00:= Rgt_not_eq _ _ Hmu0).
 elim (Rlt_0_mult_reg _ _ H00) ; intros H01 H02.
 elim ( Rlt_0_mult_reg _ _ H01) ; intros H03 H04.
 elim ( Rlt_0_mult_reg _ _ H03) ; intros _ H05 .
do 3 (split;   try assumption).
rewrite G1.
apply bound_Rn0.
rewrite Heq2.
apply Rle_trans with ((3 - 0) / 2 * (3 / 16 * b0) ) .

apply Rmult_le_compat.
unfold Rdiv; apply aux1; [apply H4|fourier].
apply aux1; [fourier|apply H_b0 with X0 f f'; assumption].
unfold Rdiv; apply Rmult_le_compat_r; [fourier|].
unfold Rminus; apply Rplus_le_compat_l.
apply Ropp_le_contravar.
rewrite Rinv_pow; [apply pow_le|apply Rgt_not_eq]; fourier.
right; trivial.
replace ((3 - 0) / 2 * (3 / 16 * b0)) with (9/32 * b0) by field.
pattern b0 at 2; replace b0 with (1*b0 ) by ring; apply Rmult_le_compat_r.
apply H_b0 with X0 f f'; assumption.
fourier.
apply Rlt_not_eq; fourier.
apply Rlt_not_eq; fourier.
intros; apply Rinv_pow; apply Rgt_not_eq; fourier.
intros; apply Rinv_pow; apply Rgt_not_eq; fourier.
Qed.


End ApproxSequence.

