Require Export mult_digit_correct.

Module MAKE_DIGIT_CORRECT(Nums : NUMS).
Module Mult_digit_correct := MULT_DIGIT_CORRECT(Nums).
Export Mult_digit_correct.

(*On devrait ne plus avoir de probleme en considérant des séries ou la somme partielle es toujours inférieure à 
  (b^2-3b-2)/(2*b^2) en valeur absolue
  le reste serait alors toujours plus petit que (b+2)/(2*b) en valeur absolue
*)

Theorem md_tech1 : 
  forall x r vx vr, 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    - (2*Z2R b*Z2R b -Z2R b-2) <= 2*Z2R b*Z2R b * (vr+vx) <= 2*Z2R b*Z2R b -Z2R b-2.
Proof.   
assert (Hbpos := base_positive).
intros x r vx vr Hrepx Hrepr Hvx Hvr.

assert (U1 : 0 < Z2R b) by IDR_to_Z.

assert (U2 :  - (Z2R b *Z2R b + 2*Z2R b) <= 
  2 * Z2R b *Z2R b* vr <= Z2R b * Z2R b + 2*Z2R b).
 replace ( - (Z2R b * Z2R b + 2 * Z2R b)) with 
   (Z2R b * (- (Z2R b + 2))) by ring.
 replace (Z2R b * Z2R b + 2 * Z2R b) with 
   (Z2R b * (Z2R b + 2)) by ring.
 replace (2*Z2R b * Z2R b * vr) with 
   (Z2R b * (2*Z2R b *vr)) by ring.
split; apply Rmult_le_compat_l; intuition.

 match goal with 
   id1 : - ?r1 <= ?r0 * vx <= ?r1,
   id2 : - ?r2 <= ?r0 * vr <= ?r2
   |- -?r3 <= ?r0*(vr+vx) <= ?r3 =>
     rewrite (Rmult_plus_distr_l r0); 
     replace r3 with (r1+r2); [ dec_tac; split; fourier| ring]
  end.
Qed.

Theorem md_tech2 : 
  forall x r vx vr, 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    Srepresents (Sadd r x) (vr+vx).
Proof.   
pose (Hbpos := base_positive).
intros x r vx vr Hrepx Hrepr Hvx Hvr.
apply Sadd_correct; assumption || omega || idtac.
generalize (md_tech1 x r vx vr Hrepx Hrepr Hvx Hvr).
intros [U1 U2].
assert (U3: -(2 * Z2R b * Z2R b) <= 2 * Z2R b * Z2R b * (vr+vx) <= 2 * Z2R b * Z2R b).
 assert (U4 : 2 * Z2R b * Z2R b - Z2R b - 2 <= 2 * Z2R b * Z2R b) by IDR_to_Z.
 split;[ apply Rle_trans with (2:= U1)|
   apply Rle_trans with (1:= U2)]; intuition.
assert (U4 : 0 < Z2R b) by IDR_to_Z.
assert (U5 : 0 < 2 * Z2R b * Z2R b) by ZR0compat.
dec_tac; split; (apply Rmult_le_reg_l with ( 2 * Z2R b * Z2R b); [assumption|]).
replace (2 * Z2R b * Z2R b * -1) with (- ( 2 * Z2R b * Z2R b)) by ring; assumption.
rewrite Rmult_1_r; assumption.
Qed.

Theorem make_digit_remainder : 
  forall x vx r vr k r', 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    (k::r') = make_digit (Sadd r x) ->
    - (Z2R b+2)<= 2*Z2R b * ((vr+vx) * Z2R b -IDR k) <= Z2R b +2.
Proof.
intros x vx r vr k r' Hrepx Hrepr Hvx Hvr.
assert (Hrepadd : Srepresents (Sadd r x) (vr+vx)).
  apply md_tech2; assumption.
generalize_all (Sadd r x) x'.
assert (Hvx': -(2*Z2R b*Z2R b -Z2R b-2) <= 2*Z2R b*Z2R b * (vr+vx) <= 2*Z2R b*Z2R b -Z2R b-2).
  apply (md_tech1 x r); assumption.
generalize_all (vr+vx) vx'.
clear Hvr Hvx Hrepx Hrepr x vx r vr.

destruct x' as [k1 [k2 x]].
unfold make_digit, Dle_bool.
compare_gt_tac G1; compare_gt_tac G2; try (compare_gt_tac G3);
  intros U0; injection U0; clear U0; intros;  
    inversion_to_variable x; subst.

assert (U1 : -Z2R b <= 2 * IDR k2+2*Z2R b) by IDR_to_Z.
assert (U2 : 2 * IDR k2+2*Z2R b<= Z2R b) by IDR_to_Z.
match goal with 
  |-  _ <= ?r <= _ =>
    replace r with (2* IDR k2 +2*r0 + 2*Z2R b) by field_IDR;
       dec_tac; split; fourier
end.

absurd(- (2 * Z2R b * Z2R b - Z2R b - 2) <= 
  2 * Z2R b * Z2R b * ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b)); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k1) with (-Z2R b+1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((-Z2R b+1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (-2 * Z2R b * Z2R b+2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 replace (- (2 * Z2R b * Z2R b - Z2R b - 2)) with 
   (- 2 * Z2R b * Z2R b + Z2R b + 2) by ring.
 assert (U1 : 2*IDR k2 < - Z2R b) by IDR_to_Z.
 dec_tac; fourier.

assert (U1 : -Z2R b <= 2 * IDR k2-2*Z2R b) by IDR_to_Z.
assert (U2 : 2 * IDR k2 -2*Z2R b<= Z2R b) by IDR_to_Z.
match goal with 
  |-  _ <= ?r <= _ =>
    replace r with (2* IDR k2 +2*r0-2*Z2R b) by field_IDR;
       dec_tac; split; fourier
end.

absurd(2 * Z2R b * Z2R b * ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b) <=
         2 * Z2R b * Z2R b - Z2R b - 2); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k1) with (Z2R b-1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((Z2R b-1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (2 * Z2R b * Z2R b-2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 assert (U1 : Z2R b < 2*IDR k2) by IDR_to_Z.
 dec_tac; fourier.

assert (U1 : -Z2R b <= 2 * IDR k2) by IDR_to_Z.
assert (U2 : 2 * IDR k2 <= Z2R b) by IDR_to_Z.
match goal with 
  |-  _ <= ?r <= _ =>
    replace r with (2* IDR k2 +2*r0) by field_IDR;
      dec_tac; split; fourier
end.
Qed.

(*
Theorem make_digit_eq : 
  forall b, (0 < IDZ b)%Z ->
    forall x k x' vx, 
      Srepresents b x vx ->
      make_digit b x = (k::x') ->
      Srepresents b x' (vx*Z2R b-IDR k)%R.
Proof.
intros b Hbpos x k x' vx Hrep.

unfold make_digit.
destruct x as [k1 [k2 x]].
simpl_IDZ.
inversion Hrep; subst.

Zbool G1; Zbool G2; try Zbool G3; intros U1; 
  injection U1; clear U1; intros U1 U2; subst;
    match goal with 
      id : Srepresents _ ?x ?r1 |- Srepresents _ ?x ?r2 => 
	replace r2 with r1 by field_IDR;
	  exact id
      | _ => idtac
    end;
    inversion_to_variable x; subst;
      match goal with 
	id : Srepresents _ x ?r1 |- Srepresents _ (?d::x) ?r2 => 
	  replace r2 with ((IDR d + r1)/Z2R b)%R by field_IDR; 
	    constructor; assumption || (simpl_IDZ; omega)
      end.
Qed.

Theorem make_digit_eq2 : 
  forall b, (0 < IDZ b)%Z ->
    forall x k x' vx vx', 
      Srepresents b x vx ->
      Srepresents b x' vx' -> 
      make_digit b x = (k::x') ->
      Srepresents b x ((IDR k + vx')/ Z2R b)%R.
Proof.
intros b Hbpos x k x' vx vx' U1 U2.

assert (U3: (-1 <= vx' <= 1)%R).
apply Srepresents_interval with (2:= U2); omega.

unfold make_digit.
destruct x as [k1 [k2 x]].
simpl_IDZ.
inversion U1.

Zbool G1; Zbool G2; try Zbool G3; intros U4; 
  injection U4; clear U4; intros U4 U5; subst k.

constructor.
rewrite U4; assumption.
assumption.
omega.

constructor.
rewrite U4; assumption.
assumption.
omega.

destruct x' as [d x''].
injection U4; clear U4; intros; inversion U2; subst.
replace ((IDR (k1 + 1)%D + (IDR (k2 - b)%D + r0) / Z2R b) / Z2R b)%R with 
   ((IDR k1+(IDR k2+r0)/Z2R b)/ Z2R b)%R by field_IDR.
inversion H1;
constructor.
constructor; (assumption|| omega).
apply Rle_div_1; [IDR_to_Z| idtac].
dec_tac; split.
assert (-Z2R b +1 <= IDR k2); [IDR_to_Z| fourier]. 
assert (IDR k2 <= Z2R b-1); [IDR_to_Z| fourier]. 
omega.

constructor.
rewrite U4; assumption.
assumption.
omega.

destruct x' as [d x''].
injection U4; clear U4; intros; inversion U2; subst.
replace ((IDR (k1 - 1)%D + (IDR (k2 + b)%D + r0) / Z2R b) / Z2R b)%R with 
  ((IDR k1+(IDR k2+r0)/Z2R b)/ Z2R b)%R by field_IDR.
inversion H1;
constructor.
constructor; (assumption|| omega).
apply Rle_div_1; [IDR_to_Z| idtac].
dec_tac; split.
assert (-Z2R b +1 <= IDR k2); [IDR_to_Z| fourier]. 
assert (IDR k2 <= Z2R b-1); [IDR_to_Z| fourier]. 
omega.
Qed.
*)

Theorem make_digit_eq : 
  forall x vx, 
    Srepresents x vx -> 
    Srepresents (make_digit x) vx.
Proof.
intros x vx Hrep.
unfold make_digit.
destruct x as [k1 [k2 x]].
unfold Dle_bool;
compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3;
  try assumption; inversion_to_variable x; subst;
    (rep_constr; [idtac| field_IDR]).

constructor.
constructor; (assumption || simpl_IDZ; omega).
apply Rle_div_1; IDR_to_Z.
fold_IDR; dec_tac; split.
apply Rle_trans with 0%R.
IDR_to_Z.
assert (-Z2R b+1 <= IDR k2)%R; [IDR_to_Z| fourier].
assert (IDR k2<= -1)%R; [IDR_to_Z| fourier].
simpl_IDZ; omega.

constructor.
constructor; (assumption || simpl_IDZ; omega).
apply Rle_div_1; IDR_to_Z.
fold_IDR; dec_tac; split.
assert (1<= IDR k2)%R; [IDR_to_Z| fourier].
apply Rle_trans with 0%R.
assert (IDR k2 <= Z2R b-1)%R; [IDR_to_Z| fourier].
IDR_to_Z.
simpl_IDZ; omega.
Qed.

Theorem make_digit_bounds  :
  forall x vx vx',
    Srepresents x vx -> 
    - 1  < vx < 1 ->
    let (x1  , x') := make_digit x in 
      Srepresents x' vx' ->
      vx = (vx' + IDR x1)/Z2R b ->
      - 1 < vx' < 1.
Proof.
assert (Hbpos := base_positive).
intros x vx vx' Hrepx Hbounds.
unfold make_digit, Dle_bool.
destruct x as [x1 [x2 x']].
compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3.

 (*case 2*x2 < -b /\  -b+1 < x1*)
  simpl_IDR; fold_IDR.
  intros Hrepx' Hvx.
  inversion Hrepx'; subst.
  simpl_IDR; fold_IDR.
  split.
    apply Rlt_div_l; [R_to_Z |].
    apply Rlt_le_trans with 0; [R_to_Z|].
    assert (- Z2R b +1  <= IDR x2); [| dec_tac; fourier]. 
    inversion Hrepx; inversion H5; IDR_to_Z.
  
    apply Rlt_div_r; [R_to_Z |].
    assert (IDR x2 < -1) by IDR_to_Z; dec_tac; fourier.

 (*case 2*x2 < -b /\  x1  = -b+1*)
  replace (IDR x1) with (- Z2R b + 1) ; [| inversion Hrepx; IDR_to_Z].
  intros Hrepx' Hvx.
  split.
    replace vx'  with (vx * Z2R b + Z2R b - 1) by (rewrite Hvx; field_Z).
    assert (-1 * Z2R b < vx * Z2R b); [| fourier].
    apply Rmult_lt_compat_r; [R_to_Z | intuition].

    inversion Hrepx'; subst.
    apply Rlt_div_r; [R_to_Z |].
    assert (IDR x2 < -1) by IDR_to_Z.
    assert (4 < Z2R b) by IDR_to_Z; dec_tac; fourier.
    
(* case  b < 2*x2 /\ x1 < b-1 *)
  simpl_IDR; fold_IDR.
  intros Hrepx' Hvx.
  inversion Hrepx'; subst.
  simpl_IDR; fold_IDR.
  split.
    apply Rlt_div_l; [R_to_Z |].
    assert (1 < IDR x2) by IDR_to_Z; dec_tac; fourier.

    apply Rlt_div_r; [R_to_Z |].
    apply Rle_lt_trans with 0; [|R_to_Z].
    assert (IDR x2 <= Z2R b - 1); [| dec_tac; fourier].
    inversion Hrepx; inversion H5; IDR_to_Z.
  
(* case  b < 2*x2 /\ x1 = b-1 *)
  replace (IDR x1) with (Z2R b - 1) ; [| inversion Hrepx; IDR_to_Z].
  intros Hrepx' Hvx.
  split.
    inversion Hrepx'; subst.
    apply Rlt_div_l; [R_to_Z |].
    assert (1 < IDR x2 ) by IDR_to_Z.
    assert (4 < Z2R b) by IDR_to_Z; dec_tac; fourier.
  
    replace vx'  with (vx * Z2R b + - Z2R b + 1) by (rewrite Hvx; field_Z).
    assert (vx * Z2R b < 1 * Z2R b); [| fourier].
    apply Rmult_lt_compat_r; [R_to_Z | intuition].

(* case - b <= 2*x2 <= b *)
  inversion 1; intros U1; subst.
  split.
    apply Rlt_div_l; [R_to_Z |].
    assert ( -Z2R b + 1 < IDR x2 ) by IDR_to_Z.
    dec_tac; fourier.

    apply Rlt_div_r; [R_to_Z |].
    assert ( IDR x2 < Z2R b - 1) by IDR_to_Z.
    dec_tac; fourier.
Qed.

Theorem make_digit_add :
  forall x k x' vx vy,
    make_digit x = (k::x') ->
    Srepresents x vx ->
    (-1 <= vx + vy <= 1 -> 
      -(Z2R b-2) <= vy*(2*Z2R b*Z2R b) <= Z2R b-2 -> 
      -1 <= (vx + vy)*Z2R b - IDR k <= 1)%R.
Proof.
pose (Hbpos := base_positive).
intros x k x' vx vy Hmd_eq Hrep Hsum Hvy.
assert (U1 : (0 < Z2R b)%R).
IDR_to_Z.
assert (U2: (0< 2*Z2R b)%R).
fourier.

destruct x as [k1 [k2 x]].
inversion_to_variable x; subst.

assert (U3 : ((-1*Z2R b +- IDR k1) *(2 * Z2R b) <= 
  IDR k2 + IDR k2 + 2 * r0 + vy * (2 * Z2R b * Z2R b) <= 
  (1*Z2R b + -IDR k1) *(2 * Z2R b))%R).
replace (IDR k2 + IDR k2 + 2 * r0 + vy * (2 * Z2R b * Z2R b))%R with 
  ((((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b + vy)*Z2R b+ -IDR k1)*(2*Z2R b))%R;
  [idtac| field_IDR].
dec_tac; split; 
  (apply Rmult_le_compat_r; [fourier|
    apply Rplus_le_compat_r; 
  apply Rmult_le_compat_r; assumption || fourier]).

replace (((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b + vy) * Z2R b - IDR k)%R with 
  (((IDR k1-IDR k)*(2*Z2R b) + (IDR (k2+k2)%D + 2*r0) + vy*(2*Z2R b*Z2R b))/
    (2*Z2R b))%R; [idtac| simpl_IDR; field_IDR].
apply Rle_div_1.
IDR_to_Z.
generalize Hmd_eq; clear Hmd_eq.
unfold make_digit, Dle_bool.

compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3;
intros U5; injection U5; clear U5; intros U5 U6; subst; simpl_IDR;
match goal with 
  |- (_ <= (?v0 - ?v0) * ?v1 + ?v2 +?v3 <= _ )%R =>
    replace ((v0-v0)*v1+v2+v3)%R  with (v2+v3)%R;[idtac|ring]
  | |- (_ <= (?v0 - (?v0+1)) * ?v1 + ?v2 +?v3 <= _ )%R =>
    replace ((v0-(v0+1))*v1+v2+v3)%R  with (-v1+(v2+v3))%R;[idtac|ring]
  | |- (_ <= (?v0 - (?v0-1)) * ?v1 + ?v2 +?v3 <= _ )%R =>
    replace ((v0-(v0-1))*v1+v2+v3)%R  with (v1+(v2+v3))%R;[idtac|ring]
  | _ => idtac
end.

split.
assert (U4: (-2*Z2R b +2 <= IDR k2 + IDR k2)%R).
IDR_to_Z.
dec_tac.
apply Rle_trans with ((2 * Z2R b) + (-(2*Z2R b) + -(2*Z2R b)))%R.
fourier.
apply Rplus_le_compat_l.
apply Rplus_le_compat; fold_IDR; fourier.

match goal with 
  |- (_ <= ?a )%R => 
    pattern a at 3; replace a with (a+((-Z2R b+2)+(Z2R b-2)))%R; [idtac| ring]  
end.
apply Rplus_le_compat_l.
assert (U4: (IDR k2 + IDR k2<= -Z2R b)%R) by IDR_to_Z.
repeat apply Rplus_le_compat; intuition; fourier.

assert (U4: (IDR k= (-Z2R b +1))%R).
IDR_to_Z.
dec_tac; split.
replace (-(2 * Z2R b))%R with ((-1 * Z2R b + -IDR k) * (2 * Z2R b))%R;
  [ assumption| rewrite U4; ring].
assert (U5: (IDR k2 + IDR k2 <= Z2R b)%R);
  [IDR_to_Z |  fold_IDR; fourier].

split.
match goal with 
  |- (?a <= _ )%R => 
    pattern a at 1; replace a with (a+((Z2R b+-2)+(-(Z2R b-2))))%R; [idtac| ring]  
end.
apply Rplus_le_compat_l.
assert (U4: (Z2R b <= (IDR k2 + IDR k2))%R);
  [IDR_to_Z| idtac].
repeat apply Rplus_le_compat; intuition; fourier.
assert (U4: (IDR k2 + IDR k2<= 2*Z2R b -2)%R).
IDR_to_Z.
dec_tac.
apply Rle_trans with (- (2 * Z2R b) + (2*Z2R b + 2*Z2R b))%R.
apply Rplus_le_compat_l.
apply Rplus_le_compat; fold_IDR; fourier.
fourier.

assert (U4: (IDR k= (Z2R b -1))%R) by IDR_to_Z.
dec_tac; split.
assert (U5: (-Z2R b <= IDR k2 + IDR k2)%R) by IDR_to_Z.
fold_IDR; fourier.
pattern (2 * Z2R b)%R at 2; 
  replace (2 * Z2R b) %R with ((1 * Z2R b + - IDR k) * (2 * Z2R b))%R;
  [ assumption| rewrite U4; ring].

assert (U4: ( - Z2R b <= IDR k2 + IDR k2 <= Z2R b)%R).
split; IDR_to_Z.
fold_IDR; dec_tac; split; fourier.
Qed.

Theorem series_bound : 
  forall b,
    (0 < b)%Z ->
    (Z2R b-2 <= 2* Z2R b*Z2R b)%R.
Proof.  
intros b' Hbpos.
IDR_to_Z.
apply Zle_trans with (1* b')%Z.
omega.
apply Zmult_le_compat_r; omega.
Qed.

Ltac case_md d r' Hmd :=
  match goal with
    |- context [ make_digit ?r ] =>
      caseEq (make_digit r); intros d r' Hmd
  end.

Ltac md_eq_hyp Hmd_eq:= 
  match goal with 
    | id1 : make_digit ?r = (?k::?r'), id2 : Srepresents ?r ?vr
      |- Srepresents (?k::_) _ =>
	assert (Hmd_eq : Srepresents (k::r') vr); 
	  [rewrite <- id1; apply make_digit_eq; assumption  | idtac]
  end.    

Ltac md_add_hyp Hmd_add:= 
  match goal with 
    | id : make_digit ?r = (?k::?r') |- Srepresents (?k::_) (?vx+?vy)%R =>
      assert (Hmd_add : (-1 <= (vx+vy)*Z2R b - IDR k <= 1)%R); 
	[apply make_digit_add with (2:=id); (assumption || omega|| idtac) | idtac]
  end.    

Ltac make_digit_tac d r' U1 U2 U3 U4 series_correct := 
  match goal with
    |- Srepresents (match  make_digit ?r with k::r' => _ end) (?v1+?v2)%R => 
      caseEq (make_digit r); intros d r' U1;
	assert (U2 : Srepresents r v1); [clear series_correct | 
	  assert (U3: Srepresents (d::r') v1) ;
	    [ clear series_correct; rewrite <- U1; apply make_digit_eq; (assumption || omega|| idtac)|
	      assert (U4: (-1<= (v1+v2)*Z2R b - IDR d<=1)%R); 
		[ clear series_correct ; apply make_digit_add with (1:=U1); 
		  (assumption|| omega ||idtac) | idtac]]]
end.

Theorem md_ultimate : 
  forall x vx r vr k r', 
    Srepresents x vx -> 
    Srepresents r vr -> 
    -(Z2R b *Z2R b-3*Z2R b -2) <= 2*Z2R b * Z2R b * vx <= Z2R b *Z2R b-3*Z2R b -2 ->
    -(Z2R b +2) <= 2*Z2R b * vr  <= Z2R b +2 ->
    make_digit (Sadd r x) = (k::r') ->
    - (Z2R b+2)<= 2*Z2R b * ((vr+vx) * Z2R b -IDR k) <= Z2R b +2 /\
    (- b < IDZ k < b)%Z /\
    Srepresents r' ((vr+vx)*Z2R b-IDR k).
Proof.
intros x vx r vr k r' Hrepx Hrepr Hvx Hvr.
assert (Hrepadd : Srepresents (Sadd r x) (vr+vx)).
  apply md_tech2; assumption.
generalize_all (Sadd r x) x'.
assert (Hvx': -(2*Z2R b*Z2R b -Z2R b-2) <= 2*Z2R b*Z2R b * (vr+vx) <= 2*Z2R b*Z2R b -Z2R b-2).
  apply (md_tech1 x r); assumption.
generalize_all (vr+vx) vx'.
clear Hvr Hvx Hrepx Hrepr x vx r vr.

destruct x' as [k1 [k2 x]].
unfold make_digit, Dle_bool;
  compare_gt_tac G1; compare_gt_tac G2; try compare_gt_tac G3;
  intros U0; injection U0; clear U0; intros;  
    inversion_to_variable x; subst.

split. 
  assert (U1 : -Z2R b <= 2 * IDR k2+2*Z2R b) by IDR_to_Z.
  assert (U2 : 2 * IDR k2+2*Z2R b<= Z2R b) by IDR_to_Z.
  match goal with 
    |-  _ <= ?r <= _ =>
      replace r with (2* IDR k2 +2*r0 + 2*Z2R b) by field_IDR;
	dec_tac; split; fourier
  end.
  
 split; [simpl_IDZ; omega| ].
 replace ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b * Z2R b - IDR (k1 - 1)%D) with
   ((IDR (k2 + Dbase)%D + r0)/ Z2R b) by field_IDR.
 constructor; assumption || simpl_IDZ; omega.

absurd(- (2 * Z2R b * Z2R b - Z2R b - 2) <= 
  2 * Z2R b * Z2R b * ((IDR k + (IDR k2 + r0) / Z2R b) / Z2R b)); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k) with (-Z2R b+1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((-Z2R b+1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (-2 * Z2R b * Z2R b+2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 replace (- (2 * Z2R b * Z2R b - Z2R b - 2)) with 
   (- 2 * Z2R b * Z2R b + Z2R b + 2) by ring.
 assert (U1 : 2*IDR k2 < - Z2R b) by IDR_to_Z.
 dec_tac; fourier.

split.
  assert (U1 : -Z2R b <= 2 * IDR k2-2*Z2R b) by IDR_to_Z.
  assert (U2 : 2 * IDR k2 -2*Z2R b<= Z2R b) by IDR_to_Z.
  match goal with 
    |-  _ <= ?r <= _ =>
      replace r with (2* IDR k2 +2*r0-2*Z2R b) by field_IDR;
	dec_tac; split; fourier
  end.
 split; [simpl_IDZ; omega| ].
 replace ((IDR k1 + (IDR k2 + r0) / Z2R b) / Z2R b * Z2R b - IDR (k1 + 1)%D) with
   ((IDR (k2 - Dbase)%D + r0)/ Z2R b) by field_IDR.
 constructor; assumption || simpl_IDZ; omega.

absurd(2 * Z2R b * Z2R b * ((IDR k + (IDR k2 + r0) / Z2R b) / Z2R b) <=
         2 * Z2R b * Z2R b - Z2R b - 2); [|intuition].
 apply RIneq.Rlt_not_le.
 replace (IDR k) with (Z2R b-1) by IDR_to_Z.
 replace (2 * Z2R b * Z2R b * ((Z2R b-1 + (IDR k2 + r0) / Z2R b) / Z2R b))
   with (2 * Z2R b * Z2R b-2*Z2R b + 2*IDR k2 + 2*r0) by field_IDR.
 assert (U1 : Z2R b < 2*IDR k2) by IDR_to_Z.
 dec_tac; fourier.

split.
  assert (U1 : -Z2R b <= 2 * IDR k2) by IDR_to_Z.
  assert (U2 : 2 * IDR k2 <= Z2R b) by IDR_to_Z.
  match goal with 
    |-  _ <= ?r <= _ =>
      replace r with (2* IDR k2 +2*r0) by field_IDR;
	dec_tac; split; fourier
  end.
  
  intuition.
  replace ((IDR k + (IDR k2 + r0) / Z2R b) / Z2R b * Z2R b - IDR k) with 
    ( (IDR k2 + r0) / Z2R b) by field_IDR; assumption.
Qed.

Ltac md_tac2 d r' vr' U1 U2 series_correct :=
  match goal with 
    |- Srepresents (series_body _ (Sadd ?r1 ?x1)) (?vr+?vx+?vs)%R => 
      unfold series_body; 
	caseEq (make_digit (Sadd r1 x1)); intros d r' U1;
	  assert(U2 :
	    - (Z2R b+2)<= 2*Z2R b * ((vr+vx) * Z2R b -IDR d) <= Z2R b +2 /\
	    (- b < IDZ d < b)%Z /\
	    Srepresents r' ((vr+vx)*Z2R b-IDR d)) ;
	  [clear series_correct; apply md_ultimate with (x:=x1) (r:=r1); 
	    assumption || (simpl_IDZ; omega) || idtac |
	  replace (vr+vx+vs) with ((IDR d+ ((vr+vx)*Z2R b-IDR d+ vs*Z2R b))/ Z2R b);
	    [ | clear series_correct; try field_IDR ]]
  end.





Theorem splitted_series_bound : 
  forall x y,
    - (Z2R b-4) <= 2* Z2R b * (x+y) <= Z2R b-4 ->
    - (Z2R b-2) <= 2* Z2R b * Z2R b* y <= Z2R b-2 ->
    - ( Z2R b* Z2R b -3* Z2R b -2) <= 2* Z2R b * Z2R b* x <= Z2R b* Z2R b -3* Z2R b -2.
Proof.
assert (Hbpos := base_positive).
intros x y Hxy Hy .

rabs_intros.

rewrite Rabs_mult.

apply Rle_trans with ( Rabs (2 * Z2R b * Z2R b) * (Rabs ( x + y) + Rabs( y))).
  apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
  generalize (Rabs_triang_inv x (-y)).
  rewrite Rabs_Ropp.
  replace (x - - y) with (x+y) by ring.
  intros; fourier.

  apply Rle_trans with (Z2R b * (Z2R b - 4) + (Z2R b - 2)).
    rewrite Rmult_plus_distr_l.
    apply Rplus_le_compat.
    replace ( Rabs (2 * Z2R b * Z2R b) * Rabs (x + y)) with 
      (Z2R b * Rabs (2 * Z2R b * (x + y))).
    apply  Rmult_le_compat_l; ZR0compat.
    repeat rewrite Rabs_mult. 
    rewrite (Rabs_right (Z2R b)); ZR0compat; ring.
    rewrite <- Rabs_mult; assumption.
    
    right; ring.
Qed.

Theorem splitted_series_bound_rabs : 
  forall x y,
    2* Z2R b * Rabs (x+y) <= Z2R b-4 ->
    2* Z2R b * Z2R b* Rabs y <= Z2R b-2 ->
    2* Z2R b * Z2R b*Rabs x <= Z2R b* Z2R b -3* Z2R b -2.
Proof.
assert (Hbpos := base_positive).
intros x y Hxy Hy .

apply Rle_trans with ( 2 * Z2R b * Z2R b * ( Rabs (x + y) + Rabs y)).
  apply Rmult_le_compat_l; ZR0compat; IDR_to_Z.
  generalize (Rabs_triang_inv x (-y)).
  rewrite Rabs_Ropp.
  replace (x - - y) with (x+y) by ring.
  intros; fourier.

  apply Rle_trans with (Z2R b * (Z2R b - 4) + (Z2R b - 2)).
    replace (2 * Z2R b * Z2R b * (Rabs (x + y) + Rabs y)) with 
      (Z2R b * (2 * Z2R b * Rabs (x + y)) + 2 * Z2R b * Z2R b * Rabs y) by ring.
    apply Rplus_le_compat.
    apply  Rmult_le_compat_l; ZR0compat.
    assumption.
    
    right; ring.
Qed.

End MAKE_DIGIT_CORRECT.
