Require Export mult_correct.
Require Import series_utils.

Module PI_CORRECT(Nums : NUMS).

Module Mult_correct := MULT_CORRECT(Nums).
Export Mult_correct.

Definition num (k1 k2 k3 : Z) : Z := 
  (k2*(8*k1+18)+k3*(3*k1+8))%Z.

Definition denom (k1 k2 k3 : Z) : Z := 
  (k1*(k1+2)*k2*k3)%Z.

Definition terme_pi k :=
  fun i => 
    let k' := (Z_of_nat (i+k)%nat) in  
      let k1 := (4*k'+1)%Z in 
	let k2 := (2^(4*k'+3))%Z in 
	  let k3 := (3^(4*k'+3))%Z in 
	    Z2R (num k1 k2 k3)/ Z2R (denom k1 k2 k3).

Theorem terme_pi_add : 
  forall a b c d,
    (b+a =d +c)%nat ->
    terme_pi a b = terme_pi c d.
Proof.    
intros a b' c d H.
unfold terme_pi.
rewrite H; ring.
Qed.

Theorem terme_pi_pos : 
  forall k n,
    (0 <= terme_pi k n).
Proof.
intros k n.
unfold terme_pi, num, denom.
ZR0compat.
Qed.

Theorem pi_series_pos : 
  forall k s, 
    infinit_sum (terme_pi k) s -> 
    0 <= s.
Proof.    
intros k s Hkpos.
apply infinit_sum_bound with (An := fun (n:nat) => 0%R) (3:= Hkpos).
intros n.
apply terme_pi_pos ; assumption.
apply infinit_sum_0.
Qed.

Theorem terme_pi_bound :
  forall k n,
    terme_pi k n <= 15 / ((4 * INR k + 3) * 2 ^ (4 * (n + k) + 1)).
Proof.
intros k n.
unfold terme_pi, num, denom.
apply Rle_trans with 
  (15 / ((4 * INR (n+k) + 3) * 2 ^ (4 * (n + k) + 1))).
generalize (n+k)%nat; clear; intros k.
match goal with 
  |- ?a / Z2R (?b*?c*?d*?e)%Z <= 15 / ?f =>
    replace (a / Z2R (b*c*d*e)%Z) with  ((a /Z2R (4*b*e)%Z)/f)
end.
apply Rdiv_le_compat_r.
ZR0compat.
apply Rle_div_r.
ZR0compat.
R_to_Z.
generalize (Zle_0_nat k).
generalize (Z_of_nat k); clear k; intros k Hk.
match goal with 
  |- (?a*?b+ ?c*?d <= _)%Z =>
    apply Zle_trans with (c*b+c*b)%Z;[
      apply Zplus_le_compat| 
	replace (c*b+c*b)%Z with ((2*b)*c)%Z by ring]
end.
apply Zmult_le_compat_r; ZR0compat.
apply Zpower_incr; ZR0compat.
apply Zmult_le_compat_l; ZR0compat.
match goal with |- (_ <= ?a*(?b*?c*?d))%Z => 
  replace (a*(b*c*d))%Z with ((a*b*c)*d)%Z by ring
end.
apply Zmult_le_compat_r; ZR0compat.
rewrite Rdiv_div_mult; ZR0compat.
apply Rdiv_eq_compat_l; ZR0compat.
match goal with 
  |- Z2R ?a*_ = Z2R (?b*?c*2^(?d+3)*?e)%Z => 
    replace (Z2R (b*c*2^(d+3)*e)%Z) with (Z2R a*Z2R (c * (2^(d+1)))%Z)
end.
apply Rmult_eq_compat_l.
R_to_Z.
rewrite inj_plus.
rewrite inj_mult.
simpl Z_of_nat.
ring.
generalize (Zle_0_nat k).
generalize (Z_of_nat k); clear k; intros k Hk.
R_to_Z.
replace (2 ^ (4 * k + 3))%Z with 
  (4*2^(4*k+1))%Z.
ring.
repeat rewrite Zpower_exp; ZR0compat.
ring.
apply Rdiv_le_compat_l; ZR0compat.
apply Rmult_le_compat_r; ZR0compat.
apply Rplus_le_compat_r.
apply Rmult_le_compat_l; ZR0compat.
apply le_INR; omega.
Qed.

Theorem pi_series_bound:
  forall k r,
    infinit_sum (terme_pi k) r -> 
      (r*((4*INR k+3)*16^k) <= 8)%R.
Proof.
intros k r Hinf.
apply Rdiv_le_l.
ZR0compat.
apply infinit_sum_bound with (Bn := fun i => 
  (15/((4*INR k+3)*(2^(4*(i+k)+1))))) (2:=Hinf); clear r Hinf.
apply terme_pi_bound.

replace (8 / ((4 * INR k + 3) * 16 ^ k)) with 
  (15 / ((4*INR k+3)*2^(4*k+1)) * (1 / (1-1/16)))%R.
apply infinit_sum_ext with 
  (fun i => (1/ 16) ^ i * (15 / ((4 * INR k + 3) * 2 ^ (4 * k + 1)))).
intros n.
replace ( 2 ^ (4 * (n + k) + 1)) with (16^n* 2 ^ (4 * k + 1)).
replace ((1 / 16) ^ n) with (1/16^n). 
field; repeat split; ZR0compat.
unfold Rdiv; repeat rewrite Rmult_1_l; apply Rinv_pow; ZR0compat.
replace (4 * (n + k) + 1)%nat with (((4 * k) + 1)+4*n)%nat by ring.
generalize (4*k+1)%nat; intros k'.
rewrite pow_add.
rewrite pow_mult.
replace (2^4) with 16; simpl; ring.
apply scal_infinit_sum.
apply infinit_sum_ext with (fun i => (1/16) ^i).
intros; trivial.
apply GP_infinite.
split; fourier.
ZR0compat.
ZR0compat.
rewrite pow_add.
rewrite pow_mult.
rewrite pow_1.
replace (2^4) with 16 by (simpl; ring).
field;split; ZR0compat.
Qed.

Theorem pi_series_bound_1 : 
  forall k s, 
    infinit_sum (terme_pi k) s ->
    s <= 1.
Proof.
intros k s Hinf.
elim (le_or_lt 1 k); intros G1.
apply Rmult_le_reg_l with ((4*INR k+3)*16^k).
ZR0compat.
apply Rle_trans with 8.
rewrite Rmult_comm.
apply pi_series_bound; assumption.
rewrite Rmult_1_r.
pattern 8 at 1; replace 8 with (1*8) by ring.
apply Rmult_le_compat; try fourier.
assert (0 <= INR k ) by ZR0compat; fourier.
apply Rle_trans with (16 ^1); [simpl; fourier| 
  apply Rle_pow; [fourier | assumption]].
assert (k=0)%nat by omega; subst.
clear G1.

assert (Hinf1 : infinit_sum (terme_pi 1%nat) (s-(terme_pi 0 0)%nat)).
change (terme_pi 0 0) with (sum_f_R0 (terme_pi 0) 0).
apply infinit_sum_ext with 
  (An := (fun i => (terme_pi 0) (i +(S 0))%nat)).
intros n; apply terme_pi_add; omega.
apply infinit_sum_step_k; assumption.
assert (s - (terme_pi 0 0) <= 1- (terme_pi 0 0)); [| fourier].

generalize (pi_series_bound _ _ Hinf1).
clear Hinf Hinf1.
rewrite pow_1.
rewrite Rmult_1_r.
replace ((4+3)*16) with (14*8) by ring.
intros H.
assert (H1 : s - terme_pi 0 0 <= 1/Z2R 14).
apply Rle_div_l; simpl; fourier.
apply Rle_trans with (1:=H1).
clear H H1.
unfold terme_pi.
rewrite Zmult_0_r.
repeat rewrite Zplus_0_l.
unfold num, denom.
change (2 ^ 3)%Z with 8%Z.
change (3 ^ 3)%Z with 27%Z.
change (8 * (8 * 1 + 18) + 27 * (3 * 1 + 8))%Z with 505%Z .
change (1 * (1 + 2) * 8 * 27)%Z with 648%Z.
replace (1-Z2R 505 / Z2R 648) with (Z2R 143/ Z2R 648).
apply Rdiv_mult_compat; R_to_Z.
field_Z.
Qed.

Theorem num_numB_eq : 
  forall k1 k2 k3,
    num (IBZ k1) (IBZ k2) (IBZ k3) = IBZ (numB k1 k2 k3). 
Proof.
intros k1 k2 k3.
unfold num, numB.
simpl_IBZ; ring.
Qed.

Theorem denom_denomB_eq : 
  forall k1 k2 k3,
    denom (IBZ k1) (IBZ k2) (IBZ k3) = IBZ (denomB k1 k2 k3). 
Proof.
intros k1 k2 k3.
unfold denom, denomB.
simpl_IBZ; ring.
Qed.

Theorem  tech_lemma_k1 : 
  forall k k1,
    (IBZ k1 = 4*Z_of_nat k+1 -> 
    IBZ (k1 + 4)%B = 4 * Z_of_nat (k + 1) + 1)%Z.
Proof.
intros k k1 Hk1.
simpl_IBZ; rewrite Hk1.
rewrite inj_plus; ring.
Qed.

Theorem  tech_lemma_k2 : 
  forall k1 k2,
    ( 0 <= IBZ k1 ->
      IBZ k2 = 2^(IBZ k1+2) -> 
      IBZ (k2 * 16)%B = 2 ^ (IBZ (k1 + 4)%B + 2))%Z.
Proof.
intros k1 k2 Hk1pos Hk2.
simpl_IBZ; rewrite Hk2.
repeat rewrite Zpower_exp; ZR0compat; ring.
Qed.

Theorem  tech_lemma_k3 : 
  forall k1 k3,
    ( 0 <= IBZ k1 ->
      IBZ k3 = 3^(IBZ k1+2) -> 
      IBZ (k3 * 81)%B = 3 ^ (IBZ (k1 + 4)%B + 2))%Z.
Proof.
intros k1 k3 Hk1pos Hk3.
simpl_IBZ; rewrite Hk3.
repeat rewrite Zpower_exp; ZR0compat; ring.
Qed.

Theorem finite_sum_pi_aux_correct1: 
  forall n k k1 k2 k3 x y z k1' k2' k3' x' y',
    ( IBZ y <> 0 ->
      IBZ k1 = 4*Z_of_nat k+1 -> 
      IBZ k2 = 2^(IBZ k1+2) -> 
      IBZ k3 = 3^(IBZ k1+2) -> 
      finite_sum_pi_aux k1 k2 k3 x y z n =
      (k1', k2', k3', x', y') -> 
      IBZ y' <> 0 /\
      IBZ k1' = 4*(Z_of_nat (k+S n))+1 /\ 
      IBZ k2' = 2^(IBZ k1'+2) /\ 
      IBZ k3' = 3^(IBZ k1'+2))%Z.
Proof.
induction n.
intros k k1 k2 k3 x y z.
simpl finite_sum_pi_aux. 
intros k1' k2' k3' x' y' Hy Hk1 Hk2 Hk3 H.
injection H; intros; subst.
assert (Hk1pos : (0 <= IBZ k1)%Z).
rewrite Hk1; ZR0compat.
repeat split.
assumption.
apply tech_lemma_k1; assumption.
apply tech_lemma_k2; assumption.
apply tech_lemma_k3; assumption.
        
intros k k1 k2 k3 x y z k1' k2' k3' x' y' Hy Hk1 Hk2 Hk3.
assert (Hk1pos : (0 <= IBZ k1)%Z).
rewrite Hk1; ZR0compat.
simpl finite_sum_pi_aux. 
replace (k+S (S n))%nat with (S k+S n)%nat by ring.
apply IHn.
do 2 rewrite mult_IBZ.
repeat apply Zmult_non0; assumption || simpl_IBZ; ZR0compat.
simpl_IBZ; rewrite inj_S; omega.
apply tech_lemma_k2; assumption.
apply tech_lemma_k3; assumption.
Qed.

Theorem finite_sum_pi_tech_lemma : 
  forall k k1 k2 k3 x y z,
    (IBZ y = IBZ z * IBZ k2 * IBZ k3  ->
      0 < IBZ z ->
      IBZ k1 = 4*Z_of_nat k+1 -> 
      IBZ k2 = 2^(IBZ k1+2) -> 
      IBZ k3 = 3^(IBZ k1+2) -> 
      (IBR (x * 1296 * ((k1 + 4) * (k1 + 4 + 2)) +
	z * numB  (k1 + 4) (k2 * 16) (k3 * 81))%B /
      IBR (y * 1296 * ((k1 + 4) * (k1 + 4 + 2)))%B = 
      IBR x / IBR y + terme_pi (S k) 0)%R)%Z.
Proof.
intros k k1 k2 k3 x y z Hy Hzpos Hk1 Hk2 Hk3. 

assert (Hk2pos : (0 < IBZ k2)%Z).
rewrite Hk2; ZR0compat.
assert (Hk3pos : (0 < IBZ k3)%Z).
rewrite Hk3; ZR0compat.
assert (Hypos : (0 < IBZ y)%Z).
rewrite Hy; ZR0compat.

unfold IBR.
rewrite plus_IBZ.
rewrite plus_Z2R.

rewrite Rdiv_plus_distr_r.
match goal with 
  |- ?a +?b = ?c+?d =>
    replace a with c by field_IBR;
      replace b with d;[ring|]
end.

repeat( rewrite mult_IBZ; replace (IBZ 1296%B) with 1296%Z by (simpl_IBZ; ring)).
rewrite <- num_numB_eq. 
rewrite Hy.
replace (IBZ z * IBZ k2 * IBZ k3 * 1296 * (IBZ (k1 + 4)%B * IBZ (k1 + 4 + 2)%B))%Z
  with (IBZ z * denom (IBZ (k1 + 4)%B) (IBZ (k2 * 16)%B) (IBZ (k3 * 81)%B))%Z.
unfold terme_pi.
rewrite plus_0_l.
replace (4 * Z_of_nat (S k)+1)%Z with (IBZ (k1+4))%B ;
  [|simpl_IBZ; rewrite Hk1; rewrite inj_S; ring].
replace (4 * Z_of_nat (S k)+3)%Z with (IBZ k1+2+4)%Z ;
  [|simpl_IBZ; rewrite Hk1; rewrite inj_S; ring].
replace (2 ^ (IBZ k1 + 2 + 4))%Z with (IBZ (k2 * 16)%B).
replace (3 ^ (IBZ k1 + 2 + 4))%Z with (IBZ (k3 * 81)%B).
simpl_Z2R.
field; unfold denom; simpl_IBZ; split; ZR0compat.
rewrite Zpower_exp;ZR0compat.
rewrite <- Hk3.
replace (3 ^ 4)%Z with (IBZ 81)%B; simpl_IBZ; ring.
rewrite Zpower_exp;ZR0compat.
rewrite <- Hk2.
replace (2 ^ 4)%Z with (IBZ 16)%B; simpl_IBZ; ring.
unfold denom.
do 2 rewrite mult_IBZ.
replace 1296%Z with (IBZ 16%B * IBZ 81%B)%Z by (simpl_IBZ;ring).
replace 2%Z with (IBZ 2)%B by (simpl_IBZ;ring).
rewrite <- plus_IBZ; ring.
Qed.

Theorem finite_sum_pi_aux_correct2 : 
  forall n k k1 k2 k3 x y z k1' k2' k3' x' y',
    ( IBZ y = IBZ z * IBZ k2 * IBZ k3  ->
      0 < IBZ z ->
      IBZ k1 = 4*Z_of_nat k+1 -> 
      IBZ k2 = 2^(IBZ k1+2) -> 
      IBZ k3 = 3^(IBZ k1+2) -> 
      finite_sum_pi_aux k1 k2 k3 x y z (S n) =
      (k1', k2', k3', x', y') -> 
      (IBR x' / IBR y' = IBR x / IBR y+ sum_f_R0 (terme_pi (S k)) n)%R )%Z.
Proof.
induction n;
intros k k1 k2 k3 x y z k1' k2' k3' x' y' Hy Hypos Hk1 Hk2 Hk3.
simpl finite_sum_pi_aux.
intros H2; injection H2; intros; subst; clear H2.
simpl.
apply finite_sum_pi_tech_lemma; assumption.
rewrite decomp_sum_S.
rewrite <- (sum_eq (terme_pi (S (k+1)))).
pattern (S n).
simpl finite_sum_pi_aux.
lazy beta.
match goal with 
  |- finite_sum_pi_aux _ _ _ ?x ?y _ _ = _ -> _ =  ?a + (?b +?c) =>
    replace (a + (b + c)) with (IBR x/ IBR y + c); [|
      replace (IBR x/ IBR y) with (a+b); [ring|]]
end.
apply IHn.
IBR_to_Z; rewrite Hy; ring.
simpl_IBZ.
repeat apply Zmult_non0; assumption || ZR0compat.
simpl_IBZ; rewrite Hk1; rewrite inj_plus; ring.
simpl_IBZ; rewrite Hk2; rewrite Hk1.
repeat rewrite Zpower_exp; omega || ring.
simpl_IBZ; rewrite Hk3; rewrite Hk1.
repeat rewrite Zpower_exp; omega || ring.
symmetry; apply finite_sum_pi_tech_lemma; assumption.
intros i Hi.
unfold terme_pi.
replace (i + S (k + 1))%nat with (S i + S k)%nat; ring.
Qed.

Theorem tech_lemma : 
  forall k k1 k2 k3,
    ( IBZ k1 = 4*Z_of_nat k+1 -> 
      IBZ k2 = 2^(IBZ k1+2) -> 
      IBZ k3 = 3^(IBZ k1+2) -> 
      (IBR (numB k1 k2 k3) / IBR (denomB k1 k2 k3))%R = terme_pi k 0)%Z.
Proof.
intros k k1 k2 k3 Hk1 Hk2 Hk3.
unfold IBR.
rewrite <- denom_denomB_eq.
rewrite <- num_numB_eq.
unfold terme_pi.
rewrite plus_0_l.
replace (4 * Z_of_nat k+3)%Z with  (4 * Z_of_nat k+1+2)%Z  by ring.
rewrite <- Hk1.
rewrite <- Hk2.
rewrite <- Hk3.
reflexivity.
Qed.

Theorem finite_sum_pi_correct: 
  forall k k1 k2 k3 step k1' k2' k3' x' y',
    ( IBZ k1 = 4*Z_of_nat k+1 -> 
      IBZ k2 = 2^(IBZ k1+2) -> 
      IBZ k3 = 3^(IBZ k1+2) -> 
      finite_sum_pi k1 k2 k3 (S step) =
      (k1', k2', k3', x', y') -> 
      (	IBR x' / IBR y' = sum_f_R0 (terme_pi k) step)%R /\
      IBZ y' <> 0 /\
      IBZ k1' = 4*Z_of_nat (k+ (S step))+1 /\ 
      IBZ k2' = 2^(IBZ k1'+2) /\ 
      IBZ k3' = 3^(IBZ k1'+2) )%Z.
Proof.
unfold finite_sum_pi.
intros k k1 k2 k3 step k1' k2' k3' x' y' Hk1 Hk2 Hk3 H.
replace (S step -1)%nat with step in H by omega.
split.
generalize H; clear H.
case step.

simpl finite_sum_pi_aux.
injection 1; intros; subst.
simpl; apply tech_lemma; assumption.

intros n H.
rewrite decomp_sum; omega || simpl.
rewrite sum_eq with (Bn := terme_pi (S k)).
rewrite <- (tech_lemma k k1 k2 k3); try assumption.

apply finite_sum_pi_aux_correct2 with (6 := H); try assumption.
rewrite <- denom_denomB_eq.
unfold denom; simpl_IBZ; ring.
simpl_IBZ; rewrite Hk1; ZR0compat.
intros i Hi.
unfold terme_pi.
replace (S i + k)%nat with (i + S k)%nat; unfold Rdiv; ring.

apply finite_sum_pi_aux_correct1  with (5 := H); assumption || idtac.
unfold denomB; simpl_IBZ.
rewrite Hk2; rewrite Hk3; rewrite Hk1.
ZR0compat.
Qed.

Theorem infinit_finit_sum_bound : 
  forall An k s,
    (forall n, (S k<= n)%nat -> 0 <= An n) -> 
    infinit_sum An s ->
    sum_f_R0 An k <= s.
Proof.    
intros An k s Hk Hinf.
assert (0<= s-sum_f_R0 An k);[|fourier].
apply infinit_sum_pos with (fun i : nat => An (i + S k)%nat).
intros n.
assert (U1 : (S k <= n+S k)%nat) by omega. 
exact (Hk (n+ S k)%nat U1).
apply infinit_sum_step_k.
exact Hinf.
Qed.

Theorem tech_series_bound : 
  forall x, 
    x*(2*Z2R b*Z2R b) <= Z2R b-2 -> 
    x <= 1.
Proof.
intros x Hx.
pose (Hbpos := base_positive).
assert (U1 : 1 <= Z2R b) by IDR_to_Z.
apply Rle_trans with ((Z2R b - 2)/ (2 * Z2R b * Z2R b)).
apply Rle_div_l.
ZR0compat.
assumption.
apply Rle_div_r_1.
ZR0compat.
apply Rle_trans with (1*Z2R b).
fourier.
apply Rmult_le_compat_r; fourier.
Qed.

Theorem pi_series_correct : 
  forall k k1 k2 k3 y step r vr s, 
    Srepresents r vr -> 
    Z2R b <= 16 ^ (S step) ->
    (0 <= IBZ y)%Z ->
    IBR y* (8/((4*INR k+3)*16^ k)) *(2*Z2R b*Z2R b) <= Z2R b-2 ->
    (IBZ k1 = 4*Z_of_nat k+1)%Z -> 
    (IBZ k2 = 2^(IBZ k1+2))%Z -> 
    (IBZ k3 = 3^(IBZ k1+2))%Z -> 
    infinit_sum (terme_pi k) s -> 
    -1 <= (IBR y*s+vr) <= 1 -> 
    Srepresents (pi_series k1 k2 k3 y (S step) r) (IBR y*s+vr).
Proof. 
pose (Hbpos := base_positive).
cofix.
intros k k1 k2 k3 y step r vr s Hrep Hstep Hypos Hy Hk1 Hk2 Hk3 Hinf Hbound.
str_unfold (pi_series k1 k2 k3 y (S step) r).
simpl.
fold_decomp.
let_simpl; intros aux q.
let_simpl; intros aux1 p.
let_simpl; intros aux2 k3'.
let_simpl; intros k1' k2'.
intros _ _ _ H; clear aux aux1 aux2.
unfold series_body.

generalize (finite_sum_pi_correct _ _ _ _ _ _ _ _ _ _ Hk1 Hk2 Hk3 H).
intros [G1 [G2 [G3 [G4 G5]]]].

assert (U1 : sum_f_R0 (terme_pi k) step <= s).
clear pi_series_correct;
abstract (apply infinit_finit_sum_bound; [| exact Hinf];
intros n _; apply terme_pi_pos).

assert (U2 : IBR y * s * (2 * Z2R b * Z2R b) <= Z2R b - 2).
clear pi_series_correct;
abstract (apply Rle_trans with (2:=Hy);
apply Rmult_le_compat_r; [unfold IDR; ZR0compat|];
apply Rmult_le_compat_l; [IBR_to_Z|];
apply Rle_div_l;[ZR0compat|];
apply pi_series_bound; assumption).

assert (R1 : IBR y * s + vr = vr + IBR y* (sum_f_R0 (terme_pi k) step) +
  (IBR y * (s - sum_f_R0 (terme_pi k) step))).
clear pi_series_correct;
abstract ( ring). 

rewrite R1.
make_digit_tac d r' Md1 Md2 Md3 Md4 pi_series_correct.

assert (U3 : 0 <= IBR y * sum_f_R0 (terme_pi k) step).
abstract (unfold IBR; ZR0compat;
apply cond_pos_sum;
apply terme_pi_pos).

assert (U4 : IBR y*sum_f_R0 (terme_pi k) step <=  IBR y*s).
abstract (apply Rmult_le_compat_l; [IBR_to_Z| assumption]).

apply Sadd_correct; assumption ||  idtac.
  abstract (assert (R2 : IBR y * sum_f_R0 (terme_pi k) step = IBR (y*p)%B / IBR q);
    [ rewrite <- G1; simpl_IBR; unfold Rdiv; ring |  
      rewrite R2;
	apply Srat_to_stream_correct; omega||idtac; 
	  rewrite <- R2; split; [fourier|];
	    apply Rle_trans with (IBR y*s); intuition;
	      apply tech_series_bound; omega || intuition]).

 abstract (split; [
   generalize (Srepresents_interval _ _ Hrep);
       intros [U8 _]; fourier
 | generalize Hbound; intros; dec_tac; fourier]).

 abstract (rewrite <- R1; assumption).

 split. 
 abstract (apply Rle_trans with 0;[
  assert (0 <= Z2R b -2) by IDR_to_Z; fourier |
  unfold IBR, IDR; ZR0compat]).
 abstract ( apply Rle_trans with (2:= U2);
   apply Rmult_le_compat_r; [unfold IDR; ZR0compat|];
   apply Rmult_le_compat_l;[IBR_to_Z|];
   assert (0 <= sum_f_R0 (terme_pi k) step);[|fourier];
   apply cond_pos_sum;
   intros; apply terme_pi_pos).

inversion Md3.

match goal with 
  |- Srepresents (?d::_) ((IDR ?d + ?r1)/Z2R ?b + IBR ?y*?r2) => 
    replace ((IDR d + r1)/Z2R b + IBR y*r2) with 
      ((IDR d + (IBR (Bbase*y)%B*r2+r1))/Z2R b) ; 
      [|clear pi_series_correct; simpl_IBR; field_IDR]
end.

assert (U4 : -1 <=
   IBR (Bbase * y)%B * (s - sum_f_R0 (terme_pi k) step) + r0<=1).
clear pi_series_correct;
abstract (rewrite <- H2 in Md4;
replace (((IDR d + r0) / Z2R b + IBR y * (s - sum_f_R0 (terme_pi k) step)) *
  Z2R b - IDR d) with 
(IBR (Bbase * y)%B * (s - sum_f_R0 (terme_pi k) step) + r0) in Md4;
[assumption|simpl_IBR;field_IDR]). 

constructor; try apply pi_series_correct with (k+(S step))%nat;
  clear pi_series_correct; try assumption.

abstract(simpl_IBZ; ZR0compat).

apply Rle_trans with (2:=Hy);
apply Rmult_le_compat_r; [unfold IDR; ZR0compat|];
simpl_IBR;
rewrite pow_add.
match goal with 
  |- ?b*?y*(8/(?r1*(?r2*?r3))) <= ?y*(8/(?r4*?r2)) =>
    replace (b*y*(8/(r1*(r2*r3)))) with 
      ((y*8/r2)*(b/(r1*r3))); [
	replace (y*(8/(r4*r2))) with 
	  ((y*8/r2)*(1/r4))|]
end.
abstract(apply Rmult_le_compat_l; unfold IBR; ZR0compat;
apply Rdiv_mult_compat; ZR0compat;
rewrite Rmult_1_l;
rewrite Rmult_comm; apply Rmult_le_compat; assumption || unfold IDR; ZR0compat;
apply Rplus_le_compat_r;apply Rmult_le_compat_l; [fourier|];
apply le_INR; omega).
abstract(field; split; ZR0compat).
abstract(field; repeat split; ZR0compat).

apply infinit_sum_ext with (An := (fun i => (terme_pi k) (i +(S step))%nat));
  [intros n; apply terme_pi_add; omega|
    apply infinit_sum_step_k; assumption].
Qed.

Theorem find_step_correct : 
  (1 <= find_step)%nat /\ 
  Z2R b <= 16 ^ find_step.
Proof.
unfold find_step.
Ddiv_tac q r Heq Hr.

split.
 assert (0 <=  N_digits (b - 1)/4)%Z; [| omega].
 ZR0compat; apply ZERO_le_N_digits.

 assert (Hbpos := base_positive).
 assert (Hbm1pos : (0 < b-1)%Z ) by omega.
 generalize (N_digits_correct (b-1) Hbm1pos).
 intros [ _ H1].
 assert (H2 : (b <=  2 ^ (N_digits ( b - 1) + 1))%Z) by omega.
 clear H1; R_to_Z.
 change (2*(2*(2*2)))%Z with (2^4)%Z.
 rewrite <- Zpower_mult ; simpl_IBZ; (ZR0compat; apply ZERO_le_N_digits)|| idtac. 
 apply Zle_trans with (1:= H2).
 apply Zle_power; [omega |].
 split; [ZR0compat; apply ZERO_le_N_digits |].
 simpl_IDZ; rewrite inj_S; rewrite Z_of_nat_Zabs_nat.
   omega.
   assert  (0 <= N_digits (b - 1))%Z; [apply ZERO_le_N_digits |omega].
Qed.

Theorem pi_div4_correct : 
  forall s,
    infinit_sum (terme_pi 0) s -> 
    Srepresents (pi_div4) s.
Proof.
pose (Hbpos := base_positive).
intros s Hinf.
unfold pi_div4.
let_simpl; intros aux q.
let_simpl; intros aux1 p.
let_simpl; intros aux2 k3'.
let_simpl; intros k1' k2'.
intros _ _ _ H; clear aux aux1 aux2.
assert (Hk1 : (IBZ 1%B =  4*Z_of_nat 0+1)%Z) .
simpl_IBZ; reflexivity.
assert (Hk2 : (IBZ 8%B = 2^(IBZ 1%B+2))%Z). 
simpl_IBZ; reflexivity.
assert (Hk3 : (IBZ 27%B = 3^(IBZ 1%B+2))%Z). 
simpl_IBZ; reflexivity.
generalize find_step_correct.
intros [U1 U2].
assert (R1 : find_step = 
  S (pred find_step)) by omega.
generalize H U1 U2; clear H U1 U2.
rewrite R1; clear R1.
generalize (pred find_step).
intros step H U1 U2.

generalize (finite_sum_pi_correct _ _ _ _ _ _ _ _ _ _ Hk1 Hk2 Hk3 H).
rewrite plus_0_l.
intros [G1 [G2 [G3 [G4 G5]]]].

assert (R2 : s = IBR 1%B * (s - IBR p/ IBR q) + IBR p / IBR q) by field_IBR.
rewrite R2.

apply pi_series_correct with (5:=G3); assumption || idtac.

apply Srat_to_stream_correct; try omega.
rewrite G1.
split.
apply Rle_trans with 0; intuition.
apply cond_pos_sum; intros; apply terme_pi_pos.
apply Rle_trans with s.
apply infinit_finit_sum_bound; [|assumption].
intros n _; apply terme_pi_pos.
apply pi_series_bound_1 with 0%nat; assumption.

simpl_IBZ; omega.

simpl_IBR; rewrite Rmult_1_l.
rewrite Rmult_comm.
rewrite Rmult_div_assoc;[|ZR0compat].
apply Rle_div_r; [ZR0compat|].
replace (2* Z2R b * Z2R b * 8) with (Z2R b * (16 *Z2R b)) by ring.
rewrite <- (Rmult_assoc (Z2R b - 2)).
apply Rmult_le_compat; [IDR_to_Z| IDR_to_Z| |].
apply Rle_trans with ((Z2R b - 2)*11); [IDR_to_Z|].
apply Rmult_le_compat_l; [IDR_to_Z|].
assert (2 <= INR (S (S step)) ); [| fourier].
apply (le_INR 2); omega.
rewrite <- tech_pow_Rmult.
apply Rmult_le_compat_l; [fourier| assumption].

rewrite G1.
apply infinit_sum_ext with 
  (An := (fun i => (terme_pi 0) (i +(S (S step)))%nat)).
intros n; apply terme_pi_add; omega.
apply infinit_sum_step_k; assumption.

rewrite <- R2.
split.
apply Rle_trans with 0; [fourier| apply pi_series_pos with 0%nat; assumption].
apply pi_series_bound_1 with 0%nat; assumption.
Qed.

Axiom pi_series : infinit_sum (terme_pi 0) (PI/4). 

Theorem pi_div4_correct' : 
  Srepresents pi_div4 (PI/4).
Proof.
apply pi_div4_correct; apply pi_series.
Qed.

Theorem pi_correct : 
  represents pi PI.
Proof.
pose (Hbpos := base_positive).
unfold pi.
replace PI with (Z2R b^1 * (IDR 4%D/Z2R b*(PI/4))).
constructor.
apply mult_digit_correct.
simpl_IDZ; omega.
apply pi_div4_correct'.
field_IDR.
Qed.

End PI_CORRECT.
