Require Export Reals.
Require Export Field.
Require Export ZArith.
Require Export Fourier.
Require Export Zwf.
Require Export Field_tac. 

(*borrowed and new proofs to apply archimede's principle*)
Theorem INR_Zabs_nat_eq :
  forall (z:Z), (0 <= z)%Z -> (IZR z = INR(Zabs_nat z))%R.
unfold Zabs_nat.
intros z; case z.
simpl.
intros; auto.
unfold IZR.
intros; auto.
compute.
intros x H'; elim H'; auto.
Qed.

Lemma pow2_larger :
  forall n, (INR n <= 2^n).
induction n.
simpl; fourier.
replace (S n) with (1+n)%nat.
rewrite pow_add; rewrite pow_1.
assert (Hlem: (forall a b, 1<= b -> a<=b -> 1+a<= 2*b)%R).
intros; fourier.
rewrite plus_INR; simpl (INR 1).
apply Hlem; auto.
apply pow_R1_Rle;fourier.
omega.
Qed.

Lemma pow_half_smaller :
 forall r, (0 < r) -> (1/(2^(Zabs_nat(up (1/r)))) <= r).
intros r Hpos.
elim (archimed (1/r)); intros Harc1 Harc2.
apply (Rmult_le_reg_l (2^(Zabs_nat (up (1/r))))).
apply pow_lt.
fourier.
replace (2 ^ Zabs_nat (up (1 / r)) * (1 / 2 ^ Zabs_nat (up (1 / r))))%R
 with 1%R.
apply Rle_trans with ((INR (Zabs_nat (up (1/r))))*r)%R.
rewrite <- INR_Zabs_nat_eq.
pattern 1%R at 1; replace 1%R with (1/r * r)%R ;[|field; auto with real].
apply Rmult_le_compat_r.
fourier.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
apply le_IZR.
apply Rle_trans with (1/r)%R.
replace (1/r)%R with (/r)%R.
simpl IZR.
assert (0 < / r)%R.
apply Rinv_0_lt_compat; auto.
unfold Rle; left; assumption.
field.
apply Rgt_not_eq; auto.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
apply Rmult_le_compat_r.
fourier.
apply pow2_larger.
field.
apply pow_nonzero.
discrR.
Qed.

Lemma archim_lt:
(forall v, 0 < v -> {n:nat| 1/(2^n) <= v/2}).
intros v Hvpos.
exists (Zabs_nat (up (1/v))+1)%nat.
destruct (archimed (1/v)) as [Hb1 _].
rewrite pow_add.
replace (1 / (2 ^ Zabs_nat (up (1 / v))*2^1))%R
 with ((1 / (2 ^ Zabs_nat (up (1 / v))))*/2)%R.
replace (v/2)%R with (v*/2)%R.
apply Rmult_le_compat_r.
assert (0 < /2)%R.
apply Rinv_0_lt_compat;auto.
fourier.
fourier.
apply pow_half_smaller.
fourier.
field.
replace (2^1)%R with 2%R.
field.
assert (2^Zabs_nat(up(1*/v)) <> 0)%R.
apply pow_nonzero;auto.
discrR.
simpl;assumption.
simpl;field.
Qed.

Lemma archim_lt2:
forall v, 0 < v -> {n:nat| (1/(2^n) < v)%R}.
intros v Hvpos.
assert (Hhalf: {n:nat| 1/(2^n) <= v/2}).
apply archim_lt; assumption.
elim Hhalf; intros n Hineq.
exists n.
assert (v/2<v).
fourier.
unfold Rle in Hineq; elim Hineq; intros.
apply Rlt_trans with (v/2); assumption.
rewrite H0.
assumption.
Qed.


(* usefull general proofs and tactics *)

(**** Tactics ***)

Ltac dec_tac :=
  match goal with 
   id: (_ <= _ <= _)%R |- _ => decompose [and] id; clear id; dec_tac
  |id: (_ < _ <= _)%R |- _ => decompose [and] id; clear id; dec_tac
  |id: (_ < _ < _)%R |- _ => decompose [and] id; clear id; dec_tac
  |id: (_ < _ <= _)%R |- _ => decompose [and] id; clear id; dec_tac
  | |- _ => idtac
  end.

Ltac caseEq f := generalize (refl_equal f); pattern f at -1; case f.

Ltac rewrite_all H :=
 match type of H with
 | ?t1 = ?t2 =>
   let rec aux H :=
     match goal with
     | id : context [t1] |- _ =>
       match type of id with
       | t1 = t2 => fail 1
       | _ => generalize id;clear id; try aux H; intro id
       end
     | _ => try rewrite H
     end in
   aux H
 end. 

Ltac generalize_all H1 H2 :=
  match goal with
    | id : context [H1] |- _ => generalize id; clear id; generalize_all H1 H2; intros id
    | _ => generalize H1; intros H2
  end.

Definition Zle_bool_total_tf :
  forall x y:Z, {Zle_bool x y = true} + {Zle_bool x y = false}.
Proof.
intros x y.
elim (Zle_bool_total x y).
tauto.
unfold Zle_bool.
case (x?=y)%Z; auto.
Defined.

Ltac Zbool H := 
  match goal with 
    |- context [Zle_bool ?a ?b] => 
      generalize (Zle_cases a b);
	elim (Zle_bool_total_tf a b); 
	  intros H; rewrite H; clear H; intro H
  end.

Theorem minus_IZR : forall m n, IZR (m - n)%Z = (IZR m - IZR n)%R.
Proof.
intros m n; rewrite Z_R_minus; auto.
Qed.

Theorem opp_IZR : forall n, IZR (- n)%Z = (- IZR n)%R.
Proof.
intros n; rewrite Ropp_Ropp_IZR; auto.
Qed.

Theorem pow_IZR : forall (z : Z) (n : nat), 
  IZR (z ^ Z_of_nat n) = IZR z ^ n.
Proof.
intros z n; rewrite pow_IZR; auto.
Qed.

Fixpoint P2R (p : positive) :=
  match p with
  | xH => 1%R
  | xO xH => 2%R
  | xO t => (2 * P2R t)%R
  | xI xH => 3%R
  | xI t => (1 + 2 * P2R t)%R
  end.

Definition Z2R n :=
  match n with
  | Zpos p => P2R p
  | Zneg p => Ropp (P2R p)
  | Z0 => R0
  end.

Lemma P2R_INR :
  forall n, P2R n = INR (nat_of_P n).
induction n ; simpl ; try (
  rewrite IHn ;
  rewrite <- (mult_INR 2) ;
  rewrite <- (nat_of_P_mult_morphism 2) ;
  change (2 * n)%positive with (xO n)).
(* xI *)
rewrite (Rplus_comm 1).
change (nat_of_P (xO n)) with (Pmult_nat n 2).
case n ; intros ; simpl ; try apply refl_equal.
case (Pmult_nat p 4) ; intros ; try apply refl_equal.
rewrite Rplus_0_l.
apply refl_equal.
apply Rplus_comm.
(* xO *)
case n ; intros ; apply refl_equal.
(* xH *)
apply refl_equal.
Qed.

Lemma Z2R_IZR :
  forall n, Z2R n = IZR n.
intro.
case n ; intros ; simpl.
apply refl_equal.
apply P2R_INR.
apply Ropp_eq_compat.
apply P2R_INR.
Qed.

Lemma plus_Z2R :
  forall m n, (Z2R (m + n) = Z2R m + Z2R n)%R.
intros.
repeat rewrite Z2R_IZR.
apply plus_IZR.
Qed.

Lemma minus_Z2R :
  forall m n, (Z2R (m - n) = Z2R m - Z2R n)%R.
intros.
repeat rewrite Z2R_IZR.
apply minus_IZR.
Qed.

Lemma mult_Z2R :
  forall m n, (Z2R (m * n) = Z2R m * Z2R n)%R.
intros.
repeat rewrite Z2R_IZR.
apply mult_IZR.
Qed.

Lemma opp_Z2R :
  forall m, (Z2R (- m) = - Z2R m)%R.
intros.
repeat rewrite Z2R_IZR.
apply Ropp_Ropp_IZR.
Qed.

Lemma Rabs_Z2R : 
  forall m, Z2R (Zabs m) = Rabs (Z2R m) .
intros.
repeat rewrite Z2R_IZR.
rewrite Rabs_Zabs; auto.
Qed.

Theorem pow_Z2R : 
  forall (z : Z) (n : nat), 
  Z2R (z ^ Z_of_nat n) = Z2R z ^ n.
Proof.
intros z n.
repeat rewrite Z2R_IZR.
rewrite pow_IZR; auto.
Qed.

Lemma Z_of_nat_Z2R :
  forall m, Z2R (Z_of_nat m) = INR m.
intros m.
repeat rewrite Z2R_IZR.
rewrite INR_IZR_INZ; auto.
Qed.

Lemma Z2R_le :
  forall m n, (m <= n)%Z -> (Z2R m <= Z2R n)%R.
intros m n.
repeat rewrite Z2R_IZR.
apply IZR_le.
Qed.

Lemma le_Z2R :
  forall m n, (Z2R m <= Z2R n)%R -> (m <= n)%Z.
intros m n.
repeat rewrite Z2R_IZR.
apply le_IZR.
Qed.

Lemma Z2R_lt :
  forall m n, (m < n)%Z -> (Z2R m < Z2R n)%R.
intros m n.
repeat rewrite Z2R_IZR.
apply IZR_lt.
Qed.

Lemma lt_Z2R :
  forall m n, (Z2R m < Z2R n)%R -> (m < n)%Z.
intros m n.
repeat rewrite Z2R_IZR.
apply lt_IZR.
Qed.

Lemma Z2R_eq :
  forall m n, (m = n)%Z -> (Z2R m = Z2R n)%R.
intros m n H.
rewrite H; auto.
Qed.

Lemma Z2R_neq :
  forall m n, (m <> n)%Z -> (Z2R m <> Z2R n)%R.
intros m n H.
repeat rewrite Z2R_IZR.
apply IZR_neq; auto.
Qed.

Ltac simpl_Z2R := 
  repeat (rewrite minus_Z2R || rewrite plus_Z2R || rewrite mult_Z2R || rewrite Z_of_nat_Z2R
    || rewrite Rabs_Z2R || rewrite opp_Z2R || rewrite pow_Z2R);  simpl Z2R.

Ltac fold_Z2R :=
  change 2%R with (Z2R 2%Z); change 0%R with (Z2R 0%Z); change 1%R with (Z2R 1%Z);
    repeat (rewrite <- plus_Z2R || rewrite <- minus_Z2R || rewrite <- Rabs_Z2R || rewrite <- opp_Z2R ||
      rewrite <- Z_of_nat_Z2R || rewrite <- mult_Z2R || rewrite <- pow_Z2R).

Ltac simpl_IZR := 
  repeat rewrite <- Z2R_IZR;
    simpl_Z2R; repeat rewrite Z2R_IZR.

Ltac fold_IZR := 
  repeat rewrite <- Z2R_IZR;
    fold_Z2R; repeat rewrite Z2R_IZR.

Ltac R_to_Z :=
  repeat rewrite <- Z2R_IZR;
  fold_Z2R;
  match goal with 
    | |- (Z2R _ < Z2R _)%R => apply Z2R_lt
    | |- (Z2R _ <= Z2R _)%R => apply Z2R_le
    | |- (Z2R _ = Z2R _)%R => apply Z2R_eq
    | |- (Z2R _ <> Z2R _)%R => apply Z2R_neq
    | |- _ => idtac
  end;
  simpl_Z2R; try omega.

Ltac field_Z := 
  simpl_Z2R; field; repeat split; R_to_Z; try omega.

Ltac make_Z2R_ineq := 
  match goal with
    | id : (?a < ?b)%Z |- _ => generalize (Z2R_lt a b id); 
      generalize id; clear id; make_Z2R_ineq
    | id : (?a <= ?b)%Z |- _ => generalize (Z2R_le a b id); 
      generalize id; clear id; make_Z2R_ineq
    | id : (?a = ?b)%Z |- _ => generalize (Z2R_eq a b id); 
      generalize id; clear id; make_Z2R_ineq
    | id : (?a <> ?b)%Z |- _ => generalize (Z2R_neq a b id); 
      generalize id; clear id; make_Z2R_ineq
    | |- _ => simpl_Z2R
  end.

(* zero *)

Theorem Rdiv_neq_0 : 
  forall x y, 
    y <> 0 -> 
    x / y <> 0 ->
    x <> 0.
Proof.
intros x y H1 H2.
replace x with (x/y*y);[| field; assumption].
apply prod_neq_R0; assumption.
Qed.

Theorem Rmult_neq_0 :
  forall x y, 
    x * y <> 0 ->
    x <> 0. 
Proof.
intros x y H1 H2. 
apply H1; subst; ring.
Qed.

Lemma Rlt_neq_0 : forall r, (0 < r)%R -> (r <> 0)%R.
intros r Hpos; replace r with (0 + r)%R.
apply tech_Rplus.
fourier.
auto.
field.
Qed.

Theorem Rdiv_le_0_compat :
  forall x y, (0 <= x -> 0 < y  -> 0 <= x / y)%R.
Proof.
intros x y H H1; replace (x / y)%R with (x * /y)%R.
apply Rmult_le_pos.
assumption.
assert (0 < /y)%R.
apply Rinv_0_lt_compat;auto.
apply Rlt_le; assumption.
field.
apply Rlt_neq_0;auto.
Qed.

Lemma Zmult_non0 :
  forall x y, (x <> 0-> y<> 0 -> x*y  <> 0)%Z.
Proof.
  intros x y H H1 H2; elim (Zmult_integral _ _ H2); intuition.
Qed.

Lemma Zmult_non_0_r :
  forall x y, x*y  <> 0 -> y <> 0.
  intros x y H H1; apply H; rewrite H1; ring.
Qed.

Theorem Zpower_1_r : 
  forall k, (k^1=k)%Z.
Proof.
intros k.
simpl; unfold Zpower_pos; simpl; ring.
Qed.

Theorem Zpower_le_0_compat :
  forall k i, (0 <= k -> 0 <= k ^ i)%Z.
Proof.
intros k i Hkpos;elim i using (well_founded_ind (Zwf_well_founded 0)).
clear i; intros i IHi.
destruct (Z_le_gt_dec 0 i) as [Hipos | Hineg].
destruct (Z_eq_dec i 0) as [Hi0 | Hin0].
subst i; compute; intros; discriminate.

replace i with (1+(i-1))%Z.
rewrite Zpower_exp.
rewrite Zpower_1_r.
apply Zmult_le_0_compat.
assumption.
apply IHi.
unfold Zwf; omega.
omega.
omega.
ring.
destruct i.
discriminate Hineg.
discriminate Hineg.
simpl; auto with zarith.
Qed.

Lemma Zpower_not_0 :
  forall x n, 
    (0 <= n ->
      x <> 0 ->
      x ^ n <> 0)%Z.
Proof.
  intros x n; elim n using (well_founded_ind (Zwf_well_founded 0)).
  clear n; intros n Hrec Hpos Hn0; elim (Z_le_gt_dec n 0).
  intros Hneg.
  replace n with 0%Z; simpl; omega.
  intros Hgt; replace n with ((n-1)+1)%Z;[idtac|ring].
  rewrite Zpower_exp; try omega.
  rewrite Zpower_1_r.
  apply Zmult_non0.
  apply Hrec.
  unfold Zwf; omega.
  omega.
  assumption.
  assumption.
Qed.

Theorem Zpow_pos_lemma : 
  forall p n m,
    (0 < m -> 
      0 < n ->
      0 < iter_pos p Z (fun x : Z => n * x) m)%Z.
Proof.
intros p; induction p;simpl; 
intros n m Hm Hn. 
apply Zmult_lt_0_compat;[ | repeat apply IHp]; assumption.
repeat apply IHp; assumption.
apply Zmult_lt_0_compat; assumption.
Qed.

Lemma Zpow_pos :
  forall i n, (0 < n -> 0 <=i -> 0 < n ^i)%Z. 
Proof.
intros i n Hn; case i.
intros; simpl; omega.
intros p _.
unfold Zpower.
unfold Zpower_pos.
apply Zpow_pos_lemma; omega.
intros p H1.
absurd ( Zneg p < 0)%Z.
omega.
apply Zlt_neg_0.
Qed.

Theorem Rdiv_lt_0_compat: 
  forall x y : R, 0 < x -> 0 < y -> 0 < x / y.
Proof.
intros.
unfold Rdiv.
apply Rmult_lt_0_compat.
assumption.
apply Rinv_0_lt_compat; assumption.
Qed.

Theorem Zdiv_le_0_compat : 
  forall a b, 
    (0 <= a -> 
    0 < b -> 
    0 <= a / b)%Z.
Proof.
intros a b Hapos Hbpos.
apply Zge_le.
apply Z_div_ge0; omega.
Qed.

Theorem Rlt_gt : 
  forall x y, 
    (x < y -> y> x)%R.
Proof. 
intros; fourier.
Qed.

Theorem Rle_0_opp : 
  forall x, 
    x <= 0 -> 
    0 <= - x.
Proof.
intros; fourier.
Qed.

Theorem Ropp_le_0 : 
  forall x, 
    0 <= x -> 
    - x <= 0.
Proof.
intros; fourier.
Qed.

Ltac R0compat := 
  match goal with 
    | |- ( 0 <= Rsqr _)%R => apply Rle_0_sqr
    | |- (0 < Rabs _)%R => apply Rabs_pos_lt
    | |- (0 <= Rabs _)%R => apply Rabs_pos
    | |- (- _ < 0)%R => apply Ropp_lt_gt_0_contravar
    | |- (0 < - _ )%R => apply Ropp_0_gt_lt_contravar
    | |- (- _ <= 0)%R => apply Ropp_le_0
    | |- (0 <= - _ )%R => apply Rle_0_opp
    | |- ( _ <> 0)%R => apply Rlt_neq_0
    | |- ( 0 <> _)%R => apply sym_not_eq
    | |- ( _ >= 0)%R => apply Rle_ge
    | |- ( _ > 0)%R  => apply Rlt_gt
    | |- (0 <  _ + _)%R => apply Rplus_le_lt_0_compat
    | |- (0 <= _ + _)%R => apply Rplus_le_le_0_compat
    | |- (0 <  _ * _)%R => apply Rmult_lt_0_compat
    | |- (0 <= _ * _)%R => apply Rmult_le_pos
    | |- (0 <= _ / _)%R => apply Rdiv_le_0_compat
    | |- (0 <  _ / _)%R => apply Rdiv_lt_0_compat
    | |- (0 <    / _)%R => apply Rinv_0_lt_compat
    | |- (0 <  _ ^ _)%R => apply pow_lt
    | |- (0 <= _ ^ _)%R => apply pow_le
    | |- (0 <  Z2R _)%R => apply (Z2R_lt 0)
    | |- (0 <= Z2R _)%R => apply (Z2R_le 0)
    | |- (0 <  IZR _)%R => apply (IZR_lt 0)
    | |- (0 <= IZR _)%R => apply (IZR_le 0)
    | |- (0 <= INR _)%R => apply pos_INR
    | |- (0 <  INR _)%R => apply lt_INR_0

  end.

Ltac Z0compat := 
  match goal with 
    | |- (_ * _ <> 0)%Z => apply Zmult_non0
    | |- (_ ^ _ <> 0)%Z => apply Zpower_not_0
    | |- (_ <> 0)%Z => apply sym_not_eq
    | |- (0 <> _)%Z => apply Zlt_not_eq
    | |- (_ >= 0)%Z => apply Zle_ge
    | |- (_ > 0)%Z  => apply Zlt_gt
    | |- (0 <= _ + _)%Z => apply Zplus_le_0_compat
    | |- (0 <  _ * _)%Z => apply Zmult_lt_0_compat
    | |- (0 <= _ * _)%Z => apply Zmult_le_0_compat
    | |- (0 <= _ / _)%Z => apply Zdiv_le_0_compat
    | |- (0 <  _ ^ _)%Z => apply Zpow_pos
    | |- (0 <= _ ^ _)%Z => apply Zpower_le_0_compat
    | |- (0 < fact _)%nat => apply lt_O_fact
  end.

Ltac ZR0compat := 
  repeat (R0compat || Z0compat || assumption || omega || fourier).

(* compat *)

Theorem Rdiv_le_compat_r : 
  forall x y z, 
    (0 < z -> 
      x <= y ->
      x / z <= y / z)%R.
Proof.
intros.
unfold Rdiv.
apply Rmult_le_compat_r.
apply Rlt_le.
apply Rinv_0_lt_compat.
assumption.
assumption.
Qed.

Theorem Rdiv_le_compat_l: 
  forall x y z : R, 
    0 <= x -> 
    0 < z -> 
    z <= y -> 
    x / y <= x / z.
Proof.
intros.
unfold Rdiv.
apply Rmult_le_compat_l; try assumption.
apply Rle_Rinv; ZR0compat.
Qed.

Theorem Rdiv_lt_compat_l:
  forall x y z : R, 
    0 < x ->
    0 < z * y -> 
    z < y ->
    x / y < x / z.
Proof.
intros x y z H1 H2 H3.
unfold Rdiv.
apply Rmult_lt_compat_l.
assumption.
apply Rinv_lt_contravar; assumption.
Qed.

Theorem Rmult_div_simpl : 
  forall x y, 
    (x <> 0 -> 
    x * (y / x) = y)%R.
Proof.
intros; field; auto.
Qed.

Theorem Rdiv_lt_compat_r:
  forall x y z : R, 
    0 < z -> 
    x < y ->
    x / z < y / z.
Proof.
intros x y z H1 H2.
apply Rmult_lt_reg_l with z; [assumption|].
repeat rewrite Rmult_div_simpl; ZR0compat.
Qed.

Theorem Rdiv_eq_compat_l: 
  forall x y z : R, 
    0 < z -> 
    y = z -> 
    x / y = x / z.
Proof.
intros; subst.
field.
ZR0compat.
Qed.

(* Rle_div *)

Theorem Rplus_minus_le : 
  forall x y z,
    (x <= z - y -> 
      x + y <= z)%R.
Proof.
intros; fourier.
Qed.

Theorem Rminus_plus_le : 
  forall x y z,
    (x - y <= z -> 
      x <= y + z)%R.
Proof.
intros; fourier.
Qed.




Theorem  Ropp_div_distr_l_reverse  :
  forall x y,
    (- x / y = -(x/y))%R.
Proof.
intros. 
unfold Rdiv.
ring.
Qed.

Theorem Ropp_div_distr_r_reverse: 
  forall r1 r2, 
    (r2 <> 0 ->
      r1 / - r2 = - (r1 / r2))%R.
Proof.
intros r1 r2 H.
unfold Rdiv.
rewrite <- Ropp_inv_permute.
apply Ropp_mult_distr_r_reverse.
assumption.
Qed.

Theorem Rdiv_plus_distr_r : 
  forall x y z,
    ((x+y)/ z = x/z+ y/z)%R.
Proof.
intros.
unfold Rdiv.
ring.
Qed.

Theorem Rdiv_div_mult : 
  forall x y z, 
    0 <> y -> 
    0 <> z ->
    x/y/z = x/(y*z).
Proof.
intros; field; split; auto.
Qed.

Theorem Rmult_div_assoc :
  forall a b c, (c <> 0 -> a * (b/c) = a*b /c)%R.
intros; field; auto.
Qed.

Theorem Rle_div_l : 
  forall x1 x2 x3, 
    (0 < x3 ->
     x1 * x3 <= x2  ->
      x1 <= x2 / x3)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hle.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
apply Rmult_le_reg_l with x3; try assumption.
replace (x3 * (x2 / x3))%R with x2;
  [rewrite Rmult_comm; assumption|field; auto].
Qed.

Theorem Rlt_div_l : 
  forall x1 x2 x3, 
    (0 < x3 ->
     x1 * x3 < x2  ->
      x1 < x2 / x3)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hlt.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
apply Rmult_lt_reg_l with x3; try assumption.
replace (x3 * (x2 / x3))%R with x2;
  [rewrite Rmult_comm; assumption|field; auto].
Qed.

Theorem Rdiv_le_l : 
  forall x1 x2 x3, 
    (0 < x3 ->
      x1 <= x2 / x3 ->
      x1 * x3 <= x2)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hle.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
replace x2 with (x2/x3*x3); [idtac| field; auto].
apply Rmult_le_compat_r; intuition.
Qed.

Theorem Rdiv_lt_l : 
  forall x1 x2 x3, 
    (0 < x3 ->
      x1 < x2 / x3 ->
      x1 * x3 < x2)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hlt.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
replace x2 with (x2/x3*x3); [idtac| field; auto].
apply Rmult_lt_compat_r; intuition.
Qed.

Theorem Rle_div_r : 
  forall x1 x2 x3, 
    (0 < x3 ->
      x1 <= x2 * x3 ->
      x1 / x3 <= x2)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hle.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
apply Rmult_le_reg_l with x3; try assumption.
replace (x3 * (x1 / x3))%R with x1;
  [rewrite Rmult_comm; assumption|field; auto].
Qed.

Theorem Rlt_div_r : 
  forall x1 x2 x3, 
    (0 < x3 ->
      x1 < x2 * x3 ->
      x1 / x3 < x2)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hlt.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
apply Rmult_lt_reg_l with x3; try assumption.
replace (x3 * (x1 / x3))%R with x1;
  [rewrite Rmult_comm; assumption|field; auto].
Qed.

Theorem Rdiv_le_r : 
  forall x1 x2 x3, 
    (0 < x3 ->
      x1 / x3 <= x2 -> 
      x1 <= x2 * x3)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hle.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
replace x1 with (x1/x3*x3); [idtac| field; auto].
apply Rmult_le_compat_r; intuition.
Qed.

Theorem Rdiv_lt_r : 
  forall x1 x2 x3, 
    (0 < x3 ->
      x1 / x3 < x2 -> 
      x1 < x2 * x3)%R.
Proof.
intros x1 x2 x3 H0ltx3 Hlt.
assert (G1: (0<>x3)%R).
apply Rlt_not_eq; assumption.
replace x1 with (x1/x3*x3); [idtac| field; auto].
apply Rmult_lt_compat_r; intuition.
Qed.

Theorem Rle_div : 
  forall x1 x2 x3 x4, 
    (0 < x4 ->
     x1 * x4 <= x2 <= x3 * x4 ->
      x1 <= x2 / x4 <= x3)%R.
Proof.
split.
apply Rle_div_l; intuition.
apply Rle_div_r; intuition.
Qed.

Theorem Rlt_div : 
  forall x1 x2 x3 x4, 
    (0 < x4 ->
     x1 * x4 < x2 < x3 * x4 ->
      x1 < x2 / x4 < x3)%R.
Proof.
split.
apply Rlt_div_l; intuition.
apply Rlt_div_r; intuition.
Qed.

Theorem Rdiv_le : 
  forall x1 x2 x3 x4, 
    (0 < x4 ->
      x1 <= x2 / x4 <= x3 ->
      x1 * x4 <= x2 <= x3 * x4)%R.
Proof.
split.
apply Rdiv_le_l; intuition.
apply Rdiv_le_r; intuition.
Qed.

Theorem Rdiv_lt : 
  forall x1 x2 x3 x4, 
    (0 < x4 ->
      x1 < x2 / x4 < x3 ->
      x1 * x4 < x2 < x3 * x4)%R.
Proof.
split.
apply Rdiv_lt_l; intuition.
apply Rdiv_lt_r; intuition.
Qed.

Theorem Rle_div_l_1 : 
  forall x1 x2, 
    (0 < x1 ->
      -x1 <= x2  ->
      -1 <= x2 / x1)%R.
Proof.
intros x1 x2 H0ltx1 Hle.
apply Rle_div_l; fourier.
Qed.

Theorem Rdiv_le_l_1 : 
  forall x1 x2, 
    (0 < x1 ->
      -1 <= x2 / x1 ->
      -x1 <= x2 )%R.
Proof.
intros x1 x2 H0ltx1 Hle.
replace (-x1)%R with (-1*x1)%R; [idtac| ring].
apply Rdiv_le_l; intuition.
Qed.

Theorem Rle_div_r_1 : 
  forall x1 x2, 
    (0 < x2 ->
      x1 <= x2  ->
      x1/ x2 <= 1)%R.
Proof.
intros x1 x2 H0ltx1 Hle.
apply Rle_div_r; intuition.
rewrite Rmult_1_l; assumption.
Qed.

Theorem Rdiv_le_r_1 : 
  forall x1 x2, 
    (0 < x2 ->
      x1 / x2 <= 1  ->
      x1 <= x2)%R.
Proof.
intros x1 x2 H0ltx1 Hle.
rewrite <- Rmult_1_l with x2.
apply Rdiv_le_r; intuition.
Qed.

Theorem Rle_div_1 : 
  forall x1 x2, 
    (0 < x1 ->
     - x1 <= x2 <= x1 ->
      -1 <= x2 / x1 <= 1)%R.
Proof.
intros x1 x2 Hx1 Hx2.
split.
apply Rle_div_l_1; intuition.
apply Rle_div_r_1; intuition.
Qed.

Theorem Rdiv_le_1 : 
  forall x1 x2, 
    (0 < x1 ->
      -1 <= x2 / x1 <= 1 ->
      - x1 <= x2 <=  x1)%R.
Proof.
intros x1 x2 Hx1 Hx2.
split.
apply Rdiv_le_l_1; intuition.
apply Rdiv_le_r_1; intuition.
Qed.

Theorem Rdiv_le_mult :
  forall r1 r2 r3,
    (0 <= r1 ->
    -r1 <= r3 <= r1 ->
    -1 <= r2 <= 1 -> 
    -r1 <= r2*r3 <= r1)%R.
Proof.
intros r1 r2 r3 Hr1pos Hr3 Hr2.
elim (Rle_or_lt 0 r2); elim (Rle_or_lt 0 r3); intros G1 G2; split.

apply Rle_trans with 0%R; [intuition | apply Rmult_le_pos; assumption].

replace r1 with (1*r1)%R; [apply Rmult_le_compat; intuition| ring].

replace (r2 * r3)%R with (-(r2*-r3))%R; [apply Ropp_le_contravar|ring].
replace r1 with (1*- -r1)%R; [apply Rmult_le_compat; intuition| ring].

apply Rle_trans with (-0)%R; [idtac| fourier].  
replace (r2 * r3)%R with (-(r2*-r3))%R; [apply Ropp_le_contravar|ring].
apply Rmult_le_pos; intuition.

replace (r2 * r3)%R with (-(-r2*r3))%R; [apply Ropp_le_contravar|ring].
replace r1 with (- - 1*r1)%R; [apply Rmult_le_compat; intuition| ring].

apply Rle_trans with (-0)%R; [idtac| fourier].  
replace (r2 * r3)%R with (-(-r2*r3))%R; [apply Ropp_le_contravar|ring].
apply Rmult_le_pos; intuition.

replace (r2 * r3)%R with (-r2*-r3)%R; [idtac|ring].
apply Rle_trans with 0%R; [intuition | apply Rmult_le_pos; intuition].

replace (r2 * r3)%R with (-r2*-r3)%R; [idtac|ring].
replace r1 with (1*r1)%R; [apply Rmult_le_compat; intuition; fourier| ring].
Qed.

Theorem Rmult_le_compat_sym : 
  forall a b x y, 
    (0 <= a -> 
      0 <= b -> 
      -a <= x <= a -> 
      -b <= y <= b -> 
      -(a*b) <= x*y <= a*b)%R.
Proof.
intros a b x y Hapos Hbpos [Hx1 Hx2]  [Hy1 Hy2].
elim (Rlt_le_dec 0 x);  
elim (Rlt_le_dec 0 y); intros G1 G2; split.

apply Rle_trans with 0%R.
assert (0<= a*b)%R; [idtac| fourier].
apply Rmult_le_pos; fourier.
apply Rmult_le_pos; fourier.

apply Rmult_le_compat; fourier.

replace (x*y)%R with (-(x*-y))%R by ring.
apply Ropp_le_contravar.
apply Rmult_le_compat; fourier.

apply Rle_trans with 0%R.
replace 0%R with (x*0)%R; [apply Rmult_le_compat_l; fourier| ring].
apply Rmult_le_pos; fourier.

replace (x*y)%R with (-(-x*y))%R by ring.
apply Ropp_le_contravar.
apply Rmult_le_compat; fourier.

apply Rle_trans with 0%R.
replace 0%R with (0*y)%R; [apply Rmult_le_compat_r; fourier| ring].
apply Rmult_le_pos; fourier.

apply Rle_trans with 0%R.
assert (0<= a*b)%R; [idtac| fourier].
apply Rmult_le_pos; fourier.
replace (x * y)%R with (-x*-y)%R by ring.
apply Rmult_le_pos; fourier.

replace (x * y)%R with (-x*-y)%R by ring.
apply Rmult_le_compat; fourier.
Qed.

Theorem Rdiv_mult_compat : 
  forall a b c d, 
    0 < b ->
    0 < d ->
    a * b <= c * d -> 
    a / d <= c / b.
Proof.
intros a b c d H1 H2 H3.
apply Rle_div_r; [assumption |].
replace (c / b * d) with ((c*d)/b) by (unfold Rdiv; ring).
apply Rle_div_l; assumption.
Qed.

Theorem pow_1_1 : 
  forall x n,
    -1 <= x <= 1 -> 
    -1 <= x^n <= 1.
Proof.
intros x n H; induction n; simpl.
split; fourier.

elim (Rle_or_lt 0 x); intros G1.
assert (H2: 0 <= x^n) by ZR0compat.
split.
apply Rle_trans with 0; [fourier|].
apply Rmult_le_pos; assumption.
replace 1 with (1*1) by ring.
apply Rmult_le_compat; intuition.

elim (Rle_or_lt 0 (x^n)); intros G2.
replace (x * x ^ n) with (-(-x*x^n)) by ring.
split.
apply Ropp_le_contravar.
replace 1 with (1*1) by ring.
apply Rmult_le_compat; intuition; fourier.
apply Rle_trans with (-0); [|fourier].
apply Ropp_le_contravar.
apply Rmult_le_pos; intuition; fourier.
replace (x * x ^ n) with (-x*-x^n) by ring.
split.
apply Rle_trans with 0; [fourier|].
apply Rmult_le_pos; fourier.
replace 1 with (1*1) by ring.
dec_tac; apply Rmult_le_compat; fourier.
Qed.

Theorem pow_minus : 
  forall x n m, 
    (n <= m)%nat ->
    x <> 0 -> 
    x ^(m-n)%nat = x^m/x^n.
Proof.
intros x n m H1 H2.
pattern m at 2.
replace m with (n + (m-n))%nat by omega.
rewrite pow_add.
field.
apply pow_nonzero.
assumption.  
Qed.

Theorem Rle_le_trans : 
  forall a b c, 
    a <= b -> 
    -a <= c <= a -> 
    -b <= c <= b.
Proof.
intros a b c Hab [Hac Hca].
split.
apply Rle_trans with (2:= Hac).
apply Ropp_le_contravar; assumption.
apply Rle_trans with a; assumption.
Qed.

(*
Theorem IZR_mult_div_eq : 
  forall a b c d, 
    (a*d=b*c ->
      b <> 0 -> 
      d <> 0 ->
      (IZR a / IZR b = IZR c / IZR d)%R)%Z.
Proof.
intros a b c d H1 H2 H3.
field_simplify_eq.
repeat rewrite <- mult_IZR; rewrite H1; auto.

change 0%R with (IZR 0)%Z.
split; apply IZR_neq; assumption.
Qed.
*)

(*** Z ****)

Theorem Zpower_two_p : 
  forall z, (two_p z = 2 ^ z)%Z.
Proof.
intros z.
unfold two_p, Zpower.
case z; clear z; auto.
intros p.
apply two_power_pos_correct.
Qed.

Theorem N_digits_correct : 
  forall b, (0 < b ->
    2 ^ (N_digits b) <= b < 2^ (N_digits b+1))%Z.
Proof.
intros b.
repeat rewrite <- Zpower_two_p.
replace (N_digits b+1)%Z with (Zsucc (N_digits b)) by ring.
unfold N_digits; case b.
intros Hbpos; simpl; absurd (1=0)%Z; omega.
intros p; generalize (log_inf_correct p); tauto.
intros p Hpos.
generalize (Zlt_neg_0 p); intros H1.
absurd (1<=0)%Z; omega.
Qed.

Theorem Zmult_le_compat_sym : 
  forall a b x y, 
    (0 <= a -> 
      0 <= b -> 
      -a <= x <= a -> 
      -b <= y <= b -> 
      -(a*b) <= x*y <= a*b)%Z.
Proof.
intros a b x y Hapos Hbpos Hx Hy.
assert (U1: (- (Z2R a * Z2R b) <= Z2R x * Z2R y <= Z2R a * Z2R b)%R).
apply Rmult_le_compat_sym; try split; R_to_Z.
split; apply le_Z2R; simpl_Z2R; dec_tac; assumption.
Qed.

Theorem Zdiv_mult_bound : 
  forall a b c, 
    (c > 0->
    0 < b -> 
    0 <= a < c -> 
    0 <= (a*b)/c < b)%Z.
Proof.
intros a b c Hcpos Hbpos [Ha1 Ha2].
split.
apply Zge_le.
apply Z_div_ge0.
omega.
apply Zle_ge.
apply Zmult_le_0_compat; omega.

generalize (Z_div_mod_eq (a*b) c Hcpos).
generalize (Z_mod_lt (a*b) c Hcpos).
generalize (a*b/c)%Z; intros k.
generalize (a*b mod c)%Z; intros r.
intros [G1 G2] G3.
make_Z2R_ineq.
intros G1 G2 G3 G4 G5 G6 G7 G8 G9 G10 G11 G12.
apply lt_Z2R.
assert (G13 : (Z2R k = (Z2R a * Z2R b-Z2R r)/Z2R c)%R).
rewrite G2; field_Z.
rewrite G13.
unfold Rminus.
rewrite Rdiv_plus_distr_r.
assert (G14: (0 < Z2R c)%R).
apply (Z2R_lt 0); omega.
assert (G15: (Z2R a/Z2R c < 1)%R).
apply Rmult_lt_reg_l with (Z2R c); [assumption | 
  rewrite Rmult_div_simpl; [fourier| R_to_Z]].
replace (Z2R a*Z2R b/Z2R c)%R with (Z2R a/Z2R c*Z2R b)%R;
  [idtac| unfold Rdiv; ring].
pattern (Z2R b) at 2; replace (Z2R b) with (1* Z2R b+ 0)%R; [idtac| ring].
apply Rplus_lt_le_compat.
apply Rmult_lt_compat_r; assumption.
apply Rle_div_r; [assumption| rewrite Rmult_0_l; fourier].
Qed.

Theorem Z_of_nat_Zabs_nat : 
  forall x, (0 <= x)%Z ->
    Z_of_nat(Zabs_nat x) = x.
Proof.
intros x Hle;destruct x.
auto.
unfold Zabs_nat.
rewrite Zpos_eq_Z_of_nat_o_nat_of_P.
auto.
assert False.
apply Hle; auto.
contradiction.
Qed.

Theorem Zpower_incr : 
  forall a b n, (0<= n -> 0<=a<=b -> a^n <= b^n)%Z.
Proof.
intros a b n.
unfold Zpower.
case n; clear n.
intros; omega.
intros p H1 H2. 
repeat rewrite (Zpower_pos_nat).
generalize (nat_of_P p).
induction n.
unfold Zpower_nat.
simpl; omega.
replace (S n) with (n + 1)%nat;[idtac|omega].
repeat rewrite Zpower_nat_is_exp.
apply Zmult_le_compat.
assumption.
unfold Zpower_nat.
simpl; omega.
clear IHn.
induction n.
unfold Zpower_nat.
simpl; omega.
replace (S n) with (n + 1)%nat;[idtac|omega].
rewrite Zpower_nat_is_exp.
apply Zmult_le_0_compat.
assumption.
unfold Zpower_nat.
simpl; omega.
unfold Zpower_nat.
simpl; omega.
intros; omega.
Qed.

Theorem Zle_power :
  forall z m n,
    (1 <= z -> 0 <= m <= n -> z ^ m <= z ^ n)%Z.
Proof.
intros z m n Hz Hmn.
apply le_Z2R.
rewrite <- (Z_of_nat_Zabs_nat m); [| omega]. 
rewrite <- (Z_of_nat_Zabs_nat n); [| omega].
simpl_Z2R.
apply Rle_pow. 
apply (Z2R_le 1); omega.
elim Hmn; clear Hmn; intros H1 H2.
elim (Zle_lt_or_eq _ _ H2); intros H3.
apply lt_le_weak.
apply Zabs_nat_lt; omega.
subst; omega.
Qed.

Theorem Zabs_nat_1 : Zabs_nat 1 = 1%nat.
auto.
Qed.

Theorem Rmult_inv : 
  forall x,
    x <> 0 -> 
    x * (1/x) =1.
Proof.
intros; field; auto.
Qed.



Ltac let_simpl :=
  match goal with 
    |- context [(match ?X with pair _ _ => _ end)] => 
      caseEq X
  end.

Theorem Zpower_mult : 
  forall a b c, 
    (0 <= b -> 
      0 <= c -> 
      a^(b*c) = (a^b)^c)%Z.
Proof.
intros a b c Hb Hc.
generalize Hc.
rewrite <- (Z_of_nat_Zabs_nat c); try assumption.
generalize (Zabs_nat c).
intros n _.
generalize n a b Hb; clear.
induction n.
intros.
rewrite Zmult_0_r.
reflexivity.
replace (S n) with (n+1)%nat by ring.
rewrite inj_plus; simpl.
intros a b Hbpos.
rewrite Zmult_plus_distr_r.
repeat rewrite Zpower_exp; try omega.
rewrite Zpower_1_r.
rewrite Zmult_1_r.
rewrite IHn; omega.
ZR0compat.
Qed.


Theorem Rmin_left : 
  forall a b, 
    a<=b -> Rmin a b = a.
Proof.
intros; unfold Rmin; elim Rle_dec; intuition.
Qed.

Theorem Rmin_right : 
  forall a b, 
    b <= a -> Rmin a b = b.
Proof.
intros; unfold Rmin; elim Rle_dec; intuition.
Qed.

Theorem Rmin_plus_r : 
  forall a b c, 
    Rmin (a+c) (b+c) = Rmin a b + c.
Proof.
intros; unfold Rmin; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmin_plus_l : 
  forall a b c, 
    Rmin (c+a) (c+b) = c + Rmin a b.
Proof.
intros; unfold Rmin; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmin_minus_r : 
  forall a b c, 
    Rmin (a-c) (b-c) = Rmin a b - c.
Proof.
unfold Rminus; intros; apply Rmin_plus_r.
Qed.

Theorem Rmin_mult_r : 
  forall a b c, 0<=c ->
    Rmin (a*c) (b*c) = Rmin a b * c.
Proof.
intros; unfold Rmin; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmin_mult_l : 
  forall a b c, 0<=c ->
    Rmin (c*a) (c*b) = c* Rmin a b.
Proof.
intros; unfold Rmin; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmin_div_r : 
  forall a b c, 0<c ->
    Rmin (a/c) (b/c) = Rmin a b / c.
Proof.
intros; unfold Rdiv; apply Rmin_mult_r.
apply Rlt_le; apply Rinv_0_lt_compat; assumption.
Qed.

Theorem Rmin_le_r : 
  forall a b c, 
    a <= c \/ b <= c -> Rmin a b <= c.
Proof.
intros; unfold Rmin; elim Rle_dec; intuition.
apply Rle_trans with b; assumption.
apply Rle_trans with a; intuition.
Qed.

Theorem Rmin_le_l : 
  forall a b c, 
    c <=  a -> c <= b -> c <= Rmin a b.
Proof.
intros; unfold Rmin; elim Rle_dec; intuition.
Qed.

Theorem Rmax_r : 
  forall a b, 
    a<=b -> Rmax a b = b.
Proof.
intros; unfold Rmax; elim Rle_dec; intuition.
Qed.

Theorem Rmax_l : 
  forall a b, 
    b <= a -> Rmax a b = a.
Proof.
intros; unfold Rmax; elim Rle_dec; intuition.
Qed.

Theorem Rmax_plus_r : 
  forall a b c, 
    Rmax (a+c) (b+c) = Rmax a b + c.
Proof.
intros; unfold Rmax; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmax_plus_l : 
  forall a b c, 
    Rmax (c+a) (c+b) = c + Rmax a b.
Proof.
intros; unfold Rmax; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmax_minus_r : 
  forall a b c, 
    Rmax (a-c) (b-c) = Rmax a b - c.
Proof.
unfold Rminus; intros; apply Rmax_plus_r.
Qed.

Theorem Rmax_mult_r : 
  forall a b c, 0<=c ->
    Rmax (a*c) (b*c) = Rmax a b * c.
Proof.
intros; unfold Rmax; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmax_mult_l : 
  forall a b c, 0<=c ->
    Rmax (c*a) (c*b) = c* Rmax a b.
Proof.
intros; unfold Rmax; do 2 elim Rle_dec; intuition.
Qed.

Theorem Rmax_div_r : 
  forall a b c, 0<c ->
    Rmax (a/c) (b/c) = Rmax a b / c.
Proof.
intros; unfold Rdiv; apply Rmax_mult_r.
apply Rlt_le; apply Rinv_0_lt_compat; assumption.
Qed.

Theorem Rmax_le_l : 
  forall a b c, 
    c <= a \/ c <= b -> c <= Rmax a b .
Proof.
intros; unfold Rmax; elim Rle_dec; intuition.
apply Rle_trans with a; assumption.
apply Rle_trans with b; intuition.
Qed.

Theorem Rmax_le_r : 
  forall a b c, 
    a <=  c -> b <= c -> Rmax a b <=c .
Proof.
intros; unfold Rmax; elim Rle_dec; intuition.
Qed.

Theorem Rabs_le_le : 
  forall x y,
    0 <= y ->
    -y <= x <= y ->
    -y <= Rabs x <= y.
Proof.
intros x y Hypos Hx.
elim (Rle_or_lt 0 x); intros U1.
rewrite Rabs_right; assumption || fourier.
dec_tac; rewrite Rabs_left; try split; fourier.
Qed.

Theorem Rabs_div :
  forall x y : R, y <> 0 ->
    Rabs (x / y) = Rabs x / Rabs y.
Proof.
intros x y Hyneq0.
unfold Rdiv.
rewrite Rabs_mult.
rewrite Rabs_Rinv; trivial.
Qed.

Theorem Rabs_le_div_r : 
  forall x y z,
    0 < y -> 
    Rabs x <= z * y ->
    Rabs (x/y) <= z .
Proof.
intros x y z Hy Hrabs.
rewrite Rabs_div; [| ZR0compat].
rewrite (Rabs_right y); [| ZR0compat].
apply Rle_div_r; assumption.
Qed.

Theorem Rle_trans_trans : 
  forall w x y z,
    w <= x -> x <= y -> y <= z ->
    w <= z.
Proof.
intros; fourier.
Qed.

Theorem Rabs_le_le_eq : 
  forall x y, 
    Rabs x <= y <-> -y <= x <= y. 
Proof.
intros x y; unfold Rabs; split; 
  elim Rcase_abs; intros H1 H2;
    split || dec_tac; fourier.
Qed.

Theorem Rabs_lt_lt_eq : 
  forall x y, 
    Rabs x < y <-> -y < x < y. 
Proof.
intros x y; unfold Rabs; split; 
  elim Rcase_abs; intros H1 H2;
    split || dec_tac; fourier.
Qed.

Ltac rabs_intros :=
  match goal with 
    | id : - ?y <= ?x <= ?y |- _ => 
      rewrite <- Rabs_le_le_eq in id; rabs_intros
    | id : - ?y < ?x < ?y |- _ => 
      rewrite <- Rabs_lt_lt_eq in id; rabs_intros
    | |- (- ?y <= ?x <= ?y) => rewrite <- Rabs_le_le_eq
    | |- (- ?y < ?x < ?y) => rewrite <- Rabs_lt_lt_eq
    | |- _ => idtac
  end.

Ltac rabs_elim :=
  match goal with 
    | id : Rabs ?x <= ?y |- _ => 
      rewrite Rabs_le_le_eq in id; rabs_elim 
    | id : Rabs ?x < ?y |- _ => 
      rewrite Rabs_lt_lt_eq in id; rabs_elim 
    | |- (Rabs ?x <= ?y) => rewrite Rabs_le_le_eq 
    | |- (Rabs ?x < ?y) => rewrite Rabs_lt_lt_eq 
    | |- _ => idtac
  end.

Theorem Rabs_add : 
  forall w x y z,
    Rabs w <= y ->
    Rabs x <= z ->
    Rabs (w+x) <= y+z.
Proof.
intros w x y z H1 H2.
apply Rle_trans with (1:= (Rabs_triang w x)).
fourier.
Qed.


