Require Export exact_newton.
Module SQRT(Nums : NUMS).

Module Newton_correct := NEWTON(Nums).
Export Newton_correct.

Section sqrt_correct.
(* Square root*)
Variable vy : R.
Variable  n : nat.
Variable y EX0: stream digit.
Variable X0 : R.

Definition b0 := Rabs (((Z2R b ^2) * X0^2 - vy )/ (2*(Z2R b ^2) * X0)).
Definition g_sqrt y x := (x/ (1+1)%D - y * ((Sinverse (2) x) / (1+1)%D))%S.
Definition f_sqrt y x := (Z2R b ) ^2 * x ^ 2 - y.
Definition f'_sqrt x := 2*(Z2R b)^2 * x.

Hypothesis Hin_disc:  included (c_disc X0 (2*b0)) (c_I (/Z2R b^2) 1).
Hypothesis H_X0: /Z2R b <= X0 < 1.                                                                                                                                                                           
Hypothesis Hn: Z2R b <= 2 ^ n.
Hypothesis Hn2: (Z2R b - 2) * Rpower 2 (INR n-1) >= b0 * (2* Z2R b * Z2R b).
Hypothesis Hny:  / (Z2R b)^2 <= vy <= 1.
Hypothesis Hrepy: Srepresents y vy.
Hypothesis HrepX0: Srepresents EX0 X0.

(*
Theorem g_f_f'2:
  forall x vx,
    0<= vx ->
    Srepresents x vx ->
    (c_disc X0 (2*b0) vx) ->
    f'_sqrt vx <> 0 -> 
    -1 <= f_sqrt vy vx / f'_sqrt vx <= 1 ->
(* 1 <= vx * Z2R b ^ 2 \/ vx * Z2R b ^ 2 <= -1 ->*)
    Srepresents (g_sqrt y x) ((f_sqrt vy vx)/ (f'_sqrt vx)).
Proof.
intros x vx Hvx Hrepx Hdisc Hdif Hineq.
 assert (Hbpos := base_positive). 
unfold g_sqrt, f_sqrt, f'_sqrt.


assert (Hone: 1 <= vx * Z2R b ^ 2).
assert (Hin:= Hin_disc vx Hdisc).
elim Hin; intros I1 I2.
apply Rmult_le_reg_l with (/(Z2R b^2)).
apply Rinv_0_lt_compat; simpl pow; ZR0compat.
replace (/ Z2R b ^ 2 * (vx * Z2R b ^ 2)) with vx;
ring_simplify; [assumption|field; ZR0compat].

replace (((Z2R b) ^2* vx ^ 2 - vy) / (2 * (Z2R b) ^2* vx)) with 
  (vx / IDR 2%D - vy * ((1 / ( (Z2R b) ^2* vx))/ IDR 2)%D).

apply Ssub_correct.
apply div_digit_correct.
simpl_IDZ; omega.
simpl_IDZ; omega.
assumption.
apply Smult_correct.
assumption.
apply div_digit_correct.
simpl_IDZ; omega.
simpl_IDZ; omega.

apply Sinverse_correct.
assumption.
left; assumption.

assert (U1 :=  Srepresents_interval _ _ Hrepx).
assert (U2 :=  Srepresents_interval _ _ Hrepy).
simpl_IDR.
rabs_intros.
pattern 1%R at 6; replace 1%R with (1/2+(1*1)/2) by field.
unfold Rminus; apply Rabs_add.
  rewrite Rabs_div; [| ZR0compat]. 
  rewrite (Rabs_right 2); [|ZR0compat].
  apply Rdiv_le_compat_r; ZR0compat.

  rewrite Rabs_Ropp.
  replace (vy * (1 / (Z2R b ^ 2 * vx) / 2)) with 
    ((vy * (1 / (Z2R b ^ 2 * vx))) / 2) by (unfold Rdiv; ring).
  rewrite Rabs_div; [| ZR0compat]. 
  rewrite (Rabs_right 2); [|ZR0compat].
  apply Rdiv_le_compat_r; [ZR0compat |].
  rewrite Rabs_mult.
  apply Rmult_le_compat; assumption || ZR0compat.
  rabs_elim; rewrite Rmult_comm.
  apply Sinverse_tech_lemma.
left; assumption.

field_IDR.
ZR0compat.
apply Rlt_le_trans with (1/Z2R b^2); [ZR0compat|].
apply Rle_div_r; ZR0compat.
Qed.
*)
Definition Ssqrt (x : Stream digit) := 
  compute_newton (g_sqrt x) (EX0)%S 2%nat n.

Lemma pr_f_sqrt vx:
forall x, c_I (-1) 1 x -> derivable_pt (f_sqrt vx) x.
Proof.
intros; unfold f_sqrt; reg.
Qed.
(*
Theorem sqrt_correct : 
  forall vx, 0 <= vx ->
    vx^2 =  vy ->
    Srepresents (Ssqrt y) (vx/Z2R b).
Proof.
assert (Hbpos:= base_positive).
intros vx Hvxpos Heq.
unfold Ssqrt.
assert (Hdif: 0 < X0 <1).
split; elim H_X0; intros; [apply Rlt_le_trans with (/Z2R b); [ZR0compat|trivial]|trivial].
assert (Hf': f'_sqrt X0 > 0).
unfold f'_sqrt;  ZR0compat; dec_tac; assumption.
assert (H0: forall (t : R) (H : c_I (-1) 1 t),
derive_pt (f_sqrt vy) t (pr_f_sqrt vy t H) = f'_sqrt t).
intros.
unfold f_sqrt; unfold pow; unfold f'_sqrt; reg.
assert (H1:forall y0 : R, c_I (-1) 1 y0 -> continuity_pt f'_sqrt y0).
intros; unfold f'_sqrt; reg.
assert (H2: forall y1 y2 : R,
c_I (-1) 1 y1 ->
c_I (-1) 1 y2 ->
Rabs (f'_sqrt y1 - f'_sqrt y2) <= 2 * Z2R b ^ 2 * Rabs (y1 - y2)).
intros; unfold f'_sqrt.
replace (2 * Z2R b ^ 2 * y1 - 2 * Z2R b ^ 2 * y2) with (2 * Z2R b ^ 2 * (y1 -  y2)) by ring.
rewrite Rabs_mult.
apply Rmult_le_compat_r; [apply Rabs_pos|rewrite Rabs_right; [right;trivial|ZR0compat]].
assert (H3:2 * / (2 * Z2R b ^ 2 * X0) * b0 * (2 * Z2R b ^ 2) <= 1).
replace (2 * ( / (2 * Z2R b ^ 2 * X0)) * b0 * (2 * Z2R b ^ 2) ) with
 (2 * (1 / X0) * b0  ); [| field; split; dec_tac;  ZR0compat].
unfold b0.
 rewrite Rabs_div; [|dec_tac;ZR0compat].
pattern Rabs at 1; rewrite Rabs_right; [|dec_tac; ZR0compat].
replace (2 * (1 / X0) * (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy) / (2 * Z2R b ^ 2 * X0)) ) with
  ( (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy) / ( Z2R b ^ 2 * X0^2)) ); 
   [|field; split; dec_tac;ZR0compat].
apply Rmult_le_reg_l with ((Z2R b ^ 2 * X0 ^ 2) );[dec_tac; ZR0compat|].
 replace (Z2R b ^ 2 * X0 ^ 2 * (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy) / (Z2R b ^ 2 * X0 ^ 2)))
 with (Rabs (Z2R b ^ 2 * X0 ^ 2 - vy)); [|field_Z; dec_tac; apply Rgt_not_eq; trivial].
rewrite Rabs_right.
rewrite Rmult_1_r; 
 pattern (Z2R b ^ 2 * X0 ^ 2 ) at 2; rewrite <-(Rplus_0_r (Z2R b ^ 2 * X0 ^ 2 )). 
unfold Rminus; apply Rplus_le_compat_l; apply Ropp_le_0; apply Rle_trans with (/Z2R b^2); 
[left; ZR0compat|dec_tac; assumption].
apply Rle_ge; apply Rle_trans with (1-vy); [dec_tac; fourier|].
unfold Rminus; apply Rplus_le_compat_r.
replace 1 with (Z2R b ^ 2 * (/Z2R b) ^ 2) by field_Z.
apply Rmult_le_compat_l; [ZR0compat|].
apply pow_incr; split; [left; ZR0compat|elim H_X0; intros; assumption].
(assert (H4: -1 < X0 < 1)).
dec_tac; split;  fourier.
assert (H5: f'_sqrt X0 <> 0).
apply Rgt_not_eq; assumption.
assert (H6: Rabs (/ f'_sqrt X0) <= / (2 * Z2R b ^ 2 * X0)).
rewrite Rabs_Rinv; [|apply Rgt_not_eq; assumption]; rewrite Rabs_right;
  [|left;assumption].
unfold f'_sqrt; right; trivial.
assert (H7: Rabs (f_sqrt vy X0 / f'_sqrt X0) <= b0).
unfold b0; unfold f_sqrt; unfold f'_sqrt; right; trivial.
assert (H8: included (c_disc X0 (2 * b0)) (c_I (-1) 1)).
apply included_trans with (c_I (/ Z2R b ^ 2) 1); [trivial|].
intros x U; elim U; intros U1 U2; split; [|trivial].
apply Rle_trans with 0; [fourier|apply Rle_trans with (/Z2R b^2); 
  [left|];ZR0compat].
assert (H9: forall (x : Stream digit) (vx0 : R),
Srepresents x vx0 ->
c_disc X0 (2 * b0) vx0 ->
f'_sqrt vx0 <> 0 ->
-1 <= f_sqrt vy vx0 / f'_sqrt vx0 <= 1 ->
Srepresents (g_sqrt y x) (f_sqrt vy vx0 / f'_sqrt vx0)). 
intros x vx0 Hq0 Hd Hq1 Hq3; apply g_f_f'2; try assumption.
assert (Hd1:= Hin_disc vx0 Hd).
elim Hd1; intros U I.
apply Rle_trans with (/Z2R b^2 ); [left; ZR0compat|assumption].

apply kanto_correct  with (f_sqrt vy ) (f'_sqrt) X0 (/(2*Z2R b^2 * X0)) 
  b0 (2*Z2R b^2)  (pr_f_sqrt vy); try assumption.

elim ( kantoro_exists_b _ _ _ X0 _ _ _ _ H3 (pr_f_sqrt vy) H0 H1 H2 H4 
  H5 H6 H7 H8  ); intros xs [Hxscv [Hxsd Hxs0]].
assert (root: f_sqrt vy (vx / Z2R b) = 0).
unfold f_sqrt; rewrite <-Heq; try field_Z .
unfold f_sqrt in Hxs0, root.

replace (Z2R b ^ 2 * xs ^ 2 ) with (Rsqr (Z2R b * xs) ) in Hxs0;[|unfold Rsqr; field].
replace (Z2R b ^ 2 * (vx / Z2R b) ^ 2 ) with (Rsqr (Z2R b * (vx / Z2R b)) )
 in root;[|unfold Rsqr;  field_Z].
assert (haux: Rsqr (Z2R b * xs) = vy ).
apply (Rplus_eq_reg_l (-vy)); ring_simplify; rewrite Rplus_comm; assumption.
assert (au: Rsqr (Z2R b * (vx / Z2R b)) = vy).
apply (Rplus_eq_reg_l (-vy)); ring_simplify; rewrite Rplus_comm; assumption.
assert (E:xs = vx/Z2R b).
apply Rsqr_inj.
  elim (Hin_disc xs).
  intros Z1 Z2.
  apply Rle_trans with (/Z2R b^2); [left;ZR0compat|assumption].
  assumption.
  ZR0compat.
  apply Rmult_eq_reg_l with (Rsqr (Z2R b)).
  repeat rewrite <-Rsqr_mult.
  rewrite haux; rewrite au; trivial.
  apply Rgt_not_eq; apply Rlt_gt. 
  apply Rsqr_pos_lt;  ZR0compat.

rewrite <- E; assumption.
Qed.


Theorem sqrt_correct' : 
    Srepresents (Ssqrt y) ((sqrt vy)/(Z2R b)).
Proof.
assert (Hbpos:= base_positive).
assert (Hy0: 0 <= vy).
elim Hny;  intros Z1 Z2.
 apply Rle_trans with (/Z2R b^2); [left;ZR0compat|assumption].
assert (H2: (sqrt vy) ^ 2 = vy).
unfold  pow.
rewrite Rmult_1_r.
 apply sqrt_sqrt; try assumption.
exact ( sqrt_correct (sqrt vy) (sqrt_positivity vy Hy0) H2 ).
Qed.

*)


End sqrt_correct.



End SQRT.




Require Import digits_ex.



Module Base_2_124 <: BASE.
Definition b := (10^3)%Z.
Theorem base_positive : (4 < b)%Z.
compute; auto.
Qed.
End Base_2_124.

Module digits_test := BIGZNUMS Base_2_124.

Module Test_computations := SQRT digits_test.
Import Test_computations.

(*
Fixpoint EXn g x0 (k n : nat)  {struct n} : Stream digit :=
  match n with 
    | O => x0
    | S n => let exn := Tronquer (EXn g x0 k n) k in
      ( exn - g (exn))%S
  end.

CoFixpoint compute_newton (g : Stream digit -> Stream digit) EX0 k n :=
  match (make_digit (EXn g EX0 k n)) with
    | x1::x' => x1::compute_newton (fun x => multb (g (x1::x))) x' (S k) n
  end.
*)
Definition y := (Srat_to_Stream 1%D (1+1)%D).

Definition invy := (Sinverse 1%nat y). 

(*Srat_to_Stream 1%D (BigZ.of_Z 50000%Z). *)

Definition g_sqrt x := 
  (let xdiv2 := x/(1+1)%D in 
    (multb (invy*(x * x * xdiv2)))- xdiv2)%S.

Definition Ssqrt n (EX0 x: Stream digit) := 
  compute_newton g_sqrt (EX0)%S  1%nat n. 

(*compute the nth digit of a real*)
Fixpoint nth (A:Set) (n:nat)(s:Stream A) {struct n}: A :=
  match n, s with 0%nat,(d::s') => d | (S p),(_::s') => nth A p s' end.

(*Time  Eval vm_compute in (nth _ 30%nat invy).*)

Time  Eval vm_compute in (nth _ 1000%nat y ).

Definition x0 := (BigZ.of_Z 7071)::Szero.

(*Srat_to_Stream (BigZ.of_Z 7071) (BigZ.of_Z 100000).*)

Definition sqrty  := (Ssqrt 2%nat y y).

(*Time  Eval vm_compute in (nth _ 0%nat sqrty).
Time  Eval vm_compute in (nth _ 1%nat sqrty).
Time  Eval vm_compute in (nth _ 2%nat sqrty).
Time  Eval vm_compute in (nth _ 3%nat sqrty).
Time  Eval vm_compute in (nth _ 0%nat sqrty).
Time  Eval vm_compute in (nth _ 1%nat sqrty).
Time  Eval vm_compute in (nth _ 2%nat sqrty).
*)

(* Finished transaction in 5. secs (4.814268u,0.080988s) *)

Definition y2 := (Srat_to_Stream 1%D (1+1)%D).

Definition invy2 := (Sinverse 1%nat y2).
 
Definition g_sqrt2 x :=
   (x-g_sqrt x)%S.
 
Definition Ssqrt2 n (EX0 x: Stream digit) := 
  compute_newton2 g_sqrt2 (EX0)%S 0%nat n. 

Definition sqrty2  := (Ssqrt2 3%nat y2 y2).

Time  Eval vm_compute in (nth _ 0%nat sqrty).
(* 70710678118654752440084436210484903928483593768847%bigZ
     : digit
Finished transaction in 2. secs (1.305802u,0.013998s) *)

Time  Eval vm_compute in (nth _ 0%nat sqrty2).
Time  Eval vm_compute in (nth _ 1%nat sqrty2).
Time  Eval vm_compute in (nth _ 0%nat sqrty2).
Time  Eval vm_compute in (nth _ 0%nat sqrty2).

(*  70710678118654752440084436210484903928483593768847%bigZ
     : digit
Finished transaction in 1. secs (1.310801u,0.016997s)
*)

Time  Eval vm_compute in (nth _ 1%nat sqrty2).
(* = 40365883398689953662392310535194251937671638207864%bigZ
     : digit
Finished transaction in 2. secs (1.871715u,0.016998s)*)
Time  Eval vm_compute in (nth _ 1%nat sqrty).
(*     = 40365883398689953662392310535194251937671638207864%bigZ
     : digit
Finished transaction in 2. secs (2.546613u,0.012998s) *)

(* Finished transaction in 2. secs (2.16567u,0.021997s) *)




70711

(-32188)%bigZ
(-13452)
Definition x1 := (x0 - (g_sqrt x0))%S.

Definition x2 := (x1 - (g_sqrt x1))%S.

Definition x3 := (x2 - (g_sqrt x2))%S.

Definition x4 := (x3 - (g_sqrt x3))%S.

Time Eval vm_compute in (nth _ 0%nat x1).
Time Eval vm_compute in (nth _ 0%nat x2).

Time Eval vm_compute in (nth _ 0%nat x3).

Time Eval vm_compute in (nth _ 0%nat x4). 
Time Eval vm_compute in (nth _ 1%nat x4).
Time Eval vm_compute in (nth _ 2%nat x4).
Time Eval vm_compute in (nth _ 3%nat x4).
Time Eval vm_compute in (nth _ 4%nat x4).


Definition x0 := (Srat_to_Stream 1%D (1+1)%D).

Definition toto := (Ssqrt 2%nat x0 x0).

Time Eval vm_compute in (nth _ 0%nat toto).
Time Eval vm_compute in (nth _ 1%nat toto).
Time Eval vm_compute in (nth _ 2%nat toto).
Time Eval vm_compute in (nth _ 3%nat toto).
Time Eval vm_compute in (nth _ 10%nat toto).




Definition g_sqrt y x := 
  (let xdiv2 := x/(1+1)%D in 
    let invy := Sinverse 1 y in 
      (multb (x * x * xdiv2*invy))- xdiv2)%S.





(*(x/ (1+1)%D - y * ((Sinverse (2) x) / (1+1)%D))%S.
*)
Definition Ssqrt (x : Stream digit) := 
  compute_newton (g_sqrt x) x 3%nat.


Check Ssqrt.


Definition x0 := (Srat_to_Stream 1%D (1+1)%D).

Definition x1 := (x0 - (g_sqrt x0 x0))%S.

Definition x2 := (x1 - (g_sqrt x0 x1))%S.

Definition x3 := (x2 - (g_sqrt x0 x2))%S.




Definition x3' := EXn (g_sqrt x0) x0 3%nat.

Time Eval vm_compute in (nth _ 0%nat x3').
Time Eval vm_compute in (nth _ 0%nat x3).


     = 70671%bigZ
     : digit
Finished transaction in 59. secs (57.823209u,0.905862s)


Eval vm_compute in (nth _ 0%nat (a - (g_sqrt a a))%S).





Eval vm_compute in (nth _ 20%nat ( Sinverse 1 (Srat_to_Stream 1%D (1+1)%D))).



Time Eval vm_compute in (




(nth _ 0%nat 



  (Ssqrt (Srat_to_Stream 1%D (1+1)%D))).








Definition deux := (1+1)%D.
Definition dix := (deux*deux*deux+deux)%D.
Definition dix := (1+1+1+1+1+1+1+1+1+1)%D

