Require Export inverse_correct.
Require Import digits_ex.

Module TACTIC (Nums : NUMS).

Module Inverse_correct := INV_CORRECT(Nums).
Export Inverse_correct.

Fixpoint bounds_aux (x : Stream digit) (e n : nat) (c d : big) {struct n}: big * big :=
  match n with 
    O => (c, d)
    | S n => 
    match x with 
      | x1::x' => 
        match e with 
          | O   => bounds_aux x' e n c (d * Bbase + big_of_digit x1)%B
          | S e => bounds_aux x' e n (c * Bbase + big_of_digit x1)%B d
        end
    end 
  end.

Definition Ebounds (x : ER) (n : nat) : big * big :=
  match x with 
    (e, x) => bounds_aux x e (e+n)%nat 0%B 0%B
  end.

Ltac R_to_pos x :=
  match x with
    | R1 => constr:(1%positive)
    | Rplus R1 R1 => constr:(2%positive)
    | Rplus R1 (Rplus R1 R1) => constr:(3%positive)
    | Rplus R1 (Rmult (Rplus R1 R1) ?x) => 
      let x' := R_to_pos x in 
	constr:(xI x')
    | Rmult (Rplus R1 R1) ?x => 
      let x' := R_to_pos x in  
	constr:(xO x')
    | _ => fail 1
   end.

Ltac R_to_Z x :=
  match x with
    | R0 => constr:(0%Z)
    | Ropp ?x => let x':= R_to_pos x in constr:(Zneg x')
    | _ =>  let x':= R_to_pos x in constr:(Zpos x')
    | _ => fail 1
  end.

Ltac R_to_Z_expr x :=
  match x with
    | ?y => let x' := R_to_Z x in constr:(x')
    | Z2R ?x => x
    | IZR ?x => x
    | Rabs ?x =>
      let x' := R_to_Z_expr x in
	constr:(Zabs x')
    | Ropp ?x =>
      let x' := R_to_Z_expr x in
	constr:(Zopp x')
    | Rplus ?x ?y =>
      let x' := R_to_Z_expr x in
	let y' := R_to_Z_expr y in
	  constr:(Zplus x' y')
    | Rminus ?x ?y =>
      let x' := R_to_Z_expr x in
	let y' := R_to_Z_expr y in
	  constr:(Zminus x' y')
    | Rmult ?x ?y =>
      let x' := R_to_Z_expr x in
	let y' := R_to_Z_expr y in
	  constr:(Zmult x' y')
    | pow ?x ?y =>
      let x' := R_to_Z_expr x in
	  constr:(Zpower x' (Z_of_nat y))
    | _ => fail 1
  end.

Ltac find_representation expr n :=
  match expr with 
    | 0%R => constr:(zero)
    | 1%R => constr:(one)
    | PI => constr:(pi)
    | (/ ?x)%R => 
      let x' := find_representation x n in 
	constr:(inverse x' n)
    | (?x / ?y)%R => 
      let x' := R_to_Z_expr x in
        let y' := R_to_Z_expr y in
          constr:(rat_to_stream (IZB x') (IZB y'))
    | (1/ ?x)%R => 
      let x' := find_representation x n in 
	constr:(inverse x' n)
    | ?x => let x' := R_to_Z_expr x in
      constr:(rat_to_stream (IZB x') 1%B)
    | Rabs ?x =>
      let x' := find_representation x n in 
	constr:(abs x')
    | (- ?x)%R => 
      let x' := find_representation x n in
	constr:(opp x')
    |  (?x + ?y)%R => 
      let x' := find_representation x n in 
	let y' := find_representation y n in 
	  constr:(add x' y')
    | (?x - ?y)%R => 
      let x' := find_representation x n in 
	let y' := find_representation y n in 
	  constr:(sub x' y')
    | (?x * ?y)%R => 
      let x' := find_representation x n in 
	let y' := find_representation y n in 
	  constr:(mult x' y') 
  end.

Ltac ineq_tac n' :=
  match goal with 
    |- (?vx < ?vy)%R =>
      let x' := find_representation vx n' in
	let y' := find_representation vy n' in
	  cut (represents x' vx); 
	    [ let Hx := fresh "Hrep" in intros Hx;
	      cut ( represents y' vy);
		[let HY := fresh "Hrep" in intros Hy;
		  apply compare_lt with (x:=x') (y:=y') (n:=n'); 
 		    [exact Hx|exact Hy|vm_compute;auto] 
		  |solve_represent n']
	      |solve_represent n']
    | |- (?vx <= ?vy)%R =>
      let x' := find_representation vx n' in
	let y' := find_representation vy n' in
	  cut (represents x' vx); 
	    [ let Hx := fresh "Hrep" in intros Hx;
	      cut ( represents y' vy);
		[let HY := fresh "Hrep" in intros Hy;
		  apply compare2_le with (x:=x') (y:=y') (n:=n'); 
 		    [exact Hx|exact Hy|vm_compute;auto] 
		    |solve_represent n']
		|solve_represent n']
  end

  with solve_represent n :=
  match goal with 
    | |- represents zero _ => apply rep_zero
    | |- represents one _ => apply rep_one
    | |- represents pi _ => apply pi_correct
    | |- represents (rat_to_stream ?x ?y) ?r => 
      replace r with (IBR x / IBR y)%R; 
	[apply rat_to_stream_correct; simpl_IBZ; ZR0compat|
	  simpl_IBR; try field]
    | |- represents (opp _) _ => apply opp_correct; solve_represent n
    | |- represents (abs _) _ => apply abs_correct; solve_represent n
    | |- represents (inverse _ _) _ => apply inverse_correct; 
      [simpl_Z2R; ineq_tac n | solve_represent n]
    | |- represents (add _ _) _ => apply add_correct; solve_represent n
    | |- represents (sub _ _) _ => apply sub_correct ; solve_represent n
    | |- represents (mult _ _) _ => apply mult_correct; solve_represent n
    | _ => idtac
  end.

End TACTIC.

